
/*!
*************************************************************************************
* \file image.c
*
* \brief
*    Code one image/slice
*
* \author
*    Main contributors (see contributors.h for copyright, address and affiliation details)
*     - Inge Lille-Langoy               <inge.lille-langoy@telenor.com>
*     - Rickard Sjoberg                 <rickard.sjoberg@era.ericsson.se>
*     - Jani Lainema                    <jani.lainema@nokia.com>
*     - Sebastian Purreiter             <sebastian.purreiter@mch.siemens.de>
*     - Byeong-Moon Jeon                <jeonbm@lge.com>
*     - Yoon-Seong Soh                  <yunsung@lge.com>
*     - Thomas Stockhammer              <stockhammer@ei.tum.de>
*     - Detlev Marpe                    <marpe@hhi.de>
*     - Guido Heising                   <heising@hhi.de>
*     - Thomas Wedi                     <wedi@tnt.uni-hannover.de>
*     - Ragip Kurceren                  <ragip.kurceren@nokia.com>
*     - Antti Hallapuro                 <antti.hallapuro@nokia.com>
*     - Alexis Michael Tourapis         <alexismt@ieee.org> 
*************************************************************************************
*/
#include "contributors.h"

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <string.h>
#include <memory.h>
#include <assert.h>

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include "global.h"

#include "refbuf.h"
#include "mbuffer.h"
#include "intrarefresh.h"
#include "fmo.h"
#include "sei.h"
#include "memalloc.h"
#include "nalu.h"
//#include "ratectl.h"
#include "mb_access.h"
#include "output.h"
#include "cabac.h"
#include "image.h"


#include "switched_filters.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"
// <FTRD Compatibility with APEC
extern MV_Competition mv_comp;
extern int ******send_index_mv_prediction;
extern int *****predModeMV;
// FTRD>
#endif

#ifdef ADAPTIVE_LOOP_FILTER
#include "adaptive_loop_filter.h"
#endif



#ifdef MB32X32
extern int mb32_stat[2][5];
extern int mb32_stat_currentpass[2][3][5];
#ifdef MB64X64
extern int mb64_stat[2][5];
extern int mb64_stat_currentpass[2][3][5];
#endif
extern void find_Mb_cluster(int level, int CurrentMbAddr, int *CurrentMbAddr_x, int *CurrentMbAddr_y, int *extend_x, int *extend_y, int *num_MB, int *nextMbAddr);
#endif


extern pic_parameter_set_rbsp_t *PicParSet[MAXPPS];

void code_a_picture(Picture *pic);
void frame_picture (Picture *frame, int method);

static int  writeout_picture(Picture *pic);

static void find_snr(void);
void find_distortion(void);

static void frame_mode_buffer (int bit_frame, float snr_frame_y, float snr_frame_u, float snr_frame_v);
static void init_frame(void);
static void put_buffer_frame(void);
static void copy_motion_vectors_MB(void);

static void PaddAutoCropBorders (int org_size_x, int org_size_y, int img_size_x, int img_size_y,
                                 int org_size_x_cr, int org_size_y_cr, int img_size_x_cr, int img_size_y_cr);
#ifdef ADAPTIVE_GOP
void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys, int xs_cr, int ys_cr);
#else
static void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys, int xs_cr, int ys_cr);
#endif

static void writeUnit(Bitstream* currStream ,int partition);

static void rdPictureCoding(void);
static void rdPictureCodingP(void);
static void rdPictureCodingB(void);


#ifdef _ADAPT_LAST_GROUP_
int *last_P_no;
int *last_P_no_frm;
#endif

static void ReportFirstframe(int tmp_time, int me_time);
static void ReportIntra(int tmp_time, int me_time);
static void ReportP(int tmp_time, int me_time);
static void ReportB(int tmp_time, int me_time);
static void ReportNALNonVLCBits(int tmp_time, int me_time);

static int CalculateFrameNumber(void);  // Calculates the next frame number

StorablePicture *enc_picture;
StorablePicture *enc_frame_picture;
StorablePicture *enc_frame_picture2;
StorablePicture *enc_frame_picture3;

//Rate control
int    QP;

const int ONE_FOURTH_TAP[3][2] =
{
  {20,20},
  {-5,-4},
  { 1, 0},
};



/*!
************************************************************************
* \brief
*    Encodes a picture
*
*    This is the main picture coding loop.. It is called by all this
*    frame and field coding stuff after the img-> elements have been
*    set up.  Not sure whether it is useful for MB-adaptive frame/field
*    coding
************************************************************************
*/
void code_a_picture(Picture *pic)
{
  unsigned int NumberOfCodedMBs = 0;
  int SliceGroup = 0;
  
  img->currentPicture = pic;
  img->currentPicture->idr_flag = ((!IMG_NUMBER)) || (input->idr_enable && (img->type == I_SLICE) && (img->curr_rd_pass==0));
  pic->no_slices = 0;
  pic->distortion_u = pic->distortion_v = pic->distortion_y = 0.0;
  
  
  // The slice_group_change_cycle can be changed here.
  // FmoInit() is called before coding each picture, frame or field
  img->slice_group_change_cycle=1;
  FmoInit(img, active_pps, active_sps);
  FmoStartPicture ();           //! picture level initialization of FMO
  
  CalculateQuantParam();
  CalculateOffsetParam();
  
  if(input->Transform8x8Mode)
  {
    CalculateQuant8Param();
    CalculateOffset8Param();
  }
  
  reset_pic_bin_count();
  img->bytes_in_picture = 0;
  
  while (NumberOfCodedMBs < img->PicSizeInMbs)       // loop over slices
  {
    // Encode one SLice Group
    while (!FmoSliceGroupCompletelyCoded (SliceGroup))
    {
      // Encode the current slice
      NumberOfCodedMBs += encode_one_slice (SliceGroup, pic, NumberOfCodedMBs);
      FmoSetLastMacroblockInSlice (img->current_mb_nr);
      // Proceed to next slice
      img->current_slice_nr++;
      stats->bit_slice = 0;
    }
    // Proceed to next SliceGroup
    SliceGroup++;
  }

  FmoEndPicture ();

  DeblockFrame (img, enc_picture->imgY, enc_picture->imgUV); //comment out to disable loop filter
}



/*!
************************************************************************
* \brief
*    Encodes one frame
************************************************************************
*/
int encode_one_frame (void)
{
  static int prev_frame_no = 0; // POC200301
  static int consecutive_non_reference_pictures = 0; // POC200301
#ifndef ADAPTIVE_GOP
int        FrameNumberInFile;
#endif

#ifdef _LEAKYBUCKET_
  extern unsigned long total_frame_buffer;
#endif  
 
  time_t ltime1;
  time_t ltime2;
  
#ifdef WIN32
  struct _timeb tstruct1;
  struct _timeb tstruct2;
#else
  struct timeb tstruct1;
  struct timeb tstruct2;
#endif
  
  int tmp_time;
  int bits_frm = 0;
  float dis_frm_y = 0.0, dis_frm_u = 0.0, dis_frm_v = 0.0;
  
  me_time=0;
  img->rd_pass = 0;
  enc_frame_picture  = NULL;
  enc_frame_picture2 = NULL;
  enc_frame_picture3 = NULL;

#ifdef WIN32
  _ftime (&tstruct1);           // start time ms
#else
  ftime (&tstruct1);
#endif
  time (&ltime1);               // start time s
  
#ifdef USE_NEW_OFFSET
  if(input->UseNewOffset)
    img->frameOffsetAvail = 0; 
#endif 


  if (input->ResendPPS && img->number !=0)
  {
    stats->bit_ctr_parametersets_n=write_PPS(0, 0);
    //stats->bit_slice += stats->bit_ctr_parametersets_n;
    stats->bit_ctr_parametersets += stats->bit_ctr_parametersets_n;
  }
  
  put_buffer_frame ();      // sets the pointers to the frame structures 
  // (and not to one of the field structures)
  init_frame ();

#ifdef ADAPTIVE_GOP
    if (input->AdaptiveGOP)
    {
      if(img->b_frame_to_code)   //for AdapGOP only call this function for HierChunk
        FrameNumberInFile = CalculateFrameNumber();
    }
    else
      FrameNumberInFile = CalculateFrameNumber();
#else
  FrameNumberInFile = CalculateFrameNumber();
#endif

  ReadOneFrame (FrameNumberInFile, input->infile_header,
    input->img_width, input->img_height, input->img_width_cr, input->img_height_cr);
  
  PaddAutoCropBorders (input->img_width, input->img_height, img->width, img->height,
    input->img_width_cr, input->img_height_cr, img->width_cr, img->height_cr);
  
  // set parameters for direct mode and deblocking filter
  img->direct_spatial_mv_pred_flag     = input->direct_spatial_mv_pred_flag;
  img->LFDisableIdc    = input->LFDisableIdc;
  img->LFAlphaC0Offset = input->LFAlphaC0Offset;
  img->LFBetaOffset    = input->LFBetaOffset;

#ifdef MB32X32
  memset(mb32_stat_currentpass[0][0], 0, sizeof(int)*5);
  memset(mb32_stat_currentpass[1][0], 0, sizeof(int)*5);
  memset(mb32_stat_currentpass[0][1], 0, sizeof(int)*5);
  memset(mb32_stat_currentpass[1][1], 0, sizeof(int)*5);
  memset(mb32_stat_currentpass[0][2], 0, sizeof(int)*5);
  memset(mb32_stat_currentpass[1][2], 0, sizeof(int)*5);

#ifdef MB64X64
  memset(mb64_stat_currentpass[0][0], 0, sizeof(int)*5);
  memset(mb64_stat_currentpass[1][0], 0, sizeof(int)*5);
  memset(mb64_stat_currentpass[0][1], 0, sizeof(int)*5);
  memset(mb64_stat_currentpass[1][1], 0, sizeof(int)*5);
  memset(mb64_stat_currentpass[0][2], 0, sizeof(int)*5);
  memset(mb64_stat_currentpass[1][2], 0, sizeof(int)*5);
#endif

#endif

  // Following code should consider optimal coding mode. Currently also does not support
  // multiple slices per frame.
  frame_ctr[img->type]++;
  snr->frame_ctr++;
  
  img->dc_org = Calc_dc_org();
  //FRAME CODING
      {
        if (input->GenerateMultiplePPS)
          active_pps = PicParSet[0];

        frame_picture (frame_pic_1, 0);


    if ((input->RDPictureIntra || img->type != I_SLICE) && input->RDPictureDecision)
    {
	  if(input->UseHPFilter == HPF_SIFO_FPO)
      {
        if(img->type == P_SLICE)
          rdPictureCodingFilterP_IPass();
		ComputeFiltersAndOffsets();  //compute frame filters based on first pass

        if(img->type == P_SLICE)
		{
          UpdateSequenceFilters_P();
		}
        else
          UpdateSequenceFilters_B();
      }
      else
      {
        if(img->type == P_SLICE)
          rdPictureCodingP();
        else
          rdPictureCodingB();
      }
    }


#if DUMP_MBTYPE 
    {
      int i;
      for (i=0; i<NUM_PIC_TYPE; i++)
        stats->bit_use_mb_type[i] += statsRDPass[img->rd_pass].bit_use_mb_type[i];
    }
#endif

#ifdef MV_COMPETITION // B-slice mv memory override bug fix by AT 2006/11/07
        if( input->RDPictureDecision )
        {
          if((img->type == B_SLICE) && (input->successive_Bframe > 1))
            Copy_MV_B_frame(enc_picture->mv, enc_picture->ref_idx);
        }
#endif  // MV_COMPETITION
      }
  
  stats->bit_ctr_emulationprevention += stats->em_prev_bits_frm;
  
  if (img->type != B_SLICE)
  {
    img->pstruct_next_P = 0;
  }
 
  // img->structure is FRAME 
    frame_mode_buffer (bits_frm, dis_frm_y, dis_frm_u, dis_frm_v);

	if (input->RDPictureDecision && img->rd_pass == 2)
		writeout_picture (frame_pic_3);
	else if (input->RDPictureDecision && img->rd_pass == 1)
		writeout_picture (frame_pic_2);
	else
		writeout_picture (frame_pic_1);


  if (frame_pic_3)
    free_slice_list(frame_pic_3);  
  if (frame_pic_2)
    free_slice_list(frame_pic_2);  
  if (frame_pic_1)
    free_slice_list(frame_pic_1);

  /*
  // Tian Dong (Sept 2002)
  // in frame mode, the newly reconstructed frame has been inserted to the mem buffer
  // and it is time to prepare the spare picture SEI payload.
  if (input->InterlaceCodingOption == FRAME_CODING
  && input->SparePictureOption && img->type != B_SLICE)
  CalculateSparePicture ();
  */
  
  if (input->PicInterlace == FRAME_CODING)
  {
    if (input->RestrictRef)
      UpdatePixelMap ();
  }
  
  if (input->Verbose != 0)
    find_snr ();
  else
  {
    snr->snr_y = 0.0;
    snr->snr_u = 0.0;
    snr->snr_v = 0.0;
    snr->sse_y = 0.0;
    snr->sse_u = 0.0;
    snr->sse_v = 0.0;
  }
  
  time (&ltime2);               // end time sec
#ifdef WIN32
  _ftime (&tstruct2);           // end time ms
#else
  ftime (&tstruct2);            // end time ms
#endif
  
  tmp_time = (int)((ltime2 * 1000 + tstruct2.millitm) - (ltime1 * 1000 + tstruct1.millitm));
  tot_time = tot_time + tmp_time;
  
        if (img->rd_pass==2)
        {
          store_picture_in_dpb(enc_frame_picture3);
          free_storable_picture(enc_frame_picture);
          free_storable_picture(enc_frame_picture2);
        }
        else if (img->rd_pass==1)
        {
          store_picture_in_dpb(enc_frame_picture2);
          free_storable_picture(enc_frame_picture);
          free_storable_picture(enc_frame_picture3);
        }
        else
        {
            store_picture_in_dpb(enc_frame_picture);
            free_storable_picture(enc_frame_picture2);
            free_storable_picture(enc_frame_picture3);
        }


#ifdef MB32X32
        mb32_stat[img->type][0] += mb32_stat_currentpass[img->type][img->rd_pass][0];
        mb32_stat[img->type][1] += mb32_stat_currentpass[img->type][img->rd_pass][1];
        mb32_stat[img->type][2] += mb32_stat_currentpass[img->type][img->rd_pass][2];
        mb32_stat[img->type][3] += mb32_stat_currentpass[img->type][img->rd_pass][3];
        mb32_stat[img->type][4] += mb32_stat_currentpass[img->type][img->rd_pass][4];
#ifdef MB64X64
        mb64_stat[img->type][0] += mb64_stat_currentpass[img->type][img->rd_pass][0];
        mb64_stat[img->type][1] += mb64_stat_currentpass[img->type][img->rd_pass][1];
        mb64_stat[img->type][2] += mb64_stat_currentpass[img->type][img->rd_pass][2];
        mb64_stat[img->type][3] += mb64_stat_currentpass[img->type][img->rd_pass][3];
        mb64_stat[img->type][4] += mb64_stat_currentpass[img->type][img->rd_pass][4];
#endif
#endif

#ifdef _LEAKYBUCKET_
  //increment counter of no. of coded frames
  total_frame_buffer++;
#endif

  // POC200301: Verify that POC coding type 2 is not used if more than one consecutive 
  // non-reference frame is requested or if decoding order is different from output order
  if (img->pic_order_cnt_type == 2)
  {
    if (!img->nal_reference_idc) consecutive_non_reference_pictures++;
    else consecutive_non_reference_pictures = 0;
    
    if (frame_no < prev_frame_no || consecutive_non_reference_pictures>1)
      error("POC type 2 cannot be applied for the coding pattern where the encoding /decoding order of pictures are different from the output order.\n", -1);
    prev_frame_no = frame_no;
  }
  
  if (stats->bit_ctr_parametersets_n!=0)
    ReportNALNonVLCBits(tmp_time, me_time);
  
  if (IMG_NUMBER == 0)
    ReportFirstframe(tmp_time,me_time);
  else
  {
    switch (img->type)
    {
    case I_SLICE:
      stats->bit_ctr_I += stats->bit_ctr - stats->bit_ctr_n;
      ReportIntra(tmp_time,me_time);
      break;
    case B_SLICE:
      stats->bit_ctr_B += stats->bit_ctr - stats->bit_ctr_n;
      ReportB(tmp_time,me_time);
      break;
    default:      // P
      stats->bit_ctr_P += stats->bit_ctr - stats->bit_ctr_n;
      ReportP(tmp_time,me_time);      
    }
  }  
  
  
  if (input->Verbose == 0)
  { 
    //for (i = 0; i <= (img->number & 0x0F); i++)
    //printf(".");
    //printf("                              \r");
    printf("Completed Encoding Frame %05d.\r",frame_no);
  }
  // Flush output statistics
  fflush(stdout);
  
  stats->bit_ctr_n = stats->bit_ctr;
  
  stats->bit_ctr_parametersets_n=0;
  
  if (IMG_NUMBER == 0)
    return 0;
  else
    return 1;


}


/*!
************************************************************************
* \brief
*    This function write out a picture
* \return
*    0 if OK,                                                         \n
*    1 in case of error
*
************************************************************************
*/
static int writeout_picture(Picture *pic)
{
  Bitstream *currStream;
  int partition, slice;
  Slice *currSlice;
  
#ifdef ADAPTIVE_LOOP_FILTER
  if(input->UseAdaptiveLoopFilter)
  {
    AdaptiveLoopFilterProcess(pic);
	TerminateSliceForALF(pic);
  }
#endif
  img->currentPicture=pic;
  
  for (slice=0; slice<pic->no_slices; slice++)
  {
    currSlice = pic->slices[slice];
    for (partition=0; partition<currSlice->max_part_nr; partition++)
    {
      currStream = (currSlice->partArr[partition]).bitstream;
      assert (currStream->bits_to_go == 8);    //! should always be the case, the 
      //! byte alignment is done in terminate_slice
      
      // write only if the partition has content
      if (currSlice->partArr[partition].bitstream->write_flag )
        writeUnit (currStream,partition);
    }           // partition loop
  }           // slice loop
  
  return 0;
}


void copy_params(void)
{
  enc_picture->frame_cropping_flag = active_sps->frame_cropping_flag;
  enc_picture->chroma_format_idc   = active_sps->chroma_format_idc;
  
  if (active_sps->frame_cropping_flag)
  {
    enc_picture->frame_cropping_rect_left_offset=active_sps->frame_cropping_rect_left_offset; 
    enc_picture->frame_cropping_rect_right_offset=active_sps->frame_cropping_rect_right_offset; 
    enc_picture->frame_cropping_rect_top_offset=active_sps->frame_cropping_rect_top_offset; 
    enc_picture->frame_cropping_rect_bottom_offset=active_sps->frame_cropping_rect_bottom_offset; 
  }
  else
  {
    enc_picture->frame_cropping_rect_left_offset=0; 
    enc_picture->frame_cropping_rect_right_offset=0; 
    enc_picture->frame_cropping_rect_top_offset=0; 
    enc_picture->frame_cropping_rect_bottom_offset=0; 
  }
}

/*!
************************************************************************
* \brief
*    Encodes a frame picture
************************************************************************
*/
void frame_picture (Picture *frame, int rd_pass)
{
  img->structure = FRAME;
  img->PicSizeInMbs = img->FrameSizeInMbs;

#if DUMP_MBTYPE
  img->curr_rd_pass = rd_pass;
  memset(statsRDPass[img->curr_rd_pass].bit_use_mb_type,    0, NUM_PIC_TYPE*sizeof(int64));
#endif

  if((rd_pass == 0) && (input->UseHPFilter == HPF_SIFO_FPO))
    img->filterParam = SIFO_FIRST_PASS_FPO;

  if (rd_pass == 2)
  {
    enc_frame_picture3  = alloc_storable_picture (img->structure, img->width, img->height, img->width_cr, img->height_cr);
    img->ThisPOC=enc_frame_picture3->poc=img->framepoc;
    enc_frame_picture3->frame_poc = img->framepoc;
    
    enc_frame_picture3->pic_num = img->frame_num;
    enc_frame_picture3->frame_num = img->frame_num;
    enc_frame_picture3->coded_frame = 1;
    enc_picture=enc_frame_picture3;
    copy_params();
  }
  else if (rd_pass == 1)
  {
    enc_frame_picture2  = alloc_storable_picture (img->structure, img->width, img->height, img->width_cr, img->height_cr);
    img->ThisPOC=enc_frame_picture2->poc=img->framepoc;
    enc_frame_picture2->frame_poc = img->framepoc;
    
    enc_frame_picture2->pic_num = img->frame_num;
    enc_frame_picture2->frame_num = img->frame_num;
    enc_frame_picture2->coded_frame = 1;
    enc_picture=enc_frame_picture2;
    copy_params();
  }
  else
  {
    enc_frame_picture  = alloc_storable_picture (img->structure, img->width, img->height, img->width_cr, img->height_cr);
    img->ThisPOC=enc_frame_picture->poc=img->framepoc;
    enc_frame_picture->frame_poc = img->framepoc;
    
    enc_frame_picture->pic_num = img->frame_num;
    enc_frame_picture->frame_num = img->frame_num;
    enc_frame_picture->coded_frame = 1;
    enc_picture=enc_frame_picture;
    copy_params();
  }
  
  
  stats->em_prev_bits_frm = 0;
  stats->em_prev_bits = &stats->em_prev_bits_frm;

  code_a_picture(frame);
  
  frame->bits_per_picture = 8 * ((((img->currentSlice)->partArr[0]).bitstream)->byte_pos);

    find_distortion ();      
    frame->distortion_y = snr->snr_y;
    frame->distortion_u = snr->snr_u;
    frame->distortion_v = snr->snr_v;
}

/*!
************************************************************************
* \brief
*    Frame Mode Buffer
************************************************************************
*/
static void frame_mode_buffer (int bit_frame, float snr_frame_y, float snr_frame_u, float snr_frame_v)
{
  put_buffer_frame ();
}


/*!
************************************************************************
* \brief
*    mmco initializations should go here
************************************************************************
*/
static void init_dec_ref_pic_marking_buffer(void)
{
  img->dec_ref_pic_marking_buffer=NULL;
}


/*!
************************************************************************
* \brief
*    Initializes the parameters for a new frame
************************************************************************
*/
static void init_frame (void)
{
  int i;
  int prevP_no, nextP_no;
  
  last_P_no = last_P_no_frm;
  
  img->current_mb_nr = 0;
  img->current_slice_nr = 0;
  stats->bit_slice = 0;
  
  img->mb_y = img->mb_x = 0;
  img->block_y = img->pix_y = img->pix_c_y = 0; 
  img->block_x = img->pix_x = img->block_c_x = img->pix_c_x = 0;
  
  // The 'slice_nr' of each macroblock is set to -1 here, to guarantee the correct encoding 
  // with FMO (if no FMO, encoding is correct without following assignment), 
  // for which MBs may not be encoded with scan order
  for(i=0;i< ((img->width/MB_BLOCK_SIZE)*(img->height/MB_BLOCK_SIZE));i++)
    img->mb_data[i].slice_nr=-1;
  
  if (img->b_frame_to_code == 0)
  {
    img->tr = start_tr_in_this_IGOP + IMG_NUMBER * (input->jumpd + 1);
    
    img->imgtr_last_P_frm = img->imgtr_next_P_frm;
    img->imgtr_next_P_frm = img->tr;
    
#ifdef _ADAPT_LAST_GROUP_
    if (input->last_frame && img->number + 1 == input->no_frames)
      img->tr = input->last_frame;
#endif
    
    if (IMG_NUMBER != 0 && input->successive_Bframe != 0)     // B pictures to encode
      nextP_tr_frm = img->tr;
    
    {
      if (img->type == I_SLICE)
      {
#ifdef _CHANGE_QP_
        //QP oscillation for secondary SP frames
        if ((input->qp2start > 0 && img->tr >= input->qp2start))
          img->qp = input->qp02;
        else
#endif 
          img->qp = input->qp0;   // set quant. parameter for I-frame
      }
      else
      {
#ifdef _CHANGE_QP_
        //QP oscillation for secondary SP frames
        if ((input->qp2start > 0 && img->tr >= input->qp2start))
          img->qp = input->qpN2 + (img->nal_reference_idc ? 0 : input->DispPQPOffset);
        else
#endif
          img->qp = input->qpN + (img->nal_reference_idc ? 0 : input->DispPQPOffset);
#if HIER_P
          img->qp += img->Q_Level + (img->Q_Level? input->HierP_QPoffset-1 : 0); 
#endif
      }
    }
    
    img->mb_y_intra = img->mb_y_upd;  //  img->mb_y_intra indicates which GOB to intra code for this frame
    
    if (input->intra_upd > 0) // if error robustness, find next GOB to update
    {
      img->mb_y_upd = (IMG_NUMBER / input->intra_upd) % (img->height / MB_BLOCK_SIZE);
    }
  }
  else
  {
    img->p_interval = input->jumpd + 1;
    prevP_no = start_tr_in_this_IGOP + (IMG_NUMBER - 1) * img->p_interval;
    nextP_no = start_tr_in_this_IGOP + (IMG_NUMBER) * img->p_interval;
    
#ifdef _ADAPT_LAST_GROUP_
    last_P_no[0] = prevP_no;
    for (i = 1; i < img->buf_cycle; i++)
      last_P_no[i] = last_P_no[i - 1] - img->p_interval;
    
    if (input->last_frame && img->number + 1 == input->no_frames)
    {
      nextP_no = input->last_frame;
      img->p_interval = nextP_no - prevP_no;
    }
#endif
    
    img->b_interval =
      ((double) (input->jumpd + 1) / (input->successive_Bframe + 1.0) );
    
    if (input->HierarchicalCoding == 3)
      img->b_interval = 1.0;
    
    if (input->HierarchicalCoding)
      img->tr = prevP_no + (int) (img->b_interval  * (double) (1 + gop_structure[img->b_frame_to_code - 1].display_no));      // from prev_P
    else      
      img->tr = prevP_no + (int) (img->b_interval * (double) img->b_frame_to_code);      // from prev_P
    
    
    if (img->tr >= nextP_no)
      img->tr = nextP_no - 1;

    if(input->HierarchicalCoding == 0)                  // without using rate control   
    {    
#ifdef _CHANGE_QP_
      //QP oscillation for secondary SP frames
      if ((input->qp2start > 0 && img->tr >= input->qp2start))
      {
        img->qp = input->qpB2;
      }
      else
#endif
      {
        img->qp = input->qpB;
      }
      
      if (img->nal_reference_idc)
      {
#ifdef _CHANGE_QP_
        //QP oscillation for secondary SP frames
        if ((input->qp2start > 0 && img->tr >= input->qp2start))
        {
          img->qp = Clip3(-img->bitdepth_luma_qp_scale,51,input->qpB2 + input->qpBRS2Offset);
        }
        else
#endif
        {
          img->qp = Clip3(-img->bitdepth_luma_qp_scale,51,input->qpB + input->qpBRSOffset);
        }
      }
    }
    else if (input->HierarchicalCoding !=0)  
    {
      // Note that _CHANGE_QP_ does not anymore work for gop_structure. Needs to be fixed
      img->qp =  gop_structure[img->b_frame_to_code - 1].slice_qp;
    }
  }
  
  UpdateSubseqInfo (img->layer);        // Tian Dong (Sept 2002)
  UpdateSceneInformation (0, 0, 0, -1); // JVT-D099, scene information SEI, nothing included by default
  
  //! Commented out by StW, needs fixing in SEI.h to keep the trace file clean
  //  PrepareAggregationSEIMessage ();
  
  img->no_output_of_prior_pics_flag = 0;
  img->long_term_reference_flag = 0;
  
  init_dec_ref_pic_marking_buffer();
}

#define Clip(min,max,val) (((val)<(min))?(min):(((val)>(max))?(max):(val)))


/*!
************************************************************************
* \brief
*    Generate Full Pel Representation
************************************************************************
*/
void GenerateFullPelRepresentation (pel_t ** Fourthpel,
                                    pel_t * Fullpel, int xsize,
                                    int ysize)
{
  int x, y, yy , y_pos;
  
  for (y = 0; y < ysize; y++)
  {
    yy = (y + IMG_PAD_SIZE)<<2;
    y_pos = y * xsize;
    for (x = 0; x < xsize; x++)
      PutPel_11 (Fullpel, y_pos + x, Fourthpel[yy][(x + IMG_PAD_SIZE)<<2]);
  }
}


/*!
************************************************************************
* \brief
*    Upsample 4 times, store them in out4x.  Color is simply copied
*
* \par Input:
*    srcy, srcu, srcv, out4y, out4u, out4v
*
* \par Side Effects_
*    Uses (writes) img4Y_tmp.  This should be moved to a static variable
*    in this module
************************************************************************/
void UnifiedOneForthPix (StorablePicture *s)
{
  int is;
  int i, j, j4;
  int ie2, je2, jj , jpad;
  int maxy = s->size_y + 2 * IMG_PAD_SIZE - 1;
  int ii, i1;
  imgpel **out4Y;
  imgpel  *ref11;
  imgpel  **imgY = s->imgY;
  int size_x_minus1 = s->size_x - 1;
  int size_y_minus1 = s->size_y - 1;
  int ypadded_size = s->size_y + 2 * IMG_PAD_SIZE;
  // don't upsample twice
  if (s->imgY_ups || s->imgY_11)
    return;
  
  s->imgY_11 = malloc ((s->size_x * s->size_y) * sizeof (imgpel));
  if (NULL == s->imgY_11)
    no_mem_exit("alloc_storable_picture: s->imgY_11");
  
  get_mem2Dpel (&(s->imgY_ups), (2*IMG_PAD_SIZE + s->size_y)*4, (2*IMG_PAD_SIZE + s->size_x)*4);
  
  if (input->WeightedPrediction || input->WeightedBiprediction || input->GenerateMultiplePPS)
  {
    s->imgY_11_w = malloc ((s->size_x * s->size_y) * sizeof (imgpel));
    if (NULL == s->imgY_11_w)
      no_mem_exit("alloc_storable_picture: s->imgY_11_w");
    get_mem2Dpel (&(s->imgY_ups_w), (2*IMG_PAD_SIZE + s->size_y)*4, (2*IMG_PAD_SIZE + s->size_x)*4);
  }
  out4Y = s->imgY_ups;
  ref11 = s->imgY_11;
  
  for (j = -IMG_PAD_SIZE; j < s->size_y + IMG_PAD_SIZE; j++)
  {
    jj = max (0, min (size_y_minus1, j));
    jpad = j + IMG_PAD_SIZE;
    for (i = -IMG_PAD_SIZE; i < s->size_x + IMG_PAD_SIZE; i++)
    {      
      is =
        (ONE_FOURTH_TAP[0][0] *
        (imgY[jj][max (0, min (size_x_minus1, i))] +
        imgY[jj][max (0, min (size_x_minus1, i + 1))]) +
        ONE_FOURTH_TAP[1][0] *
        (imgY[jj][max (0, min (size_x_minus1, i - 1))] +
        imgY[jj][max (0, min (size_x_minus1, i + 2))]) +
        ONE_FOURTH_TAP[2][0] *
        (imgY[jj][max (0, min (size_x_minus1, i - 2))] +
        imgY[jj][max (0, min (size_x_minus1, i + 3))]));
      
      img4Y_tmp[jpad][(i + IMG_PAD_SIZE) * 2] = imgY[jj][max (0, min (size_x_minus1, i))] << 10;    // 1/1 pix pos
      img4Y_tmp[jpad][(i + IMG_PAD_SIZE) * 2 + 1] = is << 5;  // 1/2 pix pos
    }
  }
  
  for (i = 0; i < (s->size_x + 2 * IMG_PAD_SIZE) * 2; i++)
  {
    ii = i << 1;
    // j = 0
    is =
      ( ONE_FOURTH_TAP[0][0] *(img4Y_tmp[0][i] + img4Y_tmp[1][i]) 
      + ONE_FOURTH_TAP[1][0] * (img4Y_tmp[0][i] + img4Y_tmp[2][i]) 
      + ONE_FOURTH_TAP[2][0] * (img4Y_tmp[0][i] + img4Y_tmp[3][i])) >> 5;
    out4Y[0][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (img4Y_tmp[0][i] + 512) >> 10);  // 1/2 pix
    out4Y[2][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (is + 512) >> 10);   // 1/2 pix
    // j = 1
    is =
      ( ONE_FOURTH_TAP[0][0] *(img4Y_tmp[1][i] + img4Y_tmp[2][i]) 
      + ONE_FOURTH_TAP[1][0] * (img4Y_tmp[0][i] + img4Y_tmp[3][i]) 
      + ONE_FOURTH_TAP[2][0] * (img4Y_tmp[0][i] + img4Y_tmp[4][i])) >> 5;
    out4Y[4][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (img4Y_tmp[1][i] + 512) >> 10);  // 1/2 pix
    out4Y[6][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (is + 512) >> 10);   // 1/2 pix
    
    for (j = 2; j < maxy - 2 ; j++)
    {
      j4 = j << 2;
      
      is =
        ( ONE_FOURTH_TAP[0][0] *(img4Y_tmp[j][i] + img4Y_tmp[j + 1][i]) 
        + ONE_FOURTH_TAP[1][0] * (img4Y_tmp[j - 1][i] + img4Y_tmp[j + 2][i]) 
        + ONE_FOURTH_TAP[2][0] * (img4Y_tmp[j - 2][i] + img4Y_tmp[j + 3][i])) >> 5;
      out4Y[j4    ][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (img4Y_tmp[j][i] + 512) >> 10);  // 1/2 pix
      out4Y[j4 + 2][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (is + 512) >> 10);   // 1/2 pix
    }
    
    for (j = maxy - 2; j < ypadded_size ; j++)
    {
      j4 = j << 2;
      
      is =
        ( ONE_FOURTH_TAP[0][0] *(img4Y_tmp[j][i] + img4Y_tmp[min (maxy, j + 1)][i]) 
        + ONE_FOURTH_TAP[1][0] * (img4Y_tmp[j - 1][i] + img4Y_tmp[min (maxy, j + 2)][i]) 
        + ONE_FOURTH_TAP[2][0] * (img4Y_tmp[j - 2][i] + img4Y_tmp[min (maxy, j + 3)][i])) >> 5;
      out4Y[j4    ][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (img4Y_tmp[j][i] + 512) >> 10);  // 1/2 pix
      out4Y[j4 + 2][ii] = (pel_t) Clip3 (0, img->max_imgpel_value, (is + 512) >> 10);   // 1/2 pix
    }
  }
  
  /* 1/4 pix */
  /* luma */
  ie2 = (s->size_x + 2 * IMG_PAD_SIZE - 1) * 4 + 2;
  je2 = (s->size_y + 2 * IMG_PAD_SIZE - 1) * 4 + 2;
  
  for (j = 0; j < je2 + 2; j += 2)
  {
    for (i = 0; i < ie2 - 2; i += 2)
    {
      /*  '-'  */
      //out4Y[j][i+1] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][i] + out4Y[j][i + 2] + 1) >> 1));
      out4Y[j][i+1] = (out4Y[j][i] + out4Y[j][i + 2] + 1) >> 1;
    }
    
    for (i = ie2 - 2; i < ie2 + 1; i += 2)
    {
      /*  '-'  */
      //out4Y[j][i+1] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][i] + out4Y[j][min (ie2, i + 2)] + 1) >> 1));
      out4Y[j][i+1] = (out4Y[j][i] + out4Y[j][min (ie2, i + 2)] + 1) >> 1;
    }
  }
  for (i = 0; i < ie2 + 2; i++)
  {
    ii = min (ie2, i + 1);
    i1 = i - 1;
    for (j = 0; j < je2 - 2; j += 2)
    {
      if ((i & 0x1) == 0)           /*  '|'  */
      {
        //if((out4Y[j][i] + out4Y[j + 2][i] + 1) >> 1 < 0 || (out4Y[j][i] + out4Y[j + 2][i] + 1) >> 1 > img->max_imgpel_value)
        //printf("Value %d\n",img->max_imgpel_value);
        //out4Y[j + 1][i] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][i] + out4Y[j + 2][i] + 1) >> 1));
        //out4Y[j + 1][i] = (pel_t) ((int) (out4Y[j][i] + out4Y[j + 2][i] + 1) >> 1);
        out4Y[j + 1][i] = (out4Y[j][i] + out4Y[j + 2][i] + 1) >> 1;
      }
      else if (((j & 0x3) == 0 && (i & 0x3) == 1) || ((j & 0x3) == 2 && (i & 0x3) == 3))           /*  '/'  */
      {
        //out4Y[j + 1][i] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][ii] + out4Y[j + 2][i1] + 1) >> 1));
        //out4Y[j + 1][i] = (pel_t) ((int) (out4Y[j][ii] + out4Y[j + 2][i1] + 1) >> 1);
        out4Y[j + 1][i] = (out4Y[j][ii] + out4Y[j + 2][i1] + 1) >> 1;
      }
      else           /*  '\'  */
      {
        //out4Y[j + 1][i] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][i1] + out4Y[j + 2][ii] + 1) >> 1));
        //out4Y[j + 1][i] = (pel_t) ((int) (out4Y[j][i1] + out4Y[j + 2][ii] + 1) >> 1);
        out4Y[j + 1][i] = (out4Y[j][i1] + out4Y[j + 2][ii] + 1) >> 1;
      }
    }
    
    for (j = je2 - 2; j < je2 + 1; j += 2)
    {
      if ((i & 0x1) == 0)           /*  '|'  */
      {
        //out4Y[j + 1][i] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][i] + out4Y[min (je2, j + 2)][i] + 1) >> 1));
        //out4Y[j + 1][i] = (pel_t) ((int) (out4Y[j][i] + out4Y[min (je2, j + 2)][i] + 1) >> 1);
        out4Y[j + 1][i] = (out4Y[j][i] + out4Y[min (je2, j + 2)][i] + 1) >> 1;
      }
      else if (((j & 0x3) == 0 && (i & 0x3) == 1) || ((j & 0x3) == 2 && (i & 0x3) == 3))           /*  '/'  */
      {
        //out4Y[j + 1][i] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][ii] + out4Y[min (je2, j + 2)][i1] + 1) >> 1));
        //out4Y[j + 1][i] = (pel_t) ((int) (out4Y[j][ii] + out4Y[min (je2, j + 2)][i1] + 1) >> 1);
        out4Y[j + 1][i] = (out4Y[j][ii] + out4Y[min (je2, j + 2)][i1] + 1) >> 1;
      }
      else           /*  '\'  */
      {
        //out4Y[j + 1][i] = (pel_t) (Clip3 (0, img->max_imgpel_value, (int) (out4Y[j][i1] + out4Y[min (je2, j + 2)][ii] + 1) >> 1));
        out4Y[j + 1][i] = (pel_t) ((int) (out4Y[j][i1] + out4Y[min (je2, j + 2)][ii] + 1) >> 1);
      }
    }
    
  }
  
  // Generate 1/1th pel representation (used for integer pel MV search)
  GenerateFullPelRepresentation (out4Y, ref11, s->size_x, s->size_y);    
}


short int sub_loc[15][2] =
{ //a     b     c     d     e     f     g     h     i     j     k     l     m     n     o
  {1,0},{2,0},{3,0},{0,1},{1,1},{2,1},{3,1},{0,2},{1,2},{2,2},{3,2},{0,3},{1,3},{2,3},{3,3}
  //0     1     2     3     4     5     6     7     8     9     10    11    12   13     14
};

short int acdl_idx[4]={0, 2, 3, 11};
short int egmo_idx[4]={4, 6, 12, 14};
short int fikn_idx[4]={5, 8, 10, 13};

short int acdl_f[4][2]  = {{0,0},{4,0},{0,0},{0,4}}; //contribution full pel location 
short int acdl_h[4][2]  = {{2,0},{2,0},{0,2},{0,2}}; //contributing half pel location {x,y}
short int egmo[4][2][2] =                            //contributing half pel location
{
  {{2,0},{0,2}},  //"e"
  {{2,0},{4,2}},  //"g"
  {{2,4},{0,2}},  //"m"
  {{2,4},{4,2}}   //"o"
}; 

short int fikn_q[4][2]  = {{2,2},{2,2},{2,2},{2,2}}; //contributing (1/2, 1/2) pel "j"
short int fikn_h[4][2]  = {{2,0},{0,2},{4,2},{2,4}}; //contributing half pel location

int get_mem2Dshortint(short int ***array2D, int rows, int columns)
{
  int i;
  
  if((*array2D      = (short int**)calloc(rows,        sizeof(short int*))) == NULL)
    no_mem_exit("get_mem2Dshortint: array2D");
  if(((*array2D)[0] = (short int* )calloc(rows*columns,sizeof(short int ))) == NULL)
    no_mem_exit("get_mem2Dshortint: array2D");
  
  for(i=1 ; i<rows ; i++)
    (*array2D)[i] =  (*array2D)[i-1] + columns  ;
  
  return rows*columns*sizeof(short int);
}
void free_mem2Dshortint(short int **array2D)
{
  if (array2D)
  {
    if (array2D[0]) 
      free (array2D[0]);
    else error ("free_mem2Dshortint: trying to free unused memory",100);
    
    free (array2D);
    
  } else
  {
    error ("free_mem2Dshortint: trying to free unused memory",100);
  }
}

void UnifiedOneForthPixWithHighPrecisionH264Filter_16 (StorablePicture *s)
{
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  int is, AF[3], full_neighbor, half1_neighbor, half2_neighbor, half_neighbor, temp;
#else
  short int is, AF[3], full_neighbor, half1_neighbor, half2_neighbor, half_neighbor, temp;
#endif
  int i, j, sub_i;
  int ii, jj;
  imgpel **out4Y;
  imgpel  *ref11;
  imgpel  **imgY = s->imgY;
  int pos_y1, pos_y2, pos_x1, pos_x2; 
  int x_sub, y_sub, sub_pos; 
  int x_max, y_max, tmp_x, tmp_y;
  
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  int **tmp4Y; 
#else
  short int **tmp4Y; 
#endif
  
  // don't upsample twice
  if (s->imgY_ups || s->imgY_11)
    return;
  
  s->imgY_11 = malloc ((s->size_x * s->size_y) * sizeof (imgpel));
  if (NULL == s->imgY_11)
    no_mem_exit("alloc_storable_picture: s->imgY_11");
  
  get_mem2Dpel (&(s->imgY_ups), (2*IMG_PAD_SIZE + s->size_y)*4, (2*IMG_PAD_SIZE + s->size_x)*4);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  get_mem2Dint (&tmp4Y, (2*IMG_PAD_SIZE + s->size_y)*4, (2*IMG_PAD_SIZE + s->size_x)*4);
#else
  get_mem2Dshortint (&tmp4Y, (2*IMG_PAD_SIZE + s->size_y)*4, (2*IMG_PAD_SIZE + s->size_x)*4);
#endif
  
  if (input->WeightedPrediction || input->WeightedBiprediction || input->GenerateMultiplePPS)
  {
    s->imgY_11_w = malloc ((s->size_x * s->size_y) * sizeof (imgpel));
    if (NULL == s->imgY_11_w)
      no_mem_exit("alloc_storable_picture: s->imgY_11_w");
    get_mem2Dpel (&(s->imgY_ups_w), (2*IMG_PAD_SIZE + s->size_y)*4, (2*IMG_PAD_SIZE + s->size_x)*4);
  }
  out4Y = s->imgY_ups;
  ref11 = s->imgY_11;
  
  // interpolate 1D half-pel first (but do not round them down yet)
  for(i=-4*IMG_PAD_SIZE; i < (IMG_PAD_SIZE+s->size_y)*4; i+=2)
    for(j=-4*IMG_PAD_SIZE; j < (IMG_PAD_SIZE+s->size_x)*4; j+=2)
    {
      x_sub = (j+4*IMG_PAD_SIZE)%4;       // x-sub-coordinate in a 4x4block
      y_sub = (i+4*IMG_PAD_SIZE)%4;       // y-sub-coordinate in a 4x4block
      sub_pos = x_sub + 4*y_sub;          // pos 1..15 in a 4x4 block
      is = 0;
      if(sub_pos)
      {
        if(sub_pos/4 == 0)    // horizontal 1/2 position 
        {
          pos_y1 = max(0, min(s->size_y-1, i/4 ) );
          
          for(jj = 0; jj < 3; jj++)
          {
            pos_x1 = max(0, min(s->size_x -1, j/4-2+jj ) );
            pos_x2 = max(0, min(s->size_x -1, j/4-2+(5-jj) ) );
            if(j<0 && pos_x2 > 0)
              pos_x2 -=1;
            is += ONE_FOURTH_TAP[2-jj][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y1][pos_x2]);
          }
          tmp4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] = is;
        }
        else if(sub_pos%4 == 0) // vertical 1/2 position
        {
          pos_x1 = max(0, min(s->size_x -1, j/4 ) );
          
          for(ii = 0; ii < 3; ii++)
          {
            pos_y1 = max(0, min(s->size_y-1, i/4-2+ii ) );
            pos_y2 = max(0, min(s->size_y-1, i/4-2+(5-ii) ) );
            if(i<0 && pos_y2 > 0)
              pos_y2 -=1;
            is += ONE_FOURTH_TAP[2-ii][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y2][pos_x1]);
          }
          tmp4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] = is;
        }
      }
      else
        tmp4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] =
        imgY[max(0,min(s->size_y-1,i/4))][max(0,min(s->size_x-1,j/4))];
    }
    
    // fill the (1/2,1/2) positions
    for(i=-4*IMG_PAD_SIZE+2; i < (IMG_PAD_SIZE+s->size_y)*4; i+=4)
      for(j=-4*IMG_PAD_SIZE+2; j < (IMG_PAD_SIZE+s->size_x)*4; j+=4)
      {
        is = 0;
        //AF[0]=A+F; AF[1]=B+E; AF[2]=C+D;
        pos_x1 = j;
        for(ii = 0; ii < 3; ii++)
        {
          pos_y1 = max(-4*IMG_PAD_SIZE, min((s->size_y+IMG_PAD_SIZE-1)*4, (i-2)-2*4+4*ii ) ); //max(0, min((s->size_y)*4, (i-2)-2*4+4*ii ) );
          pos_y2 = max(-4*IMG_PAD_SIZE, min((s->size_y+IMG_PAD_SIZE-1)*4, (i-2)-2*4+4*(5-ii) ) ); //max(0, min((s->size_y)*4, (i-2)-2*4+4*(5-ii) ) );        
          
          AF[ii] = tmp4Y[pos_y1+4*IMG_PAD_SIZE][pos_x1+4*IMG_PAD_SIZE]+tmp4Y[pos_y2+4*IMG_PAD_SIZE][pos_x1+4*IMG_PAD_SIZE];
        }
        
        AF[0] -= AF[1]; //(A+F)-(B+E)
        AF[0] >>= 2;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        if (img->BitDepthIncrease==0)
        {
          AF[2] = max(-2371, min(18640, AF[2]));
        }
#else
        AF[2] = max(-2371, min(18640, AF[2]));
#endif
        AF[1] = AF[2] - AF[1];   // (C+D)-(B+E)
        AF[0] += AF[1];
        AF[0] >>= 2;
        AF[0] += AF[2];
        
        tmp4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] = AF[0];
      }
      
      
      // interpolate quarter-pel and shift down half-pel/full-pel at the same time
      x_max = (s->size_x+2*IMG_PAD_SIZE)*4-1;//(s->size_x+2*IMG_PAD_SIZE)*4-2;
      y_max = (s->size_y+2*IMG_PAD_SIZE)*4-1;//(s->size_y+2*IMG_PAD_SIZE)*4-2;
      for(i=-4*IMG_PAD_SIZE; i < (IMG_PAD_SIZE+s->size_y)*4; i+=4)
        for(j=-4*IMG_PAD_SIZE; j < (IMG_PAD_SIZE+s->size_x)*4; j+=4)
        {
          int full_x, full_y, half1_x, half1_y, half2_x, half2_y, half_x, half_y, sub_x, sub_y;
          // full-pel
          temp = tmp4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE ];
          out4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] = 
            (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          
          // a, c, d, l
          for(sub_i=0; sub_i<4; sub_i++)
          {
            sub_x = sub_loc[acdl_idx[sub_i]][0];
            sub_y = sub_loc[acdl_idx[sub_i]][1];
            
            full_x =acdl_f[sub_i][0];
            full_y =acdl_f[sub_i][1];
            tmp_x = j+4*IMG_PAD_SIZE + full_x;
            if(tmp_x >x_max) 
              tmp_x = x_max -3;
            tmp_y = i+4*IMG_PAD_SIZE + full_y;
            if(tmp_y > y_max)
              tmp_y = y_max -3;
            
            full_neighbor = tmp4Y[tmp_y][tmp_x];
            
            half_x = acdl_h[sub_i][0];
            half_y = acdl_h[sub_i][1];
            tmp_x = j+4*IMG_PAD_SIZE + half_x;
            if(tmp_x >x_max)
              tmp_x = x_max -3;
            tmp_y = i+4*IMG_PAD_SIZE + half_y;
            if(tmp_y > y_max)
              tmp_y = y_max -3;
            
            half_neighbor = tmp4Y[tmp_y][tmp_x];
            
            temp = ((full_neighbor<<5) + half_neighbor + 32) >> 6;
            out4Y[i+4*IMG_PAD_SIZE +sub_y][j+4*IMG_PAD_SIZE+sub_x] = 
              (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          }                
          
          // h,
          temp = (tmp4Y[i+4*IMG_PAD_SIZE + 2][j+4*IMG_PAD_SIZE] + 16)>>5;
          out4Y[i+4*IMG_PAD_SIZE + 2][j+4*IMG_PAD_SIZE] = 
            (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          
          // b 
          temp = (tmp4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE + 2] + 16)>>5;
          out4Y[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE + 2] = 
            (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          
          // j
          temp = (tmp4Y[i+4*IMG_PAD_SIZE + 2][j+4*IMG_PAD_SIZE + 2] + 32)>>6;
          out4Y[i+4*IMG_PAD_SIZE + 2][j+4*IMG_PAD_SIZE + 2] = 
            (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          
          // e, g, m, o
          for(sub_i=0; sub_i<4; sub_i++)
          {
            sub_x = sub_loc[egmo_idx[sub_i]][0];
            sub_y = sub_loc[egmo_idx[sub_i]][1];
            
            half1_x =egmo[sub_i][0][0];
            half1_y =egmo[sub_i][0][1];
            tmp_x = j+4*IMG_PAD_SIZE + half1_x;
            tmp_y = i+4*IMG_PAD_SIZE + half1_y;
            if(tmp_x >x_max)
              tmp_x = x_max -3;
            if(tmp_y > y_max)
              tmp_y = y_max -3;
            
            half1_neighbor = tmp4Y[tmp_y][tmp_x];
            
            half2_x = egmo[sub_i][1][0];
            half2_y = egmo[sub_i][1][1];
            tmp_x = j+4*IMG_PAD_SIZE + half2_x;
            tmp_y = i+4*IMG_PAD_SIZE + half2_y;
            if(tmp_x >x_max)
              tmp_x = x_max -3;
            if(tmp_y > y_max)
              tmp_y = y_max -3;
            half2_neighbor = tmp4Y[tmp_y][tmp_x];
            
            temp = (half1_neighbor + half2_neighbor + 32) >> 6;
            
            out4Y[i+4*IMG_PAD_SIZE +sub_y][j+4*IMG_PAD_SIZE+sub_x] = 
              (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          }                
          
          // i, k,f, n            
          for(sub_i=0; sub_i<4; sub_i++)
          {
            sub_x = sub_loc[fikn_idx[sub_i]][0];
            sub_y = sub_loc[fikn_idx[sub_i]][1];
            
            half_x = fikn_h[sub_i][0];
            half_y = fikn_h[sub_i][1];
            tmp_x = j+4*IMG_PAD_SIZE + half_x;
            tmp_y = i+4*IMG_PAD_SIZE + half_y;
            if(tmp_x >x_max)
              tmp_x = x_max -3;
            if(tmp_y > y_max)
              tmp_y = y_max -3;
            
            half_neighbor = tmp4Y[tmp_y][tmp_x];
            
            temp = (half_neighbor + (tmp4Y[i+4*IMG_PAD_SIZE + 2][j+4*IMG_PAD_SIZE + 2]>>1) + 32) >> 6;
            
            out4Y[i+4*IMG_PAD_SIZE +sub_y][j+4*IMG_PAD_SIZE+sub_x] = 
              (pel_t)Clip3(0,(1<<img->bitdepth_luma)-1,temp); 
          }                
    }
    
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    free_mem2Dint(tmp4Y); 
#else
    free_mem2Dshortint(tmp4Y); 
#endif
    
    // Generate 1/1th pel representation (used for integer pel MV search)
    GenerateFullPelRepresentation (out4Y, ref11, s->size_x, s->size_y);    
}


#ifdef USE_NEW_OFFSET
void SetFrameOffset(void)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int x, y, xj, yi, temp, valOrg;
  int mvx=0,  mvy=0;
  int ref_frame=0;
  int subblock=0;
  int x_orig, y_orig;       
  int x_pos, y_pos;
  int out4Y_width  = (img->width+2*IMG_PAD_SIZE)*4-1;  
  int out4Y_height = (img->height+2*IMG_PAD_SIZE)*4-1;
  
  if(IS_INTRA(currMB)) //intra macroblocks are not used for calculation of the filter coeffs.
    return;
  
  x_orig = MB_BLOCK_SIZE*(img->block_x/4);
  y_orig = MB_BLOCK_SIZE*(img->block_y/4);
  for(subblock = 0; subblock < 16; subblock++)
  {
    x = x_orig+4*(subblock%4);
    y = y_orig+4*(subblock/4);
    mvx = enc_picture->mv[LIST_0][y/4][x/4][0];
    mvy = enc_picture->mv[LIST_0][y/4][x/4][1];
    ref_frame = enc_picture->ref_idx[LIST_0][y/4][x/4];
    if(ref_frame != -1)
    {
      for(yi = 0; yi < 4; yi++)
      {    //y
        for(xj = 0; xj < 4; xj++)
        {  //x
          valOrg   = imgY_org[y+yi][x+xj];

          y_pos = max(0,min(out4Y_height,4*(y+yi)+4*IMG_PAD_SIZE+mvy));
          x_pos = max(0,min(out4Y_width, 4*(x+xj)+4*IMG_PAD_SIZE+mvx));
          temp=listX[LIST_0][ref_frame]->imgY_ups[y_pos][x_pos];
          img->frameOffsetTotal[LIST_0][ref_frame]+=(valOrg-temp);
          img->frameOffsetCount[LIST_0][ref_frame]++;          

#ifdef USE_LMS     
          img->ref_sum       [LIST_0][ref_frame] += temp;
          img->ref_square_sum[LIST_0][ref_frame] += temp*temp;
          img->ref_cross     [LIST_0][ref_frame] += temp*valOrg;
          img->curr_sum      [LIST_0][ref_frame] += valOrg;
#endif
        }
      }
    }

    mvx = enc_picture->mv[LIST_1][y/4][x/4][0];
    mvy = enc_picture->mv[LIST_1][y/4][x/4][1];
    ref_frame = enc_picture->ref_idx[LIST_1][y/4][x/4];
    if(ref_frame != -1)
    {
      for(yi = 0; yi < 4; yi++)
      {    //y
        for(xj = 0; xj < 4; xj++)
        {  //x
          valOrg   = imgY_org[y+yi][x+xj];

          y_pos = max(0,min(out4Y_height,4*(y+yi)+4*IMG_PAD_SIZE+mvy));
          x_pos = max(0,min(out4Y_width, 4*(x+xj)+4*IMG_PAD_SIZE+mvx));
          temp=listX[LIST_1][ref_frame]->imgY_ups[y_pos][x_pos];
          img->frameOffsetTotal[LIST_1][ref_frame]+=(valOrg-temp);
          img->frameOffsetCount[LIST_1][ref_frame]++;          

#ifdef USE_LMS     
          img->ref_sum        [LIST_1][ref_frame] += temp;
          img->ref_square_sum [LIST_1][ref_frame] += temp*temp;
          img->ref_cross      [LIST_1][ref_frame] += temp*valOrg;
          img->curr_sum       [LIST_1][ref_frame] += valOrg;
#endif
        }
      }
    }
  }
}



void CalcFrameOffset(void)
{
  int frame, list, offset; 
  double temp;
  int numlists  = (img->type == B_SLICE) ? 2 : 1;
#ifdef USE_LMS     
  int default_weight = 1<<6;
  double ref_sum, ref_square_sum, curr_sum, ref_cross;
#endif
  
  for(list = 0; list < numlists; list++)
    for(frame = 0; frame < listXsize[list]; frame++)
    {
#ifdef USE_LMS     
      img->frameOffset_mse[list][frame] = 0;
      img->frameWeight_mse[list][frame] = default_weight;
#endif
      
      temp=(double)img->frameOffsetTotal[list][frame];
      
      if (img->frameOffsetCount[list][frame]>0)
      {
        offset=(int)(fabs(temp)/(double)img->frameOffsetCount[list][frame]+0.5);
        if (img->frameOffsetTotal[list][frame]>=0)
        {
          img->frameOffset[list][frame]=offset;
        }
        else
        {
          img->frameOffset[list][frame]=-offset;
        }
        
#ifdef USE_LMS     
        ref_sum = (double)img->ref_sum[list][frame] / img->frameOffsetCount[list][frame];
        ref_square_sum = (double)img->ref_square_sum[list][frame]/img->frameOffsetCount[list][frame];
        curr_sum = (double)img->curr_sum[list][frame]/img->frameOffsetCount[list][frame];
        ref_cross = (double)img->ref_cross[list][frame]/img->frameOffsetCount[list][frame];
        
        temp = ref_sum * ref_sum - ref_square_sum;
        if((int)temp != 0)
        {
          img->frameWeight_mse[list][frame] = (int)(default_weight*(ref_sum * curr_sum - ref_cross)/temp + 0.5);
          img->frameOffset_mse[list][frame] = (int)((-ref_square_sum * curr_sum + ref_sum * ref_cross)/temp + 0.5);
        }
#endif
      }
    }
}

void ResetFrameOffset(void)
{
  int frame, list;
  
  for(list = 0; list < 2; list++)
    for(frame = 0; frame < MAX_REFERENCE_PICTURES; frame++)
    {
      img->frameOffsetTotal[list][frame] = 0;
      img->frameOffsetCount[list][frame] = 0;
      
#ifdef USE_LMS     
      img->ref_sum[list][frame] = 0;
      img->ref_square_sum[list][frame] = 0;
      img->ref_cross[list][frame] = 0;
      img->curr_sum[list][frame] = 0;
#endif
    }
}

#endif


int64 compute_SSE(imgpel **imgRef, imgpel **imgSrc, int yRef, int ySrc, int xRef, int xSrc, int ySize, int xSize)
{
  int i, j;
  imgpel *lineRef, *lineSrc;
  int64 distortion = 0;

#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  int increase = img->BitDepthIncrease;
  if(increase)
  {
    for (j = 0; j < ySize; ++j)
    {
      lineRef = &imgRef[yRef+j][xRef];    
      lineSrc = &imgSrc[ySrc+j][xSrc];

      for (i = 0; i < xSize; ++i)
      {
        distortion += img->quad[Pel_12to8[*lineRef++] - Pel_12to8[*lineSrc++]];
      }
    }
  }
  else
#endif
  {
    for (j = 0; j < ySize; ++j)
    {
      lineRef = &imgRef[yRef+j][xRef];    
      lineSrc = &imgSrc[ySrc+j][xSrc];
      for (i = 0; i < xSize; ++i)
      {
        distortion += img->quad[( *lineRef++ - *lineSrc++ )];
      }
    }
  }
  return distortion;
}
/*!
************************************************************************
* \brief
*    Find SNR for all three components
************************************************************************
*/
static void find_snr (void)
{
  int i;
  int64 diff_y=0, diff_u=0, diff_v=0;
  int impix;
  int impix_cr;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  unsigned int max_pix_value_sqd = (img->BitDepthIncrease)? ((1<<input->InputBitDepth)-1)*((1<<input->InputBitDepth)-1)
    : img->max_imgpel_value * img->max_imgpel_value;
  unsigned int max_pix_value_sqd_uv = (img->BitDepthIncreaseChroma)? ((1<<input->InputBitDepth)-1)*((1<<input->InputBitDepth)-1)
    : img->max_imgpel_value_uv * img->max_imgpel_value_uv;
#else
  unsigned int max_pix_value_sqd = img->max_imgpel_value * img->max_imgpel_value;
  unsigned int max_pix_value_sqd_uv = img->max_imgpel_value_uv * img->max_imgpel_value_uv;
#endif
  
  //  Calculate  PSNR for Y, U and V.
  
  //     Luma.
  impix = input->img_height * input->img_width;
  impix_cr = input->img_height_cr * input->img_width_cr;
  
 
    imgY_org  = imgY_org_frm;
    imgUV_org = imgUV_org_frm;
    
    diff_y = 0;

    diff_y = compute_SSE(imgY_org    , enc_picture->imgY     , 0, 0, 0, 0, input->img_height   , input->img_width   );
    diff_u = compute_SSE(imgUV_org[0], enc_picture->imgUV[0], 0, 0, 0, 0, input->img_height_cr, input->img_width_cr);
    diff_v = compute_SSE(imgUV_org[1], enc_picture->imgUV[1], 0, 0, 0, 0, input->img_height_cr, input->img_width_cr);

  snr->sse_y = (float)diff_y;
  snr->sse_u = (float)diff_u;
  snr->sse_v = (float)diff_v;
  
#if ZEROSNR
  if (diff_y == 0)
    diff_y = 1;
  if (diff_u == 0)
    diff_u = 1;
  if (diff_v == 0)
    diff_v = 1; 
#endif
  
  //  Collecting SNR statistics
  if (diff_y != 0)
  {
    snr->snr_y = (float) (10 * log10 (max_pix_value_sqd * (double)((double) impix    / diff_y)));         // luma snr for current frame
    if (img->yuv_format != YUV400)
    {
      snr->snr_u = (float) (10 * log10 (max_pix_value_sqd_uv * (double)((double) impix_cr / diff_u)));   // u croma snr for current frame, 1/4 of luma samples
      snr->snr_v = (float) (10 * log10 (max_pix_value_sqd_uv * (double)((double) impix_cr / diff_v)));   // v croma snr for current frame, 1/4 of luma samples
    }
    else
    {
      snr->snr_u = 0.0;
      snr->snr_v = 0.0;
    }
  }
  
  
  if (img->number == 0)
  {
    snr->snr_y1 = snr->snr_y;  // keep luma snr for first frame
    snr->snr_u1 = snr->snr_u;  // keep croma u snr for first frame
    snr->snr_v1 = snr->snr_v;  // keep croma v snr for first frame
    snr->snr_ya = snr->snr_y1;
    snr->snr_ua = snr->snr_u1;
    snr->snr_va = snr->snr_v1;
    // sse stats
    snr->msse_y = snr->sse_y;
    snr->msse_u = snr->sse_u;
    snr->msse_v = snr->sse_v;
    for (i=0; i<5; i++)
    {
      snr->snr_yt[i] = 0.0;
      snr->snr_ut[i] = 0.0;
      snr->snr_vt[i] = 0.0;
    }
  }
  else
  {
    //int total_frames = img->number + frame_ctr[B_SLICE];
    int total_frames = snr->frame_ctr - 1;
    
    snr->snr_ya = (float) (snr->snr_ya * total_frames + snr->snr_y) / (total_frames + 1); // average snr luma for all frames inc. first
    snr->snr_ua = (float) (snr->snr_ua * total_frames + snr->snr_u) / (total_frames + 1); // average snr u croma for all frames inc. first
    snr->snr_va = (float) (snr->snr_va * total_frames + snr->snr_v) / (total_frames + 1); // average snr v croma for all frames inc. first
    snr->msse_y = (float) (snr->msse_y * total_frames + snr->sse_y) / (total_frames + 1); // average mse luma for all frames inc. first
    snr->msse_u = (float) (snr->msse_u * total_frames + snr->sse_u) / (total_frames + 1); // average mse u croma for all frames inc. first
    snr->msse_v = (float) (snr->msse_v * total_frames + snr->sse_v) / (total_frames + 1); // average mse v croma for all frames inc. first  
  }
  
  snr->snr_yt[img->type] = (float) (snr->snr_yt[img->type] * (frame_ctr[img->type] - 1) + snr->snr_y) / ( frame_ctr[img->type] );  // average luma snr for img->type coded frames
  snr->snr_ut[img->type] = (float) (snr->snr_ut[img->type] * (frame_ctr[img->type] - 1) + snr->snr_u) / ( frame_ctr[img->type] );  // average chroma u snr for img->type coded frames
  snr->snr_vt[img->type] = (float) (snr->snr_vt[img->type] * (frame_ctr[img->type] - 1) + snr->snr_v) / ( frame_ctr[img->type] );  // average chroma v snr for img->type coded frames
}

/*!
************************************************************************
* \brief
*    Find distortion for all three components
************************************************************************
*/

void find_distortion (void)
{
  int64 diff_y=0, diff_u=0, diff_v=0;
  int impix;
  
  //  Calculate  PSNR for Y, U and V.
  
  //     Luma.
  impix = input->img_height * input->img_width;
  
    imgY_org   = imgY_org_frm;
    imgUV_org  = imgUV_org_frm;
    
  diff_y  = compute_SSE(imgY_org    , enc_picture->imgY     , 0, 0, 0, 0, input->img_height   , input->img_width   );
  diff_u  = compute_SSE(imgUV_org[0], enc_picture->imgUV[0] , 0, 0, 0, 0, input->img_height_cr, input->img_width_cr);
  diff_v  = compute_SSE(imgUV_org[1], enc_picture->imgUV[1] , 0, 0, 0, 0, input->img_height_cr, input->img_width_cr);

  // Calculate real PSNR at find_snr_avg()
  snr->snr_y = (float) diff_y;
  snr->snr_u = (float) diff_u;
  snr->snr_v = (float) diff_v;
}


/*!
************************************************************************
* \brief
*    Just a placebo
************************************************************************
*/
Boolean dummy_slice_too_big (int bits_slice)
{
  return FALSE;
}

static void copy_motion_vectors_MB (void)
{
  int i,j,k,l;
  
  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
    {
      for (k = 0; k < img->max_num_references; k++)
      {
        for (l = 0; l < 9; l++)
        {
          img->all_mv[j][i][LIST_0][k][l][0] = rdopt->all_mv[j][i][LIST_0][k][l][0];
          img->all_mv[j][i][LIST_0][k][l][1] = rdopt->all_mv[j][i][LIST_0][k][l][1];
          
          img->all_mv[j][i][LIST_1][k][l][0] = rdopt->all_mv[j][i][LIST_1][k][l][0];
          img->all_mv[j][i][LIST_1][k][l][1] = rdopt->all_mv[j][i][LIST_1][k][l][1];
          
          img->pred_mv[j][i][LIST_0][k][l][0] = rdopt->pred_mv[j][i][LIST_0][k][l][0];
          img->pred_mv[j][i][LIST_0][k][l][1] = rdopt->pred_mv[j][i][LIST_0][k][l][1];
          
          img->pred_mv[j][i][LIST_1][k][l][0] = rdopt->pred_mv[j][i][LIST_1][k][l][0];
          img->pred_mv[j][i][LIST_1][k][l][1] = rdopt->pred_mv[j][i][LIST_1][k][l][1];
        }
      }
    }
  }
}


static void ReportNALNonVLCBits(int tmp_time, int me_time)
{
  
  //! Need to add type (i.e. SPS, PPS, SEI etc).
  if (input->Verbose != 0)
    printf ("%04d(NVB)%8d \n", frame_no, stats->bit_ctr_parametersets_n);
}
static void ReportFirstframe(int tmp_time,int me_time)
{
  
  if (input->Verbose == 1)
  {
    printf ("%04d(IDR)%8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", img->nal_reference_idc,img->rd_pass); 
  }
  else if (input->Verbose == 2)
  {
    printf ("%04d(IDR)%8d %1d %2d %7.3f %7.3f %7.3f %9d %7d    %3s %5d   %2d %2d  %d   %d\n",
      frame_no,(int) (stats->bit_ctr - stats->bit_ctr_n),0,
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", intras, img->num_ref_idx_l0_active, img->num_ref_idx_l1_active,img->rd_pass, img->nal_reference_idc);
  }
  
  stats->bit_ctr_I = stats->bit_ctr;
  stats->bit_ctr = 0;
  
}


static void ReportIntra(int tmp_time, int me_time)
{
  if (input->Verbose == 1)
  {
    if (img->currentPicture->idr_flag == 1)
      printf ("%04d(IDR)%8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", img->nal_reference_idc, img->rd_pass);
    else
      printf ("%04d(I)  %8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", img->nal_reference_idc, img->rd_pass);
  }
  else if (input->Verbose == 2)
  {
    if (img->currentPicture->idr_flag == 1)
      printf ("%04d(IDR)%8d %1d %2d %7.3f %7.3f %7.3f %9d %7d    %3s %5d   %2d %2d  %d   %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 0,
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", intras, img->num_ref_idx_l0_active, img->num_ref_idx_l1_active,img->rd_pass, img->nal_reference_idc); 
    else
      printf ("%04d(I)  %8d %1d %2d %7.3f %7.3f %7.3f %9d %7d    %3s %5d   %2d %2d  %d   %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 0,
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", intras, img->num_ref_idx_l0_active, img->num_ref_idx_l1_active,img->rd_pass, img->nal_reference_idc);
  }
}

static void ReportB(int tmp_time, int me_time)
{
  if (input->Verbose == 1)
  {
    printf ("%04d(B)  %8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n),
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time,me_time,
      "FRM", img->nal_reference_idc,img->rd_pass);
  }
  else if (input->Verbose == 2)
  {
    printf ("%04d(B)  %8d %1d %2d %7.3f %7.3f %7.3f %9d %7d    %3s %5d %1d %2d %2d  %d   %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), active_pps->weighted_bipred_idc,
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time,me_time,
      "FRM",intras,img->direct_spatial_mv_pred_flag, img->num_ref_idx_l0_active, img->num_ref_idx_l1_active,img->rd_pass, img->nal_reference_idc);
  }
}


static void ReportP(int tmp_time, int me_time)
{            
  if (input->Verbose == 1)
  {
#if HIER_P
    if(img->KeyP)
      printf ("%04d(P)  %8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
        frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 
        img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
        "FRM", img->nal_reference_idc,img->rd_pass);
    else
      printf ("%04d(p%1d) %8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
        frame_no,img->Q_Level+input->HierP_QPoffset-1, (int) (stats->bit_ctr - stats->bit_ctr_n), 
        img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
        "FRM", img->nal_reference_idc,img->rd_pass);
#else
      printf ("%04d(P)  %8d   %2d %7.3f %7.3f %7.3f %9d %7d    %3s    %d  %d\n",
        frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), 
        img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
        "FRM", img->nal_reference_idc,img->rd_pass);
#endif
  }
  else if (input->Verbose == 2)
  {
    printf ("%04d(P)  %8d %1d %2d %7.3f %7.3f %7.3f %9d %7d    %3s %5d   %2d %2d  %d   %d\n",
      frame_no, (int) (stats->bit_ctr - stats->bit_ctr_n), active_pps->weighted_pred_flag, 
      img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, me_time,
      "FRM", intras, img->num_ref_idx_l0_active, img->num_ref_idx_l1_active, img->rd_pass, img->nal_reference_idc);
  }
}

/*!
************************************************************************
* \brief
*    Padding of automatically added border for picture sizes that are not
*     multiples of macroblock/macroblock pair size
*
* \param org_size_x
*    original image horizontal size (luma)
* \param org_size_y
*    original image vertical size (luma)
* \param img_size_x
*    coded image horizontal size (luma)
* \param img_size_y
*    code image vertical size (luma)
* \param org_size_x_cr
*    original image horizontal size (chroma)
* \param org_size_y_cr
*    original image vertical size (chroma)
* \param img_size_x_cr
*    coded image horizontal size (chroma)
* \param img_size_y_cr
*    code image vertical size (chroma)
************************************************************************
*/
static void PaddAutoCropBorders (int org_size_x, int org_size_y, int img_size_x, int img_size_y,
                                 int org_size_x_cr, int org_size_y_cr, int img_size_x_cr, int img_size_y_cr)
{
  int x, y;
  
  //padding right border
  for (y=0; y<org_size_y; y++)
    for (x=org_size_x; x<img_size_x; x++)
      imgY_org_frm [y][x] = imgY_org_frm [y][x-1];
    
    //padding bottom border
    for (y=org_size_y; y<img_size_y; y++)
      for (x=0; x<img_size_x; x++)
        imgY_org_frm [y][x] = imgY_org_frm [y-1][x];
      
      
      if (img->yuv_format != YUV400)
      {
        //padding right border
        for (y=0; y<org_size_y_cr; y++)
          for (x=org_size_x_cr; x<img_size_x_cr; x++)
          {
            imgUV_org_frm [0][y][x] = imgUV_org_frm [0][y][x-1];
            imgUV_org_frm [1][y][x] = imgUV_org_frm [1][y][x-1];
          }
          
          //padding bottom border
          for (y=org_size_y_cr; y<img_size_y_cr; y++)
            for (x=0; x<img_size_x_cr; x++)
            {
              imgUV_org_frm [0][y][x] = imgUV_org_frm [0][y-1][x];
              imgUV_org_frm [1][y][x] = imgUV_org_frm [1][y-1][x];
            }
      }
}


/*!
************************************************************************
* \brief
*    Calculates the absolute frame number in the source file out
*    of various variables in img-> and input->
* \return
*    frame number in the file to be read
* \par side effects
*    global variable frame_no updated -- dunno, for what this one is necessary
************************************************************************
*/
static int CalculateFrameNumber(void)
{
  if (img->b_frame_to_code)
  {
    if (input->HierarchicalCoding)
      frame_no = start_tr_in_this_IGOP + (IMG_NUMBER - 1) * (input->jumpd + 1) + (int) (img->b_interval * (double) (1 + gop_structure[img->b_frame_to_code - 1].display_no));
    else      
      frame_no = start_tr_in_this_IGOP + (IMG_NUMBER - 1) * (input->jumpd + 1) + (int) (img->b_interval * (double) img->b_frame_to_code);

#if AdaptiveCADENCE
    if(input->HierB_AdaptiveCadence)
    {
    if (input->HierarchicalCoding)
      frame_no = start_tr_in_this_IGOP + (img->number - 1 + (int) (img->b_interval * (double) (1 + gop_structure[img->b_frame_to_code - 1].display_no)));
    else
      frame_no = start_tr_in_this_IGOP + (img->number - 1 + (int) (img->b_interval * (double) img->b_frame_to_code));
    }
#endif
  }
  else
  {
    frame_no = start_tr_in_this_IGOP + IMG_NUMBER * (input->jumpd + 1);
#if AdaptiveCADENCE
  if(input->HierB_AdaptiveCadence)
  {
    if(IMG_NUMBER)
      frame_no =  (img->number + input->jumpd); 
  }
#endif

#ifdef _ADAPT_LAST_GROUP_
    if (input->last_frame && img->number + 1 == input->no_frames)
      frame_no = input->last_frame;
#endif
  }
  
  return frame_no;
}


/*!
************************************************************************
* \brief
*    Convert file read buffer to source picture structure
************************************************************************
*/
void buf2img ( imgpel** imgX,           //!< Pointer to image plane
              unsigned char* buf,      //!< Buffer for file output
              int size_x,              //!< horizontal size of picture
              int size_y,              //!< vertical size of picture
              int symbol_size_in_bytes //!< number of bytes in file used for one pixel
              )
{
  int i,j;
  
  unsigned short tmp16, ui16;
  unsigned long  tmp32, ui32;
  
  if (symbol_size_in_bytes> sizeof(imgpel))
  {
    error ("Source picture has higher bit depth than imgpel data type. Please recompile with larger data type for imgpel.", 500);
  }
  
  if (( sizeof(char) == sizeof (imgpel)) && ( sizeof(char) == symbol_size_in_bytes))
  {
    // imgpel == pixel_in_file == 1 byte -> simple copy
    for(j=0;j<size_y;j++)
      memcpy(imgX[j], buf+j*size_x, size_x);
  }
  else
  {
    // sizeof (imgpel) > sizeof(char)
    if (testEndian())
    {
      // big endian
      switch (symbol_size_in_bytes)
      {
      case 1:
        {
          for(j=0;j<size_y;j++)
            for(i=0;i<size_x;i++)
            {
              imgX[j][i]= buf[i+j*size_x];
            }
            break;
        }
      case 2:
        {
          for(j=0;j<size_y;j++)
            for(i=0;i<size_x;i++)
            {
              memcpy(&tmp16, buf+((i+j*size_x)*2), 2);
              ui16  = (tmp16 >> 8) | ((tmp16&0xFF)<<8);
              imgX[j][i] = (imgpel) ui16;
            }
            break;
        }
      case 4:
        {
          for(j=0;j<size_y;j++)
            for(i=0;i<size_x;i++)
            {
              memcpy(&tmp32, buf+((i+j*size_x)*4), 4);
              ui32  = ((tmp32&0xFF00)<<8) | ((tmp32&0xFF)<<24) | ((tmp32&0xFF0000)>>8) | ((tmp32&0xFF000000)>>24);
              imgX[j][i] = (imgpel) ui32;
            }
        }
      default:
        {
          error ("reading only from formats of 8, 16 or 32 bit allowed on big endian architecture", 500);
          break;
        }
      }
    }
    else
    {
      // little endian
      for (j=0; j < size_y; j++)
        for (i=0; i < size_x; i++)
        {
          imgX[j][i]=0;
          memcpy(&(imgX[j][i]), buf +((i+j*size_x)*symbol_size_in_bytes), symbol_size_in_bytes);
        }
    }
  }
}


#ifdef  INTERNAL_BIT_DEPTH_INCREASE
/*!
************************************************************************
* \brief
*    Pixel Bit-Depth Increase
************************************************************************
*/
void ImgBitDepthIncrease ( imgpel** imgX,           //!< Pointer to image plane
                          int size_x,              //!< horizontal size of picture
                          int size_y,              //!< vertical size of picture
                          int input_bit_depth,
                          int bit_depth_increase
                          )
{
  int i,j;
  
  for(j=0;j<size_y;j++)
    for(i=0;i<size_x;i++)
    {
      imgX[j][i] = imgX[j][i]<<bit_depth_increase;
    }
    
}
#endif


/*!
************************************************************************
* \brief
*    Reads one new frame from file
*
* \param FrameNoInFile
*    Frame number in the source file
* \param HeaderSize
*    Number of bytes in the source file to be skipped
* \param xs
*    horizontal size of frame in pixels
* \param ys
*    vertical size of frame in pixels
* \param xs_cr
*    horizontal chroma size of frame in pixels
* \param ys_cr
*    vertical chroma size of frame in pixels
************************************************************************
*/
#ifdef ADAPTIVE_GOP
void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys, int xs_cr, int ys_cr)
#else
static void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys, int xs_cr, int ys_cr)
#endif
{
  unsigned int symbol_size_in_bytes = img->pic_unit_size_on_disk/8;
  
  const int imgsize_y = xs*ys;
  const int imgsize_uv = xs_cr*ys_cr;
  
  const int bytes_y = imgsize_y * symbol_size_in_bytes;
  const int bytes_uv = imgsize_uv * symbol_size_in_bytes;
  
  const int64 framesize_in_bytes = bytes_y + 2*bytes_uv;
  unsigned char *buf;
  
  Boolean rgb_input = (input->rgb_input_flag==1 && input->yuv_format==3);
  
  assert (p_in != -1);
  
  // KS: this buffer should actually be allocated only once, but this is still much faster than the previous version
  if (NULL==(buf = malloc (xs*ys * symbol_size_in_bytes))) no_mem_exit("ReadOneFrame: buf");
  
  // skip Header
  if (lseek (p_in, HeaderSize, SEEK_SET) != HeaderSize)
  {
    error ("ReadOneFrame: cannot fseek to (Header size) in p_in", -1);
  }
  
  // skip starting frames
  if (lseek (p_in, framesize_in_bytes * input->start_frame, SEEK_CUR) == -1) 
  {
    snprintf(errortext, ET_SIZE, "ReadOneFrame: cannot advance file pointer in p_in beyond frame %d\n", input->start_frame);
    error (errortext,-1);
  } 
  
  // seek to current frame 
  if (lseek (p_in, framesize_in_bytes * FrameNoInFile, SEEK_CUR) == -1) 
  {
    snprintf(errortext, ET_SIZE, "ReadOneFrame: cannot advance file pointer in p_in beyond frame %d\n", input->start_frame + FrameNoInFile);
    error (errortext,-1);
  }
  
  // Here we are at the correct position for the source frame in the file.  Now
  // read it.
  if (img->pic_unit_size_on_disk%8 == 0)
  {
    if(rgb_input)
      lseek (p_in, framesize_in_bytes/3, SEEK_CUR);
    
    if (read(p_in, buf, bytes_y) != bytes_y)
    {
      printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_y);
      report_stats_on_error();
      exit (-1);
    }
    
    buf2img(imgY_org_frm, buf, xs, ys, symbol_size_in_bytes);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    if (img->BitDepthIncrease)
    {
      ImgBitDepthIncrease(imgY_org_frm, xs, ys, input->BitDepthLuma, img->BitDepthIncrease);
    }
#endif
    
    if (img->yuv_format != YUV400)
    {
      if (read(p_in, buf, bytes_uv) != bytes_uv)
      {
        printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_y);
        report_stats_on_error();
        exit (-1);
      }
      buf2img(imgUV_org_frm[0], buf, xs_cr, ys_cr, symbol_size_in_bytes);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      if (img->BitDepthIncreaseChroma)
      {
        ImgBitDepthIncrease(imgUV_org_frm[0], xs_cr, ys_cr, input->BitDepthChroma, img->BitDepthIncreaseChroma);
      }
#endif
      
      if(rgb_input)
        lseek (p_in, -framesize_in_bytes, SEEK_CUR);
      
      if (read(p_in, buf, bytes_uv) != bytes_uv)
      {
        printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_y);
        report_stats_on_error();
        exit (-1);
      }
      buf2img(imgUV_org_frm[1], buf, xs_cr, ys_cr, symbol_size_in_bytes);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      if (img->BitDepthIncreaseChroma)
      {
        ImgBitDepthIncrease(imgUV_org_frm[1], xs_cr, ys_cr, input->BitDepthChroma, img->BitDepthIncreaseChroma);
      }
#endif
      
      if(rgb_input)
        lseek (p_in, framesize_in_bytes*2/3, SEEK_CUR);
    }
  }
  else
  {
    printf ("ReadOneFrame (NOT IMPLEMENTED): pic unit size on disk must be divided by 8");
    exit (-1);
  }
  free (buf);
  
}


/*!
************************************************************************
* \brief
*    point to frame coding variables 
************************************************************************
*/
static void put_buffer_frame(void)
{
  imgY_org  = imgY_org_frm;
  imgUV_org = imgUV_org_frm;  
}

/*!
************************************************************************
* \brief
*    Writes a NAL unit of a partition or slice
************************************************************************
*/

static void writeUnit(Bitstream* currStream, int partition)
{
  const int buffer_size = 500 + img->FrameSizeInMbs * (128 + 256 * img->bitdepth_luma + 512 * img->bitdepth_chroma);
  // KS: this is approx. max. allowed code picture size
  NALU_t *nalu;
  assert (currStream->bits_to_go == 8);
  nalu = AllocNALU(buffer_size);
  nalu->startcodeprefix_len = 2+(img->current_mb_nr == 0?ZEROBYTES_SHORTSTARTCODE+1:ZEROBYTES_SHORTSTARTCODE);
  //printf ("nalu->startcodeprefix_len %d\n", nalu->startcodeprefix_len);
  nalu->len = currStream->byte_pos +1;            // add one for the first byte of the NALU
  //printf ("nalu->len %d\n", nalu->len);
  memcpy (&nalu->buf[1], currStream->streamBuffer, nalu->len-1);
  if (img->currentPicture->idr_flag)
  {
    nalu->nal_unit_type = NALU_TYPE_IDR;
    nalu->nal_reference_idc = NALU_PRIORITY_HIGHEST;
  }
  else if (img->type == B_SLICE)
  {
    //different nal header for different partitions
    if(input->partition_mode == 0)
    {
      nalu->nal_unit_type = NALU_TYPE_SLICE;
    }
    else
    {
      nalu->nal_unit_type = NALU_TYPE_DPA +  partition;
    }
    
    if (img->nal_reference_idc !=0)
    {
      nalu->nal_reference_idc = NALU_PRIORITY_HIGH;
    }
    else
    {
      nalu->nal_reference_idc = NALU_PRIORITY_DISPOSABLE;
    }
  }
  else   // non-b frame, non IDR slice
  {
    //different nal header for different partitions
    if(input->partition_mode == 0)
    {
      nalu->nal_unit_type = NALU_TYPE_SLICE;
    }
    else
    {
      nalu->nal_unit_type = NALU_TYPE_DPA +  partition;
    }
    if (img->nal_reference_idc !=0)
    {
      nalu->nal_reference_idc = NALU_PRIORITY_HIGH;
    }
    else
    {
      nalu->nal_reference_idc = NALU_PRIORITY_DISPOSABLE;
    }
  }      
  nalu->forbidden_bit = 0;
  stats->bit_ctr += WriteNALU (nalu);
  
  FreeNALU(nalu);
}


static void rdPictureCoding(void)
{
  int second_qp = img->qp, rd_qp = img->qp;
  int previntras = intras;
  int prevtype = img->type;
  int skip_encode = 0;
  pic_parameter_set_rbsp_t *sec_pps;
  int pps_id;

  if (img->type!=I_SLICE && input->GenerateMultiplePPS)
  {
    if (img->type==P_SLICE)
    {
      if (test_wp_P_slice(0) == 1)
      {
        active_pps = PicParSet[1];
      }
      else
      {
        skip_encode = input->RDPSliceWeightOnly;
        active_pps = PicParSet[0];
        img->qp-=1;
      }
    }
    else
    {
        if (test_wp_B_slice(1) == 1)//Implicit wp based on POC...skip this pass if weights are (0.5,0.5)
          active_pps = PicParSet[2];		
        else
          skip_encode = 1;
    }
  }
  else        
  {
      img->qp-=1;
  }
  
  sec_pps = active_pps;
  second_qp = img->qp;
  
  if (skip_encode)
  {
    img->rd_pass = 0;
    enc_frame_picture2 = NULL;
  }
  else
  {
    frame_picture (frame_pic_2,1);
    img->rd_pass=picture_coding_decision(frame_pic_1, frame_pic_2, rd_qp);
  }
  //      update_rd_picture_contexts (img->rd_pass); 
  if (img->rd_pass==0)
  {
    enc_picture=enc_frame_picture;
    if (img->type!=I_SLICE && input->GenerateMultiplePPS)
    { 
      img->qp=rd_qp;
      active_pps = PicParSet[0];
    }
    else       
    {
      img->qp=rd_qp;
    }
    intras = previntras;
    frame_pic = frame_pic_1;
  }
  else
  {
    previntras = intras;
    frame_pic = frame_pic_2;
  }

  pps_id = active_pps->pic_parameter_set_id;
  // Final Encoding pass - note that we should 
  // make this more flexible in a later version.
  
  if (img->type!=I_SLICE && input->GenerateMultiplePPS)
  {
    skip_encode = 0;
    img->qp    = rd_qp;
      if (img->type == P_SLICE && input->GenerateMultiplePPS && (intras * 100 )/img->FrameSizeInMbs >=75)
      {
        img->type=I_SLICE;
        active_pps = PicParSet[0];
      }
      else if (img->type==P_SLICE)
      {
        if (test_wp_P_slice(1) == 1)
        {
          active_pps = PicParSet[1];
        }
        else
        {
          skip_encode = input->RDPSliceWeightOnly;
          active_pps = PicParSet[0];
          img->qp+=1;
        }
      }
      else
      {
        if (test_wp_B_slice(0) == 1)
        {
          active_pps = PicParSet[1];
        }
        else
        {
          skip_encode = input->RDBSliceWeightOnly;
          if(!input->UseRDO_Q)
            img->qp = rd_qp + (img->nal_reference_idc ? - 1 : 1);
        }      
      }
  }
  else 
  {
    active_pps = PicParSet[0];
    img->qp    = (rd_qp + 1);
  }
  
	if (skip_encode || ((img->type==B_SLICE) && (pps_id==active_pps->pic_parameter_set_id) && (input->UseRDO_Q)))
  {
    enc_frame_picture3 = NULL;
    img->qp = rd_qp;
  }
  else
  {
    frame_picture (frame_pic_3,2);
    if (img->rd_pass==0)
      img->rd_pass  = 2*picture_coding_decision(frame_pic_1, frame_pic_3, rd_qp);
    else
      img->rd_pass +=   picture_coding_decision(frame_pic_2, frame_pic_3, rd_qp);
  }
  
  //update_rd_picture_contexts (img->rd_pass); 
  if (img->rd_pass==0)
  {
    enc_picture = enc_frame_picture;
    img->type   = prevtype;
    active_pps  = PicParSet[0];
    img->qp     = rd_qp;
    intras      = previntras;
  }
  else if (img->rd_pass==1)
  {
    enc_picture = enc_frame_picture2;
    img->type   = prevtype;
    active_pps  = sec_pps;
    img->qp     = second_qp;
    intras      = previntras;
  }       
}

static void rdPictureCodingP(void)
{
  int second_qp = img->qp, rd_qp = img->qp;
  int previntras = intras;
  int prevtype = img->type;
  int skip_encode = 0;
  pic_parameter_set_rbsp_t *sec_pps;

  if (input->GenerateMultiplePPS)
  {
    if (test_wp_P_slice(0) == 1)            //explicit weights
    {
      active_pps = PicParSet[1];
    }
    else
    {
      skip_encode = 1;
      active_pps = PicParSet[0];
    }
  }

  sec_pps = active_pps;
  second_qp = img->qp;

  if (skip_encode)
  {
    img->rd_pass = 0;
    enc_frame_picture2 = NULL;
  }
  else
  {
    frame_picture (frame_pic_2,1);
    img->rd_pass=picture_coding_decision(frame_pic_1, frame_pic_2, rd_qp);
  }
  //      update_rd_picture_contexts (img->rd_pass); 
  if (img->rd_pass==0)
  {
    enc_picture=enc_frame_picture;
    if (input->GenerateMultiplePPS)
    { 
      active_pps = PicParSet[0];
    }
    img->qp=rd_qp;
    intras = previntras;
    frame_pic = frame_pic_1;
  }
  else
  {
    previntras = intras;
    frame_pic = frame_pic_2;
  }

  // Final Encoding pass - note that we should 
  // make this more flexible in a later version.

  if (input->GenerateMultiplePPS)
  {
    skip_encode = 0;
    img->qp    = rd_qp;

    if ((intras * 100 )/img->FrameSizeInMbs >=75)
    {
      img->type=I_SLICE;
      active_pps = PicParSet[0];
    }
    else
    {
      if (test_wp_P_slice(1) == 1)        //offsets
      {
        active_pps = PicParSet[1];
      }
      else
      {
        skip_encode = 1;
        active_pps = PicParSet[0];
      }
    }
  }

  if (skip_encode)
  {
    enc_frame_picture3 = NULL;
    img->qp = rd_qp;
  }
  else
  {
    frame_picture (frame_pic_3,2);
    if (img->rd_pass==0)
      img->rd_pass  = 2*picture_coding_decision(frame_pic_1, frame_pic_3, rd_qp);
    else
      img->rd_pass +=   picture_coding_decision(frame_pic_2, frame_pic_3, rd_qp);
  }

  //update_rd_picture_contexts (img->rd_pass); 
  if (img->rd_pass==0)
  {
    enc_picture = enc_frame_picture;
    img->type   = prevtype;
    active_pps  = PicParSet[0];
    img->qp     = rd_qp;
    intras      = previntras;
  }
  else if (img->rd_pass==1)
  {
    enc_picture = enc_frame_picture2;
    img->type   = prevtype;
    active_pps  = sec_pps;
    img->qp     = second_qp;
    intras      = previntras;
  }       
}

static void rdPictureCodingB(void)
{
//  int second_qp = img->qp;
  int rd_qp = img->qp;
  int previntras = intras;
  int skip_encode = 1;
  pic_parameter_set_rbsp_t *sec_pps;
  int explicitWeights;

  img->rd_pass = 0;
  sec_pps = active_pps;

  // First pass
  img->rdPassWeights=1;
  explicitWeights = test_wp_B_slice(0);    // check weights only
  if (explicitWeights != 64)
  {
    active_pps = PicParSet[1];
    frame_picture (frame_pic_2,1);
    img->rd_pass = picture_coding_decision(frame_pic_1, frame_pic_2, rd_qp);
    previntras=(img->rd_pass==1) ? intras : previntras;
  }
  
  sec_pps = active_pps;
  
  // Second pass 
  img->rdPassWeights=0;

  skip_encode = 0;
  if (test_wp_B_slice(1) == 1)            // check offset
  { 
    active_pps = PicParSet[1];
  }
  else
  {
    active_pps = PicParSet[0];
    skip_encode = 1;
  } 

  if (skip_encode)
  {
    enc_frame_picture3 = NULL;
    img->qp = rd_qp;
  }
  else
  {
    frame_picture (frame_pic_3,2);
    if (img->rd_pass==0)
      img->rd_pass  = 2*picture_coding_decision(frame_pic_1, frame_pic_3, rd_qp);
    else
      img->rd_pass +=   picture_coding_decision(frame_pic_2, frame_pic_3, rd_qp);
  }

  if (img->rd_pass==0)
  {
    enc_picture = enc_frame_picture;
    frame_pic   = frame_pic_1;
    active_pps  = sec_pps;
    img->qp     = rd_qp;
    intras      = previntras;
  }
  else if (img->rd_pass==1)
  {
    enc_picture = enc_frame_picture2;
    frame_pic   = frame_pic_2;
    active_pps  = sec_pps;
    img->qp     = rd_qp;
    intras      = previntras;
  }   

}




