
/*!
***********************************************************************
*  \mainpage
*     This is the H.264/AVC encoder reference software. For detailed documentation
*     see the comments in each file.
*
*  \author
*     The main contributors are listed in contributors.h
*
*  \version
*     JM 11.0 (FRExt)
*
*  \note
*     tags are used for document system "doxygen"
*     available at http://www.doxygen.org
*/
/*!
*  \file
*     lencod.c
*  \brief
*     H.264/AVC reference encoder project main()
*  \author
*   Main contributors (see contributors.h for copyright, address and affiliation details)
*   - Inge Lille-Langoy               <inge.lille-langoy@telenor.com>
*   - Rickard Sjoberg                 <rickard.sjoberg@era.ericsson.se>
*   - Stephan Wenger                  <stewe@cs.tu-berlin.de>
*   - Jani Lainema                    <jani.lainema@nokia.com>
*   - Byeong-Moon Jeon                <jeonbm@lge.com>
*   - Yoon-Seong Soh                  <yunsung@lge.com>
*   - Thomas Stockhammer              <stockhammer@ei.tum.de>
*   - Detlev Marpe                    <marpe@hhi.de>
*   - Guido Heising                   <heising@hhi.de>
*   - Valeri George                   <george@hhi.de>
*   - Karsten Suehring                <suehring@hhi.de>
*   - Alexis Michael Tourapis         <alexismt@ieee.org>
***********************************************************************
*/

#include "contributors.h"

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <sys/timeb.h>

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include "global.h"

#ifdef RDO_Q
#include "cabac.h"
#endif

#include "configfile.h"
#include "memalloc.h"
#include "intrarefresh.h"
#include "fmo.h"
#include "sei.h"
#include "parset.h"
#include "image.h"
#include "output.h"
#include "fast_me.h"
#include "simplified_fast_me.h"
#include "explicit_gop.h"
#include "epzs.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern MV_Competition mv_comp;
#endif
#ifdef ADAPTIVE_LOOP_FILTER
#include "adaptive_loop_filter.h"
#endif
#define VERSION "11.0"

#define JM          "11 (kta2.0)"
#define EXT_VERSION "(kta2.0)"



InputParameters inputs,      *input = &inputs;
ImageParameters images,      *img   = &images;
StatParameters  statistics,  *stats = &statistics;
SNRParameters   snrs,        *snr   = &snrs;
Decoders decoders, *decs=&decoders;

#if DUMP_MBTYPE 
StatParameters  statsRDPass[3];
#endif


#ifdef _LEAKYBUCKET_
unsigned long total_frame_buffer = 0;
#endif

#ifdef _ADAPT_LAST_GROUP_
int initial_Bframes = 0;
#endif

Boolean In2ndIGOP = FALSE;
int    start_frame_no_in_this_IGOP = 0;
int    start_tr_in_this_IGOP = 0;
int    FirstFrameIn2ndIGOP=0;
int    cabac_encoding = 0;
int    frame_statistic_start;
extern ColocatedParams *Co_located;

void Init_Motion_Search_Module (void);
void Clear_Motion_Search_Module (void);
void report_frame_statistic(void);
void SetLevelIndices(void);
void encode_sequence(void);

#ifdef ADAPTIVE_GOP
extern void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys, int xs_cr, int ys_cr);
void encode_sequence_adaptiveGOP(void);
short detectFADE(int step);
#endif

short**** geo_mv[GEO_PART_NUM][2];
short****** saved_all_mv;
short****** saved_pred_mv;
extern int skipped_geo;
extern char ****geo_mask[3];//[mb_ext_level]
extern char ****geo_4x4mask[3];//[mb_ext_level]
extern char ****geo_filter[3];//[mb_ext_level]
extern char ****geo_filterUV[3];//[mb_ext_level]
void  init_geo();
int get_mem_mv_geo (short ***** mv);
void free_mem_mv_geo (short****);
void free_mem4Dchar(char ****array4D, int idx, int frames );
int geo_stat[2][3];
extern char ****saved_partition_candidate;


void init_stats(void)
{
  stats->successive_Bframe = input->successive_Bframe;
  stats->bit_ctr_I = 0;
  stats->bit_ctr_P = 0;
  stats->bit_ctr_B = 0;
  snr->snr_y = 0.0;
  snr->snr_u = 0.0;
  snr->snr_v = 0.0;
  snr->snr_y1 = 0.0;
  snr->snr_u1 = 0.0;
  snr->snr_v1 = 0.0;
  snr->snr_ya = 0.0;
  snr->snr_ua = 0.0;
  snr->snr_va = 0.0;
  snr->sse_y  = 0.0;
  snr->sse_u  = 0.0;
  snr->sse_v  = 0.0;
  snr->msse_y = 0.0;
  snr->msse_u = 0.0;
  snr->msse_v = 0.0;
  snr->frame_ctr = 0;
}

#ifdef MB32X32
extern int mb32_stat[2][5];
extern int mb32_stat_currentpass[2][3][5];    //[img type][encoding pass][mode]
#ifdef MB64X64
extern int mb64_stat[2][5];
extern int mb64_stat_currentpass[2][3][5];
#endif
#endif


/*!
***********************************************************************
* \brief
*    Main function for encoder.
* \param argc
*    number of command line arguments
* \param argv
*    command line arguments
* \return
*    exit code
***********************************************************************
*/
int main(int argc,char **argv)
{
  int i, j, prob;
  giRDOpt_B8OnlyFlag = 0;
  p_dec = p_in = -1;
  p_stat = p_log = p_trace = NULL;
  
  frame_statistic_start = 1;
  
  Configure (argc, argv);

  Init_QMatrix();
  Init_QOffsetMatrix();
  
  AllocNalPayloadBuffer();
  
  init_poc();
  GenerateParameterSets();
  SetLevelIndices();
  
  init_img();
  init_geo();

  frame_pic_1= malloc_picture();

#ifdef MV_COMPETITION
  init_MV_Competition();
#endif

    frame_pic_2 = malloc_picture();
  if (input->RDPictureDecision)
  {
    frame_pic_3 = malloc_picture();
  }
  
  init_rdopt ();
  
  if (input->HierarchicalCoding )
  {
    init_gop_structure();
    if (input->HierarchicalCoding == 3)
    {
      interpret_gop_structure();
    }
    else
    {
      create_hierarchy();
    }
#if AdaptiveCADENCE
    if(input->HierB_AdaptiveCadence)
    {
      int i;
      int max_gopsize = input->HierarchicalCoding != 3 ? input->successive_Bframe  : input->jumpd;
      max_gopsize = max(12,max_gopsize);

      for(i=0; i<max_gopsize; i++)
      {
        gop_structure[i].slice_type         = 1;    //always B frame
        gop_structure[i].hierarchy_layer    = 0;
        gop_structure[i].hierarchyPocDelta  = 0;
      }
    }
#endif
  }  
  
  dpb.init_done = 0;
  init_dpb();
  init_out_buffer();
  init_stats();
  
  enc_picture = enc_frame_picture = NULL;
  
  init_global_buffers();
  
  create_context_memory ();
  
  Init_Motion_Search_Module ();
  
  information_init();
  
#ifdef RDO_Q
  if(input->UseRDO_Q)
    precalculate_unary_exp_golomb_level();
#endif
  
  // Init frame type counter. Only supports single slice per frame.
  memset(frame_ctr, 0, 5 * sizeof(int));
  img->last_valid_reference = 0;
  
  tot_time=0;                 // time for total encoding session
  
#ifdef _ADAPT_LAST_GROUP_
  if (input->last_frame > 0)
    input->no_frames = 1 + (input->last_frame + input->jumpd) / (input->jumpd + 1);
  initial_Bframes = input->successive_Bframe;
#endif
  
  PatchInputNoFrames();
  
  // Write sequence header (with parameter sets)
  stats->bit_ctr_parametersets = 0;
  stats->bit_slice = start_sequence();
  stats->bit_ctr_parametersets += stats->bit_ctr_parametersets_n;
  start_frame_no_in_this_IGOP = 0;


//============================================
#ifdef ADAPTIVE_GOP
  if(input->AdaptiveGOP && (input->successive_Bframe!=0 && input->HierarchicalCoding==3))
      encode_sequence_adaptiveGOP();
  else
#if HIER_P
    img->Q_Level = input->HierP_Level;
#endif
    encode_sequence();
#else 
  encode_sequence();  
#endif
//============================================

  // terminate sequence
  terminate_sequence();

  flush_dpb();
  
  close(p_in);
  if (-1!=p_dec)
    close(p_dec);
  if (p_trace)
    fclose(p_trace);
  
  Clear_Motion_Search_Module ();
  
  FmoUninit();
  
  if (input->HierarchicalCoding)
    clear_gop_structure ();
  
  // free structure for rd-opt. mode decision
  clear_rdopt ();

  // report everything
  report();

#if DUMP_MBTYPE
  {
    int64 total_bits  = (stats->bit_ctr_I + stats->bit_ctr_P + stats->bit_ctr_B + stats->bit_ctr_parametersets);
    int64 B_mbtype = stats->bit_use_mb_type[B_SLICE];
    double p1,p2;
    if(stats->bit_ctr_B)
    {
      p1 = B_mbtype*100.0/stats->bit_ctr_B;
      p2 = B_mbtype*100.0/total_bits;
      fprintf(stdout,"\nBits for B_mb_type = %" FORMAT_OFF_T  "  ",B_mbtype);
      fprintf(stdout,"(%5.2f%% of B-slice %5.2f%% of Total bits) \n", p1, p2);
    }
  }
#endif


  //printf("number of MB using geometry partition = %d\n", stat_geo);
  printf("number of MB did not try geometry partition = %d out of %d\n", skipped_geo,  
    (input->no_frames + (input->successive_Bframe*(input->no_frames-1)) - 1) * img->FrameSizeInMbs);


    printf("prob of mb using geometry partition\n");
    for(i=0; i<2; i++)
    {
      printf("slice type = %d\n", i);
      for(j=0; j<3; j++)
      {
        if(i==0)
          prob = (int)(geo_stat[i][j]*100.0/((input->no_frames-1)*img->FrameSizeInMbs/(1<<(2*j))));
        else
          prob = (int)(geo_stat[i][j]*100.0/(input->successive_Bframe*(input->no_frames-1)*img->FrameSizeInMbs/(1<<(2*j))));
        printf("  Mb_ext=%d [geo=%d(prob %d%%)]\n", j, geo_stat[i][j], (int)prob);
      }
      printf("\n");
    }
  if(img->geo_scheme != 0)
  {
    int i;
    for(i=0; i<3; i++)
    {
      free_mem4Dchar(geo_mask[i], ((16<<i)>>1), (1<<THETA_BITS));
      free_mem4Dchar(geo_4x4mask[i], ((16<<i)>>1), (1<<THETA_BITS));
      free_mem4Dchar(geo_filter[i], ((16<<i)>>1), (1<<THETA_BITS));
      free_mem4Dchar(geo_filterUV[i], ((16<<i)>>1), (1<<THETA_BITS));
    }

    free_mem4Dchar(saved_partition_candidate, img->FrameSizeInMbs, 3);
  }

  free_picture (frame_pic_1);
  if (input->RDPictureDecision)
  {
    free_picture (frame_pic_2);
    free_picture (frame_pic_3);
  }

#ifdef MV_COMPETITION
  close_MV_Competition();
#endif

  free_dpb();
  free_colocated(Co_located);
  uninit_out_buffer();
  
  free_global_buffers();
 
  // free image mem
  free_img ();
  free_context_memory ();
  FreeNalPayloadBuffer();
  FreeParameterSets();
  return 0;                         //encode JM73_FME version
}






/*!
***********************************************************************
* \brief
*    Terminates and reports statistics on error.
* 
***********************************************************************
*/
void report_stats_on_error(void)
{
  input->no_frames=img->number;
  terminate_sequence();
  
  flush_dpb();
  
  close(p_in);
  if (-1!=p_dec)
    close(p_dec);
  
  if (p_trace)
    fclose(p_trace);
  
  Clear_Motion_Search_Module ();
  
  FmoUninit();
  
  if (input->HierarchicalCoding)
    clear_gop_structure ();
  
  // free structure for rd-opt. mode decision
  clear_rdopt ();

  if (input->ReportFrameStats)
    report_frame_statistic();
  
  // report everything
  report();
  
  free_picture (frame_pic_1);
  free_dpb();
  free_colocated(Co_located);
  uninit_out_buffer();
  
  free_global_buffers();
  
  // free image mem
  free_img ();
  free_context_memory ();
  FreeNalPayloadBuffer();
  FreeParameterSets();
}

/*!
***********************************************************************
* \brief
*    Initializes the POC structure with appropriate parameters.
* 
***********************************************************************
*/
void init_poc()
{
  //the following should probably go in sequence parameters
  // frame poc's increase by 2
  
  img->pic_order_cnt_type=input->pic_order_cnt_type;
  
  img->delta_pic_order_always_zero_flag=0;
  img->num_ref_frames_in_pic_order_cnt_cycle= 1;
  
  if (input->BRefPictures == 1)
  {
    img->offset_for_non_ref_pic  =   0;
    img->offset_for_ref_frame[0] =   2;
  }
  else
  {
    img->offset_for_non_ref_pic  =  -2*(input->successive_Bframe);
    img->offset_for_ref_frame[0] =   2*(input->successive_Bframe+1);
  }
  
  img->pic_order_present_flag=0;
  img->delta_pic_order_cnt_bottom = 0;

}


/*!
***********************************************************************
* \brief
*    Initializes the img->nz_coeff
* \par Input:
*    none
* \par  Output:
*    none
* \ side effects
*    sets omg->nz_coef[][][][] to -1
***********************************************************************
*/
void CAVLC_init(void)
{
  unsigned int i, k, l;
  
  for (i=0;i < img->PicSizeInMbs; i++)
    for (k=0;k<4;k++)
      for (l=0;l < (4 + (unsigned int)img->num_blk8x8_uv);l++)
        img->nz_coeff[i][k][l]=0;
}


/*!
***********************************************************************
* \brief
*    Initializes the Image structure with appropriate parameters.
* \par Input:
*    Input Parameters struct inp_par *inp
* \par  Output:
*    Image Parameters struct img_par *img
***********************************************************************
*/
void init_img()
{
  int i,j;
  int byte_abs_range;
  
  static int mb_width_cr[4] = {0,8, 8,16};
  static int mb_height_cr[4]= {0,8,16,16};
  
  img->yuv_format = input->yuv_format;
  
  //pel bitdepth init
  img->bitdepth_luma   = input->BitDepthLuma;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  img->BitDepthIncrease = input->BitDepthLuma - input->InputBitDepth;
#endif
  
  img->bitdepth_luma_qp_scale   = 6*(img->bitdepth_luma   - 8);
  img->bitdepth_lambda_scale    = 2*(img->bitdepth_luma   - 8);
  
  img->dc_pred_value_luma   = 1<<(img->bitdepth_luma - 1);
  img->max_imgpel_value = (1<<img->bitdepth_luma) - 1;
  
  if (img->yuv_format != YUV400)  
  {
    img->bitdepth_chroma      = input->BitDepthChroma;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    img->BitDepthIncreaseChroma = input->BitDepthChroma - input->InputBitDepth;
#endif
    img->dc_pred_value_chroma = 1<<(img->bitdepth_chroma - 1);
    img->max_imgpel_value_uv  = (1<<img->bitdepth_chroma) - 1;
    img->num_blk8x8_uv        = (1<<img->yuv_format)&(~(0x1));
    img->num_cdc_coeff        = img->num_blk8x8_uv<<1;
    img->mb_cr_size_x         = (img->yuv_format==YUV420 || img->yuv_format==YUV422)? 8:16;
    img->mb_cr_size_y         = (img->yuv_format==YUV444 || img->yuv_format==YUV422)? 16:8;
    
    img->bitdepth_chroma_qp_scale = 6*(img->bitdepth_chroma - 8);
    
    img->chroma_qp_offset[0] = active_pps->cb_qp_index_offset;
    img->chroma_qp_offset[1] = active_pps->cr_qp_index_offset;
  }
  else
  {
    img->bitdepth_chroma     = 0;
    img->max_imgpel_value_uv = 0;
    img->num_blk8x8_uv       = 0;
    img->num_cdc_coeff       = 0;
    img->mb_cr_size_x        = 0;
    img->mb_cr_size_y        = 0;
    
    img->bitdepth_chroma_qp_scale = 0;
    img->bitdepth_chroma_qp_scale = 0;
    
    img->chroma_qp_offset[0] = 0;
    img->chroma_qp_offset[1] = 0;
  }
  
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  if (input->InputBitDepth != input->BitDepthLuma ||input->InputBitDepth != input->BitDepthLuma)
  {
    img->pic_unit_size_on_disk = (input->InputBitDepth > 8)? 16:8;
  }
  else
  {
    if((img->bitdepth_luma > img->bitdepth_chroma) || img->yuv_format == YUV400)
      img->pic_unit_size_on_disk = (img->bitdepth_luma > 8)? 16:8;
    else
      img->pic_unit_size_on_disk = (img->bitdepth_chroma > 8)? 16:8;
  }
#else
  if((img->bitdepth_luma > img->bitdepth_chroma) || img->yuv_format == YUV400)
    img->pic_unit_size_on_disk = (img->bitdepth_luma > 8)? 16:8;
  else
    img->pic_unit_size_on_disk = (img->bitdepth_chroma > 8)? 16:8;
#endif
  
  img->num_ref_frames = active_sps->num_ref_frames;
  img->max_num_references   = active_sps->num_ref_frames ;
  
  img->buf_cycle = input->num_ref_frames;
  
  img->DeblockCall = 0;
  
  //  img->framerate=INIT_FRAME_RATE;   // The basic frame rate (of the original sequence)
  img->framerate=(float) input->FrameRate;   // The basic frame rate (of the original sequence)
  
  
  get_mem_mv (&(img->pred_mv));
  get_mem_mv (&(img->all_mv));
#ifdef MB32X32
  get_mem_mv (&(img->all_mv32));
#endif
  get_mem_mv(&(saved_all_mv));
  get_mem_mv(&(saved_pred_mv));
  for(i=0; i<GEO_PART_NUM; i++)
    for(j=0; j<2; j++)
      get_mem_mv_geo(&(geo_mv[i][j]));
#ifdef RDO_Q
  if(input->UseRDO_Q)
  {
#ifdef MB32X32
    int i;
    for(i=0; i<(1<<(2*MAX_MB_EXT_LEVEL)); i++)
    {
      get_mem_mv (&(img->all_mv8[i]));
      get_mem_mv (&(img->all_mv4[i]));

      get_mem_mv (&(img->bipred_mv1_8[i]));
      get_mem_mv (&(img->bipred_mv2_8[i]));
      get_mem_mv (&(img->bipred_mv1_4[i]));
      get_mem_mv (&(img->bipred_mv2_4[i]));
    }
    for(i=0; i<5; i++)
    {
      get_mem_mv (&(img->all32_mv8[i]));
      get_mem_mv (&(img->all32_mv4[i]));
      get_mem_mv (&(img->bipred32_mv1_8[i]));
      get_mem_mv (&(img->bipred32_mv2_8[i]));
      get_mem_mv (&(img->bipred32_mv1_4[i]));
      get_mem_mv (&(img->bipred32_mv2_4[i]));
    }
#else
    get_mem_mv (&(img->all_mv8));
    get_mem_mv (&(img->all_mv4));
#endif
  }
#endif
  
  if (input->BiPredMotionEstimation)
  {
    get_mem_mv (&(img->bipred_mv1));
    get_mem_mv (&(img->bipred_mv2));
  }
  
  get_mem_ACcoeff (&(img->cofAC));
  get_mem_DCcoeff (&(img->cofDC));
  
  if(img->max_imgpel_value > img->max_imgpel_value_uv)
    byte_abs_range = (img->max_imgpel_value + 1) * 2;
  else
    byte_abs_range = (img->max_imgpel_value_uv + 1) * 2;
  
  if ((img->quad = (int*)calloc (byte_abs_range, sizeof(int))) == NULL)
    no_mem_exit ("init_img: img->quad");
  img->quad+=byte_abs_range/2;
  for (i=0; i < byte_abs_range/2; ++i)
  {
    img->quad[i]=img->quad[-i]=i*i;
  }
  
  img->width    = (input->img_width+img->auto_crop_right);
  img->height   = (input->img_height+img->auto_crop_bottom);
  if (img->yuv_format != YUV400)
  {
    img->width_cr = img->width/(16/mb_width_cr[img->yuv_format]);
    img->height_cr= img->height/(16/mb_height_cr[img->yuv_format]);
    
    input->img_width_cr  = input->img_width/(16/mb_width_cr[img->yuv_format]);
    input->img_height_cr = input->img_height/(16/mb_height_cr[img->yuv_format]);
  }
  else
  {
    img->width_cr = 0;
    img->height_cr= 0;
    
    input->img_width_cr  = 0;
    input->img_height_cr = 0;
  }
  img->height_cr_frame = img->height_cr;
  
  img->PicWidthInMbs    = (input->img_width+img->auto_crop_right)/MB_BLOCK_SIZE;
  img->FrameHeightInMbs = (input->img_height+img->auto_crop_bottom)/MB_BLOCK_SIZE;
  img->FrameSizeInMbs   = img->PicWidthInMbs * img->FrameHeightInMbs;
  
  img->PicHeightInMapUnits = img->FrameHeightInMbs;
  
  if(((img->mb_data) = (Macroblock *) calloc(img->FrameSizeInMbs,sizeof(Macroblock))) == NULL)
    no_mem_exit("init_img: img->mb_data");
  
  if(input->UseConstrainedIntraPred)
  {
    if(((img->intra_block) = (int*)calloc(img->FrameSizeInMbs,sizeof(int))) == NULL)
      no_mem_exit("init_img: img->intra_block");
  }
  
  get_mem2D((byte***)&(img->ipredmode), img->height/BLOCK_SIZE, img->width/BLOCK_SIZE);        //need two extra rows at right and bottom
  get_mem2D((byte***)&(img->ipredmode8x8), img->height/BLOCK_SIZE, img->width/BLOCK_SIZE);     // help storage for ipredmode 8x8, inserted by YV
  
  get_mem2D((byte***)&(rddata_top_frame_mb.ipredmode), img->height/BLOCK_SIZE, img->width/BLOCK_SIZE);
  
  // CAVLC mem
  get_mem3Dint(&(img->nz_coeff), img->FrameSizeInMbs, 4, 4+img->num_blk8x8_uv);
  
  
  get_mem2Ddb_offset(&(img->lambda_md), 10, 52 + img->bitdepth_luma_qp_scale,img->bitdepth_luma_qp_scale);
  get_mem2Ddb_offset(&(img->lambda_me), 10, 52 + img->bitdepth_luma_qp_scale,img->bitdepth_luma_qp_scale);
  get_mem2Dint_offset(&(img->lambda_mf), 10, 52 + img->bitdepth_luma_qp_scale,img->bitdepth_luma_qp_scale);
  //get_mem2Ddouble(&(img->lambda_md), 10, 52 + img->bitdepth_luma_qp_scale);
  
  CAVLC_init();
  
  for (i=0; i < img->width/BLOCK_SIZE; i++)
    for (j=0; j < img->height/BLOCK_SIZE; j++)
    {
      img->ipredmode[j][i]=-1;
      img->ipredmode8x8[j][i]=-1;
    }
    
    img->mb_y_upd=0;
    
 
    InitSEIMessages();  // Tian Dong (Sept 2002)
    
    // Initialize filtering parameters. If sending parameters, the offsets are 
    // multiplied by 2 since inputs are taken in "div 2" format.
    // If not sending parameters, all fields are cleared 
    if (input->LFSendParameters)
    {
      input->LFAlphaC0Offset <<= 1;
      input->LFBetaOffset <<= 1;
    }
    else
    {
      input->LFDisableIdc = 0;
      input->LFAlphaC0Offset = 0;
      input->LFBetaOffset = 0;
    }
#ifdef EIGHTH_PEL
	img->mv_res=input->mv_res;
#endif
}

/*!
***********************************************************************
* \brief
*    Free the Image structures
* \par Input:
*    Image Parameters struct img_par *img
***********************************************************************
*/
void free_img ()
{
  int i, j;
  CloseSEIMessages(); // Tian Dong (Sept 2002)
  free_mem_mv (img->pred_mv);
  free_mem_mv (img->all_mv);
#ifdef MB32X32
  free_mem_mv (img->all_mv32);
#endif
  for(i=0; i<GEO_PART_NUM; i++)
    for(j=0; j<2; j++)
      free_mem_mv_geo(geo_mv[i][j]);

    
  free_mem_mv(saved_all_mv);
  free_mem_mv(saved_pred_mv);


#ifdef RDO_Q
  if(input->UseRDO_Q)
  {
#ifdef MB32X32
    int i;
    for(i=0; i<(1<<(2*MAX_MB_EXT_LEVEL)); i++)
    {
      free_mem_mv (img->all_mv8[i]);
      free_mem_mv (img->all_mv4[i]);
      free_mem_mv (img->bipred_mv1_8[i]);
      free_mem_mv (img->bipred_mv2_8[i]);
      free_mem_mv (img->bipred_mv1_4[i]);
      free_mem_mv (img->bipred_mv2_4[i]);
    }
    for(i=0; i<5; i++)
    {
      free_mem_mv (img->all32_mv8[i]);
      free_mem_mv (img->all32_mv4[i]);
      free_mem_mv (img->bipred32_mv1_8[i]);
      free_mem_mv (img->bipred32_mv2_8[i]);
      free_mem_mv (img->bipred32_mv1_4[i]);
      free_mem_mv (img->bipred32_mv2_4[i]);
    }
#else
    free_mem_mv (img->all_mv8);
    free_mem_mv (img->all_mv4);

#endif
  }
#endif
  if (input->BiPredMotionEstimation)
  {
    free_mem_mv (img->bipred_mv1);
    free_mem_mv (img->bipred_mv2);
  }
  
  free_mem_ACcoeff (img->cofAC);
  free_mem_DCcoeff (img->cofDC);
 
  if(img->max_imgpel_value > img->max_imgpel_value_uv)
    free (img->quad-(img->max_imgpel_value + 1));
  else
    free (img->quad-(img->max_imgpel_value_uv + 1));
}


/*!
************************************************************************
* \brief
*    Allocates the picture structure along with its dependent
*    data structures
* \return
*    Pointer to a Picture
************************************************************************
*/

Picture *malloc_picture()
{
  Picture *pic;
  if ((pic = calloc (1, sizeof (Picture))) == NULL) no_mem_exit ("malloc_picture: Picture structure");
  //! Note: slice structures are allocated as needed in code_a_picture
  return pic;
}

/*!
************************************************************************
* \brief
*    Frees a picture
* \param
*    pic: POinter to a Picture to be freed
************************************************************************
*/


void free_picture(Picture *pic)
{
  if (pic != NULL)
  {
    free_slice_list(pic);
    free (pic);
  }
}


/*!
************************************************************************
* \brief
*    Reports frame statistical data to a stats file
************************************************************************
*/
void report_frame_statistic()
{
  FILE *p_stat_frm = NULL;
  static int   last_mode_use[NUM_PIC_TYPE][MAXMODE];
  static int   last_b8_mode_0[NUM_PIC_TYPE][2];
  static int   last_mode_chroma_use[4];
  static int64   last_bit_ctr_n = 0;
  int i;
  char name[20];
  int bitcounter;
  
#ifndef WIN32
  time_t now;
  struct tm *l_time;
  char string[1000];
#else
  char timebuf[128];
#endif
  
  
  // write to log file
  if ((p_stat_frm=fopen("stat_frame.dat","r"))==0)                      // check if file exist
  {
    if ((p_stat_frm=fopen("stat_frame.dat","a"))==NULL)            // append new statistic at the end
    {
      snprintf(errortext, ET_SIZE, "Error open file %s  \n","stat_frame.dat.dat");
      error(errortext, 500);
    }
    else                                            // Create header for new log file
    {
      fprintf(p_stat_frm," --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
      fprintf(p_stat_frm,"|            Encoder statistics. This file is generated during first encoding session, new sessions will be appended                                                                                                                                                                                                                                                                                                                                                              |\n");
      fprintf(p_stat_frm," --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
    }
  }
  else
  {
    fclose (p_stat_frm);
    if ((p_stat_frm=fopen("stat_frame.dat","a"))==NULL)            // File exist,just open for appending
    {
      snprintf(errortext, ET_SIZE, "Error open file %s  \n","stat_frame.dat.dat");
      error(errortext, 500);
    }
  }
  
  if (frame_statistic_start)
  {
    fprintf(p_stat_frm,"|  ver        | Date  | Time  |    Sequence        |Frm | QP |P/MbInt|   Bits   |  SNRY  |  SNRU  |  SNRV  |  I4  |  I8  | I16  | IC0  | IC1  | IC2  | IC3  | PI4  | PI8  | PI16 |  P0  |  P1  |  P2  |  P3  | P1*8*| P1*4*| P2*8*| P2*4*| P3*8*| P3*4*|  P8  | P8:4 | P4*8*| P4*4*| P8:5 | P8:6 | P8:7 | BI4  | BI8  | BI16 |  B0  |  B1  |  B2  |  B3  | B0*8*| B0*4*| B1*8*| B1*4*| B2*8*| B2*4*| B3*8*| B3*4*|  B8  | B8:0 |B80*8*|B80*4*| B8:4 | B4*8*| B4*4*| B8:5 | B8:6 | B8:7 |\n");
    fprintf(p_stat_frm," --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
  } 
  
  //report
  fprintf(p_stat_frm,"|%4s/%s", VERSION, EXT_VERSION); 
  
#ifdef WIN32
  _strdate( timebuf );
  fprintf(p_stat_frm,"| %1.5s |",timebuf );
  
  _strtime( timebuf);
  fprintf(p_stat_frm," % 1.5s |",timebuf);
#else
  now = time ((time_t *) NULL); // Get the system time and put it into 'now' as 'calender time'
  time (&now);
  l_time = localtime (&now);
  strftime (string, sizeof string, "%d-%b-%Y", l_time);
  fprintf(p_stat_frm,"| %1.5s |",string );
  
  strftime (string, sizeof string, "%H:%M:%S", l_time);
  fprintf(p_stat_frm," %1.5s |",string );
#endif
  
  for (i=0;i<20;i++)
    name[i]=input->infile[i+max(0,(int) (strlen(input->infile)-20))]; // write last part of path, max 20 chars
  fprintf(p_stat_frm,"%20.20s|",name);
  
  fprintf(p_stat_frm,"%3d |",frame_no);
  
  fprintf(p_stat_frm,"%3d |",img->qp);
  
  fprintf(p_stat_frm,"  %d/%d  |",input->PicInterlace, input->MbInterlace);
  
  
  if (img->frame_num == 0)
  {
    bitcounter = (int) stats->bit_ctr_I;
  }
  else
  {
    bitcounter = (int) (stats->bit_ctr_n - last_bit_ctr_n);
    last_bit_ctr_n = stats->bit_ctr_n;
  }
  
  //report bitrate
  fprintf(p_stat_frm, " %9d|", bitcounter);
  
  //report snr's
  fprintf(p_stat_frm, " %2.4f| %2.4f| %2.4f|", snr->snr_y, snr->snr_u, snr->snr_v);
  
  //report modes
  //I-Modes
  //fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[I_SLICE][I4MB] - last_mode_use[I_SLICE][I4MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[I_SLICE][I4MB] - last_mode_use[I_SLICE][I4MB]);  
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[I_SLICE][I8MB] - last_mode_use[I_SLICE][I8MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[I_SLICE][I16MB] - last_mode_use[I_SLICE][I16MB]);
  
  //chroma intra mode
  fprintf(p_stat_frm, " %5d|",stats->intra_chroma_mode[0] - last_mode_chroma_use[0]);
  fprintf(p_stat_frm, " %5d|",stats->intra_chroma_mode[1] - last_mode_chroma_use[1]);
  fprintf(p_stat_frm, " %5d|",stats->intra_chroma_mode[2] - last_mode_chroma_use[2]);
  fprintf(p_stat_frm, " %5d|",stats->intra_chroma_mode[3] - last_mode_chroma_use[3]);
  
  //P-Modes
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][I4MB] - last_mode_use[P_SLICE][I4MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][I8MB] - last_mode_use[P_SLICE][I8MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][I16MB] - last_mode_use[P_SLICE][I16MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][0   ] - last_mode_use[P_SLICE][0   ]);
  
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][1   ] - last_mode_use[P_SLICE][1   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][2   ] - last_mode_use[P_SLICE][2   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][3   ] - last_mode_use[P_SLICE][3   ]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[0][1]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[0][1]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[0][2]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[0][2]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[0][3]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[0][3]);
  
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][P8x8] - last_mode_use[P_SLICE][P8x8]);
  //  fprintf(p_stat_frm, " %5d|",stats->b8_mode_0_use[P_SLICE][0]  - last_b8_mode_0[P_SLICE ][0]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][4   ] - last_mode_use[P_SLICE][4   ]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[0][4]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[0][4]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][5   ] - last_mode_use[P_SLICE][5   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][6   ] - last_mode_use[P_SLICE][6   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[P_SLICE][7   ] - last_mode_use[P_SLICE][7   ]);
  
  //B-Modes
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][I4MB] - last_mode_use[B_SLICE][I4MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][I8MB] - last_mode_use[B_SLICE][I8MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][I16MB] - last_mode_use[B_SLICE][I16MB]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][0   ] - last_mode_use[B_SLICE][0   ]);
  /*
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][1   ] - last_mode_use[B_SLICE][1   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][2   ] - last_mode_use[B_SLICE][2   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][3   ] - last_mode_use[B_SLICE][3   ]);
  */
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][1   ] - last_mode_use[B_SLICE][1   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][2   ] - last_mode_use[B_SLICE][2   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][3   ] - last_mode_use[B_SLICE][3   ]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[1][0]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[1][0]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[1][1]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[1][1]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[1][2]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[1][2]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[1][3]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[1][3]);
  
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][P8x8] - last_mode_use[B_SLICE][P8x8]);
  fprintf(p_stat_frm, " %5d|",(stats->b8_mode_0_use [B_SLICE][0]+stats->b8_mode_0_use [B_SLICE][1]) - (last_b8_mode_0[B_SLICE][0]+last_b8_mode_0[B_SLICE][1]));
  fprintf(p_stat_frm, " %5d|",stats->b8_mode_0_use [B_SLICE][1] - last_b8_mode_0[B_SLICE][1]);
  fprintf(p_stat_frm, " %5d|",stats->b8_mode_0_use [B_SLICE][0] - last_b8_mode_0[B_SLICE][0]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][4   ] - last_mode_use[B_SLICE][4   ]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_8x8[1][4]);
  fprintf(p_stat_frm, " %5d|",stats->mode_use_transform_4x4[1][4]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][5   ] - last_mode_use[B_SLICE][5   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][6   ] - last_mode_use[B_SLICE][6   ]);
  fprintf(p_stat_frm, " %5" FORMAT_OFF_T  "|",stats->mode_use[B_SLICE][7   ] - last_mode_use[B_SLICE][7   ]);
  
  fprintf(p_stat_frm, "\n");
  
  //save the last results
  memcpy(last_mode_use[I_SLICE],stats->mode_use[I_SLICE], MAXMODE *  sizeof(int));
  memcpy(last_mode_use[P_SLICE],stats->mode_use[P_SLICE], MAXMODE *  sizeof(int));
  memcpy(last_mode_use[B_SLICE],stats->mode_use[B_SLICE], MAXMODE *  sizeof(int));
  memset(stats->mode_use_transform_8x8,0, 2 * MAXMODE *  sizeof(int));
  memset(stats->mode_use_transform_4x4,0, 2 * MAXMODE *  sizeof(int));
  memcpy(last_b8_mode_0[B_SLICE],stats->b8_mode_0_use[B_SLICE], 2 *  sizeof(int));
  memcpy(last_mode_chroma_use,stats->intra_chroma_mode, 4 *  sizeof(int));
  
  frame_statistic_start = 0;
  fclose(p_stat_frm);
}


/*!
************************************************************************
* \brief
*    Reports the gathered information to appropriate outputs
* \par Input:
*    struct inp_par *inp,                                            \n
*    struct img_par *img,                                            \n
*    struct stat_par *stats,                                          \n
*    struct stat_par *stats                                           \n
*
* \par Output:
*    None
************************************************************************
*/
void report()
{
  int64 bit_use[NUM_PIC_TYPE][2] ;
  int i,j;
  char name[30];
  int64 total_bits;
  float frame_rate;
  double mean_motion_info_bit_use[2] = {0.0};
  
#ifndef WIN32
  time_t now;
  struct tm *l_time;
  char string[1000];
#else
  char timebuf[128];
#endif

#ifdef _LEAKYBUCKET_
  fprintf(stdout,"-----------------------------------------------------------------------------------\n");
  printf(" Total Frames:  %ld (%d) \n", total_frame_buffer, input->no_frames);
#endif
#if AdaptiveCADENCE
  printf(" Total Frames: I(%d)   P(%d)   B(%d)\n", frame_ctr[I_SLICE],frame_ctr[P_SLICE],frame_ctr[B_SLICE]);
  stats->successive_Bframe = input->successive_Bframe;
#endif

  bit_use[I_SLICE][0] = frame_ctr[I_SLICE];
  bit_use[P_SLICE][0] = max(frame_ctr[P_SLICE],1);
  bit_use[B_SLICE][0] = frame_ctr[B_SLICE];
  
  //  Accumulate bit usage for inter and intra frames
  for (j=0;j<NUM_PIC_TYPE;j++)
  {
    bit_use[j][1] = 0;
  }
  
  for (j=0;j<NUM_PIC_TYPE;j++)
  {
    for(i=0; i<MAXMODE; i++)
      bit_use[j][1] += stats->bit_use_mode    [j][i]; 
    
    bit_use[j][1]+=stats->bit_use_header[j];
    bit_use[j][1]+=stats->bit_use_mb_type[j];
    bit_use[j][1]+=stats->tmp_bit_use_cbp[j];
    bit_use[j][1]+=stats->bit_use_coeffY[j];
    bit_use[j][1]+=stats->bit_use_coeffC[j];
    bit_use[j][1]+=stats->bit_use_delta_quant[j];
    bit_use[j][1]+=stats->bit_use_stuffingBits[j];
  }
  
  frame_rate = (img->framerate *(float)(stats->successive_Bframe + 1)) / (float) (input->jumpd+1);
  
  //! Currently adding NVB bits on P rate. Maybe additional stats info should be created instead and added in log file
  stats->bitrate_I = (stats->bit_ctr_I)*(frame_rate)/(float) (input->no_frames + frame_ctr[B_SLICE]);
  stats->bitrate_P = (stats->bit_ctr_P)*(frame_rate)/(float) (input->no_frames + frame_ctr[B_SLICE]);
  stats->bitrate_B = (stats->bit_ctr_B)*(frame_rate)/(float) (input->no_frames + frame_ctr[B_SLICE]);    
  
  fprintf(stdout,"-----------------------------------------------------------------------------------\n");
  if (input->Verbose != 0)
  {
    fprintf(stdout,  " Freq. for encoded bitstream       : %1.0f\n",img->framerate/(float)(input->jumpd+1));
    switch (input->hadamard)
    {
    case 2:
      fprintf(stdout," Hadamard transform                : Used for QPel\n");
      break;
    case 1:
      fprintf(stdout," Hadamard transform                : Used\n");    
      break;
    default:
      fprintf(stdout," Hadamard transform                : Not used\n");
      break;
    }
    
    fprintf(stdout,  " Image format                      : %dx%d\n",input->img_width,input->img_height);
    
    if(input->intra_upd)
      fprintf(stdout," Error robustness                  : On\n");
    else
      fprintf(stdout," Error robustness                  : Off\n");
    fprintf(stdout,  " Search range                      : %d\n",input->search_range);
    
    fprintf(stdout,  " Total number of references        : %d\n",input->num_ref_frames);
    fprintf(stdout,  " References for P slices           : %d\n",input->P_List0_refs? input->P_List0_refs:input->num_ref_frames);
    if(stats->successive_Bframe != 0)
    {
      fprintf(stdout," List0 references for B slices     : %d\n",input->B_List0_refs? input->B_List0_refs:input->num_ref_frames);
      fprintf(stdout," List1 references for B slices     : %d\n",input->B_List1_refs? input->B_List1_refs:input->num_ref_frames);
    }
    
    fprintf(stdout,  " Total encoding time for the seq.  : %.3f sec (%.2f fps)\n",tot_time*0.001, 1000.0*(input->no_frames + frame_ctr[B_SLICE])/tot_time);
    fprintf(stdout,  " Total ME time for sequence        : %.3f sec \n",me_tot_time*0.001);
    
    // B pictures
    fprintf(stdout,  " Sequence type                     :" );
    
    if(stats->successive_Bframe>0 && input->HierarchicalCoding) 
    {
      fprintf(stdout, " Hierarchy (QP: I %d, P %d, B %d) \n",
        input->qp0, input->qpN, input->qpB);
    }
    else if(stats->successive_Bframe>0) 
    {
      char seqtype[80];
      int i,j;
      
      strcpy (seqtype,"I");
      
      for (j=0;j<2;j++)
      {
        for (i=0;i<stats->successive_Bframe;i++)
        {
          if (input->BRefPictures)
            strncat(seqtype,"-RB",max (0, (int) (79-strlen(seqtype))));
          else
            strncat(seqtype,"-B",max (0, (int) (79-strlen(seqtype))));
        }
        strncat(seqtype,"-P",max (0, (int) (79-strlen(seqtype))));
      }
      if (input->BRefPictures)
        fprintf(stdout, " %s (QP: I %d, P %d, RB %d) \n", seqtype,input->qp0, input->qpN, Clip3(0,51,input->qpB+input->qpBRSOffset));
      else
        fprintf(stdout, " %s (QP: I %d, P %d, B %d) \n", seqtype,input->qp0, input->qpN, input->qpB);
    }
    else if(stats->successive_Bframe==0 && input->sp_periodicity==0) fprintf(stdout, " IPPP (QP: I %d, P %d) \n",   input->qp0, input->qpN);
    
    else fprintf(stdout, " I-P-P-SP-P (QP: I %d, P %d, SP (%d, %d)) \n",  input->qp0, input->qpN, input->qpsp, input->qpsp_pred);
    
    // report on entropy coding  method
    if (input->symbol_mode == UVLC)
      fprintf(stdout," Entropy coding method             : CAVLC\n");
    else
      fprintf(stdout," Entropy coding method             : CABAC\n");
    
    fprintf(stdout,  " Profile/Level IDC                 : (%d,%d)\n",input->ProfileIDC,input->LevelIDC);
    
    if (input->FMEnable==1)
      fprintf(stdout,  " Motion Estimation Scheme          : HEX\n");    
    else if (input->FMEnable==2)
      fprintf(stdout,  " Motion Estimation Scheme          : SHEX\n");
    else if (input->FMEnable == 3)
    {
      fprintf(stdout,  " Motion Estimation Scheme          : EPZS\n");
      EPZSOutputStats(stdout, 0);
    }
    else
      fprintf(stdout,  " Motion Estimation Scheme          : Full Search\n");
    
    
    
#ifdef _FULL_SEARCH_RANGE_
    if (input->full_search == 2)
      fprintf(stdout," Search range restrictions         : none\n");
    else if (input->full_search == 1)
      fprintf(stdout," Search range restrictions         : older reference frames\n");
    else
      fprintf(stdout," Search range restrictions         : smaller blocks and older reference frames\n");
#endif
    
    if (input->rdopt)
      fprintf(stdout," RD-optimized mode decision        : used\n");
    else
      fprintf(stdout," RD-optimized mode decision        : not used\n");
    
    switch(input->partition_mode)
    {
    case PAR_DP_1:
      fprintf(stdout," Data Partitioning Mode            : 1 partition \n");
      break;
    case PAR_DP_3:
      fprintf(stdout," Data Partitioning Mode            : 3 partitions \n");
      break;
    default:
      fprintf(stdout," Data Partitioning Mode            : not supported\n");
      break;
    }
    
    switch(input->of_mode)
    {
    case PAR_OF_ANNEXB:
      fprintf(stdout," Output File Format                : H.264 Bit Stream File Format \n");
      break;
    default:
      fprintf(stdout," Output File Format                : not supported\n");
      break;
    }
    // Residue Color Transform
      fprintf(stdout," Residue Color Transform           : not supported\n");
  }
  
  fprintf(stdout,"-----------------------------------------------------------------------------------\n");
  fprintf(stdout,  " Total encoding time for the seq.  : %.3f sec (%.2f fps)\n",tot_time*0.001, 1000.0*(input->no_frames + frame_ctr[B_SLICE])/tot_time);
  fprintf(stdout,  " Total ME time for sequence        : %.3f sec \n",me_tot_time*0.001);
  fprintf(stdout,"------------------ Average data all frames  ---------------------------------------\n");
  if (input->Verbose != 0)
  {
    int  impix = input->img_height * input->img_width;
    int  impix_cr = input->img_height_cr * input->img_width_cr;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    unsigned int max_pix_value_sqd = (img->BitDepthIncrease)? ((1<<input->InputBitDepth)-1)*((1<<input->InputBitDepth)-1)
      : img->max_imgpel_value * img->max_imgpel_value;
    unsigned int max_pix_value_sqd_uv = (img->BitDepthIncreaseChroma)? ((1<<input->InputBitDepth)-1)*((1<<input->InputBitDepth)-1)
      : img->max_imgpel_value_uv * img->max_imgpel_value_uv;
#else
    unsigned int max_pix_value_sqd = img->max_imgpel_value * img->max_imgpel_value;
    unsigned int max_pix_value_sqd_uv = img->max_imgpel_value_uv * img->max_imgpel_value_uv;
#endif
    float csnr_y = (float) (10 * log10 (max_pix_value_sqd * 
      (double)((double) impix / (snr->msse_y == 0.0? 1.0 : snr->msse_y))));  
    float csnr_u = (float) (10 * log10 (max_pix_value_sqd_uv * 
      (double)((double) impix_cr / (snr->msse_u == 0.0? 1.0 : snr->msse_u))));  
    float csnr_v = (float) (10 * log10 (max_pix_value_sqd_uv * 
      (double)((double) impix_cr / (snr->msse_v == 0.0? 1.0 : snr->msse_v))));  
    
    fprintf(stdout," PSNR Y(dB)                        : %5.4f\n",snr->snr_ya);
    fprintf(stdout," PSNR U(dB)                        : %5.4f\n",snr->snr_ua);
    fprintf(stdout," PSNR V(dB)                        : %5.4f\n",snr->snr_va);
    fprintf(stdout," cSNR Y(dB)                        : %5.2f (%5.2f)\n",csnr_y,snr->msse_y/impix);
    fprintf(stdout," cSNR U(dB)                        : %5.2f (%5.2f)\n",csnr_u,snr->msse_u/impix_cr);
    fprintf(stdout," cSNR V(dB)                        : %5.2f (%5.2f)\n",csnr_v,snr->msse_v/impix_cr);
  }
  else
    fprintf(stdout,  " Total encoding time for the seq.  : %.3f sec (%.2f fps)\n",tot_time*0.001, 1000.0*(input->no_frames + frame_ctr[B_SLICE])/tot_time);
  
  if(frame_ctr[B_SLICE]!=0)
  {
    fprintf(stdout, " Total bits                        : %" FORMAT_OFF_T  " (I %" FORMAT_OFF_T  ", P %" FORMAT_OFF_T  ", B %" FORMAT_OFF_T  " NVB %d) \n",
      total_bits=stats->bit_ctr_P + stats->bit_ctr_I + stats->bit_ctr_B + stats->bit_ctr_parametersets, 
      stats->bit_ctr_I, stats->bit_ctr_P, stats->bit_ctr_B, stats->bit_ctr_parametersets);
    
    frame_rate = (img->framerate *(float)(stats->successive_Bframe + 1)) / (float) (input->jumpd+1);
    //    stats->bitrate= ((float) total_bits * frame_rate)/((float) (input->no_frames + frame_ctr[B_SLICE]));
    stats->bitrate= ((float) total_bits * frame_rate)/((float)(frame_ctr[I_SLICE] + frame_ctr[P_SLICE] + frame_ctr[B_SLICE]));
    
    fprintf(stdout, " Bit rate (kbit/s)  @ %2.2f Hz     : %5.2f\n", frame_rate, stats->bitrate/1000);
    
  }
  else if (input->sp_periodicity==0)
  {
    fprintf(stdout, " Total bits                        : %" FORMAT_OFF_T  " (I %" FORMAT_OFF_T  ", P %" FORMAT_OFF_T  ", NVB %d) \n",
      total_bits=stats->bit_ctr_P + stats->bit_ctr_I + stats->bit_ctr_parametersets, stats->bit_ctr_I, stats->bit_ctr_P, stats->bit_ctr_parametersets);
    
    
    frame_rate = img->framerate / ( (float) (input->jumpd + 1) );
    stats->bitrate= ((float) total_bits * frame_rate)/((float) input->no_frames );
    
    fprintf(stdout, " Bit rate (kbit/s)  @ %2.2f Hz     : %5.2f\n", frame_rate, stats->bitrate/1000);
  }
  else
  {
    fprintf(stdout, " Total bits                        : %" FORMAT_OFF_T  " (I %" FORMAT_OFF_T  ", P %" FORMAT_OFF_T  ", NVB %d) \n",
      total_bits=stats->bit_ctr_P + stats->bit_ctr_I + stats->bit_ctr_parametersets, stats->bit_ctr_I, stats->bit_ctr_P, stats->bit_ctr_parametersets);
    
    
    frame_rate = img->framerate / ( (float) (input->jumpd + 1) );
    stats->bitrate= ((float) total_bits * frame_rate)/((float) input->no_frames );
    
    fprintf(stdout, " Bit rate (kbit/s)  @ %2.2f Hz     : %5.2f\n", frame_rate, stats->bitrate/1000);
  }
  
  fprintf(stdout, " Bits to avoid Startcode Emulation : %d \n", stats->bit_ctr_emulationprevention);
  fprintf(stdout, " Bits for parameter sets           : %d \n", stats->bit_ctr_parametersets);
  
  fprintf(stdout,"-----------------------------------------------------------------------------------\n");
  fprintf(stdout,"Exit JM %s encoder ver %s ", JM, VERSION);
  fprintf(stdout,"\n");
  
  // status file
  if ((p_stat=fopen("stats.dat","wt"))==0)
  {
    snprintf(errortext, ET_SIZE, "Error open file %s", "stats.dat");
    error(errortext, 500);
  }
  fprintf(p_stat," -------------------------------------------------------------- \n");
  fprintf(p_stat,"  This file contains statistics for the last encoded sequence   \n");
  fprintf(p_stat," -------------------------------------------------------------- \n");
  fprintf(p_stat,   " Sequence                     : %s\n",input->infile);
  fprintf(p_stat,   " No.of coded pictures         : %4d\n",input->no_frames+frame_ctr[B_SLICE]);
  fprintf(p_stat,   " Freq. for encoded bitstream  : %4.0f\n",frame_rate);
  
  fprintf(p_stat,   " I Slice Bitrate(kb/s)        : %6.2f\n", stats->bitrate_I/1000);
  fprintf(p_stat,   " P Slice Bitrate(kb/s)        : %6.2f\n", stats->bitrate_P/1000);
  // B pictures
  if(stats->successive_Bframe != 0)
    fprintf(p_stat,   " B Slice Bitrate(kb/s)        : %6.2f\n", stats->bitrate_B/1000);
  fprintf(p_stat,   " Total Bitrate(kb/s)          : %6.2f\n", stats->bitrate/1000);
  
  switch (input->hadamard)
  {
  case 2:
    fprintf(p_stat," Hadamard transform           : Used for QPel\n");
    break;
  case 1:
    fprintf(p_stat," Hadamard transform           : Used\n");
    break;
  default:
    fprintf(p_stat," Hadamard transform           : Not used\n");
    break;
  }
  
  fprintf(p_stat,  " Image format                 : %dx%d\n",input->img_width,input->img_height);
  
  if(input->intra_upd)
    fprintf(p_stat," Error robustness             : On\n");
  else
    fprintf(p_stat," Error robustness             : Off\n");
  
  fprintf(p_stat,  " Search range                 : %d\n",input->search_range);
  
  fprintf(p_stat,   " Total number of references   : %d\n",input->num_ref_frames);
  fprintf(p_stat,   " References for P slices      : %d\n",input->P_List0_refs? input->P_List0_refs:input->num_ref_frames);
  if(stats->successive_Bframe != 0)
  {
    fprintf(p_stat, " List0 refs for B slices      : %d\n",input->B_List0_refs? input->B_List0_refs:input->num_ref_frames);
    fprintf(p_stat, " List1 refs for B slices      : %d\n",input->B_List1_refs? input->B_List1_refs:input->num_ref_frames);
  }
  
  if (input->symbol_mode == UVLC)
    fprintf(p_stat,   " Entropy coding method        : CAVLC\n");
  else
    fprintf(p_stat,   " Entropy coding method        : CABAC\n");
  
  fprintf(p_stat,   " Profile/Level IDC            : (%d,%d)\n",input->ProfileIDC,input->LevelIDC);
  
  if (input->FMEnable == 3)
    EPZSOutputStats(p_stat, 1);
  
#ifdef _FULL_SEARCH_RANGE_
  if (input->full_search == 2)
    fprintf(p_stat," Search range restrictions    : none\n");
  else if (input->full_search == 1)
    fprintf(p_stat," Search range restrictions    : older reference frames\n");
  else
    fprintf(p_stat," Search range restrictions    : smaller blocks and older reference frames\n");
#endif
  if (input->rdopt)
    fprintf(p_stat," RD-optimized mode decision   : used\n");
  else
    fprintf(p_stat," RD-optimized mode decision   : not used\n");
  
  fprintf(p_stat," ---------------------|----------------|---------------|\n");
  fprintf(p_stat,"     Item             |     Intra      |   All frames  |\n");
  fprintf(p_stat," ---------------------|----------------|---------------|\n");
  fprintf(p_stat," SNR Y(dB)            |");
  fprintf(p_stat," %5.2f          |",snr->snr_y1);
  fprintf(p_stat," %5.2f         |\n",snr->snr_ya);
  fprintf(p_stat," SNR U/V (dB)         |");
  fprintf(p_stat," %5.2f/%5.2f    |",snr->snr_u1,snr->snr_v1);
  fprintf(p_stat," %5.2f/%5.2f   |\n",snr->snr_ua,snr->snr_va);
  
  // QUANT.
  fprintf(p_stat," Average quant        |");
  fprintf(p_stat," %5d          |",absm(input->qp0));
  fprintf(p_stat," %5.2f         |\n",(float)stats->quant1/max(1.0,(float)stats->quant0));
  
  fprintf(p_stat,"\n ---------------------|----------------|---------------|---------------|\n");
  fprintf(p_stat,"     SNR              |        I       |       P       |       B       |\n");
  fprintf(p_stat," ---------------------|----------------|---------------|---------------|\n");
  fprintf(p_stat," SNR Y(dB)            |      %5.3f    |     %5.3f    |     %5.3f    |\n",
    snr->snr_yt[I_SLICE],snr->snr_yt[P_SLICE],snr->snr_yt[B_SLICE]);
  fprintf(p_stat," SNR U(dB)            |      %5.3f    |     %5.3f    |     %5.3f    |\n",
    snr->snr_ut[I_SLICE],snr->snr_ut[P_SLICE],snr->snr_ut[B_SLICE]);
  fprintf(p_stat," SNR V(dB)            |      %5.3f    |     %5.3f    |     %5.3f    |\n",
    snr->snr_vt[I_SLICE],snr->snr_vt[P_SLICE],snr->snr_vt[B_SLICE]);
  
  
  // MODE
  fprintf(p_stat,"\n ---------------------|----------------|\n");
  fprintf(p_stat,"   Intra              |   Mode used    |\n");
  fprintf(p_stat," ---------------------|----------------|\n");
  
  fprintf(p_stat," Mode 0  intra 4x4    |  %5" FORMAT_OFF_T  "         |\n",stats->mode_use[I_SLICE][I4MB ]);
  fprintf(p_stat," Mode 1  intra 8x8    |  %5" FORMAT_OFF_T  "         |\n",stats->mode_use[I_SLICE][I8MB ]);
  fprintf(p_stat," Mode 2+ intra 16x16  |  %5" FORMAT_OFF_T  "         |\n",stats->mode_use[I_SLICE][I16MB]);
  
  fprintf(p_stat,"\n ---------------------|----------------|-----------------|\n");
  fprintf(p_stat,"   Inter              |   Mode used    | MotionInfo bits |\n");
  fprintf(p_stat," ---------------------|----------------|-----------------|");
  fprintf(p_stat,"\n Mode  0  (copy)      |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[P_SLICE][0   ],(double)stats->bit_use_mode[P_SLICE][0   ]/(double)bit_use[P_SLICE][0]);
  fprintf(p_stat,"\n Mode  1  (16x16)     |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[P_SLICE][1   ],(double)stats->bit_use_mode[P_SLICE][1   ]/(double)bit_use[P_SLICE][0]);
  fprintf(p_stat,"\n Mode  2  (16x8)      |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[P_SLICE][2   ],(double)stats->bit_use_mode[P_SLICE][2   ]/(double)bit_use[P_SLICE][0]);
  fprintf(p_stat,"\n Mode  3  (8x16)      |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[P_SLICE][3   ],(double)stats->bit_use_mode[P_SLICE][3   ]/(double)bit_use[P_SLICE][0]);
  fprintf(p_stat,"\n Mode  4  (8x8)       |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[P_SLICE][P8x8],(double)stats->bit_use_mode[P_SLICE][P8x8]/(double)bit_use[P_SLICE][0]);
  fprintf(p_stat,"\n Mode  5  intra 4x4   |  %5" FORMAT_OFF_T  "         |-----------------|",stats->mode_use[P_SLICE][I4MB]);
  fprintf(p_stat,"\n Mode  6  intra 8x8   |  %5" FORMAT_OFF_T  "         |",stats->mode_use[P_SLICE][I8MB]);
  fprintf(p_stat,"\n Mode  7+ intra 16x16 |  %5" FORMAT_OFF_T  "         |",stats->mode_use[P_SLICE][I16MB]);
  mean_motion_info_bit_use[0] = (double)(stats->bit_use_mode[P_SLICE][0] + stats->bit_use_mode[P_SLICE][1] + stats->bit_use_mode[P_SLICE][2] 
    + stats->bit_use_mode[P_SLICE][3] + stats->bit_use_mode[P_SLICE][P8x8])/(double) bit_use[P_SLICE][0]; 
  
  // B pictures
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
  {
    
    fprintf(p_stat,"\n\n ---------------------|----------------|-----------------|\n");
    fprintf(p_stat,"   B frame            |   Mode used    | MotionInfo bits |\n");
    fprintf(p_stat," ---------------------|----------------|-----------------|");
    fprintf(p_stat,"\n Mode  0  (copy)      |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[B_SLICE][0   ],(double)stats->bit_use_mode[B_SLICE][0   ]/(double)frame_ctr[B_SLICE]);
    fprintf(p_stat,"\n Mode  1  (16x16)     |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[B_SLICE][1   ],(double)stats->bit_use_mode[B_SLICE][1   ]/(double)frame_ctr[B_SLICE]);
    fprintf(p_stat,"\n Mode  2  (16x8)      |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[B_SLICE][2   ],(double)stats->bit_use_mode[B_SLICE][2   ]/(double)frame_ctr[B_SLICE]);
    fprintf(p_stat,"\n Mode  3  (8x16)      |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[B_SLICE][3   ],(double)stats->bit_use_mode[B_SLICE][3   ]/(double)frame_ctr[B_SLICE]);
    fprintf(p_stat,"\n Mode  4  (8x8)       |  %5" FORMAT_OFF_T  "         |    %8.2f     |",stats->mode_use[B_SLICE][P8x8],(double)stats->bit_use_mode[B_SLICE][P8x8]/(double)frame_ctr[B_SLICE]);
    fprintf(p_stat,"\n Mode  5  intra 4x4   |  %5" FORMAT_OFF_T  "         |-----------------|",stats->mode_use[B_SLICE][I4MB]);
    fprintf(p_stat,"\n Mode  6  intra 8x8   |  %5" FORMAT_OFF_T  "         |",stats->mode_use[B_SLICE][I8MB]);
    fprintf(p_stat,"\n Mode  7+ intra 16x16 |  %5" FORMAT_OFF_T  "         |",stats->mode_use[B_SLICE][I16MB]);
    mean_motion_info_bit_use[1] = (double)(stats->bit_use_mode[B_SLICE][0] + stats->bit_use_mode[B_SLICE][1] + stats->bit_use_mode[B_SLICE][2] 
      + stats->bit_use_mode[B_SLICE][3] + stats->bit_use_mode[B_SLICE][P8x8])/(double) frame_ctr[B_SLICE]; 
  }
  
  fprintf(p_stat,"\n\n ---------------------|----------------|----------------|----------------|\n");
  fprintf(p_stat,"  Bit usage:          |      Intra     |      Inter     |    B frame     |\n");
  fprintf(p_stat," ---------------------|----------------|----------------|----------------|\n");
  
  fprintf(p_stat," Header               |");
  fprintf(p_stat," %10.2f     |",(float) stats->bit_use_header[I_SLICE]/bit_use[I_SLICE][0]);
  fprintf(p_stat," %10.2f     |",(float) stats->bit_use_header[P_SLICE]/bit_use[P_SLICE][0]);
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," %10.2f     |",(float) stats->bit_use_header[B_SLICE]/frame_ctr[B_SLICE]);
  else fprintf(p_stat," %10.2f     |", 0.);
  fprintf(p_stat,"\n");
  
  fprintf(p_stat," Mode                 |");
  fprintf(p_stat," %10.2f     |",(float)stats->bit_use_mb_type[I_SLICE]/bit_use[I_SLICE][0]);
  fprintf(p_stat," %10.2f     |",(float)stats->bit_use_mb_type[P_SLICE]/bit_use[P_SLICE][0]);
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," %10.2f     |",(float)stats->bit_use_mb_type[B_SLICE]/frame_ctr[B_SLICE]);
  else fprintf(p_stat," %10.2f     |", 0.);
  fprintf(p_stat,"\n");
  
  fprintf(p_stat," Motion Info          |");
  fprintf(p_stat,"        ./.     |");
  fprintf(p_stat," %10.2f     |",mean_motion_info_bit_use[0]);
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," %10.2f     |",mean_motion_info_bit_use[1]);
  else fprintf(p_stat," %10.2f     |", 0.);
  fprintf(p_stat,"\n");
  
  fprintf(p_stat," CBP Y/C              |");
  fprintf(p_stat," %10.2f     |", (float)stats->tmp_bit_use_cbp[I_SLICE]/bit_use[I_SLICE][0]);
  fprintf(p_stat," %10.2f     |", (float)stats->tmp_bit_use_cbp[P_SLICE]/bit_use[P_SLICE][0]);
  if(stats->successive_Bframe!=0 && bit_use[B_SLICE][0]!=0)
    fprintf(p_stat," %10.2f     |", (float)stats->tmp_bit_use_cbp[B_SLICE]/bit_use[B_SLICE][0]);
  else fprintf(p_stat," %10.2f     |", 0.);
  fprintf(p_stat,"\n");
  
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," Coeffs. Y            | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_coeffY[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_coeffY[P_SLICE]/bit_use[P_SLICE][0], (float)stats->bit_use_coeffY[B_SLICE]/frame_ctr[B_SLICE]);
  else
    fprintf(p_stat," Coeffs. Y            | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_coeffY[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_coeffY[P_SLICE]/(float)bit_use[P_SLICE][0], 0.);
  
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," Coeffs. C            | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_coeffC[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_coeffC[P_SLICE]/bit_use[P_SLICE][0], (float)stats->bit_use_coeffC[B_SLICE]/frame_ctr[B_SLICE]);
  else
    fprintf(p_stat," Coeffs. C            | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_coeffC[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_coeffC[P_SLICE]/bit_use[P_SLICE][0], 0.);
  
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," Delta quant          | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_delta_quant[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_delta_quant[P_SLICE]/bit_use[P_SLICE][0], (float)stats->bit_use_delta_quant[B_SLICE]/frame_ctr[B_SLICE]);
  else
    fprintf(p_stat," Delta quant          | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_delta_quant[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_delta_quant[P_SLICE]/bit_use[P_SLICE][0], 0.);
  
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," Stuffing Bits        | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_stuffingBits[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_stuffingBits[P_SLICE]/bit_use[P_SLICE][0], (float)stats->bit_use_stuffingBits[B_SLICE]/frame_ctr[B_SLICE]);
  else
    fprintf(p_stat," Stuffing Bits        | %10.2f     | %10.2f     | %10.2f     |\n",
    (float)stats->bit_use_stuffingBits[I_SLICE]/bit_use[I_SLICE][0], (float)stats->bit_use_stuffingBits[P_SLICE]/bit_use[P_SLICE][0], 0.);
  
  
  
  fprintf(p_stat," ---------------------|----------------|----------------|----------------|\n");
  
  fprintf(p_stat," average bits/frame   |");
  
  fprintf(p_stat," %10.2f     |", (float) bit_use[I_SLICE][1]/(float) bit_use[I_SLICE][0] );
  fprintf(p_stat," %10.2f     |", (float) bit_use[P_SLICE][1]/(float) bit_use[P_SLICE][0] );
  
  if(stats->successive_Bframe!=0 && frame_ctr[B_SLICE]!=0)
    fprintf(p_stat," %10.2f     |", (float) bit_use[B_SLICE][1]/ (float) frame_ctr[B_SLICE] );
  else fprintf(p_stat," %10.2f     |", 0.);
  
  fprintf(p_stat,"\n");
  fprintf(p_stat," ---------------------|----------------|----------------|----------------|\n");

  fclose(p_stat);
  
  // write to log file
  if ((p_log=fopen("log.dat","r"))==0)                      // check if file exist
  {
    if ((p_log=fopen("log.dat","a"))==NULL)            // append new statistic at the end
    {
      snprintf(errortext, ET_SIZE, "Error open file %s  \n","log.dat");
      error(errortext, 500);
    }
    else                                            // Create header for new log file
    {
      fprintf(p_log," ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
      fprintf(p_log,"|            Encoder statistics. This file is generated during first encoding session, new sessions will be appended                                                                                                                                                                                                                |\n");
      fprintf(p_log," ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
      fprintf(p_log,"|    ver      | Date  | Time  |         Sequence             | #Img |P/MbInt| QPI| QPP| QPB| Format  |Iperiod| #B | FMES | Hdmd | S.R |#Ref | Freq |Coding|RD-opt|Intra upd|8x8Tr|AIF|FDIF|APEC|AQM|AQP|MVRes|MVQ| SNRY 1| SNRU 1| SNRV 1| SNRY N| SNRU N| SNRV N|#Bitr I|#Bitr P|#Bitr B|#Bitr IPB|     Total Time   |      Me Time     |\n");
      fprintf(p_log," ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
      //FDIF - Fast Decision on Interpolation Filter
    }
  }
  else
  {
    fclose (p_log);
    if ((p_log=fopen("log.dat","a"))==NULL)            // File exist,just open for appending
    {
      snprintf(errortext, ET_SIZE, "Error open file %s  \n","log.dat");
      error(errortext, 500);
    }
  }
  fprintf(p_log,"|%s/%-4s", VERSION, EXT_VERSION);
  
#ifdef WIN32
  _strdate( timebuf );
  fprintf(p_log,"| %1.5s |",timebuf );
  
  _strtime( timebuf);
  fprintf(p_log," % 1.5s |",timebuf);
#else
  now = time ((time_t *) NULL); // Get the system time and put it into 'now' as 'calender time'
  time (&now);
  l_time = localtime (&now);
  strftime (string, sizeof string, "%d-%b-%Y", l_time);
  fprintf(p_log,"| %1.5s |",string );
  
  strftime (string, sizeof string, "%H:%M:%S", l_time);
  fprintf(p_log," %1.5s |",string );
#endif
  
  for (i=0;i<30;i++)
    name[i]=input->infile[i+max(0,((int)strlen(input->infile))-30)]; // write last part of path, max 20 chars
  fprintf(p_log,"%30.30s|",name);
  
  fprintf(p_log,"%5d |",input->no_frames);
  fprintf(p_log,"  %d/%d  |",input->PicInterlace, input->MbInterlace);
  fprintf(p_log," %-3d|",input->qp0);
  fprintf(p_log," %-3d|",input->qpN);
  fprintf(p_log," %-3d|",input->qpB);
  
  fprintf(p_log,"%4dx%-4d|",input->img_width,input->img_height);
  
  fprintf(p_log,"  %3d  |",input->intra_period);
  fprintf(p_log,"%3d |",stats->successive_Bframe); 
  
  if (input->FMEnable==1)
    fprintf(p_log,"  HEX |");
  else if (input->FMEnable==2)
    fprintf(p_log," SHEX |");
  else if (input->FMEnable == 3)
    fprintf(p_log," EPZS |");
  else
    fprintf(p_log,"  OFF |");
  
  switch (input->hadamard)
  {
  case 2:
    fprintf(p_log,"  QPL |");
    break;
  case 1:
    fprintf(p_log,"  ON  |");
    break;
  default:
    fprintf(p_log,"  OFF |");
    break;
  }
  fprintf(p_log," %3d |",input->search_range );
  
  fprintf(p_log," %2d  |",input->num_ref_frames);
  
  
  //  fprintf(p_log," %3d  |",img->framerate/(input->jumpd+1));
  fprintf(p_log," %5.2f|",(img->framerate *(float) (stats->successive_Bframe + 1)) / (float)(input->jumpd+1));
  
  if (input->symbol_mode == UVLC)
    fprintf(p_log," CAVLC|");
  else
    fprintf(p_log," CABAC|");
  
  fprintf(p_log,"   %d  |",input->rdopt);
  
  if (input->intra_upd==1)
    fprintf(p_log,"   ON    |");
  else
    fprintf(p_log,"   OFF   |");
  
  fprintf(p_log,"  %d  |",input->Transform8x8Mode);

#ifdef EIGHTH_PEL
  fprintf(p_log," 1/%d |", input->mv_res?8:4);
#else
//1/4th pel
  fprintf(p_log," 1/4 |");	
#endif
//MV_COMPETITION
#ifdef MV_COMPETITION
  fprintf(p_log," %d |", input->mv_competition);
#else
  fprintf(p_log," 0 |");   
#endif
  
  fprintf(p_log,"%7.3f|",snr->snr_y1);
  fprintf(p_log,"%7.3f|",snr->snr_u1);
  fprintf(p_log,"%7.3f|",snr->snr_v1);
  fprintf(p_log,"%7.3f|",snr->snr_ya);
  fprintf(p_log,"%7.3f|",snr->snr_ua);
  fprintf(p_log,"%7.3f|",snr->snr_va);
  /*
  fprintf(p_log,"%-5.3f|",snr->snr_yt[I_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_ut[I_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_vt[I_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_yt[P_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_ut[P_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_vt[P_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_yt[B_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_ut[B_SLICE]);
  fprintf(p_log,"%-5.3f|",snr->snr_vt[B_SLICE]);
  */
  fprintf(p_log,"%7.0f|",stats->bitrate_I);
  fprintf(p_log,"%7.0f|",stats->bitrate_P);
  fprintf(p_log,"%7.0f|",stats->bitrate_B);
  fprintf(p_log,"%9.0f|",stats->bitrate);
  
  fprintf(p_log,"   %12d   |", tot_time);
  fprintf(p_log,"   %12d   |\n", me_tot_time);
  
  fclose(p_log);
  
  p_log=fopen("data.txt","a");
  
  if(stats->successive_Bframe != 0 && frame_ctr[B_SLICE] != 0) // B picture used
  {
    fprintf(p_log, "%3d %2d %2d %2.2f %2.2f %2.2f %5" FORMAT_OFF_T  " "
      "%2.2f %2.2f %2.2f %5d "
      "%2.2f %2.2f %2.2f %5" FORMAT_OFF_T  " %5" FORMAT_OFF_T  " %.3f\n",
      input->no_frames, input->qp0, input->qpN,
      snr->snr_y1,
      snr->snr_u1,
      snr->snr_v1,
      stats->bit_ctr_I,
      0.0,
      0.0,
      0.0,
      0,
      snr->snr_ya,
      snr->snr_ua,
      snr->snr_va,
      (stats->bit_ctr_I+stats->bit_ctr)/(input->no_frames+frame_ctr[B_SLICE]),
      stats->bit_ctr_B/frame_ctr[B_SLICE],
      (double)0.001*tot_time/(input->no_frames+frame_ctr[B_SLICE]));
  }
  else
  {
    if (input->no_frames!=0)
      fprintf(p_log, "%3d %2d %2d %2.2f %2.2f %2.2f %5" FORMAT_OFF_T  " "
      "%2.2f %2.2f %2.2f %5d "
      "%2.2f %2.2f %2.2f %5" FORMAT_OFF_T  " %5d %.3f\n",
      input->no_frames, input->qp0, input->qpN,
      snr->snr_y1,
      snr->snr_u1,
      snr->snr_v1,
      stats->bit_ctr_I,
      0.0,
      0.0,
      0.0,
      0,
      snr->snr_ya,
      snr->snr_ua,
      snr->snr_va,
      (stats->bit_ctr_I+stats->bit_ctr)/input->no_frames,
      0,
      (double)0.001*tot_time/input->no_frames);
  }
  
  fclose(p_log);
  
  if (input->ReportFrameStats)
  {
    if ((p_log=fopen("stat_frame.dat","a"))==NULL)            // append new statistic at the end
    {
      snprintf(errortext, ET_SIZE, "Error open file %s  \n","stat_frame.dat.dat");
      //    error(errortext, 500);
    }
    else
    {
      fprintf(p_log," --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- \n");
      fclose(p_log);  
    }
  }

#ifdef MB32X32
  {
    int i, j;
    double prob;

    printf("\n");
    printf("=================  Skip   32x32   32x16   16x32   16x16  ==========\n");
    for(i=0; i<2; i++)
    {
      if(i==0)
        printf("MB 32x32 P slice  ");
      else if (input->successive_Bframe)
        printf("MB 32x32 B slice  ");

      for(j=0; j<5; j++)
      { 
        if(i==0)
        {
          prob = mb32_stat[i][j]*100.0/((input->no_frames-1)*img->FrameSizeInMbs/4.0);
          printf("%3d%%    ", (int)prob);
        }
        else if (input->successive_Bframe)
        {
          prob = mb32_stat[i][j]*100.0/(input->successive_Bframe*(input->no_frames-1)*img->FrameSizeInMbs/4.0);
          printf("%3d%%    ", (int)prob);
        }
      }
      printf("\n");
    }

#ifdef MB64X64
    printf("=================  Skip   64x64   64x32   32x64   32x32  ==========\n");
    for(i=0; i<2; i++)
    {
      if(i==0)
        printf("MB 64x64 P slice  ");
      else if (input->successive_Bframe)
        printf("MB 64x64 B slice  ");

      for(j=0; j<5; j++)
      { 
        if(i==0)
        {
          prob = mb64_stat[i][j]*100.0/((input->no_frames-1)*img->FrameSizeInMbs/16.0);
          printf("%3d%%    ", (int)prob);
        }
        else if (input->successive_Bframe)
        {
          prob = mb64_stat[i][j]*100.0/(input->successive_Bframe*(input->no_frames-1)*img->FrameSizeInMbs/16.0);
          printf("%3d%%    ", (int)prob);
        }
      }
      printf("\n");
    }
#endif
  }
#endif

}



/*!
************************************************************************
* \brief
*    Prints the header of the protocol.
* \par Input:
*    struct inp_par *inp
* \par Output:
*    none
************************************************************************
*/
void information_init(void)
{
  char yuv_types[4][10]= {"YUV 4:0:0","YUV 4:2:0","YUV 4:2:2","YUV 4:4:4"};
  if (input->Verbose == 0 || input->Verbose  == 1)
    printf("------------------------------- JM %s %s ----------------------------------\n",VERSION, EXT_VERSION);    
  else   
    printf("------------------------------- JM %s %s ------------------------------------------\n",VERSION, EXT_VERSION);
  printf(" Input YUV file                    : %s \n",input->infile);
  printf(" Output H.264 bitstream            : %s \n",input->outfile);
  if (p_dec != -1)
    printf(" Output YUV file                   : %s \n",input->ReconFile);
  printf(" YUV Format                        : %s \n", &yuv_types[img->yuv_format][0]);//img->yuv_format==YUV422?"YUV 4:2:2":(img->yuv_format==YUV444)?"YUV 4:4:4":"YUV 4:2:0");
#if AdaptiveCADENCE  
  if(input->HierB_AdaptiveCadence)
    printf(" Frames to be encoded              : %d\n", input->no_frames);
  else
    printf(" Frames to be encoded I-P/B        : %d/%d\n", input->no_frames, (input->successive_Bframe*(input->no_frames-1)));
#else
  printf(" Frames to be encoded I-P/B        : %d/%d\n", input->no_frames, (input->successive_Bframe*(input->no_frames-1)));
#endif
  printf(" PicInterlace / MbInterlace        : %d/%d\n", input->PicInterlace, input->MbInterlace);
  printf(" Transform8x8Mode                  : %d\n", input->Transform8x8Mode);
#ifdef EIGHTH_PEL
  printf(" Motion vector resolution          : 1/%i-pel\n",input->mv_res ? 8 : 4 );
#endif

#ifdef MV_COMPETITION
  printf(" MV Competition                    : %i\n",input->mv_competition);
  if (input->mv_competition == 1)
  {
      printf("\t\t\t\tPredictors for Skip     : 01100000\n");
      printf("\t\t\t\tPredictors for P vectors: 10000010\n");
    if (input->successive_Bframe>0)
      printf("\t\t\t\tPredictors for B vectors: 10000010\n");
  }
  if (input->mv_competition > 1)
  {
      printf("\t\t\t\tPredictors for Skip     : %s\n", input->predictors_skip);
      printf("\t\t\t\tPredictors for P vectors: %s\n", input->predictors_mvp);
    if (input->successive_Bframe>0)
      printf("\t\t\t\tPredictors for B vectors: %s\n", input->predictors_mvb);
  }
#endif

#ifdef USE_INTRA_MDDT
  printf(" INTRA MDDT                        : %i\n",input->UseIntraMDDT);
#endif

#ifdef RDO_Q
  printf(" RDO_Q                             : %i\n",input->UseRDO_Q);
#endif  
#ifdef USE_NEW_OFFSET
  printf(" USE_NEW_OFFSET                    : %i\n",input->UseNewOffset);
#endif

  printf(" HP_FILTER                         : %i\n",input->UseHPFilter);
  printf(" NEW_WEIGHTS                       : 1\n");

#if B_MBTYPE
  printf(" B_MBTYPE                          : 1\n");
#else
  printf(" B_MBTYPE                          : 0\n");
#endif

  printf(" SCALED_MVPREDICTOR                : 1\n");

#if P_DIRECT
  printf(" P_DIRECT                          : 1\n");
#else
  printf(" P_DIRECT                          : 0\n");
#endif
#ifdef ADAPTIVE_LOOP_FILTER
  printf(" Adaptive Loop Filter              : %i\n",input->UseAdaptiveLoopFilter);
#endif


#if AdaptiveCADENCE
  printf(" Adaptive Cadence                  : %i\n",input->HierB_AdaptiveCadence);
#endif

  printf(" Tx4=0 Tx8=0 Tx16=1 for BigBlocks  : %i\n",input->DisableSmallTransInLargeMB);

#ifdef INTERNAL_BIT_DEPTH_INCREASE
    printf(" BIT DEPTH (Luma,Chroma)           : (%i,%i)\n",input->BitDepthLuma,input->BitDepthChroma);
#endif


  switch (input->Verbose) 
  {
  case 1:
    printf("-----------------------------------------------------------------------------------\n");    
    printf("  Frame  Bit/pic    QP   SnrY    SnrU    SnrV    Time(ms) MET(ms) Frm/Fld Ref \n");
    printf("-----------------------------------------------------------------------------------\n");
    break;
  case 2:
    printf("---------------------------------------------------------------------------------------------\n");
    printf("  Frame  Bit/pic WP QP   SnrY    SnrU    SnrV    Time(ms) MET(ms) Frm/Fld   I D L0 L1 RDP Ref\n");
    printf("---------------------------------------------------------------------------------------------\n");
    break;
  case 0:
  default:
    printf("-------------------------------------------------------------------------------\n");
    printf("\nEncoding. Please Wait.\n\n");
    break;
  }
}

/*!
************************************************************************
* \brief
*    memory allocation for original picture buffers
************************************************************************
*/
int init_orig_buffers(void)
{
  int memory_size = 0;
  
  // allocate memory for reference frame buffers: imgY_org_frm, imgUV_org_frm
  memory_size += get_mem2Dpel(&imgY_org_frm, img->height, img->width);
  
  if (img->yuv_format != YUV400)
    memory_size += get_mem3Dpel(&imgUV_org_frm, 2, img->height_cr, img->width_cr);
  
  return memory_size;
}

/*!
************************************************************************
* \brief
*    Dynamic memory allocation of frame size related global buffers
*    buffers are defined in global.h, allocated memory must be freed in
*    void free_global_buffers()
* \par Input:
*    Input Parameters struct inp_par *inp,                            \n
*    Image Parameters struct img_par *img
* \return Number of allocated bytes
************************************************************************
*/
int init_global_buffers(void)
{
  int j,memory_size=0;
#ifdef _ADAPT_LAST_GROUP_
  extern int *last_P_no_frm;
  
  if ((last_P_no_frm = (int*)malloc(2*img->max_num_references*sizeof(int))) == NULL)
    no_mem_exit("init_global_buffers: last_P_no");
#endif
    
#if defined MB32X32
  memory_size += get_mem2Dint(&img->cofAC16x16,2,257);
  memory_size += get_mem3Dint(&img->cofAC16x8, 2, 2, 129);
  memory_size += get_mem3Dint(&img->cofAC8x16, 2, 2, 129);
#endif
  memory_size += init_orig_buffers(); 
    
    
    if ((yPicPos = (unsigned int*)malloc(img->FrameSizeInMbs * sizeof(unsigned int))) == NULL)
      no_mem_exit("init_global_buffers: yPicPos");  
    memory_size += img->FrameSizeInMbs * sizeof(unsigned int);
    if ((xPicPos = (unsigned int*)malloc(img->FrameSizeInMbs * sizeof(unsigned int))) == NULL)
      no_mem_exit("init_global_buffers: xPicPos");
    memory_size += img->FrameSizeInMbs * sizeof(unsigned int);
    
    for (j=0;j< (int) img->FrameSizeInMbs;j++)
    {
      xPicPos[j] = (j % img->PicWidthInMbs);
      yPicPos[j] = (j / img->PicWidthInMbs);
    }
    
    if (input->WeightedPrediction || input->WeightedBiprediction || input->GenerateMultiplePPS)
    {
      // Currently only use up to 20 references. Need to use different indicator such as maximum num of references in list
      memory_size += get_mem3Dint(&wp_weight,6,MAX_REFERENCE_PICTURES,3);
      memory_size += get_mem3Dint(&wp_offset,6,MAX_REFERENCE_PICTURES,3);
      memory_size += get_mem4Dint(&wbp_weight, 6, MAX_REFERENCE_PICTURES, MAX_REFERENCE_PICTURES, 3);
      memory_size += get_mem3Dint(&explicit_weight,6,MAX_REFERENCE_PICTURES,3);
    }
    
    // allocate memory for reference frames of each block: refFrArr
    
    if(input->successive_Bframe!=0 || input->BRefPictures> 0)
    {    
      memory_size += get_mem3D     ((byte ****)(void*)(&direct_ref_idx), 2, img->height >> BLOCK_SHIFT, img->width >> BLOCK_SHIFT);
      memory_size += get_mem2Dshort(&direct_pdir, img->height >> BLOCK_SHIFT, img->width >> BLOCK_SHIFT);
    }
    
    // allocate memory for temp quarter pel luma frame buffer: img4Y_tmp
    // int img4Y_tmp[576][704];  (previously int imgY_tmp in global.h)
    memory_size += get_mem2Dint(&img4Y_tmp, img->height+2*IMG_PAD_SIZE, (img->width+2*IMG_PAD_SIZE)*4);
    
    if (input->RestrictRef)
    {
      memory_size += get_mem2D(&pixel_map, img->height,img->width);
      memory_size += get_mem2D(&refresh_map, img->height/8,img->width/8);
    }
    

    memory_size += EPZSInit();
#ifdef ADAPTIVE_LOOP_FILTER
    if(input->UseAdaptiveLoopFilter)
    {
      memory_size += InitALFGlobalBuffers();
    }
#endif

    if(img->BitDepthIncrease)
    {
      int max       = 1<<input->BitDepthLuma;
      int max_half  = 1<<(img->BitDepthIncrease-1);
      if ((Pel_12to8 = (imgpel*)malloc(max * sizeof(imgpel))) == NULL)
        no_mem_exit("init_global_buffers: Pel_12to8");
      memory_size += max * sizeof(imgpel);

      for (j=0; j<max; j++)
      {
        Pel_12to8[j] = (imgpel) Clip3(0,(1<<input->InputBitDepth)-1,(j + max_half)>>img->BitDepthIncrease);
      }
    }

    return (memory_size);
}


/*!
************************************************************************
* \brief
*    Free allocated memory of original picture buffers
************************************************************************
*/
void free_orig_planes(void)
{
  free_mem2Dpel(imgY_org_frm);      // free ref frame buffers
  
  if (img->yuv_format != YUV400)
    free_mem3Dpel(imgUV_org_frm, 2);
}


/*!
************************************************************************
* \brief
*    Free allocated memory of frame size related global buffers
*    buffers are defined in global.h, allocated memory is allocated in
*    int get_mem4global_buffers()
* \par Input:
*    Input Parameters struct inp_par *inp,                             \n
*    Image Parameters struct img_par *img
* \par Output:
*    none
************************************************************************
*/
void free_global_buffers(void)
{
#ifdef _ADAPT_LAST_GROUP_
  extern int *last_P_no_frm;
  free (last_P_no_frm);
#endif
  
#if defined MB32X32
  free_mem2Dint(img->cofAC16x16);
  free_mem3Dint(img->cofAC16x8, 2);
  free_mem3Dint(img->cofAC8x16, 2);
#endif
  free_orig_planes();
  
  free(yPicPos);
  free(xPicPos);
  // Free Qmatrices and offsets
  free_QMatrix();
  free_QOffsets();

  if (input->WeightedPrediction || input->WeightedBiprediction || input->GenerateMultiplePPS)
  {
    free_mem3Dint(wp_weight,6);
    free_mem3Dint(wp_offset,6);
    free_mem4Dint(wbp_weight,6,MAX_REFERENCE_PICTURES);
    free_mem3Dint(explicit_weight,6);
  }
  
  if(stats->successive_Bframe!=0 || input->BRefPictures> 0)
  {
    free_mem3D     ((byte ***)direct_ref_idx,2);
    free_mem2Dshort(direct_pdir);
  } // end if B frame
  
  
  free_mem2Dint(img4Y_tmp);    // free temp quarter pel frame buffer
  
  // free mem, allocated in init_img()
  // free intra pred mode buffer for blocks
  free_mem2D((byte**)img->ipredmode);
  free_mem2D((byte**)img->ipredmode8x8);
  free(img->mb_data);
  
  free_mem2D((byte**)rddata_top_frame_mb.ipredmode);
  
  if(input->UseConstrainedIntraPred)
  {
    free (img->intra_block);
  }
  
  if (input->RestrictRef)
  {
    free(pixel_map[0]);
    free(pixel_map);
    free(refresh_map[0]);
    free(refresh_map);
  }
  
 
  free_mem3Dint(img->nz_coeff, img->FrameSizeInMbs);
  
  free_mem2Ddb_offset(img->lambda_md,img->bitdepth_luma_qp_scale);
  free_mem2Ddb_offset(img->lambda_me,img->bitdepth_luma_qp_scale);
  free_mem2Dint_offset(img->lambda_mf,img->bitdepth_luma_qp_scale);
  

  EPZSDelete();

#ifdef ADAPTIVE_LOOP_FILTER
  if(input->UseAdaptiveLoopFilter)
  {
    FreeALFGlobalBurrers();
  }
#endif

  if(img->BitDepthIncrease)
    free(Pel_12to8);
}

/*!
************************************************************************
* \brief
*    Allocate memory for mv
* \par Input:
*    Image Parameters struct img_par *img                             \n
*    int****** mv
* \return memory size in bytes
************************************************************************
*/
int get_mem_mv (short ******* mv)
{
  int i, j, k, l, m;
  
#ifdef MB32X32
  if ((*mv = (short******)calloc(4<<MAX_MB_EXT_LEVEL,sizeof(short*****))) == NULL)
    no_mem_exit ("get_mem_mv: mv");

  for (i=0; i<(4<<MAX_MB_EXT_LEVEL); i++)
  {
    if (((*mv)[i] = (short*****)calloc(4<<MAX_MB_EXT_LEVEL,sizeof(short****))) == NULL)
      no_mem_exit ("get_mem_mv: mv");

    for (j=0; j<(4<<MAX_MB_EXT_LEVEL); j++)
    {
      if (((*mv)[i][j] = (short****)calloc(2,sizeof(short***))) == NULL)
        no_mem_exit ("get_mem_mv: mv");
      for (k=0; k<2; k++)
      {
        if (((*mv)[i][j][k] = (short***)calloc(img->max_num_references,sizeof(short**))) == NULL)
          no_mem_exit ("get_mem_mv: mv");
        for (l=0; l<img->max_num_references; l++)
        {
          if (((*mv)[i][j][k][l] = (short**)calloc(9,sizeof(short*))) == NULL)
            no_mem_exit ("get_mem_mv: mv");
          if (((*mv)[i][j][k][l][0] = (short*)calloc(2*9,sizeof(short))) == NULL)
            no_mem_exit ("get_mem_mv: mv");
          for (m=1; m<9; m++)
            (*mv)[i][j][k][l][m] = (*mv)[i][j][k][l][m - 1] + 2;
        }
      }
    }
  }
  return (4<<MAX_MB_EXT_LEVEL)*(4<<MAX_MB_EXT_LEVEL)*img->max_num_references*9*2*sizeof(short);
#else

  if ((*mv = (short******)calloc(4,sizeof(short*****))) == NULL)
    no_mem_exit ("get_mem_mv: mv");
  for (i=0; i<4; i++)
  {
    if (((*mv)[i] = (short*****)calloc(4,sizeof(short****))) == NULL)
      no_mem_exit ("get_mem_mv: mv");
    for (j=0; j<4; j++)
    {
      if (((*mv)[i][j] = (short****)calloc(2,sizeof(short***))) == NULL)
        no_mem_exit ("get_mem_mv: mv");
      for (k=0; k<2; k++)
      {
        if (((*mv)[i][j][k] = (short***)calloc(img->max_num_references,sizeof(short**))) == NULL)
          no_mem_exit ("get_mem_mv: mv");
        for (l=0; l<img->max_num_references; l++)
        {
          if (((*mv)[i][j][k][l] = (short**)calloc(9,sizeof(short*))) == NULL)
            no_mem_exit ("get_mem_mv: mv");
          if (((*mv)[i][j][k][l][0] = (short*)calloc(2*9,sizeof(short))) == NULL)
            no_mem_exit ("get_mem_mv: mv");
          for (m=1; m<9; m++)
            (*mv)[i][j][k][l][m] = (*mv)[i][j][k][l][m - 1] + 2;
        }
      }
    }
  }
  return 4*4*img->max_num_references*9*2*sizeof(short);
#endif
}


/*!
************************************************************************
* \brief
*    Free memory from mv
* \par Input:
*    int****** mv
************************************************************************
*/
void free_mem_mv (short****** mv)
{
  int i, j, k, l;
#ifdef MB32X32  
  
  for (i=0; i<(4<<MAX_MB_EXT_LEVEL); i++)
  {
    for (j=0; j<(4<<MAX_MB_EXT_LEVEL); j++)
    {
      for (k=0; k<2; k++)
      {
        for (l=0; l<img->max_num_references; l++)
        {
          free (mv[i][j][k][l][0]);
          free (mv[i][j][k][l]);
        }
        free (mv[i][j][k]);
      }
      free (mv[i][j]);
    }
    free (mv[i]);
  }
  free (mv);

#else
  
  for (i=0; i<4; i++)
  {
    for (j=0; j<4; j++)
    {
      for (k=0; k<2; k++)
      {
        for (l=0; l<img->max_num_references; l++)
        {
          free (mv[i][j][k][l][0]);
          free (mv[i][j][k][l]);
        }
        free (mv[i][j][k]);
      }
      free (mv[i][j]);
    }
    free (mv[i]);
  }
  
  free (mv);
#endif
}


/*!
************************************************************************
* \brief
*    Allocate memory for AC coefficients
************************************************************************
*/
int get_mem_ACcoeff (int***** cofAC)
{
  int i, j, k;
  int num_blk8x8 = 4 + img->num_blk8x8_uv;
  
  if ((*cofAC = (int****)calloc (num_blk8x8, sizeof(int***))) == NULL)              no_mem_exit ("get_mem_ACcoeff: cofAC");
  for (k=0; k<num_blk8x8; k++)
  {
    if (((*cofAC)[k] = (int***)calloc (4, sizeof(int**))) == NULL)         no_mem_exit ("get_mem_ACcoeff: cofAC");
    for (j=0; j<4; j++)
    {
      if (((*cofAC)[k][j] = (int**)calloc (2, sizeof(int*))) == NULL)      no_mem_exit ("get_mem_ACcoeff: cofAC");
      for (i=0; i<2; i++)
      {
        if (((*cofAC)[k][j][i] = (int*)calloc (65, sizeof(int))) == NULL)  no_mem_exit ("get_mem_ACcoeff: cofAC"); // 18->65 for ABT
      }
    }
  }
  return num_blk8x8*4*2*65*sizeof(int);// 18->65 for ABT
}

/*!
************************************************************************
* \brief
*    Allocate memory for DC coefficients
************************************************************************
*/
int get_mem_DCcoeff (int**** cofDC)
{
  int j, k;
  
  if ((*cofDC = (int***)calloc (3, sizeof(int**))) == NULL)           no_mem_exit ("get_mem_DCcoeff: cofDC");
  for (k=0; k<3; k++)
  {
    if (((*cofDC)[k] = (int**)calloc (2, sizeof(int*))) == NULL)      no_mem_exit ("get_mem_DCcoeff: cofDC");
    for (j=0; j<2; j++)
    {
      if (((*cofDC)[k][j] = (int*)calloc (65, sizeof(int))) == NULL)  no_mem_exit ("get_mem_DCcoeff: cofDC"); // 18->65 for ABT
    }
  }
  return 3*2*65*sizeof(int); // 18->65 for ABT
}


/*!
************************************************************************
* \brief
*    Free memory of AC coefficients
************************************************************************
*/
void free_mem_ACcoeff (int**** cofAC)
{
  int i, j, k;
  
  for (k=0; k<4+img->num_blk8x8_uv; k++)
  {
    for (i=0; i<4; i++)
    {
      for (j=0; j<2; j++)
      {
        free (cofAC[k][i][j]);
      }
      free (cofAC[k][i]);
    }
    free (cofAC[k]);
  }
  free (cofAC);
}

/*!
************************************************************************
* \brief
*    Free memory of DC coefficients
************************************************************************
*/
void free_mem_DCcoeff (int*** cofDC)
{
  int i, j;
  
  for (j=0; j<3; j++)
  {
    for (i=0; i<2; i++)
    {
      free (cofDC[j][i]);
    }
    free (cofDC[j]);
  }
  free (cofDC);
}

/*!
************************************************************************
* \brief
*    Do some initialization work for encoding the 2nd IGOP
************************************************************************
*/
void process_2nd_IGOP(void)
{
  Boolean FirstIGOPFinished = FALSE;
  if ( img->number == input->no_frames-1 )
    FirstIGOPFinished = TRUE;
  if (input->NumFrameIn2ndIGOP==0) return;
  if (!FirstIGOPFinished || In2ndIGOP) return;
  In2ndIGOP = TRUE;
  
  //  img->number = -1;
  start_frame_no_in_this_IGOP = input->no_frames;
  start_tr_in_this_IGOP = (input->no_frames-1)*(input->jumpd+1) +1;
  input->no_frames = input->no_frames + input->NumFrameIn2ndIGOP;
  
  /*  reset_buffers();
  
    frm->picbuf_short[0]->used=0;
    frm->picbuf_short[0]->picID=-1;
    frm->picbuf_short[0]->lt_picID=-1;
  frm->short_used = 0; */
}


/*!
************************************************************************
* \brief
*    Set the image type for I,P and SP pictures (not B!)
************************************************************************
*/
void SetImgType(void)
{
#if AdaptiveCADENCE
    int intra_refresh;
  if(input->HierB_AdaptiveCadence)
  {
    if(img->number)
      intra_refresh = input->intra_period == 0 ? (img->number == 0) : (((img->number + input->jumpd)%input->intra_period) == 0);
    else
      intra_refresh = 1;
  }
  else
    intra_refresh = input->intra_period == 0 ? (IMG_NUMBER == 0) : ((IMG_NUMBER%input->intra_period) == 0);
#else
  int intra_refresh = input->intra_period == 0 ? (IMG_NUMBER == 0) : ((IMG_NUMBER%input->intra_period) == 0);
#endif
  
  if (intra_refresh)
  {
    img->type = I_SLICE;        // set image type for first image to I-frame
  }
  else
  {
    img->type = (input->BRefPictures == 2) ? B_SLICE : P_SLICE;      
  }  
}


void SetLevelIndices(void)
{
  switch(active_sps->level_idc)
  {
  case 10:         
    img->LevelIndex=0;
    break;
  case 11:         
    if (active_sps->constrained_set3_flag == 0)
      img->LevelIndex=2;
    else
      img->LevelIndex=1;
    break;
  case 12:         
    img->LevelIndex=3;
    break;
  case 13:         
    img->LevelIndex=4;
    break;
  case 20:         
    img->LevelIndex=5;
    break;
  case 21:         
    img->LevelIndex=6;
    break;
  case 22:         
    img->LevelIndex=7;
    break;
  case 30:         
    img->LevelIndex=8;
    break;
  case 31:         
    img->LevelIndex=9;
    break;
  case 32:         
    img->LevelIndex=10;
    break;
  case 40:         
    img->LevelIndex=11;
    break;
  case 41:         
    img->LevelIndex=12;
    break;
  case 42:         
    if (active_sps->profile_idc <= 88)
      img->LevelIndex=13;
    else
      img->LevelIndex=14;
    break;
  case 50:         
    img->LevelIndex=15;
    break;
  case 51:
  default:
    img->LevelIndex=16;
    break;
  }
}

void encode_sequence()
{
  int primary_disp = 0;
#if HIER_P
  int loop,j;
#endif
#if AdaptiveCADENCE
  int new_HierB_GOP, orig_HierB_GOP, th, lastGOP=0;
  orig_HierB_GOP = input->successive_Bframe;
  th = orig_HierB_GOP >> 1;
//  input->no_frames = ((input->no_frames-1)*(input->successive_Bframe+1))+1;
#endif

  for (img->number=0; img->number < input->no_frames; img->number++)
  {
#if AdaptiveCADENCE
  if(input->HierB_AdaptiveCadence)
  {
    if((input->no_frames-img->number-1) <= orig_HierB_GOP)
      lastGOP=(input->no_frames-img->number)-1;
    if(IMG_NUMBER && input->HierarchicalCoding==3)
    {
      if(lastGOP)
      {
        printf("                        Last GOP : %1d \n",lastGOP);
        create_adaptive_hierarchy(lastGOP);
      }
      else
      {
        if(input->intra_period)
        {
          new_HierB_GOP = input->intra_period - (img->number%input->intra_period);
          //if(orig_HierB_GOP-th <= new_HierB_GOP &&  new_HierB_GOP <= orig_HierB_GOP+th+1)
          if(new_HierB_GOP <= orig_HierB_GOP)
          {
            printf("                             GOP : %1d \n",new_HierB_GOP);
            create_adaptive_hierarchy(new_HierB_GOP);
          }
          else
            create_adaptive_hierarchy(orig_HierB_GOP);
        }
        else
          create_adaptive_hierarchy(orig_HierB_GOP);
      }
    }
  }
#endif

#if HIER_P
  //all even numbered frames are KeyP.
  img->KeyP = ((img->number)%(1<<input->HierP_Level)) ? 0 : 1 ;
  //========P0 P3 P2 P3 P1 P3 P2 P3 P0===================
  if(input->HierP_Level)
  {
    img->Q_Level=input->HierP_Level;
    loop=1<<(input->HierP_Level);
    if(img->number%loop==0)
      img->Q_Level=0;
    else
    {
      if((img->number%loop)%2 !=0)
        img->Q_Level=input->HierP_Level;
      else
      {
        j=0;
        while(loop>1)
        {
          loop /= 2;
          j++;
          if(img->number%loop==0)
          {
            img->Q_Level=j;
            break;
          }
        }
      }
    }
  }
  //===========================
#endif

    //img->nal_reference_idc = 1;
    if (input->intra_period)
      img->nal_reference_idc = ((IMG_NUMBER % input->intra_period) && input->DisposableP) ? (img->number + 1)% 2 : 1;
    else
      img->nal_reference_idc = (img->number && input->DisposableP) ? (img->number + 1)% 2 : 1;
    
    //much of this can go in init_frame()
    //poc for this frame 
#if AdaptiveCADENCE
    if(input->HierB_AdaptiveCadence)
    {
      if(img->number)
        img->framepoc = 2 * (img->number + input->jumpd); 
    }
    else
      img->framepoc = (input->intra_period && input->idr_enable ? IMG_NUMBER % input->intra_period : IMG_NUMBER) * (2*(input->jumpd+1)); 
#else
    img->framepoc = (input->intra_period && input->idr_enable ? IMG_NUMBER % input->intra_period : IMG_NUMBER) * (2*(input->jumpd+1)); 
#endif
    
    //frame_num for this frame
    //if (input->BRefPictures== 0 || input->successive_Bframe == 0 || img-> number < 2)
    if ((input->BRefPictures != 1 &&  input->HierarchicalCoding == 0) || input->successive_Bframe == 0 || img-> number < 2)// ||  input->HierarchicalCoding == 0)
    {
      if (input->intra_period && input->idr_enable)
      {
        img->frame_num =  ((IMG_NUMBER - primary_disp)  % input->intra_period ) % (1 << (log2_max_frame_num_minus4 + 4)); 
        if (IMG_NUMBER % input->intra_period  == 0)
        {
          img->frame_num = 0;
          primary_disp   = 0;
        }
      }
      else
        img->frame_num = (IMG_NUMBER - primary_disp) % (1 << (log2_max_frame_num_minus4 + 4)); 
    }
    else 
    {
#if AdaptiveCADENCE
      if(input->HierB_AdaptiveCadence)
      {
          if(last_coded_frame_was_reference)
            img->frame_num ++;
      }
#endif
      if (input->intra_period && input->idr_enable)
      {
        if (0== (img->number % input->intra_period))
        {
          img->frame_num=0;
          primary_disp   = 0;
        }
      }
      img->frame_num %= (1 << (log2_max_frame_num_minus4 + 4)); 
    }
    
    //the following is sent in the slice header
    img->delta_pic_order_cnt[0]=0;
    
    if (input->BRefPictures == 1)
    {
      if (img->number)
      {
        img->delta_pic_order_cnt[0]=+2 * input->successive_Bframe;
      }
    }
    
    SetImgType();
   
#ifdef _ADAPT_LAST_GROUP_
    if (input->successive_Bframe && input->last_frame && IMG_NUMBER+1 == input->no_frames)
    {                                           
      int bi = (int)((float)(input->jumpd+1)/(input->successive_Bframe+1.0)+0.499999);
      
      input->successive_Bframe = (input->last_frame-(img->number-1)*(input->jumpd+1))/bi-1;
      
      //about to code the last ref frame, adjust delta poc         
      img->delta_pic_order_cnt[0]= -2*(initial_Bframes - input->successive_Bframe);
      img->framepoc += img->delta_pic_order_cnt[0];
    }
#endif
    
    // which layer the image belonged to?
    if ( IMG_NUMBER % (input->NumFramesInELSubSeq+1) == 0 )
      img->layer = 0;
    else
      img->layer = 1;
    
    encode_one_frame(); // encode one I- or P-frame
    
    if (img->type == I_SLICE && input->EnableOpenGOP)
      img->last_valid_reference = img->ThisPOC;

    if (input->ReportFrameStats)
      report_frame_statistic();
    
    if (img->nal_reference_idc == 0)
    {
      primary_disp ++;
      img->frame_num -= 1;
      img->frame_num %= (1 << (log2_max_frame_num_minus4 + 4));
    }    

    encode_enhancement_layer();
#if AdaptiveCADENCE
  if(input->HierB_AdaptiveCadence)
  {
    if(img->number)
      img->number += input->successive_Bframe;
    last_coded_frame_was_reference = (input->successive_Bframe)? 0 : 1;   //if any B frame is coded, then the last coded B frame will always be nonref
  }
#endif
    
    process_2nd_IGOP();
  }//for loop ends
}

#ifdef ADAPTIVE_GOP
/*! 
*************************************************************************************
*   Based on decision by detectFADE, this function actually converts current P frame to B frame
*   and picks the next P frame and code it first and then encode the newly converted P->B frame
*    - Rahul Panchal                      <rpanchal@qualcomm.com>
*************************************************************************************
*/
void encode_sequence_adaptiveGOP()
{
  int step = (input->jumpd+1);

  for (img->number=0; img->number < input->no_frames; img->number++)
  {
    isAdapGOP=0;
    img->type         = (img->number==0) ? 2:0;
    FrameNumberInFile = img->number * step;

    if(img->number > 1 && img->number < input->no_frames-2)   //first and last GOP must be 8...
    {
      isAdapGOP = detectFADE(step);
    }

    //========Encode P========    
    if(isAdapGOP)
    {
      FrameNumberInFile = FrameNumberInFile + step;
      img->type         = P_SLICE;
    }
    frame_no          = FrameNumberInFile;
    img->framepoc     = FrameNumberInFile * 2;

    //frame_num for this frame
    if (img-> number < 2)
      img->frame_num = (IMG_NUMBER) % (1 << (log2_max_frame_num_minus4 + 4)); 
    else 
      img->frame_num %= (1 << (log2_max_frame_num_minus4 + 4)); 

    img->nal_reference_idc = 1;
    img->layer = 0;

    encode_one_frame();               // encode one I- or P-frame

    if (img->type == I_SLICE && input->EnableOpenGOP)
      img->last_valid_reference = img->ThisPOC;

    last_coded_frame_was_reference = 1;
    //========Encode B========
    if(isAdapGOP)
    {
      img->number++;
      FrameNumberInFile = FrameNumberInFile - step;
      img->type         = B_SLICE;

      frame_no          = FrameNumberInFile;
      img->framepoc     = FrameNumberInFile * 2;

      img->frame_num++;
      img->frame_num %= (1 << (log2_max_frame_num_minus4 + 4));

      img->nal_reference_idc = 1;
      img->layer = 0;

      encode_one_frame();               // encode one B-frame with lambda and QP of P frame
    if (img->type == I_SLICE && input->EnableOpenGOP)
      img->last_valid_reference = img->ThisPOC;

      last_coded_frame_was_reference = 1;
      img->number--;
      isAdapGOP = 0;                    //set to zero so that HierB uses normal lambda and QP and not that of P frame
      encode_enhancement_layer();       //encode hiearchial frames
      img->number++;
      last_coded_frame_was_reference = 0;
    }
    encode_enhancement_layer();         //encode hiearchial frames
  }//for loop ends
}

/*! 
*************************************************************************************
*   Detects if there is fading and converts P to B slice
*    - Rahul Panchal                      <rpanchal@qualcomm.com>
*************************************************************************************
*/
short detectFADE(int step)
{

  int i,j;
  int64 diff23=0,diff13=0,diff21=0,diff=0;
  double alpha=0;

  imgpel **imgY1,**imgY2,**imgY3, imgY22;
  get_mem2Dpel(&imgY1, img->height, img->width);
  get_mem2Dpel(&imgY2, img->height, img->width);
  get_mem2Dpel(&imgY3, img->height, img->width);

  ReadOneFrame (FrameNumberInFile - step, input->infile_header, input->img_width, input->img_height, input->img_width_cr, input->img_height_cr);
  for (i = 0; i < img->height; i++)
    memcpy(imgY1[i], imgY_org_frm[i], img->width * sizeof(imgpel));
  ReadOneFrame (FrameNumberInFile    , input->infile_header, input->img_width, input->img_height, input->img_width_cr, input->img_height_cr);
  for (i = 0; i < img->height; i++)
    memcpy(imgY2[i], imgY_org_frm[i], img->width * sizeof(imgpel));
  ReadOneFrame (FrameNumberInFile + step, input->infile_header, input->img_width, input->img_height, input->img_width_cr, input->img_height_cr);
  for (i = 0; i < img->height; i++)
    memcpy(imgY3[i], imgY_org_frm[i], img->width * sizeof(imgpel));

  
  //method to calculate "alpha" will be improved in future...not so important right now

  for (i = 0; i < img->height; i++)
  {
    for (j = 0; j < img->width; j++)
    {
      diff23 += abs(imgY2[i][j]-imgY3[i][j]);
      diff21 += abs(imgY2[i][j]-imgY1[i][j]);
      diff13 += abs(imgY1[i][j]-imgY3[i][j]);
    }
  }
  if(diff13!=0)
    alpha = (double)diff23/diff13;

  for (i = 0; i < img->height; i++)
  {
    for (j = 0; j < img->width; j++)
    {
      imgY22 = (imgpel)(alpha*imgY1[i][j] + (1-alpha)*imgY3[i][j]);
      diff += abs(imgY2[i][j]-imgY22);
    }
  }

  isAdapGOP = 0;
  if((diff*input->AdaptiveGOP_threshold) < min(diff21,diff23))    //   2*diff < min(diff13,diff23)
    isAdapGOP = 1;

  free_mem2Dpel(imgY1);  
  free_mem2Dpel(imgY2);  
  free_mem2Dpel(imgY3);  

  return isAdapGOP;
}
#endif
