
/*!
 ************************************************************************
 * \file refbuf.c
 *
 * \brief
 *    Declarations of teh reference frame buffer types and functions
 ************************************************************************
 */

#include <stdlib.h>
#include <memory.h>
#include "global.h"

#include "refbuf.h"

#ifdef EIGHTH_PEL_DIRECT
#include "mbuffer.h"
#endif

/*!
 ************************************************************************
 * \brief
 *    Reference buffer write routines
 ************************************************************************
 */
void PutPel_14 (pel_t **Pic, int y, int x, pel_t val)
{
  Pic [y][x] = val;
}

void PutPel_11 (pel_t *Pic, int pel_pos, pel_t val)
{
  Pic [pel_pos] = val;
}

/*!
 ************************************************************************
 * \note
 *    The following functions returning line are NOT reentrant!  Use a buffer
 *    provided by the caller to change that (but it costs a memcpy()...
 ************************************************************************
 */
static pel_t line[16];

#ifdef EIGHTH_PEL
int i, maxx, xfloor, yfloor, xx, yy, xq, yq;
static int buf_switch;
pel_t *out;
#endif
pel_t *FastLine16Y_11 (pel_t *Pic, int y, int x, int height, int width)
{
  return &Pic [y*width+x];
}


pel_t *UMVLine16Y_11 (pel_t *Pic, int y, int x, int height, int width)
{
  int i, maxx;
  pel_t *Picy;

  Picy = &Pic [max(0,min(height-1,y)) * width];

  if (x < 0) {                    // Left edge ?

    maxx = min(0,x+16);
    for (i = x; i < maxx; i++)
      line[i-x] = Picy [0];       // Replicate left edge pixel

    maxx = x+16;
    for (i = 0; i < maxx; i++)    // Copy non-edge pixels
      line[i-x] = Picy [i];
  }
  else if (x > width-16)  {  // Right edge ?

    maxx = width;
    for (i = x; i < maxx; i++)
      line[i-x] = Picy [i];       // Copy non-edge pixels

    maxx = x+16;
    for (i = max(width,x); i < maxx; i++)
      line[i-x] = Picy [width-1];  // Replicate right edge pixel
  }
  else                            // No edge
    return &Picy [x];

  return line;
}


pel_t *FastLineX (int dummy, pel_t* Pic, int y, int x, int height, int width)
{
  return Pic + y*width + x;
}

static pel_t umv_line[2][16];

pel_t *UMVLineX_impl (int size, pel_t* Pic, int y, int x, int height, int width, int line_idx)
{
  int i;
  pel_t *Picy;

  Picy = Pic + max(0,min(height-1,y)) * width;

  if (x < 0)                            // Left edge
  {   
    for (i = x; i < min(0,x+size); i++)
    {
      umv_line[line_idx][i-x] = Picy [0];             // Replicate left edge pixel
    }

    memcpy(&umv_line[line_idx][min(-x,15)],Picy,max(x+size,0) * sizeof(pel_t)); // Copy non-edge pixels
  }
  else if (x > width-size)         // Right edge
  {
    memcpy(umv_line[line_idx],&Picy[x], max((width - x),0) * sizeof(pel_t)); // Copy non-edge pixels
    for (i = max(width,x); i < x+size; i++)
    {
      umv_line[line_idx][i-x] = Picy [width-1];  // Replicate right edge pixel
    }
  }
  else                                  // No edge
  {
    return Picy + x;
  }

  return umv_line[line_idx];
}

pel_t *UMVLineX (int size, pel_t* Pic, int y, int x, int height, int width)
{
  return UMVLineX_impl (size, Pic, y, x, height, width, 0);
}

pel_t *UMVLineX2 (int size, pel_t* Pic, int y, int x, int height, int width)
{
  return UMVLineX_impl (size, Pic, y, x, height, width, 1);
}


/*!
 ************************************************************************
 * \brief
 *    Reference buffer, 1/4 pel
 ************************************************************************
 */
pel_t UMVPelY_14 (pel_t **Pic, int y, int x, int height4, int width4)
{
  int ypos = (y < 0 ? y&3 : (y >  height4 ? height4+(y&3) : y));
  int xpos = (x < 0 ? x&3 : (x >  width4 ? width4+(x&3) : x));

  return Pic [ypos][xpos];
}

pel_t FastPelY_14 (pel_t **Pic, int y, int x, int height, int width)
{
  return Pic [y][x];
}



pel_t *FastLine4X (pel_t **Pic, int y, int x, int height, int width)
{
  return &Pic [y][x];
}

static pel_t line4[2][64];
#ifdef EIGHTH_PEL
static pel_t line41[64];
#endif

/*!
 ************************************************************************
 * \brief
 *    Reference buffer, 1/4 pel
 ************************************************************************
 */
pel_t *UMVLine4X_impl (pel_t **Pic, int y, int x, int height4, int width4, int line_idx)
{
  int i, xx;

  int xpos, ypos = (y < 0 ? y&3 : (y >  height4 ? height4+(y&3) : y));

  for (i=0; i< 64; i+=4)
  {
    xx = x + i;
    xpos = (xx < 0 ? (xx) &3 : ((xx) >  width4 ? width4+((xx) &3) : (xx) ));
    line4[line_idx][i] = Pic [ypos][xpos];
  }
  return line4[line_idx];
}

pel_t *UMVLine4X (pel_t **Pic, int y, int x, int height4, int width4)
{
  return UMVLine4X_impl (Pic, y, x, height4, width4, 0);
}

pel_t *UMVLine4X2 (pel_t **Pic, int y, int x, int height4, int width4)
{
  return UMVLine4X_impl (Pic, y, x, height4, width4, 1);
}



#ifdef EIGHTH_PEL

pel_t *UMVLine4X8 (pel_t **Pic, int y, int x, int height4, int width4)
{
  /* Maximum values (padding included) */
  int maxx8 = ( width4 - 4 ) << 1;
  int maxy8 = (height4 - 4 ) << 1;

  // quick hack to use two different buffers for b-frames
  buf_switch = ( buf_switch == 0 ) ? 16 : 0;
  out = line41 + buf_switch;


  if (y < 0)
    yfloor = (y&7) >> 1;
  else if (y > maxy8)
    yfloor = ( maxy8 + (y&7)) >> 1;
  else
    yfloor = y>>1;

  yy = y&1;
  xx = x&1;

  for( i=0; i < 4; i++, out+=4, x+=8)
  {
    if (x < 0)
      xfloor = (x&7) >> 1;
    else if (x > maxx8)
      xfloor = (maxx8 + (x&7)) >> 1;
    else
      xfloor = x>>1;

    if( xx && yy ) // are x and y on an eighth-pel position?
      *out = ( Pic[yfloor][xfloor] + Pic[yfloor  ][xfloor+1] + Pic[yfloor+1][xfloor] + Pic[yfloor+1][xfloor+1]) >> 2;
    else if( xx ) // only x is on an eighth-pel position
      *out = ( Pic[yfloor][xfloor] + Pic[yfloor  ][xfloor+1] ) >> 1;
    else if( yy ) // only y is on an eighth-pel position
      *out = ( Pic[yfloor][xfloor] + Pic[yfloor+1][xfloor  ] ) >> 1;
    else // quarter-pel position
      *out = Pic[yfloor][xfloor];
  }
  return out-16;
}

pel_t *FastLine4X8 (pel_t **Pic, int y, int x, int height4, int width4)
{
  xfloor = x>>1; yfloor = y>>1; // quarter-pel units
  xx = x&1; yy = y&1; // is eighth-pel position?

  xq = x&7; // eighth-pel sub position
  yq = y&7;

  // quick hack to use 2 different buffers for b-frames
  buf_switch = ( buf_switch == 0 ) ? 16 : 0;
  out = line41 + buf_switch;

  for( i=0; i < 4; i++, out+=4, x+=8, xfloor+=4)
  {
    if( xx && yy ) // are x and y on an eighth-pel position?
      *out = ( Pic[yfloor][xfloor] + Pic[yfloor  ][xfloor+1] + Pic[yfloor+1][xfloor] + Pic[yfloor+1][xfloor+1]) >> 2;
    else if( xx ) // only x is on an eighth-pel position
      *out = ( Pic[yfloor][xfloor] + Pic[yfloor  ][xfloor+1] ) >> 1;
    else if( yy ) // only y is on an eighth-pel position
      *out = ( Pic[yfloor][xfloor] + Pic[yfloor+1][xfloor  ] ) >> 1;
    else // quarter-pel position
      *out = Pic[yfloor][xfloor];
  }
  return out-16;
}

#ifdef EIGHTH_PEL_DIRECT

static void interp_block_direct_X(StorablePicture *ref_picture, int pres_x, int pres_y, const int *coefx, int max_x, int max_y, int block[16][16], int size_x, int size_y, int symmetric)
{
  int i, j, x;
  int result;

  const int *filtP;
  imgpel *imgP;

  // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
  if ((pres_y >= 0) && (pres_y < max_y - size_y) && (pres_x >= 3) && (pres_x < max_x - 3 - size_x))
  { //Image Non-Border safe positions...so dont need Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefx;
          imgP = &(ref_picture->imgY_filt[pres_y+j][pres_x+i-3]);
          for (result = 0, x = -3; x < 5; x++)
            result += (*filtP++) * (*imgP++);
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefx;
          for (result = 0, x = -3; x <= 0; x++)
            result += (*filtP++) * (ref_picture->imgY_filt[pres_y+j][pres_x+i+x] +
                                    ref_picture->imgY_filt[pres_y+j][pres_x+i-x+1]);
          block[j][i] = result;
        }
    }
  }
  else
  { //Image Border positions...so needs Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          for (result = 0, x = -3; x < 5; x++)
            result += ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * coefx[x+3];
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          for (result = 0, x = -3; x <= 0; x++)
            result += coefx[x+3] * (ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
                                    ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);

          block[j][i] = result;
        }
    }
  }
}

static void interp_block_direct_Y(StorablePicture *ref_picture, int pres_x, int pres_y, const int *coefy, int max_x, int max_y, int block[16][16], int size_x,int size_y, int symmetric)
{
  int i, j, y;
  int result;

  const int *filtP;

  // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
  if ((pres_y >= 3) && (pres_y < max_y - 3 - size_y) && (pres_x >= 0) && (pres_x < max_x - size_x))
  { //Image Non-Border safe positions...so dont need Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -3; y < 5; y++)
            result += ref_picture->imgY_filt[pres_y+j+y][pres_x+i] * (*filtP++);
          block[j][i] = result;

        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -3; y <=0; y++)
            result += (*filtP++) * (ref_picture->imgY_filt[pres_y+j+y][pres_x+i] +
                                    ref_picture->imgY_filt[pres_y+j-y+1][pres_x+i]);
          block[j][i] = result;
        }
    }
  }
  else
  { //Image Border positions...so needs Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -3; y < 5; y++)
            result += ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] * (*filtP++);
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -3; y <= 0; y++)
            result += (*filtP++) * (ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
                                    ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
          block[j][i] = result;
        }
    }
  }

}


void get_UMVLine4X8_direct(StorablePicture *ref_picture, int x_pos, int y_pos, int block[16][16], int size_x, int size_y)
{
  int i, j, x, y;

  static const int COEF[3][8] =
  {
    { -3, 12, -37, 229,  71, -21,  6, -1 },
    { -3, 12, -39, 158, 158, -39, 12, -3 },
    { -1,  6, -21,  71, 229, -37, 12, -3 }
  };


  static const int COEF_EIGHTH_PEL[4][8] =
  {
    { -3, 12, -37, 485,  71, -21,  6, -1 },
    { -6, 24, -76, 387, 229, -60, 18, -4 },
    { -4, 18, -60, 229, 387, -76, 24, -6 },
    { -1,  6, -21,  71, 485, -37, 12, -3 }
  };



  int round8 = 128;
  int round9 = 256;
  int round16 = (1<<15);
  int round17 = (1<<16);
  int round18 = (1<<17);

  int dx = x_pos & 7;
  int dy = y_pos & 7;

  int pres_x = x_pos >> 3;
  int pres_y = y_pos >> 3;

  int max_x = ref_picture->size_x - 1;
  int max_y = ref_picture->size_y - 1;

  int tmp[16][16+8];
  int result;

  int filt_index;
  const int *filtP;
  imgpel *imgP;
  int *intP;

  // This function is called only if either dx or dy is odd (1/8th pel). So not all of the possibilities are covered over here

  if (dy == 0)
  { // horizontal interpolation only (1/8-pel)
    // dx has to be in 1/8-pel location (dx&1)
    interp_block_direct_X(ref_picture, pres_x, pres_y, COEF_EIGHTH_PEL[dx>>1], max_x, max_y, block, size_x, size_y, 0);
    for (j = 0; j < size_y; j++)
      for (i = 0; i < size_x; i++)
        block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round9) >> 9); // Filter sums up to 512 for 1/8 pel

  }
  else if (dx == 0)
  { // vertical interpolation only (1/8-th pel)
    // dy has to be an 8th pel location (dx&1)
    interp_block_direct_Y(ref_picture, pres_x, pres_y, COEF_EIGHTH_PEL[dy>>1], max_x, max_y, block, size_x, size_y, 0);
    for (j = 0; j < size_y; j++)
      for (i = 0; i < size_x; i++)
        block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round9) >> 9); // Filter sums up to 512 for 1/8 pel
  }
  else if (!(dx&1))
  {  // horizontal 1/4-pel, vertical 1/8-th pel
    filt_index = (dx>>1)-1;
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so dont need Clipping
      if (filt_index != 1)
      {

        for (j = -3; j < (size_y+5); j++)
          for (i = 0; i < size_x; i++)
          {
            imgP = &(ref_picture->imgY_filt[pres_y+j][pres_x+i-3]);
            filtP = COEF[filt_index];
            for (tmp[i][j+3] = 0, x = -3; x < 5; x++)
              tmp[i][j+3] += (*imgP++) * (*filtP++); // indices (i,j) reversed to facilitate further filtering
          }

      }
      else
      {

        for (j = -3; j < (size_y+5); j++)
          for (i = 0; i < size_x; i++)
          {
            filtP = COEF[filt_index];
            for (tmp[i][j+3] = 0, x = -3; x <= 0; x++)
              tmp[i][j+3] += (*filtP++) * (ref_picture->imgY_filt[pres_y+j][pres_x+i+x] + ref_picture->imgY_filt[pres_y+j][pres_x+i-x+1]);
                // indices (i,j) reversed to facilitate further filtering
          }

      }
    }
    else
    {//Image Border positions...so needs Clipping
      if (filt_index != 1)
      {

        for (j = -3; j < (size_y+5); j++)
          for (i = 0; i < size_x; i++)
          {

            for (tmp[i][j+3] = 0, x = -3; x < 5; x++)
              tmp[i][j+3] += ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j )][Clip3(0, max_x, pres_x+i+x )]*COEF[filt_index][x+3];
                // indices (i,j) reversed to facilitate further filtering
          }

      }
      else
      {

        for (j = -3; j < (size_y+5); j++)
          for (i = 0; i < size_x; i++)
          {
            for (tmp[i][j+3] = 0, x = -3; x <= 0; x++)
              tmp[i][j+3] += COEF[filt_index][3+x] * (ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
                                                      ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);
                // indices (i,j) reversed to facilitate further filtering
          }

      }
    }
    // At this point the nominal bit-depth has been increased by 8 (1/4-pel filtering, Filter sums up to 256).

    // Vertical 1/8-pel (dx&1).
    filt_index = (dy>>1);
    for (j = 0; j < size_y; j++)
      for (i = 0; i < size_x; i++)
      {

        intP = &(tmp[i][j]);
        filtP = COEF_EIGHTH_PEL[filt_index];
        for (result = 0, y = -3; y < 5; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round17) >> 17);
        // 1/8-pel filtering, Filter sums up to 512. Total nominal bit-depth increased by 17.

      }
  }
  else if (!(dy&1))
  { // vertical 1/4-pel and horizontal 1/8-pel interpolation
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so don't need Clipping
      filt_index = (dy>>1)-1;
      if (filt_index != -1)
      {

        for (j = 0; j < size_y; j++) // y
          for (i = -3; i < size_x+5; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF[filt_index];
            for (tmp[j][i+3] = 0, y = -3; y < 5; y++)
              tmp[j][i+3] += (*filtP++) * ref_picture->imgY_filt[pres_y+j+y][pres_x+i];
          }

      }
      else
      {

        for (j = 0; j < size_y; j++) // y
          for (i = -3; i < size_x+5; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF[filt_index];
            for (tmp[j][i+3] = 0, y = -3; y <= 0; y++)
              tmp[j][i+3] += (*filtP++) * (ref_picture->imgY_filt[pres_y+j+y][pres_x+i] +
                                           ref_picture->imgY_filt[pres_y+j-y+1][pres_x+i]);
          }

      }

    }
    else
    { //Image Border positions...so needs Clipping
      filt_index = (dy>>1)-1;
      if (filt_index != -1)
      {

        for (j = 0; j < size_y; j++) // y
          for (i = -3; i < size_x+5; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF[filt_index];

            for (tmp[j][i+3] = 0, y = -3; y < 5; y++)
              tmp[j][i+3] += (*filtP++) * ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)];
          }

      }
      else
      {

        for (j = 0; j < size_y; j++) // y
          for (i = -3; i < size_x + 5; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF[filt_index];

            for (tmp[j][i+3] = 0, y = -3; y <= 0; y++)
              tmp[j][i+3] += (*filtP++) * (ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
                                           ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
          }


      }
    }
    // At this point the nominal bit-depth has been increased by 8 (1/4-pel filtering, Filter sums up to 256).

    // horizontal 1/8 pel position filtering
    filt_index = (dx>>1);
    for (j = 0; j < size_y; j++)
      for (i = 0; i < size_x; i++)
      {

        intP = &(tmp[j][i]);
        filtP = COEF_EIGHTH_PEL[filt_index];
        for (result = 0, y = -3; y < 5; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round17) >> 17);
          // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.

      }
  }       
  else
  { // 1/8-1/8 positions
    // Horizontal 1/8 pel position filtering
    filt_index = (dx>>1);
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so dont need Clipping

      for (j = -3; j < (size_y+5); j++)
        for (i = 0; i < size_x; i++)
        {
          imgP = &(ref_picture->imgY_filt[pres_y+j][pres_x+i-3]);
          filtP = COEF_EIGHTH_PEL[filt_index];
          for (tmp[i][j+3] = 0, x = -3; x < 5; x++)
            tmp[i][j+3] += (*imgP++) * (*filtP++);
        }

    }
    else
    {//Image Border positions...so needs Clipping

      for (j = -3; j < (size_y+5); j++)
        for (i = 0; i < size_x; i++)
        {
          for (tmp[i][j+3] = 0, x = -3; x < 5; x++)
            tmp[i][j+3] += ref_picture->imgY_filt[Clip3(0, max_y, pres_y+j )][Clip3(0, max_x, pres_x+i+x )]*COEF_EIGHTH_PEL[filt_index][x+3];
        }

    }
    // At this point the nominal bit-depth has been increased by 9 (1/8-pel filtering, Filter sums up to 512).

    // vertical 1/8 pel position filtering
    filt_index = (dy>>1);
    for (j = 0; j < size_y; j++)
      for (i = 0; i < size_x; i++)
      {

        intP = &(tmp[i][j]);
        filtP = COEF_EIGHTH_PEL[filt_index];
        for (result = 0, y = -3; y < 5; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round18) >> 18);
      }

  }
}

#endif // EIGHTH_PEL_DIRECT

#endif



