/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <memory.h>

#include "global.h"
#include "mb_access.h"

static int FirstMBInSlice[MAXSLICEGROUPIDS];
unsigned int PicSizeInMapUnits;


/*!
************************************************************************
* \brief
*    StartPicture: initializes FMO at the begin of each new picture
*
* \par Input:
*    None
************************************************************************
*/
int StartPicture ()
{
  int i;
  for (i=0; i<MAXSLICEGROUPIDS; i++)
    FirstMBInSlice[i] = 0;
  return 0;
}

/*!
************************************************************************
* \brief
*    FmoGetNextMBBr: Returns the MB-Nr (in scan order) of the next
*    MB in the (FMO) Slice, -1 if the SliceGroup is finished
*
* \par Input:
*    CurrentMbNr
************************************************************************
*/
int GetNextMBNr (int CurrentMbNr)
{
  ++CurrentMbNr;

  if (CurrentMbNr >= (int)img->PicSizeInMbs)
    return -1;    // No further MB in this slice (could be end of picture)
  else
    return CurrentMbNr;
}

/*!
************************************************************************
* \brief
*    FmoGetNextMBBr: Returns the MB-Nr (in scan order) of the next
*    MB in the (FMO) Slice, -1 if the SliceGroup is finished
*
* \par Input:
*    CurrentMbNr
************************************************************************
*/
int GetPreviousMBNr (int CurrentMbNr)
{
  CurrentMbNr--;

  if (CurrentMbNr < 0)
    return -1;    // No previous MB in this slice
  else
    return CurrentMbNr;
}

/*!
************************************************************************
* \brief
*    FmoGetLastCodedMBOfSlice: Returns the MB-Nr (in scan order) of 
*    the last MB of the slice group
*
* \par Input:
*    SliceGroupID
* \par Return
*    MB Nr in case of success (is always >= 0)
*    -1 if the SliceGroup doesn't exist
************************************************************************
*/
int GetLastCodedMBOfSliceGroup (int SliceGroupID)
{
  return (img->PicSizeInMbs-1);
}


void SetLastMacroblockInSlice ( int mb)
{
  // called by terminate_slice(), writes the last processed MB into the
  // FirstMBInSlice[MAXSLICEGROUPIDS] array.  GetFirstMacroblockInSlice()
  // uses this info to identify the first uncoded MB in each slice group
  int currSliceGroup = 0;
  assert (mb >= 0);
  mb = GetNextMBNr (mb);   // The next (still uncoded) MB, or -1 if SG is finished
  FirstMBInSlice[currSliceGroup] = mb;
}

int GetFirstMacroblockInSlice ( int SliceGroup)
{
  return FirstMBInSlice[SliceGroup];
  // returns the first uncoded MB in each slice group, -1 if there is no
  // more to do in this slice group
}


/*!
************************************************************************
* \brief
*    returns 1 if the macroblock at the given address is available
************************************************************************
*/
int mb_is_available(int mbAddr, int currMbAddr)
{
  if ((mbAddr < 0) || (mbAddr > ((int)img->PicSizeInMbs - 1)))
    return 0;

  // the following line checks both: slice number and if the mb has been decoded
  if (!img->DeblockCall)
  {
    if (img->mb_data[mbAddr].slice_nr != img->mb_data[currMbAddr].slice_nr)
      return 0;
  }

  return 1;
}


/*!
************************************************************************
* \brief
*    Checks the availability of neighboring macroblocks of
*    the current macroblock for prediction and context determination;
************************************************************************
*/
void CheckAvailabilityOfNeighbors(void)
{
  const int mb_nr = img->current_mb_nr;
  Macroblock *currMB = &img->mb_data[mb_nr];

  // mark all neighbors as unavailable
  currMB->mb_available_up   = NULL;
  currMB->mb_available_left = NULL;

  currMB->mbAddrA = mb_nr - 1;
  currMB->mbAddrB = mb_nr - img->PicWidthInMbs;
  currMB->mbAddrC = mb_nr - img->PicWidthInMbs + 1;
  currMB->mbAddrD = mb_nr - img->PicWidthInMbs - 1;

  currMB->mbAvailA = mb_is_available(currMB->mbAddrA, mb_nr) && ((mb_nr % img->PicWidthInMbs)!=0);
  currMB->mbAvailB = mb_is_available(currMB->mbAddrB, mb_nr);
  currMB->mbAvailC = mb_is_available(currMB->mbAddrC, mb_nr) && (((mb_nr+1) % img->PicWidthInMbs)!=0);
  currMB->mbAvailD = mb_is_available(currMB->mbAddrD, mb_nr) && ((mb_nr % img->PicWidthInMbs)!=0);

  if (currMB->mbAvailA) currMB->mb_available_left = &(img->mb_data[currMB->mbAddrA]);
  if (currMB->mbAvailB) currMB->mb_available_up   = &(img->mb_data[currMB->mbAddrB]);
}


/*!
************************************************************************
* \brief
*    returns the x and y macroblock coordinates for a given MbAddress
************************************************************************
*/
void get_mb_block_pos (int mb_addr, int *x, int*y)
{
  //*x = (mb_addr % img->PicWidthInMbs);
  //*y = (mb_addr / img->PicWidthInMbs);
  *x = xPicPos[mb_addr];
  *y = yPicPos[mb_addr];
}


/*!
************************************************************************
* \brief
*    returns the x and y sample coordinates for a given MbAddress
************************************************************************
*/
void get_mb_pos (int mb_addr, int *x, int*y)
{
  get_mb_block_pos(mb_addr, x, y);

  (*x) *= MB_BLOCK_SIZE;
  (*y) *= MB_BLOCK_SIZE;
}


/*!
************************************************************************
* \brief
*    get neighbouring positions for non-aff coding
* \param curr_mb_nr
*   current macroblock number (decoding order)
* \param xN
*    input x position
* \param yN
*    input y position
* \param luma
*    1 if luma coding, 0 for chroma
* \param pix
*    returns position informations
************************************************************************
*/
void getNonAffNeighbour(unsigned int curr_mb_nr, int xN, int yN, int luma, PixelPos *pix)
{
  Macroblock *currMb = &img->mb_data[curr_mb_nr];
  int maxW, maxH;

  if (luma)
  {
    maxW = 16;
    maxH = 16;
  }
  else
  {
    assert(img->yuv_format != 0);
    maxW = img->mb_cr_size_x;
    maxH = img->mb_cr_size_y;
  }

  if ((xN<0)&&(yN<0))
  {
    pix->mb_addr   = currMb->mbAddrD;
    pix->available = currMb->mbAvailD;
  }
  else if ((xN<0)&&((yN>=0)&&(yN<maxH)))
  {
    pix->mb_addr  = currMb->mbAddrA;
    pix->available = currMb->mbAvailA;
  }
  else if (((xN>=0)&&(xN<maxW))&&(yN<0))
  {
    pix->mb_addr  = currMb->mbAddrB;
    pix->available = currMb->mbAvailB;
  }
  else if (((xN>=0)&&(xN<maxW))&&((yN>=0)&&(yN<maxH)))
  {
    pix->mb_addr  = curr_mb_nr;
    pix->available = 1;
  }
  else if ((xN>=maxW)&&(yN<0)) // MB32X32 happens when xN==32, yN==-1  
  {
    pix->mb_addr  = currMb->mbAddrC;
    pix->available = currMb->mbAvailC;
  }
  else 
  {
    pix->available = 0;
  }

  if (pix->available || img->DeblockCall)
  {
    pix->x     = (xN & (maxW - 1));
    pix->pos_x = (pix->x + xPicPos[ pix->mb_addr ] * maxW);
    pix->y     = (yN & (maxH - 1));    
    pix->pos_y = (pix->y + yPicPos[ pix->mb_addr ] * maxH);
  }
}

/*!
************************************************************************
* \brief
*    get neighbouring positions. MB AFF is automatically used from img structure
* \param curr_mb_nr
*   current macroblock number (decoding order)
* \param xN
*    input x position
* \param yN
*    input y position
* \param luma
*    1 if luma coding, 0 for chroma
* \param pix
*    returns position informations
************************************************************************
*/
void getNeighbour(int curr_mb_nr, int xN, int yN, int luma, PixelPos *pix)
{
  if (curr_mb_nr<0)
    error ("getNeighbour: invalid macroblock number", 100);

  getNonAffNeighbour(curr_mb_nr, xN, yN, luma, pix);
}


/*!
************************************************************************
* \brief
*    get neighbouring  get neighbouring 4x4 luma block
* \param curr_mb_nr
*   current macroblock number (decoding order)
* \param block_x
*    input x block position
* \param block_y
*    input y block position
* \param rel_x
*    relative x position of neighbor
* \param rel_y
*    relative y position of neighbor
* \param pix
*    returns position informations
************************************************************************
*/
void getLuma4x4Neighbour (int curr_mb_nr, int block_x, int block_y, int rel_x, int rel_y, PixelPos *pix)
{
  int x = 4* block_x + rel_x;
  int y = 4* block_y + rel_y;

  getNeighbour(curr_mb_nr, x, y, 1, pix);

  if (pix->available)
  {
    pix->x >>= 2;
    pix->y >>= 2;
    pix->pos_x >>= 2;
    pix->pos_y >>= 2;
  }
}


/*!
************************************************************************
* \brief
*    get neighbouring 4x4 chroma block
* \param curr_mb_nr
*   current macroblock number (decoding order)
* \param block_x
*    input x block position
* \param block_y
*    input y block position
* \param rel_x
*    relative x position of neighbor
* \param rel_y
*    relative y position of neighbor
* \param pix
*    returns position informations
************************************************************************
*/
void getChroma4x4Neighbour (int curr_mb_nr, int block_x, int block_y, int rel_x, int rel_y, PixelPos *pix)
{
  int x = 4* block_x + rel_x;
  int y = 4* block_y + rel_y;

  getNeighbour(curr_mb_nr, x, y, 0, pix);

  if (pix->available)
  {
    pix->x >>= 2;
    pix->y >>= 2;
    pix->pos_x >>= 2;
    pix->pos_y >>= 2;
  }
}

int Not_MB16_3(int mb_nr)
{
  int mb_nr_x, mb_nr_y;
  mb_nr_x = xPicPos[mb_nr];
  mb_nr_y = yPicPos[mb_nr];

  if(mb_nr_x%2==1 && mb_nr_y%2==1)
  {
    if(mb_nr_x < (int)(img->PicWidthInMbs-1) )
    {
      return 0;
    }
  }
  return 1;
}
