/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>

#include "global.h"
#include "vlc.h"

#if TRACE
#define SYMTRACESTRING(s) strncpy(sym.tracestring,s,TRACESTRING_SIZE)
#else
#define SYMTRACESTRING(s) // do nothing
#endif


/*! 
*************************************************************************************
* \brief
*    ue_v, writes an ue(v) syntax element, returns the length in bits
*
* \param tracestring
*    the string for the trace file
* \param value
*    the value to be coded
*  \param bitstream
*    the target bitstream the value should be coded into
*
* \return
*    Number of bits used by the coded syntax element
*
* \ note
*    This function writes always the bit buffer for the progressive scan flag, and
*    should not be used (or should be modified appropriately) for the interlace crap
*    When used in the context of the Parameter Sets, this is obviously not a
*    problem.
*
*************************************************************************************
*/
int ue_v (char *tracestring, int value, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym=&symbol;
  sym->value1 = value;
  sym->value2 = 0;

  assert (bitstream->streamBuffer != NULL);

  ue_linfo(sym->value1,sym->value2,&(sym->len),&(sym->inf));
  symbol2uvlc(sym);

  writeUVLC2buffer (sym, bitstream);

#if TRACE
  strncpy(sym->tracestring,tracestring,TRACESTRING_SIZE);
  trace2out (sym);
#endif

  return (sym->len);
}


/*! 
*************************************************************************************
* \brief
*    se_v, writes an se(v) syntax element, returns the length in bits
*
* \param tracestring
*    the string for the trace file
* \param value
*    the value to be coded
*  \param bitstream
*    the target bitstream the value should be coded into
*
* \return
*    Number of bits used by the coded syntax element
*
* \ note
*    This function writes always the bit buffer for the progressive scan flag, and
*    should not be used (or should be modified appropriately) for the interlace crap
*    When used in the context of the Parameter Sets, this is obviously not a
*    problem.
*
*************************************************************************************
*/
int se_v (char *tracestring, int value, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym=&symbol;
  sym->value1 = value;
  sym->value2 = 0;

  assert (bitstream->streamBuffer != NULL);

  se_linfo(sym->value1,sym->value2,&(sym->len),&(sym->inf));
  symbol2uvlc(sym);

  writeUVLC2buffer (sym, bitstream);

#if TRACE
  strncpy(sym->tracestring,tracestring,TRACESTRING_SIZE);
  trace2out (sym);
#endif

  return (sym->len);
}


/*! 
*************************************************************************************
* \brief
*    u_1, writes a flag (u(1) syntax element, returns the length in bits, 
*    always 1
*
* \param tracestring
*    the string for the trace file
* \param value
*    the value to be coded
*  \param bitstream
*    the target bitstream the value should be coded into
*
* \return
*    Number of bits used by the coded syntax element (always 1)
*
* \ note
*    This function writes always the bit buffer for the progressive scan flag, and
*    should not be used (or should be modified appropriately) for the interlace crap
*    When used in the context of the Parameter Sets, this is obviously not a
*    problem.
*
*************************************************************************************
*/
int u_1 (char *tracestring, int value, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym=&symbol;

  sym->bitpattern = value;
  sym->len = 1;
  sym->value1 = value;

  assert (bitstream->streamBuffer != NULL);

  writeUVLC2buffer(sym, bitstream);

#if TRACE
  strncpy(sym->tracestring,tracestring,TRACESTRING_SIZE);
  trace2out (sym);
#endif

  return (sym->len);
}


/*! 
*************************************************************************************
* \brief
*    u_v, writes a n bit fixed length syntax element, returns the length in bits, 
*
* \param n
*    length in bits
* \param tracestring
*    the string for the trace file
* \param value
*    the value to be coded
*  \param bitstream
*    the target bitstream the value should be coded into
*
* \return
*    Number of bits used by the coded syntax element 
*
* \ note
*    This function writes always the bit buffer for the progressive scan flag, and
*    should not be used (or should be modified appropriately) for the interlace crap
*    When used in the context of the Parameter Sets, this is obviously not a
*    problem.
*
*************************************************************************************
*/

int u_v (int n, char *tracestring, int value, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym=&symbol;

  sym->bitpattern = value;
  sym->len = n;
  sym->value1 = value;

  assert (bitstream->streamBuffer != NULL);

  writeUVLC2buffer(sym, bitstream);

#if TRACE
  strncpy(sym->tracestring,tracestring,TRACESTRING_SIZE);
  trace2out (sym);
#endif

  return (sym->len);
}


/*!
************************************************************************
* \brief
*    mapping for ue(v) syntax elements
* \param ue
*    value to be mapped
* \param dummy
*    dummy parameter
* \param info
*    returns mapped value
* \param len
*    returns mapped value length
************************************************************************
*/
void ue_linfo(int ue, int dummy, int *len,int *info)
{
  int i,nn;

  nn=(ue+1)/2;

  for (i=0; i < 16 && nn != 0; i++)
  {
    nn /= 2;
  }
  *len= 2*i + 1;
  *info=ue+1-(1<<i);
}


/*!
************************************************************************
* \brief
*    mapping for se(v) syntax elements
* \param se
*    value to be mapped
* \param dummy
*    dummy parameter
* \param len
*    returns mapped value length
* \param info
*    returns mapped value
************************************************************************
*/
void se_linfo(int se, int dummy, int *len,int *info)
{

  int i,n,sign,nn;

  sign=0;

  if (se <= 0)
  {
    sign=1;
  }
  n=absm(se) << 1;

  //  n+1 is the number in the code table.  Based on this we find length and info

  nn=n/2;
  for (i=0; i < 16 && nn != 0; i++)
  {
    nn /= 2;
  }
  *len=i*2 + 1;
  *info=n - (1 << i) + sign;
}

/*!
************************************************************************
* \brief
*    Makes code word and passes it back
*    A code word has the following format: 0 0 0 ... 1 Xn ...X2 X1 X0.
*
* \par Input:
*    Info   : Xn..X2 X1 X0                                             \n
*    Length : Total number of bits in the codeword
************************************************************************
*/
// NOTE this function is called with sym->inf > (1<<(sym->len/2)).  The upper bits of inf are junk
int symbol2uvlc(SyntaxElement *sym)
{
  int suffix_len=sym->len/2;  
  assert (suffix_len<32);
  sym->bitpattern = (1<<suffix_len)|(sym->inf&((1<<suffix_len)-1));
  return 0;
}

/*!
************************************************************************
* \brief
*    writes UVLC code to the appropriate buffer
************************************************************************
*/
void  writeUVLC2buffer(SyntaxElement *se, Bitstream *currStream)
{

  int i;
  unsigned int mask = 1 << (se->len-1);
  assert ((se->len-1)<32);

  // Add the new bits to the bitstream.
  // Write out a byte if it is full
  for (i=0; i<se->len; i++)
  {
    currStream->byte_buf <<= 1;
    if (se->bitpattern & mask)
      currStream->byte_buf |= 1;
    currStream->bits_to_go--;
    mask >>= 1;
    if (currStream->bits_to_go==0)
    {
      currStream->bits_to_go = 8;
      currStream->streamBuffer[currStream->byte_pos++]=currStream->byte_buf;
      currStream->byte_buf = 0;
    }
  }
}



/*!
************************************************************************
* \brief
*    Makes code word and passes it back
*
* \par Input:
*    Info   : Xn..X2 X1 X0                                             \n
*    Length : Total number of bits in the codeword
************************************************************************
*/

int symbol2vlc(SyntaxElement *sym)
{
  int info_len = sym->len;

  // Convert info into a bitpattern int
  sym->bitpattern = 0;

  // vlc coding
  while(--info_len >= 0)
  {
    sym->bitpattern <<= 1;
    sym->bitpattern |= (0x01 & (sym->inf >> info_len));
  }
  return 0;
}

/*!
************************************************************************
* \brief
*    puts the less than 8 bits in the byte buffer of the Bitstream into
*    the streamBuffer.  
*
* \param
*   currStream: the Bitstream the alignment should be established
*
************************************************************************
*/
void writeVlcByteAlign(Bitstream* currStream)
{
  if (currStream->bits_to_go < 8)
  { // trailing bits to process
    currStream->byte_buf = (currStream->byte_buf <<currStream->bits_to_go) | (0xff >> (8 - currStream->bits_to_go));
    stats->bit_use_stuffingBits[img->type]+=currStream->bits_to_go;
    currStream->streamBuffer[currStream->byte_pos++]=currStream->byte_buf;
    currStream->bits_to_go = 8;
  }
}
