
/*!
*************************************************************************************
* \file macroblock.c
*
* \brief
*    Process one macroblock
*
* \author
*    Main contributors (see contributors.h for copyright, address and affiliation details)
*    - Inge Lille-Langoy               <inge.lille-langoy@telenor.com>
*    - Rickard Sjoberg                 <rickard.sjoberg@era.ericsson.se>
*    - Jani Lainema                    <jani.lainema@nokia.com>
*    - Sebastian Purreiter             <sebastian.purreiter@mch.siemens.de>
*    - Detlev Marpe                    <marpe@hhi.de>
*    - Thomas Wedi                     <wedi@tnt.uni-hannover.de>
*    - Ragip Kurceren                  <ragip.kurceren@nokia.com>
*    - Alexis Michael Tourapis         <alexismt@ieee.org>
*************************************************************************************
*/

#include "contributors.h"

#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <memory.h>

#include "global.h"

#include "elements.h"
#include "macroblock.h"
#include "refbuf.h"
#include "fmo.h"
#include "vlc.h"
#include "image.h"
#include "mb_access.h"
#include "cabac.h"
#include "transform8x8.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern int skip_mode;
extern int ******send_index_mv_prediction;
extern MV_Competition mv_comp;
#endif

extern const byte QP_SCALE_CR[52] ;


//Rate control
extern int DELTA_QP,DELTA_QP2;
extern int QP,QP2;


int writeDeltaQP(int rdopt, int cbp);
int writeCBP_Transform_size (int rdopt, int mb32);
int writeLumaCoeff (int rdopt, int mb32);


#ifdef MB32X32
void CheckAvailabilityOfNeighbors_Z(void);
int write_mb_type(Macroblock *MB, int mb_ext_level);
void writeCBP_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
void writeCBP_CABAC_chroma(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
void writeCBP_CABAC_luma1bit(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
extern Macroblock  MB32;
extern int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
extern int delta_qp_sent;
extern int eos_bit_32;
		extern int only_16x16_transform;
	#ifdef MB64X64
		extern Macroblock  MB64;
		extern int   cbp_32[1<<(2*(MAX_MB_EXT_LEVEL-1))], cbp_32_RDCost[1<<(2*(MAX_MB_EXT_LEVEL-1))];
	#endif
#endif

extern int no32x32;

#ifdef RDO_Q
void SetMotionVectorPredictor (short  pmv[2], char   **refPic, short  ***tmp_mv, short  ref_frame, int list, int block_x, int block_y, int blockshape_x, int blockshape_y);
#endif

extern int coeff_cost4x4[4][4];
extern int rho_bits[3];//[mb_ext_level]
extern int geo_stat[2][3];
extern int rho_bits[3];//[mb_ext_level]
extern char ****geo_mask[3];
extern char **mask;
extern int best8x8pdir_geo[GEO_PART_NUM][2];

void writeMB_geoInfo_CABAC (SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
int MBType2Value_geo (Macroblock* currMB);
int writeMotionInfo2NAL_geo (int geo_mode, int mb_ext_level);
void writeMB_geoInfo_CABAC (SyntaxElement *se, EncodingEnvironmentPtr eep_dp);

int writelumaCBPTransform_size(int BigBlock);
int write_chroma_cbp();
#ifdef MB32X32
int writeLumaCoefBigBlock(Macroblock *currMB, int mb32, int rdopt);
#else
int writeLumaCoefBigBlock(Macroblock *currMB, int rdopt);
#endif


short roundInt(short mv,int mvres)
{
  short roundMV=mv, sign;
  sign=mv>0? 1 : -1;
  roundMV*=sign;
  if (mvres ==1)
	  roundMV=roundMV/2;
  else if (mvres ==2)
	  roundMV=roundMV/4;
  roundMV*=sign;
  return(roundMV);
}

/*!
************************************************************************
* \brief
*    updates the coordinates for the next macroblock to be processed
*
* \param mb_addr
*    macroblock address in scan order
************************************************************************
*/
void set_MB_parameters (int mb_addr)
{
  img->current_mb_nr = mb_addr;
  
  get_mb_block_pos(mb_addr, &img->mb_x, &img->mb_y);
  
  img->block_x = img->mb_x << 2;
  img->block_y = img->mb_y << 2;
  
  img->pix_x   = img->block_x << 2;
  img->pix_y   = img->block_y << 2;
  
  img->opix_x   = img->pix_x;
  img->opix_y   = img->block_y << 2;
  img->mb_data[mb_addr].list_offset = 0;

  
  if (img->yuv_format != YUV400)
  {
    img->pix_c_x = (img->mb_cr_size_x * img->pix_x) >> 4;
    img->pix_c_y = (img->mb_cr_size_y * img->pix_y) >> 4;
    
    img->opix_c_x = (img->mb_cr_size_x * img->opix_x) >> 4;
    img->opix_c_y = (img->mb_cr_size_y * img->opix_y) >> 4;
  }
  //  printf ("set_MB_parameters: mb %d,  mb_x %d,  mb_y %d\n", mb_addr, img->mb_x, img->mb_y);
}


/*!
************************************************************************
* \brief
*    clip to range [0..max luma pel value]
************************************************************************
*/
int clip1a(int a)
{
  return ((a)>img->max_imgpel_value?img->max_imgpel_value:((a)<0?0:(a)));
}


/*!
************************************************************************
* \brief
*    clip to range [0..max chroma pel value]
************************************************************************
*/
int clip1a_chr(int a)
{
  return ((a)>img->max_imgpel_value_uv?img->max_imgpel_value_uv:((a)<0?0:(a)));
}


/*!
************************************************************************
* \brief
*    updates the coordinates and statistics parameter for the
*    next macroblock
************************************************************************
*/
void proceed2nextMacroblock(void)
{
#if TRACE
  int use_bitstream_backing = (input->slice_mode == FIXED_RATE || input->slice_mode == CALLBACK);
#endif
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;
  
  if (bitCount[BITS_TOTAL_MB] > 128 + 256 * img->bitdepth_luma + 2 * img->mb_cr_size_y * img->mb_cr_size_x * img->bitdepth_chroma)
    printf("Warning!!! Number of bits (%d) of macroblock_layer() data seems to exceed defined limit.\n", bitCount[BITS_TOTAL_MB]);
  
#if TRACE
  if (p_trace)
  {
    fprintf(p_trace, "\n*********** Pic: %i (I/P) MB: %i Slice: %i **********\n\n", frame_no, img->current_mb_nr, img->current_slice_nr);
    if(use_bitstream_backing)
      fprintf(p_trace, "\n*********** Pic: %i (I/P) MB: %i Slice: %i **********\n\n", frame_no, img->current_mb_nr, img->current_slice_nr);
    // Write out the tracestring for each symbol
    for (i=0; i<currMB->currSEnr; i++)
      trace2out(&(img->MB_SyntaxElements[i]));
  }
#endif
  // Update the statistics
#if DUMP_MBTYPE 
  statsRDPass[img->curr_rd_pass].bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
#else
  stats->bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
#endif
  stats->bit_use_coeffY[img->type]       += bitCount[BITS_COEFF_Y_MB] ;
  stats->tmp_bit_use_cbp[img->type]      += bitCount[BITS_CBP_MB];
  stats->bit_use_coeffC[img->type]       += bitCount[BITS_COEFF_UV_MB];
  stats->bit_use_delta_quant[img->type]  += bitCount[BITS_DELTA_QUANT_MB];
  

  if (IS_INTRA(currMB))
  {
    ++stats->intra_chroma_mode[currMB->c_ipred_mode];
    
    if ((currMB->cbp&15) != 0)
    {
      if (currMB->luma_transform_size_8x8_flag)
        ++stats->mode_use_transform_8x8[img->type][currMB->mb_type];
      else
        ++stats->mode_use_transform_4x4[img->type][currMB->mb_type];
    }
  }
  
  ++stats->mode_use[img->type][currMB->mb_type];
  
#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    if(currMB->mb_type == 0)
    {
      if(!currMB->send_index_predictor_skip)
        ++stats->predictor_for_skip_trans[img->type];
      else
        ++stats->predictor_for_skip_use[img->type][currMB->best_predictor_for_skip];
    }
    stats->bit_use_pred_for_skip[img->type]+= bitCount[BITS_SKIP_PRED];
    
    if(img->type == P_SLICE)
      stats->bit_use_pred_for_mv_inter[img->type]+= bitCount[BITS_MV_PRED];
  }
#endif

  stats->bit_use_mode[img->type][currMB->mb_type]+= bitCount[BITS_INTER_MB];
  
  if (img->type != I_SLICE)
  {
    if (currMB->mb_type == P8x8)
    {
      for(i=0;i<4;i++)
      {
        if (currMB->b8mode[i] > 0)
          ++stats->mode_use[img->type][currMB->b8mode[i]];
        else
          ++stats->b8_mode_0_use[img->type][currMB->luma_transform_size_8x8_flag];
        
        if (currMB->b8mode[i]==4)
        {
          if ((currMB->luma_transform_size_8x8_flag && (currMB->cbp&15) != 0) || input->Transform8x8Mode == 2)
            ++stats->mode_use_transform_8x8[img->type][4];
          else
            ++stats->mode_use_transform_4x4[img->type][4];
        }
      }
    }
    else if (currMB->mb_type >= 0 && currMB->mb_type <=3 && ((currMB->cbp&15) != 0))
    {
      if (currMB->luma_transform_size_8x8_flag)
        ++stats->mode_use_transform_8x8[img->type][currMB->mb_type];
      else
        ++stats->mode_use_transform_4x4[img->type][currMB->mb_type];
    }
  }
  
  // Statistics
  if ((img->type == P_SLICE))
  {
    ++stats->quant0;
    stats->quant1 += currMB->qp;      // to find average quant for inter frames
  }
}

void set_chroma_qp(Macroblock *currMB)
{
  int i;
  for (i=0; i<2; i++)
  {
    currMB->qpc[i] = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, currMB->qp + img->chroma_qp_offset[i] );
    currMB->qpc[i] = currMB->qpc[i] < 0 ? currMB->qpc[i] : QP_SCALE_CR[currMB->qpc[i]];
  }
}

/*!
************************************************************************
* \brief
*    initializes the current macroblock
*
* \param mb_addr
*    macroblock address in scan order
************************************************************************
*/
void start_macroblock(int mb_addr)
{
  int i,j,l;
  int use_bitstream_backing = (input->slice_mode == FIXED_RATE || input->slice_mode == CALLBACK);
  Macroblock *currMB = &img->mb_data[mb_addr];
  Slice *curr_slice = img->currentSlice;
  DataPartition *dataPart;
  Bitstream *currStream;
  EncodingEnvironmentPtr eep;
  int prev_mb;

  currMB->geo_mode = -1;

  
  set_MB_parameters (mb_addr);
  
  if(use_bitstream_backing)
  {
    if ((mb_addr&1)==0) // KS: MB AFF -> store stream positions for 
      //       first macroblock only
    {
      // Keep the current state of the bitstreams
      if(!img->cod_counter)
        for (i=0; i<curr_slice->max_part_nr; i++)
        {
          dataPart = &(curr_slice->partArr[i]);
          currStream = dataPart->bitstream;
          currStream->stored_bits_to_go = currStream->bits_to_go;
          currStream->stored_byte_pos   = currStream->byte_pos;
          currStream->stored_byte_buf   = currStream->byte_buf;
          
          if (input->symbol_mode ==CABAC)
          {
            eep = &(dataPart->ee_cabac);
            eep->ElowS            = eep->Elow;
            eep->ErangeS          = eep->Erange;
            eep->EbufferS         = eep->Ebuffer;
            eep->Ebits_to_goS     = eep->Ebits_to_go;
            eep->Ebits_to_followS = eep->Ebits_to_follow;
            eep->EcodestrmS       = eep->Ecodestrm;
            eep->Ecodestrm_lenS   = eep->Ecodestrm_len;
            eep->CS               = eep->C;
            eep->ES               = eep->E;
          }
        }
    }
  }
  
  // Save the slice number of this macroblock. When the macroblock below
  // is coded it will use this to decide if prediction for above is possible
  currMB->slice_nr = img->current_slice_nr;
  
  // Initialize delta qp change from last macroblock. Feature may be used for future rate control
  currMB->qpsp       = img->qpsp;
  
  prev_mb = FmoGetPreviousMBNr(img->current_mb_nr);
#if (defined MB32X32)
  if(img->type==I_SLICE)
  {
    Slice* currSlice = img->currentSlice;
    if (prev_mb>-1)
    {
      currMB->prev_qp = img->mb_data[prev_mb].qp;
      if (img->mb_data[prev_mb].slice_nr == img->current_slice_nr)
        currMB->prev_delta_qp = img->mb_data[prev_mb].delta_qp;
      else
        currMB->prev_delta_qp = 0;
    }
    else
    {
      currMB->prev_qp = currSlice->qp;
      currMB->prev_delta_qp = 0;
    }
    currMB->qp = img->qp = img->masterQP;
  }
  else
  {
    if (prev_mb>-1)
    {
      currMB->prev_qp       = Prev_QP;       //global variable
      currMB->prev_delta_qp = Prev_Delta_QP;  //global variable
    }
    else
    {
      currMB->prev_qp       = img->masterQP;
      currMB->prev_delta_qp = 0;
    }
  }
  currMB->delta_qp = currMB->qp - currMB->prev_qp;
  set_chroma_qp (currMB);
#endif
  // Initialize counter for MB symbols
  currMB->currSEnr=0;
  
  // loop filter parameter
  if (active_pps->deblocking_filter_control_present_flag)
  {
    currMB->LFDisableIdc    = img->LFDisableIdc;
    currMB->LFAlphaC0Offset = img->LFAlphaC0Offset;
    currMB->LFBetaOffset    = img->LFBetaOffset;
  }
  else
  {
    currMB->LFDisableIdc    = 0;
    currMB->LFAlphaC0Offset = 0;
    currMB->LFBetaOffset    = 0;
  }
  
  // If MB is next to a slice boundary, mark neighboring blocks unavailable for prediction
#ifdef MB32X32    
  if(img->type != I_SLICE) 
    CheckAvailabilityOfNeighbors_Z();
  else
#endif
	CheckAvailabilityOfNeighbors();
  
  if (input->symbol_mode == CABAC)
    CheckAvailabilityOfNeighborsCABAC();
  
  // Reset vectors and reference indices before doing motion search in motion_search().
  for (l=0; l<2; l++)
  {
    for (j=img->block_y; j < img->block_y + BLOCK_MULTIPLE; j++)
    {
      memset(&enc_picture->ref_idx[l][j][img->block_x], -1, BLOCK_MULTIPLE * sizeof(char));     
      memset(enc_picture->mv      [l][j][img->block_x], 0, 2 * BLOCK_MULTIPLE * sizeof(short));
      for (i=img->block_x; i < img->block_x + BLOCK_MULTIPLE; i++)
        enc_picture->ref_pic_id[l][j][i]= -1;
    }
  }
  
  // Reset syntax element entries in MB struct
  currMB->mb_type      = 0;
  currMB->cbp_blk      = 0;
  currMB->cbp          = 0;
  currMB->cbp_bits     = 0;
  currMB->c_ipred_mode = DC_PRED_8;
  
  memset (currMB->mvd, 0, BLOCK_CONTEXT * sizeof(int));  
  memset (currMB->intra_pred_modes, DC_PRED, MB_BLOCK_PARTITIONS * sizeof(char)); // changing this to char would allow us to use memset
  memset (currMB->intra_pred_modes8x8, DC_PRED, MB_BLOCK_PARTITIONS * sizeof(char));
  
  //initialize the whole MB as INTRA coded
  //Blocks are set to notINTRA in write_one_macroblock
  if (input->UseConstrainedIntraPred)
  {
    img->intra_block[img->current_mb_nr] = 1;
  }
  
  // Initialize bitcounters for this macroblock
  if(prev_mb < 0) // No slice header to account for
  {
    currMB->bitcounter[BITS_HEADER] = 0;
  }
  else if (currMB->slice_nr == img->mb_data[prev_mb].slice_nr) // current MB belongs to the
    // same slice as the last MB
  {
    currMB->bitcounter[BITS_HEADER] = 0;
  }
  
  currMB->bitcounter[BITS_MB_MODE       ] = 0;
  currMB->bitcounter[BITS_COEFF_Y_MB    ] = 0;
  currMB->bitcounter[BITS_INTER_MB      ] = 0;
  currMB->bitcounter[BITS_CBP_MB        ] = 0;
  currMB->bitcounter[BITS_DELTA_QUANT_MB] = 0;
  currMB->bitcounter[BITS_COEFF_UV_MB   ] = 0;
  
#ifdef MV_COMPETITION
  currMB->bitcounter[BITS_SKIP_PRED     ] = 0;
  currMB->bitcounter[BITS_MV_PRED       ] = 0;
#endif

//#ifdef _FAST_FULL_ME_
//  if(!input->FMEnable)
//    ResetFastFullIntegerSearch ();
//#endif
}



/*!
************************************************************************
* \brief
*    terminates processing of the current macroblock depending
*    on the chosen slice mode
************************************************************************
*/
void terminate_macroblock( Boolean *end_of_slice,      //!< returns true for last macroblock of a slice, otherwise false
                          Boolean *recode_macroblock  //!< returns true if max. slice size is exceeded an macroblock must be recoded in next slice
                          )
{
  int i;
  Slice *currSlice = img->currentSlice;
  Macroblock    *currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  int *partMap = assignSE2partition[input->partition_mode];
  DataPartition *dataPart;
  Bitstream *currStream;
  int rlc_bits=0;
  EncodingEnvironmentPtr eep;
  int use_bitstream_backing = (input->slice_mode == FIXED_RATE || input->slice_mode == CALLBACK);
  int new_slice;
  static int skip = FALSE;
  
  
  // if previous mb in the same slice group has different slice number as the current, it's the
  // the start of new slice
  new_slice=0;
  if ( (img->current_mb_nr==0) || (FmoGetPreviousMBNr(img->current_mb_nr)<0) )
    new_slice=1;
  else if( img->mb_data[FmoGetPreviousMBNr(img->current_mb_nr)].slice_nr != img->current_slice_nr )
    new_slice=1;
  
  *recode_macroblock=FALSE;
  
  switch(input->slice_mode)
  {
  case NO_SLICES:
    currSlice->num_mb++;
    *recode_macroblock = FALSE;
    if ((currSlice->num_mb) == (int)img->PicSizeInMbs) // maximum number of MBs reached
      *end_of_slice = TRUE;
    
    // if it's end of current slice group, slice ends too
    *end_of_slice |= (img->current_mb_nr == FmoGetLastCodedMBOfSliceGroup (FmoMB2SliceGroup (img->current_mb_nr)));
    
    break;
  case FIXED_MB:
    // For slice mode one, check if a new slice boundary follows
    currSlice->num_mb++;
    *recode_macroblock = FALSE;
    //! Check end-of-slice group condition first
    *end_of_slice = (img->current_mb_nr == FmoGetLastCodedMBOfSliceGroup (FmoMB2SliceGroup (img->current_mb_nr)));
    //! Now check maximum # of MBs in slice
    *end_of_slice |= (currSlice->num_mb >= input->slice_argument);
    
    break;
    
    // For slice modes two and three, check if coding of this macroblock
    // resulted in too many bits for this slice. If so, indicate slice
    // boundary before this macroblock and code the macroblock again
  case FIXED_RATE:
    // in case of skip MBs check if there is a slice boundary
    // only for UVLC (img->cod_counter is always 0 in case of CABAC)
    if(img->cod_counter)
    {
      // write out the skip MBs to know how many bits we need for the RLC
      currSE->value1 = img->cod_counter;
      currSE->value2 = 0;
      currSE->mapping = ue_linfo;
      currSE->type = SE_MBTYPE;
      dataPart = &(currSlice->partArr[partMap[currSE->type]]);
      
      dataPart->writeSyntaxElement(  currSE, dataPart);
      rlc_bits=currSE->len;
      
      currStream = dataPart->bitstream;
      // save the bitstream as it would be if we write the skip MBs
      currStream->bits_to_go_skip  = currStream->bits_to_go;
      currStream->byte_pos_skip    = currStream->byte_pos;
      currStream->byte_buf_skip    = currStream->byte_buf;
      // restore the bitstream
      currStream->bits_to_go = currStream->stored_bits_to_go;
      currStream->byte_pos = currStream->stored_byte_pos;
      currStream->byte_buf = currStream->stored_byte_buf;
      skip = TRUE;
    }
    //! Check if the last coded macroblock fits into the size of the slice
    //! But only if this is not the first macroblock of this slice
    if (!new_slice)
    {
      if(slice_too_big(rlc_bits))
      {
        *recode_macroblock = TRUE;
        *end_of_slice = TRUE;
      }
      else if(!img->cod_counter)
        skip = FALSE;
    }
    // maximum number of MBs
    
    // check if current slice group is finished
    if ((*recode_macroblock == FALSE) && (img->current_mb_nr == FmoGetLastCodedMBOfSliceGroup (FmoMB2SliceGroup (img->current_mb_nr)))) 
    {
      *end_of_slice = TRUE;
      if(!img->cod_counter)
        skip = FALSE;
    }
    
    //! (first MB OR first MB in a slice) AND bigger that maximum size of slice
    if (new_slice && slice_too_big(rlc_bits))
    {
      *end_of_slice = TRUE;
      if(!img->cod_counter)
        skip = FALSE;
    }
    if (!*recode_macroblock)
      currSlice->num_mb++;
    break;
    
  case  CALLBACK:
    if (img->current_mb_nr > 0 && !new_slice)
    {
      if (currSlice->slice_too_big(rlc_bits))
      {
        *recode_macroblock = TRUE;
        *end_of_slice = TRUE;
      }
    }
    
    if ( (*recode_macroblock == FALSE) && (img->current_mb_nr == FmoGetLastCodedMBOfSliceGroup (FmoMB2SliceGroup (img->current_mb_nr)))) 
      *end_of_slice = TRUE;
    break;
    
  default:
    snprintf(errortext, ET_SIZE, "Slice Mode %d not supported", input->slice_mode);
    error(errortext, 600);
  }
  
  if(*recode_macroblock == TRUE)
  {
    // Restore everything
    for (i=0; i<currSlice->max_part_nr; i++)
    {
      dataPart = &(currSlice->partArr[i]);
      currStream = dataPart->bitstream;
      currStream->bits_to_go = currStream->stored_bits_to_go;
      currStream->byte_pos  = currStream->stored_byte_pos;
      currStream->byte_buf  = currStream->stored_byte_buf;
      if (input->symbol_mode == CABAC)
      {
        eep = &(dataPart->ee_cabac);
        eep->Elow            = eep->ElowS;
        eep->Erange          = eep->ErangeS;
        eep->Ebuffer         = eep->EbufferS;
        eep->Ebits_to_go     = eep->Ebits_to_goS;
        eep->Ebits_to_follow = eep->Ebits_to_followS;
        eep->Ecodestrm       = eep->EcodestrmS;
        eep->Ecodestrm_len   = eep->Ecodestrm_lenS;
        eep->C               = eep->CS;
        eep->E               = eep->ES;       
      }
    }
  }
  
  if (input->symbol_mode == UVLC)
  {
    // Skip MBs at the end of this slice
    dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
    if(*end_of_slice == TRUE  && skip == TRUE) 
    { 
      // only for Slice Mode 2 or 3
      // If we still have to write the skip, let's do it!
      if(img->cod_counter && *recode_macroblock == TRUE) // MB that did not fit in this slice
      { 
        // If recoding is true and we have had skip, 
        // we have to reduce the counter in case of recoding
        img->cod_counter--;
        if(img->cod_counter)
        {
          currSE->value1 = img->cod_counter;
          currSE->value2 = 0;
          currSE->mapping = ue_linfo;
          currSE->type = SE_MBTYPE;
          dataPart->writeSyntaxElement(  currSE, dataPart);
          currMB->currSEnr ++;
#if TRACE
          snprintf(currSE->tracestring, TRACESTRING_SIZE, "Final MB runlength = %3d",img->cod_counter); 
#endif
          rlc_bits=currSE->len;
          currMB->bitcounter[BITS_MB_MODE]+=rlc_bits;
          img->cod_counter = 0;
        }
      }
      else //! MB that did not fit in this slice anymore is not a Skip MB
      {
        currStream = dataPart->bitstream;
        // update the bitstream
        currStream->bits_to_go = currStream->bits_to_go_skip;
        currStream->byte_pos  = currStream->byte_pos_skip;
        currStream->byte_buf  = currStream->byte_buf_skip;
        
        // update the statistics
        img->cod_counter = 0;
        skip = FALSE;
      }
    }
    
    // Skip MBs at the end of this slice for Slice Mode 0 or 1
    if(*end_of_slice == TRUE && img->cod_counter && !use_bitstream_backing)
    {
      currSE->value1 = img->cod_counter;
      currSE->value2 = 0;
      currSE->mapping = ue_linfo;
      currSE->type = SE_MBTYPE;
      dataPart->writeSyntaxElement(  currSE, dataPart);
      currMB->currSEnr ++;
#if TRACE
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Final MB runlength = %3d",img->cod_counter); 
#endif
      
      rlc_bits=currSE->len;
      currMB->bitcounter[BITS_MB_MODE]+=rlc_bits;
      img->cod_counter = 0;
    }
  }
}

/*!
*****************************************************************************
*
* \brief 
*    For Slice Mode 2: Checks if one partition of one slice exceeds the 
*    allowed size
* 
* \return
*    FALSE if all Partitions of this slice are smaller than the allowed size
*    TRUE is at least one Partition exceeds the limit
*
* \par Side effects
*    none
*
* \date
*    4 November 2001
*
* \author
*    Tobias Oelbaum      drehvial@gmx.net
*****************************************************************************/

int slice_too_big(int rlc_bits)
{
  Slice *currSlice = img->currentSlice;
  DataPartition *dataPart;
  Bitstream *currStream;
  EncodingEnvironmentPtr eep;
  int i;
  int size_in_bytes;
  
  //! UVLC
  if (input->symbol_mode == UVLC)
  {
    for (i=0; i<currSlice->max_part_nr; i++)
    {
      dataPart = &(currSlice->partArr[i]);
      currStream = dataPart->bitstream;
      size_in_bytes = currStream->byte_pos /*- currStream->tmp_byte_pos*/;
      
      if (currStream->bits_to_go < 8)
        size_in_bytes++;
      if (currStream->bits_to_go < rlc_bits)
        size_in_bytes++;
      if(size_in_bytes > input->slice_argument)
        return TRUE;
    }
  }
  
  //! CABAC
  if (input->symbol_mode ==CABAC)
  {
    for (i=0; i<currSlice->max_part_nr; i++)
    {
      dataPart= &(currSlice->partArr[i]);
      eep = &(dataPart->ee_cabac);
      
      if( arienco_bits_written(eep) > (input->slice_argument*8))
        return TRUE;
    }
  }
  return FALSE;
}

#ifdef EIGHTH_PEL_DIRECT
int interp_block[16][16];
extern void get_UMVLine4X8_direct(StorablePicture *ref_picture, int x_pos, int y_pos, int block[16][16], int size_x, int size_y);
#endif

static pel_t *(*get_line) (pel_t**, int, int, int, int);
/*!
************************************************************************
* \brief
*    Predict one component of a 4x4 Luma block
************************************************************************
*/
void OneComponentLumaPrediction4x4 ( int*   mpred,          //!< array of prediction values (row by row)
                                    int    pic_pix_x,      //!< absolute horizontal coordinate of 4x4 block
                                    int    pic_pix_y,      //!< absolute vertical   coordinate of 4x4 block
                                    short* mv,             //!< motion vector
                                    short  ref,            //!< reference frame 
                                    StorablePicture **list //!< reference picture list
                                    )
{
  pel_t** ref_pic = list[ref]->imgY_ups;
  int img_width =((list[ref]->size_x + 2*IMG_PAD_SIZE - 1)<<2);
  int img_height=((list[ref]->size_y + 2*IMG_PAD_SIZE - 1)<<2);
  pel_t   *ref_line;
  int j;
  int i0,j0;
#ifdef EIGHTH_PEL  
  int mv_shift= img->mv_res;
#endif  
  
#ifdef EIGHTH_PEL_DIRECT
  int i, dx, dy;
  unsigned int *refP;
#endif
  
  int mv0, mv1;
  mv0 = mv[0];
  mv1 = mv[1];
  
#ifdef MV_COMPETITION
  if ((skip_mode== TRUE) &&  (input->mv_competition > 0))
  {
    mv0 = mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][0];
    mv1 = mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][1];
  }
#endif
#ifdef EIGHTH_PEL  
  if (img->mv_res)
  {
    j0      = (pic_pix_y << 3) + mv1 + IMG_PAD_SIZE_TIMES4*2;
    i0      = (pic_pix_x << 3) + mv0 + IMG_PAD_SIZE_TIMES4*2;
  }
  else
#endif
  {
  j0      = (pic_pix_y << 2) + mv1 + IMG_PAD_SIZE_TIMES4;
  i0      = (pic_pix_x << 2) + mv0 + IMG_PAD_SIZE_TIMES4;
  }


#ifdef EIGHTH_PEL
  if (img->mv_res)
  {
    if ((i0 > 0) && (i0 < (((img_width - 4*BLOCK_SIZE)<< 1) - 1))  && (j0 > 0) && (j0 < (((img_height - 4*BLOCK_SIZE)<< 1) - 1)))
      get_line = FastLine4X8;
    else
      get_line = UMVLine4X8;
  }
  else
#endif
  {
  if ((i0 > 0) && (i0 < img_width - 4*BLOCK_SIZE - 1) && (j0 > 0) && (j0 < img_height - 4*BLOCK_SIZE - 1))
    get_line = FastLine4X;
  else
    get_line = UMVLine4X;
  }
  
#ifdef EIGHTH_PEL_DIRECT

  // Calculate the interpolated values for the entire 4x4 subblock.
  dx = i0&1;
  dy = j0&1;
  
  if ((dx || dy) && img->mv_res) // At least i0 or j0 is 1/8-pel and img->mv_res is non-zero
  {
    get_UMVLine4X8_direct(list[ref], (i0 - (IMG_PAD_SIZE<<3)), (j0 - (IMG_PAD_SIZE<<3)), interp_block, 4, 4);
    for (j=0; j < 4; j++)
    {
      refP = &interp_block[j][0];
      for (i=0; i < 4; i++)
      {
        *mpred++ = *refP++;
      }
    }
  }
  else
#endif // EIGHTH_PEL_DIRECT
  
  {
#ifdef EIGHTH_PEL
    for (j = j0; j < j0 + (MB_BLOCK_SIZE<<mv_shift); j+= (BLOCK_SIZE<<mv_shift))
#else
    for (j = j0; j < j0 + (MB_BLOCK_SIZE); j+=BLOCK_SIZE)
#endif
    {
      ref_line = get_line (ref_pic, j, i0, img_height, img_width);
      *mpred++ = *ref_line;
      *mpred++ = *(ref_line + 4);
      *mpred++ = *(ref_line + 8);
      *mpred++ = *(ref_line + 12);
    }
}
}

/*!
************************************************************************
* \brief
*    copy foward/backward prediction values of one component of a 4x4 Luma block
************************************************************************
*/
void copyblock4x4 ( int* mpred,                       //!< array of prediction values (row by row)
                   int block[BLOCK_SIZE][BLOCK_SIZE] //!< target block
                   )        
{
  *mpred++ = block[0][0];
  *mpred++ = block[1][0];
  *mpred++ = block[2][0];
  *mpred++ = block[3][0];
  *mpred++ = block[0][1];
  *mpred++ = block[1][1];
  *mpred++ = block[2][1];
  *mpred++ = block[3][1];
  *mpred++ = block[0][2];
  *mpred++ = block[1][2];
  *mpred++ = block[2][2];
  *mpred++ = block[3][2];
  *mpred++ = block[0][3];
  *mpred++ = block[1][3];
  *mpred++ = block[2][3];
  *mpred++ = block[3][3];
}

/*!
************************************************************************
* \brief
*    Predict one 4x4 Luma block
************************************************************************
*/
void LumaPrediction4x4 ( int   block_x,    //!< relative horizontal block coordinate of 4x4 block
                        int   block_y,    //!< relative vertical   block coordinate of 4x4 block
                        int   p_dir,      //!< prediction direction (0=list0, 1=list1, 2=bipred)
                        int   fw_mode,    //!< list0 prediction mode (1-7, 0=DIRECT if bw_mode=0)
                        int   bw_mode,    //!< list1 prediction mode (1-7, 0=DIRECT if fw_mode=0)
                        short fw_ref_idx, //!< reference frame for list0 prediction (-1: Intra4x4 pred. with fw_mode)
                        short bw_ref_idx  //!< reference frame for list1 prediction 
                        )
{
  static int fw_pred[16];
  static int bw_pred[16];
  
  int  i, j;
  int  block_x4  = block_x+4;
  int  block_y4  = block_y+4;
  int  pic_opix_x = img->opix_x + block_x;
  int  pic_opix_y = img->opix_y + block_y;
  int  bx        = block_x >> 2;
  int  by        = block_y >> 2;
  int* fpred     = fw_pred;
  int* bpred     = bw_pred;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  
  int  apply_weights = ( (active_pps->weighted_pred_flag && (img->type== P_SLICE )) ||
    (active_pps->weighted_bipred_idc && (img->type== B_SLICE)));  
  short**** mv_array = img->all_mv[by][bx];
  
  img->bipred_rounding_control = (img->nal_reference_idc!=0);

  if (currMB->bi_pred_me && fw_ref_idx == 0 && bw_ref_idx == 0 && p_dir == 2 && fw_mode==1 && bw_mode==1)
  {
    mv_array = currMB->bi_pred_me == 1? img->bipred_mv1[by][bx] : img->bipred_mv2[by][bx];
  }
  
  if ((p_dir==0)||(p_dir==2))
  {
    OneComponentLumaPrediction4x4 (fw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_0][fw_ref_idx][fw_mode], fw_ref_idx, listX[0+currMB->list_offset]);   
  }
  
  if ((p_dir==1)||(p_dir==2))
  { 
    OneComponentLumaPrediction4x4 (bw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_1][bw_ref_idx][bw_mode], bw_ref_idx, listX[1+currMB->list_offset]);   
  }
  
  if (apply_weights)
  {
    
    if (p_dir==2)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          if(input->UseHPFilter != 0)
          {
            img->mpr[j][i] = clip1a(((wbp_weight[0][fw_ref_idx][bw_ref_idx][0] * *fpred++ + 
                                      wbp_weight[1][fw_ref_idx][bw_ref_idx][0] * *bpred++ + 
                                      (wp_offset[0][fw_ref_idx][0] + wp_offset[1][bw_ref_idx][0])*2*wp_luma_round +
                                      2*wp_luma_round*(1-img->bipred_rounding_control)) >> (luma_log_weight_denom + 1)) 
                                    ); 
          }
          else
          {
            img->mpr[j][i] = clip1a(((wbp_weight[0][fw_ref_idx][bw_ref_idx][0] * *fpred++ + 
                                      wbp_weight[1][fw_ref_idx][bw_ref_idx][0] * *bpred++ + 
                                      2*wp_luma_round) >> (luma_log_weight_denom + 1)) + 
                                     ((wp_offset[0][fw_ref_idx][0] + wp_offset[1][bw_ref_idx][0] + 1)>>1)); 
          }

    }
    else if (p_dir==0)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = clip1a(((wp_weight[0][fw_ref_idx][0] * *fpred++  + wp_luma_round) >> luma_log_weight_denom) +
          wp_offset[0][fw_ref_idx][0] );
    }
    else // p_dir==1
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = clip1a(((wp_weight[1][bw_ref_idx][0] * *bpred++  + wp_luma_round) >> luma_log_weight_denom) +
          wp_offset[1][bw_ref_idx][0] );
    }
  }
  else
  {
    if (p_dir==2)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          if(input->UseHPFilter != 0)
          {
            img->mpr[j][i] = (*fpred++ + *bpred++ + 1 - img->bipred_rounding_control) >> 1; 
          }
          else
          {
            img->mpr[j][i] = (*fpred++ + *bpred++ + 1) >> 1; 
          }
    }
    else if (p_dir==0)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = *fpred++;
    }
    else // p_dir==1
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = *bpred++;
    }
  }
}

/*!
************************************************************************
* \brief
*    Predict one 4x4 Luma block
************************************************************************
*/
void LumaPrediction4x4Bi ( int   block_x,    //!< relative horizontal block coordinate of 4x4 block
                          int   block_y,    //!< relative vertical   block coordinate of 4x4 block
                          int   p_dir,      //!< prediction direction (0=list0, 1=list1, 2=bidir)
                          int   fw_mode,    //!< list0 prediction mode (1-7, 0=DIRECT if bw_mode=0)
                          int   bw_mode,    //!< list1 prediction mode (1-7, 0=DIRECT if fw_mode=0)
                          short fw_ref_idx, //!< reference frame for list0 prediction (-1: Intra4x4 pred. with fw_mode)
                          short bw_ref_idx, //!< reference frame for list1 prediction 
                          int   list        //!< current list for prediction.
                          )
{
  static int fw_pred[16];
  static int bw_pred[16];
  
  int  i, j;
  int  block_x4  = block_x+4;
  int  block_y4  = block_y+4;
  int  pic_opix_x = img->opix_x + block_x;
  int  pic_opix_y = img->opix_y + block_y;
  int  bx        = block_x >> 2;
  int  by        = block_y >> 2;
  int* fpred     = fw_pred;
  int* bpred     = bw_pred;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  
  int  apply_weights = ( (active_pps->weighted_pred_flag && (img->type == P_SLICE)) ||
    (active_pps->weighted_bipred_idc && (img->type == B_SLICE)));  
  short   ****mv_array = list ? img->bipred_mv1[by][bx] : img->bipred_mv2[by][bx];
  
  img->bipred_rounding_control = (img->nal_reference_idc!=0);
  
  if ((p_dir==0)||(p_dir==2))
  {
    OneComponentLumaPrediction4x4 (fw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_0][fw_ref_idx][fw_mode], fw_ref_idx, listX[0+currMB->list_offset]);   
  }
  
  if ((p_dir==1)||(p_dir==2))
  { 
    OneComponentLumaPrediction4x4 (bw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_1][bw_ref_idx][bw_mode], bw_ref_idx, listX[1+currMB->list_offset]);   
  }

  if (apply_weights)
  {
    if (p_dir==2)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          if(input->UseHPFilter != 0)
          {
            img->mpr[j][i] = clip1a(((wbp_weight[0][fw_ref_idx][bw_ref_idx][0] * *fpred++ + 
                                      wbp_weight[1][fw_ref_idx][bw_ref_idx][0] * *bpred++ + 
                                      (wp_offset[0][fw_ref_idx][0] + wp_offset[1][bw_ref_idx][0])*2*wp_luma_round+
                                      2*wp_luma_round*(1-img->bipred_rounding_control)) >> (luma_log_weight_denom + 1))
                                    ); 
          }
          else
          {
            img->mpr[j][i] = clip1a(((wbp_weight[0][fw_ref_idx][bw_ref_idx][0] * *fpred++ + 
                                     wbp_weight[1][fw_ref_idx][bw_ref_idx][0] * *bpred++ + 
                                     2*wp_luma_round) >> (luma_log_weight_denom + 1)) + 
                                     ((wp_offset[0][fw_ref_idx][0] + wp_offset[1][bw_ref_idx][0] + 1)>>1)); 
          }
    }
    else if (p_dir==0)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)                   
          img->mpr[j][i] = clip1a(((wp_weight[0][fw_ref_idx][0] * *fpred++  + wp_luma_round) >> luma_log_weight_denom) +
          wp_offset[0][fw_ref_idx][0] );
    }
    else // p_dir==1
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = clip1a(((wp_weight[1][bw_ref_idx][0] * *bpred++  + wp_luma_round) >> luma_log_weight_denom) 
          + wp_offset[1][bw_ref_idx][0] );
    }
  }
  else
  {
    if (p_dir==2)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          if(input->UseHPFilter != 0)
          {
            img->mpr[j][i] = (*fpred++ + *bpred++ + 1 - img->bipred_rounding_control) / 2; 
          }
          else
          {
            img->mpr[j][i] = (*fpred++ + *bpred++ + 1) / 2; 
          }
    }
    else if (p_dir==0)
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = *fpred++;
    }
    else // p_dir==1
    {
      for   (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = *bpred++;
    }
  }
}



/*!
************************************************************************
* \brief
*    Residual Coding of an 8x8 Luma block (not for intra)
*
* \return
*    coefficient cost
************************************************************************
*/
int LumaResidualCoding8x8 (int mbi_x, 
                           int mbi_y, 
                           int mode, 
                           int mode_idx,
                           int   *cbp,        //!< Output: cbp (updated according to processed 8x8 luminance block)
                           int64 *cbp_blk,    //!< Output: block cbp (updated according to processed 8x8 luminance block)
                           int   block8x8,    //!< block number of 8x8 block
                           short p_dir,       //!< prediction direction
                           int   fw_mode,     //!< list0 prediction mode (1-7, 0=DIRECT)
                           int   bw_mode,     //!< list1 prediction mode (1-7, 0=DIRECT)
                           short fw_refframe, //!< reference picture for list0 prediction
                           short bw_refframe  //!< reference picture for list0 prediction
                           )
{
  int    block_y, block_x, pic_pix_y, pic_pix_x, i, j, nonzero = 0, cbp_blk_mask;
  int    coeff_cost = 0;
  int    mb_y       = (block8x8 >> 1) << 3;
  int    mb_x       = (block8x8 & 0x01) << 3;
//  int    pix_y;
  int    cbp_mask   = 1 << block8x8;
  int    bxx, byy;                   // indexing curr_blk

  Macroblock* currMB = &img->mb_data[img->current_mb_nr];
  //set transform size
  int    need_8x8_transform = currMB->luma_transform_size_8x8_flag;
  static imgpel backup_mpr[100][64][64];  

  int    skipped    = (fw_mode == 0 && bw_mode == 0 && (img->type != B_SLICE));
#if P_DIRECT
         skipped    = 0;    //to allow tranform for mode=0 in P slice when img->NoResidueDirect=0 to behave like Pdirect...Rahul
#endif

  //===== loop over 4x4 blocks =====
  for (byy=0, block_y=mb_y; block_y<mb_y+8; byy+=4, block_y+=4)
  {
    pic_pix_y = img->opix_y + block_y;
    for (bxx=0, block_x=mb_x; block_x<mb_x+8; bxx+=4, block_x+=4)
    {
      pic_pix_x = img->opix_x + block_x;
      cbp_blk_mask = (block_x>>2) + block_y;
 
      if(isfirstQP || (mode<=0))
      {
        //===== prediction of 4x4 block =====
        LumaPrediction4x4 (block_x, block_y, p_dir, fw_mode, bw_mode, fw_refframe, bw_refframe);
        if(mode==1 || mode == 2 || mode == 3 || mode == 4 || mode == 5 || mode==6 || mode ==7)
        {
          int mb64_x = mbi_x<<4;
          int mb64_y = mbi_y<<4;

          if(mode == 4 || mode == 5 || mode==6 || mode ==7)//for submacroblock_mode_decision
            mode_idx = mode+25;

          for(j=block_y; j<block_y+4; j++)
            for(i=block_x; i<block_x+4; i++)
              backup_mpr[mode_idx][mb64_y+j][mb64_x+i] = img->mpr[j][i];
        }
      }
      else 
      {//!isfirstQP && mode>0 && mode!=4
        int mb64_x = mbi_x<<4;
        int mb64_y = mbi_y<<4;

          if(mode == 4 || mode == 5 || mode==6 || mode ==7)//for submacroblock_mode_decision
            mode_idx = mode+25;

        for(j=block_y; j<block_y+4; j++)
          for(i=block_x; i<block_x+4; i++)
            img->mpr[j][i] = backup_mpr[mode_idx][mb64_y+j][mb64_x+i];

      }

      //===== get displaced frame difference ======
        if(!need_8x8_transform)
        {
          //===== get displaced frame difference ======                
          for (j=0; j<4; j++)
            for (i=0; i<4; i++)
              img->m7[j][i] = imgY_org[pic_pix_y + j][pic_pix_x + i] - img->mpr[j+block_y][i+block_x];

          //===== DCT, Quantization, inverse Quantization, IDCT, Reconstruction =====      
          if ( (img->NoResidueDirect != 1 && !skipped  ) || ((img->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1) )
          {
            //===== DCT, Quantization, inverse Quantization, IDCT, Reconstruction =====
#ifdef USE_INTRA_MDDT
            nonzero = dct_luma   (block_x, block_y, &coeff_cost, 0, -1);
#else
            nonzero = dct_luma   (block_x, block_y, &coeff_cost, 0);
#endif
            if (nonzero)
            {
              (*cbp_blk) |= (int64)1 << cbp_blk_mask;  // one bit for every 4x4 block
              (*cbp)     |= cbp_mask;           // one bit for the 4x4 blocks of an 8x8 block
            }
          }
        }
        else    // store the 4x4 residual to build 8x8 residual block...and later 8x8 transform will be applied on 8x8 residual block
        {
          for (j=0; j<4; j++)
            for (i=0; i<4; i++)
              img->m7[j+byy][i+bxx] = imgY_org[pic_pix_y + j][pic_pix_x+i] - img->mpr[j+block_y][i+block_x];
        }
    }
  }//end===== loop over 4x4 blocks =====


  if(need_8x8_transform)
  {
    if (img->NoResidueDirect != 1 && !skipped)
    {
#ifdef USE_INTRA_MDDT
        nonzero = dct_luma8x8   (block8x8, &coeff_cost, 0, -1);
#else
        nonzero = dct_luma8x8   (block8x8, &coeff_cost, 0);
#endif
    
      if (nonzero)
      {
        (*cbp_blk) |= 51 << (4*block8x8-2*(block8x8 & 0x01)); // corresponds to 110011, as if all four 4x4 blocks contain coeff, shifted to block position
        (*cbp)     |= cbp_mask;           // one bit for the 4x4 blocks of an 8x8 block
      }
    }
  }
  
  /*
  The purpose of the action below is to prevent that single or 'expensive' coefficients are coded.
  With 4x4 transform there is larger chance that a single coefficient in a 8x8 or 16x16 block may be nonzero.
  A single small (level=1) coefficient in a 8x8 block will cost: 3 or more bits for the coefficient,
  4 bits for EOBs for the 4x4 blocks,possibly also more bits for CBP.  Hence the total 'cost' of that single
  coefficient will typically be 10-12 bits which in a RD consideration is too much to justify the distortion improvement.
  The action below is to watch such 'single' coefficients and set the reconstructed block equal to the prediction according
  to a given criterium.  The action is taken only for inter luma blocks.
  
    Notice that this is a pure encoder issue and hence does not have any implication on the standard.
    coeff_cost is a parameter set in dct_luma() and accumulated for each 8x8 block.  If level=1 for a coefficient,
    coeff_cost is increased by a number depending on RUN for that coefficient.The numbers are (see also dct_luma()): 3,2,2,1,1,1,0,0,...
    when RUN equals 0,1,2,3,4,5,6, etc.
    If level >1 coeff_cost is increased by 9 (or any number above 3). The threshold is set to 3. This means for example:
    1: If there is one coefficient with (RUN,level)=(0,1) in a 8x8 block this coefficient is discarded.
    2: If there are two coefficients with (RUN,level)=(1,1) and (4,1) the coefficients are also discarded
    sum_cnt_nonz is the accumulation of coeff_cost over a whole macro block.  If sum_cnt_nonz is 5 or less for the whole MB,
    all nonzero coefficients are discarded for the MB and the reconstructed block is set equal to the prediction.
  */

  if (img->NoResidueDirect != 1 && !skipped && coeff_cost <= _LUMA_COEFF_COST_ &&
    ((img->qp + img->bitdepth_luma_qp_scale)!=0 || img->lossless_qpprime_flag==0))
  {
    coeff_cost  = 0;
    (*cbp)     &=  (63 - cbp_mask);
    (*cbp_blk) &= ~(51 << (4*block8x8-2*(block8x8 & 0x01)));
   
      for (j=mb_y; j<mb_y+8; j++)
        memcpy(&enc_picture->imgY[img->pix_y + j][img->pix_x + mb_x], &img->mpr[j][mb_x], 2 * BLOCK_SIZE * sizeof(imgpel));
  }

  return coeff_cost;
}


/*!
************************************************************************
* \brief
*    Set mode parameters and reference frames for an 8x8 block
************************************************************************
*/
void SetModesAndRefframe (int b8, short* p_dir, int* fw_mode, int* bw_mode, short* fw_ref, short* bw_ref)
{
  Macroblock* currMB = &img->mb_data[img->current_mb_nr];
  int         j      = 2*(b8>>1);
  int         i      = 2*(b8 & 0x01);

  *fw_mode = *bw_mode = *fw_ref = *bw_ref = -1;
  
  *p_dir  = currMB->b8pdir[b8];
  


  if (img->type!=B_SLICE)
  {
    *fw_ref = enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i];
    *bw_ref = 0;
    *fw_mode  = currMB->b8mode[b8];
    *bw_mode  = 0;
  }
  else
  {
    if (currMB->b8pdir[b8]==-1)
    {
      *fw_ref   = -1;
      *bw_ref   = -1;
      *fw_mode  =  0;
      *bw_mode  =  0;
    }
    else if (currMB->b8pdir[b8]==0)
    {
      *fw_ref   = enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i];
      *bw_ref   = 0;
      *fw_mode  = currMB->b8mode[b8];
      *bw_mode  = 0;
    }
    else if (currMB->b8pdir[b8]==1)
    {
      *fw_ref   = 0;
      *bw_ref   = enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i];
      *fw_mode  = 0;
      *bw_mode  = currMB->b8mode[b8];
    }
    else
    {
      *fw_ref   = enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i];
      *bw_ref   = enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i];
      *fw_mode  = currMB->b8mode[b8];
      *bw_mode  = currMB->b8mode[b8];
    }
  }
}

#ifdef MB32X32
void ComputeResidue (imgpel **curImg, int img_m7[16][16], int mb_y, int mb_x, int opix_y, int opix_x, int width, int height)
#else
static void ComputeResidue (imgpel **curImg, imgpel mpr[16][16], int img_m7[16][16], int mb_y, int mb_x, int opix_y, int opix_x, int width, int height)
#endif
{
  static imgpel *imgOrg, *imgPred;
  static int    *m7;
  int i, j;

  for (j = mb_y; j < mb_y + height; j++)
  {
    imgOrg = &curImg[opix_y + j][opix_x];    
#ifdef MB32X32
    imgPred = &(img->mpr[j][mb_x]);
#else
    imgPred = &mpr[j][mb_x];
#endif
    m7 = &img_m7[j][mb_x]; 
    for (i = 0; i < width; i++)
    {
      *m7++ = *imgOrg++ - *imgPred++;
    }
  }
}

static void LumaResidualCodingBigBlock(int mbi_x, int mbi_y, int mode, int mode_idx)
{
  int block8x8;
  int fw_mode, bw_mode;
  short p_dir, refframe, bw_ref;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int nonzero; 
  int pic_pix_y, pic_pix_x;
  int byy, block_y;
  int bxx, block_x;
  int mb_x, mb_y;
  static imgpel backup_mpr[100][64][64];  

  currMB->cbp     = 0 ;
  currMB->cbp_blk = 0 ;

  assert(currMB->luma_transform_size_8x8_flag != 0);

  for (block8x8=0; block8x8<4; block8x8++)
  {    
    mb_y       = (block8x8 >> 1) << 3;
    mb_x       = (block8x8 & 0x01) << 3;

    SetModesAndRefframe (block8x8, &p_dir, &fw_mode, &bw_mode, &refframe, &bw_ref);

    if (((p_dir == 0 || p_dir == 2 )&& fw_mode < 5) || ((p_dir == 1 || p_dir == 2 ) && bw_mode < 5))
    {
      for (byy=0, block_y=mb_y; block_y<mb_y+8; byy+=4, block_y+=4)
      {
        pic_pix_y = img->opix_y + block_y;
        
        for (bxx=0, block_x=mb_x; block_x<mb_x+8; bxx+=4, block_x+=4)
        {
          pic_pix_x = img->opix_x + block_x;
          if(isfirstQP || (mode<=0 || mode>=4))
          {
            //===== prediction of 4x4 block =====
            LumaPrediction4x4 (block_x, block_y, p_dir, fw_mode, bw_mode, refframe, bw_ref);
            if(mode>0 && mode<4)
            {
              int i,j;
              int mb64_x = mbi_x<<4;
              int mb64_y = mbi_y<<4;
              for(j=block_y; j<block_y+4; j++)
                for(i=block_x; i<block_x+4; i++)
                  backup_mpr[mode_idx][mb64_y+j][mb64_x+i] = img->mpr[j][i];
            }
          }
          else
          {//!isfirstQP && mode>0 && mode<4
            int i,j;
            int mb64_x = mbi_x<<4;
            int mb64_y = mbi_y<<4;
            for(j=block_y; j<block_y+4; j++)
              for(i=block_x; i<block_x+4; i++)
                img->mpr[j][i] = backup_mpr[mode_idx][mb64_y+j][mb64_x+i];
          }
        }
      }
      //Rahul---img->mpr[0-15][0-15] contains 16x16 prediction
      //===== compute prediction residual ======     
#ifdef MB32X32
      ComputeResidue (imgY_org, img->m7, mb_y, mb_x, img->opix_y, img->opix_x + mb_x, 8, 8);
#else
      ComputeResidue (imgY_org, img->mpr, img->m7, mb_y, mb_x, img->opix_y, img->opix_x + mb_x, 8, 8);
#endif
    }
  }

#ifdef MB32X32
  if( (currMB->mb_type < 2 && currMB->luma_transform_size_8x8_flag == 2) || only_16x16_transform==1  )
#else
  if(currMB->mb_type < 2 && currMB->luma_transform_size_8x8_flag == 2)    // 16x16 transform for 16x16 partition or direct
#endif
  {
      nonzero = dct_luma_16x16_new();
      if (nonzero)
      {
        currMB->cbp_blk = 0xffff; // all sixteen 4x4 blocks contain coeff
        currMB->cbp = 0xf;        // alll 8x8 blocks contain coeff
      }
  }
  else if(currMB->mb_type == 2 && currMB->luma_transform_size_8x8_flag == 2)    // 16x8 transform
  {
    // top block
    nonzero = dct_luma_16x8(0);
    if (nonzero)
    {
      currMB->cbp_blk |= 0x00ff; // yye: double check
      currMB->cbp |= 0x3;        // yye: double check
    }
    // bottom block
    nonzero = dct_luma_16x8(1);
    if (nonzero)
    {
      currMB->cbp_blk |= 0xff00; // yye: double check
      currMB->cbp |= 0xc;        // yye: double check
    }
  }
  else if(currMB->mb_type == 3 && currMB->luma_transform_size_8x8_flag == 2)    // 8x16 transform
  {
    // top block
    nonzero = dct_luma_8x16(0);
    if (nonzero)
    {
      currMB->cbp_blk |= 0x3333; // yye: double check
      currMB->cbp |= 0x5;        // yye: double check
    }
    // bottom block
    nonzero = dct_luma_8x16(1);
    if (nonzero)
    {
      currMB->cbp_blk |= 0xcccc; // yye: double check
      currMB->cbp |= 0xa;        // yye: double check
    }
  }

}



/*!
************************************************************************
* \brief
*    Residual Coding of a Luma macroblock (not for intra)
************************************************************************
*/
void LumaResidualCoding (int mbi_x, int mbi_y, int mode, int mode_idx)
{
  int j,block8x8;
  int fw_mode, bw_mode;
  short p_dir, refframe;
  int sum_cnt_nonz;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  
  if(((currMB->mb_type >= 1 && currMB->mb_type <= 3) && currMB->luma_transform_size_8x8_flag == 2) ||
#if P_DIRECT
    (currMB->mb_type == 0                         && !img->NoResidueDirect && currMB->luma_transform_size_8x8_flag == 2))
#else
    (currMB->mb_type == 0 && img->type == B_SLICE && !img->NoResidueDirect && currMB->luma_transform_size_8x8_flag == 2))
#endif
  {
    LumaResidualCodingBigBlock(mbi_x, mbi_y, mode, mode_idx);
    return;
  }

  currMB->cbp     = 0 ;
  currMB->cbp_blk = 0 ;
  sum_cnt_nonz    = 0 ;
  
  for (block8x8=0; block8x8<4; block8x8++)
  {
    short bw_ref;
    SetModesAndRefframe (block8x8, &p_dir, &fw_mode, &bw_mode, &refframe, &bw_ref);
    
    sum_cnt_nonz += LumaResidualCoding8x8 (mbi_x, mbi_y, mode, mode_idx, &(currMB->cbp), &(currMB->cbp_blk), block8x8,
      p_dir, fw_mode, bw_mode, refframe, bw_ref);
  }
  
  if (sum_cnt_nonz <= _LUMA_MB_COEFF_COST_ )
  {
    currMB->cbp     &= 0xfffff0 ;
    currMB->cbp_blk &= 0xff0000 ;
      for (j=0; j < MB_BLOCK_SIZE; j++)
        memcpy(&enc_picture->imgY[img->pix_y+j][img->pix_x], img->mpr[j], MB_BLOCK_SIZE * sizeof (imgpel));
  }
}


/*!
************************************************************************
* \brief
*    Makes the decision if 8x8 tranform will be used (for RD-off)
************************************************************************
*/
int TransformDecision (int block_check, int *cost)
{
  int    block_y, block_x, pic_pix_y, pic_pix_x, i, j, k;
  int    mb_y, mb_x, block8x8;
  int    fw_mode, bw_mode;
  short  p_dir, fw_ref, bw_ref;
  int    num_blks;
  int    cost8x8=0, cost4x4=0;
  int    diff[64], *diff_ptr;
  
  if(block_check==-1)
  {
    block8x8=0;
    num_blks=4;
  }
  else
  {
    block8x8=block_check;
    num_blks=block_check+1;
  }
  
  for (; block8x8<num_blks; block8x8++)
  {
    SetModesAndRefframe (block8x8, &p_dir, &fw_mode, &bw_mode, &fw_ref, &bw_ref);
    
    mb_y = (block8x8 >> 1) << 3;
    mb_x = (block8x8 & 0x01) << 3;
    //===== loop over 4x4 blocks =====
    k=0;
    for (block_y=mb_y; block_y<mb_y+8; block_y+=4)
    {
      pic_pix_y = img->opix_y + block_y;
      
      for (block_x=mb_x; block_x<mb_x+8; block_x+=4)
      {
        pic_pix_x = img->opix_x + block_x;
        
        //===== prediction of 4x4 block =====
        LumaPrediction4x4 (block_x, block_y, p_dir, fw_mode, bw_mode, fw_ref, bw_ref);
        
        //===== get displaced frame difference ======
        diff_ptr=&diff[k];
        for (j=0; j<4; j++)
        {
          for (i=0; i<4; i++, k++)            
            diff[k] = imgY_org[pic_pix_y+j][pic_pix_x+i] - img->mpr[j+block_y][i+block_x];
        }
        cost4x4 += SATD (diff_ptr, input->hadamard);
      }
    }
    cost8x8 += SATD8X8 (diff, input->hadamard);
  }
  
  if(input->Transform8x8Mode==2) //always allow 8x8 transform
    return 1;
  else if(cost8x8<cost4x4)
    return 1;
  else
  {
    *cost = (*cost-cost8x8+cost4x4);
    return 0;
  }
}

static void get_fullpel_blockUV(imgpel **  ref, int x_pres, int y_pres, int max_x, int max_y, int block[32][32], int size_x, int size_y)
{
  int i, j;

  for (j = 0; j < size_y; j++)
    for (i = 0; i < size_x; i++)
      block[j][i] = ref[max(0,min(y_pres+j,max_y))][max(0,min(x_pres+i,max_x))];
}

static void interp_blockUV_direct_X(imgpel **ref, int pres_x, int pres_y, const int *coefx, int max_x, int max_y, int block[32][32], int size_x, int size_y, int symmetric)
{
  int i, j, x;
  int result;

  const int *filtP;
  imgpel *imgP;

  // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
  if ((pres_y >= 0) && (pres_y < max_y - size_y) && (pres_x >= 3) && (pres_x < max_x - 3 - size_x))
  { //Image Non-Border safe positions...so dont need Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefx;
          imgP = &(ref[pres_y+j][pres_x+i-2]);
          for (result = 0, x = -2; x < 4; x++)
            result += (*filtP++) * (*imgP++);
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefx;
          for (result = 0, x = -2; x <= 0; x++)
            result += (*filtP++) * (ref[pres_y+j][pres_x+i+x] + ref[pres_y+j][pres_x+i-x+1]);
          block[j][i] = result;
        }
    }
  }
  else
  { //Image Border positions...so needs Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          for (result = 0, x = -2; x < 4; x++)
            result += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * coefx[x+2];
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          for (result = 0, x = -2; x <= 0; x++)
            result += coefx[x+2] * (ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
                                    ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);

          block[j][i] = result;
        }
    }
  }
}

static void interp_blockUV_direct_Y(imgpel **ref, int pres_x, int pres_y, const int *coefy, int max_x, int max_y, int block[32][32], int size_x,int size_y, int symmetric)
{
  int i, j, y;
  int result;

  const int *filtP;

  // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
  if ((pres_y >= 3) && (pres_y < max_y - 3 - size_y) && (pres_x >= 0) && (pres_x < max_x - size_x))
  { //Image Non-Border safe positions...so dont need Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y < 4; y++)
            result += ref[pres_y+j+y][pres_x+i] * (*filtP++);
          block[j][i] = result;

        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y <=0; y++)
            result += (*filtP++) * (ref[pres_y+j+y][pres_x+i] +
                                    ref[pres_y+j-y+1][pres_x+i]);
          block[j][i] = result;
        }
    }
  }
  else
  { //Image Border positions...so needs Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y < 4; y++)
            result += ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] * (*filtP++);
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y <= 0; y++)
            result += (*filtP++) * (ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
                                    ref[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
          block[j][i] = result;
        }
    }
  }

}

void get_eighthpel_blockUV_direct(imgpel **ref, int x_pos, int y_pos, ImageParameters *img, int block[32][32], int geometry, int size_x, int size_y)
{
  int i, j;

  static const int COEF_6tap[3][6] =
  {
    {  2,  -5, 28,  9, -3, 1 },
    {  2,  -6, 20, 20, -6, 2 },
    {  1,  -3,  9, 28, -5, 2 }
  };

  static const int COEF_6tap_EIGHTH_PEL[4][6] =
  {
    {  2, -5,  60,  9, -3,  1 },
    {  4, -11, 48, 29, -9,  3 },
    {  3, -9,  29, 48, -11, 4 },
    {  1, -3,   9, 60, -5,  2 }
  };

/*
// Filters summing up to 256
  static const int COEF_6tap[3][6] =
  {
    { 11, -36, 226,  71, -24,  8 },
    { 10, -40, 158, 158, -40, 10 },
    { 8, -24,  71, 226, -36, 11 }
  };

  static const int COEF_6tap_EIGHTH_PEL[4][6] =
  {
    { 11, -36, 482,  71, -24,  8 },
    { 21, -76, 384, 229, -64, 18 },
    { 18, -64, 229, 384, -76, 21 },
    { 8,  -24,  71, 482, -36, 11 }
  };
*/

  int chroma_size_x = geometry?1:size_x; // If the geometry mode is ON, only 1x1 block is processed. Otherwise size_x x size_y. 
  int chroma_size_y = geometry?1:size_y; // If the geometry mode is ON, only 1x1 block is processed. Otherwise size_x x size_y. 

#ifdef EIGHTH_PEL
  int dx = (x_pos>>img->mv_res) & 7, x = 0;
  int dy = (y_pos>>img->mv_res) & 7, y = 0;
  int pres_x = x_pos >> (3+img->mv_res);
  int pres_y = y_pos >> (3+img->mv_res);
#else
  int dx = x_pos & 7, x = 0;
  int dy = y_pos & 7, y = 0;
  int pres_x = x_pos >> 3;
  int pres_y = y_pos >> 3;
#endif

  //int dx = x_pos & 7;
  //int dy = y_pos & 7;

  //int pres_x = x_pos >> 3;
  //int pres_y = y_pos >> 3;


  int tmp[32][32+16]; // 16 is a conservative choice
  int result;

  int max_x = img->width/2 - 1;
  int max_y = img->height/2 - 1;

  int round5 = 16;
  int round6 = 32;
  int round10 = (1<<9);
  int round11 = (1<<10);
  int round12 = (1<<11);

  int filt_index, symmetric;
  const int *filtP;
  imgpel *imgP;
  int *intP;


  // choose filter depending on the sub-pel position
  if (dx == 0 && dy == 0)
  { // full-pel position, just return full-pel block
    get_fullpel_blockUV(ref, pres_x, pres_y, max_x, max_y, block, chroma_size_x, chroma_size_y);
  }
  else if (dy == 0)
  { // horizontal interpolation only
    if (dx&1)
    { // 1/8-pel position
      interp_blockUV_direct_X(ref, pres_x, pres_y, COEF_6tap_EIGHTH_PEL[dx>>1], max_x, max_y, block, chroma_size_x, chroma_size_y, 0);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round6) >> 6); // Filter sums up to 64 for 1/8 pel
    }
    else
    {
      filt_index = (dx>>1)-1;
      symmetric = (filt_index == 1);
      interp_blockUV_direct_X(ref, pres_x, pres_y, COEF_6tap[filt_index], max_x, max_y, block, chroma_size_x, chroma_size_y, symmetric);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round5) >> 5); // Filter sums up to 32 for 1/4 pel
    }
  }
  else if (dx == 0)
  { // vertical interpolation only
    if (dy&1)
    { // 1/8-pel position (vertical)
      interp_blockUV_direct_Y(ref, pres_x, pres_y, COEF_6tap_EIGHTH_PEL[dy>>1], max_x, max_y, block, chroma_size_x, chroma_size_y, 0);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round6) >> 6); // Filter sums up to 64 for 1/8 pel
    }
    else
    {
      filt_index = (dy>>1)-1;
      symmetric = (filt_index == 1);
      interp_blockUV_direct_Y(ref, pres_x, pres_y, COEF_6tap[filt_index], max_x, max_y, block, size_x, size_y, symmetric);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round5) >> 5); // Filter sums up to 32 for 1/4 pel
    }
  }
  else if (!(dx&1))
  {  // horizontal 1/4-pel
    filt_index = (dx>>1)-1;
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so dont need Clipping
      if (filt_index != 1)
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            imgP = &(ref[pres_y+j][pres_x+i-2]);
            filtP = COEF_6tap[filt_index];
            for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
              tmp[i][j+2] += (*imgP++) * (*filtP++); // indices (i,j) reversed to facilitate further filtering
          }
      }
      else
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            filtP = COEF_6tap[filt_index];
            for (tmp[i][j+2] = 0, x = -2; x <= 0; x++)
              tmp[i][j+2] += (*filtP++) * (ref[pres_y+j][pres_x+i+x] + ref[pres_y+j][pres_x+i-x+1]);
                // indices (i,j) reversed to facilitate further filtering
          }
      }
    }
    else
    {//Image Border positions...so needs Clipping
      if (filt_index != 1)
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
              tmp[i][j+2] += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x )] * COEF_6tap[filt_index][x+2];
                // indices (i,j) reversed to facilitate further filtering
          }
      }
      else
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            for (tmp[i][j+2] = 0, x = -2; x <= 0; x++)
              tmp[i][j+2] += COEF_6tap[filt_index][2+x] * (ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
                                                           ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);
                // indices (i,j) reversed to facilitate further filtering
          }
      }
    }
    // At this point the nominal bit-depth has been increased by 5 (1/4-pel filtering, Filter sums up to 32).

    if (dy&1)
    { // 1/8-pel position (vertical)
      filt_index = (dy>>1);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
        {
          intP = &(tmp[i][j]);
          filtP = COEF_6tap_EIGHTH_PEL[filt_index];
          for (result = 0, y = -2; y < 4; y++)
            result += (*filtP++) * (*intP++);
          block[j][i] = Clip3(0, img->max_imgpel_value, (result + round11) >> 11);
            // 1/8-pel filtering, Filter sums up to 512. Total nominal bit-depth increased by 17.
        }
    }
    else
    { // 1/4-pel position (vertical)
      filt_index = (dy>>1)-1;
      if (filt_index != 1)
      {
        for (j = 0; j < chroma_size_y; j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            intP = &(tmp[i][j]);
            filtP = COEF_6tap[filt_index];
            for (result = 0, y = -2; y < 4; y++)
              result += (*filtP++) * (*intP++);
            block[j][i] = Clip3(0, img->max_imgpel_value, (result + round10) >> 10);
              // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.
          }
      }
      else
      {
        for (j = 0; j < chroma_size_y; j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            filtP = COEF_6tap[filt_index];
            for (result = 0, y = -2; y <= 0; y++)
              result += (*filtP++) * (tmp[i][j+y+2] + tmp[i][j-y+1+2]);
            block[j][i] = Clip3(0, img->max_imgpel_value, (result + round10) >> 10);
              // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.
          }
      }
    }
  }
  else if (!(dy&1))
  { // vertical 1/4-pel and horizontal 1/8-pel interpolation
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so don't need Clipping
      filt_index = (dy>>1)-1;
      if (filt_index != -1)
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x+4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y < 4; y++)
              tmp[j][i+2] += (*filtP++) * ref[pres_y+j+y][pres_x+i];
          }
      }
      else
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x+4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y <= 0; y++)
              tmp[j][i+2] += (*filtP++) * (ref[pres_y+j+y][pres_x+i] +
                                           ref[pres_y+j-y+1][pres_x+i]);
          }
      }
    }
    else
    { //Image Border positions...so needs Clipping
      filt_index = (dy>>1)-1;
      if (filt_index != -1)
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x+4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y < 4; y++)
              tmp[j][i+2] += (*filtP++) * ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)];
          }
      }
      else
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x + 4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y <= 0; y++)
              tmp[j][i+2] += (*filtP++) * (ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
                                           ref[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
          }
      }
    }
    // At this point the nominal bit-depth has been increased by 8 (1/4-pel filtering, Filter sums up to 256).

    // horizontal 1/8 pel position filtering
    filt_index = (dx>>1);
    for (j = 0; j < chroma_size_y; j++)
      for (i = 0; i < chroma_size_x; i++)
      {
        intP = &(tmp[j][i]);
        filtP = COEF_6tap_EIGHTH_PEL[filt_index];
        for (result = 0, y = -2; y < 4; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round11) >> 11);
          // 1/8-pel filtering, Filter sums up to 64. Total nominal bit-depth increased by 11.
      }
  }       
  else
  { // 1/8-1/8 positions
    // Horizontal 1/8 pel position filtering
    filt_index = (dx>>1);
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so dont need Clipping
      for (j = -2; j < (chroma_size_y+4); j++)
        for (i = 0; i < chroma_size_x; i++)
        {
          imgP = &(ref[pres_y+j][pres_x+i-2]);
          filtP = COEF_6tap_EIGHTH_PEL[filt_index];
          for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
            tmp[i][j+2] += (*imgP++) * (*filtP++);
        }
    }
    else
    {//Image Border positions...so needs Clipping
      for (j = -2; j < (chroma_size_y+4); j++)
        for (i = 0; i < chroma_size_x; i++)
        {
          for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
            tmp[i][j+2] += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * COEF_6tap_EIGHTH_PEL[filt_index][x+2];
        }
    }
    // At this point the nominal bit-depth has been increased by 6 (1/8-pel filtering, Filter sums up to 64).

    // vertical 1/8 pel position filtering
    filt_index = (dy>>1);
    for (j = 0; j < chroma_size_y; j++)
      for (i = 0; i < chroma_size_x; i++)
      {
        intP = &(tmp[i][j]);
        filtP = COEF_6tap_EIGHTH_PEL[filt_index];
        for (result = 0, y = -2; y < 4; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round12) >> 12);
      }

  }
}


/*!
************************************************************************
* \brief
*    Predict one component of a chroma 4x4 block
************************************************************************
*/
void OneComponentChromaPrediction4x4 (int*        mpred,      //!< array to store prediction values
                                      int         block_c_x,  //!< horizontal pixel coordinate of 4x4 block
                                      int         block_c_y,  //!< vertical   pixel coordinate of 4x4 block
                                      short****** mv,         //!< motion vector array
                                      int         list_idx,   //!< reference picture list
                                      short       ref,        //!< reference index
                                      int         blocktype,  //!< block type
                                      int         uv)         //!< chroma component
{
  int     i, j, ii, jj;
  short*  mvb;
#ifdef EIGHTH_PEL  
  int mv_shift= img->mv_res;
#endif  

#ifdef EIGHTH_PEL
  int     f1_x = (64/img->mb_cr_size_x) << mv_shift;
  int     f1_y = (64/img->mb_cr_size_y) << mv_shift;
#else
  int     f1_x = (64/img->mb_cr_size_x);
  int     f1_y = (64/img->mb_cr_size_y);
#endif
  
  int     list_offset = img->mb_data[img->current_mb_nr].list_offset;  
  int     jjx, iix;
  int     mb_cr_y_div4 = img->mb_cr_size_y>>2;
  int     mb_cr_x_div4 = img->mb_cr_size_x>>2;
  int     jpos;
  
  StorablePicture **list = listX[list_idx + list_offset];
  pel_t** refimage = list[ref]->imgUV[uv];
  int blockUV[32][32];

  unsigned short m=0;
  int geometry=1;
  
  for (j=block_c_y; j < block_c_y + BLOCK_SIZE; j++)
  {
    jjx = j/mb_cr_y_div4;
    jpos = (j + img->opix_c_y)*f1_y;//if img->mb_cr_size_y=8, f1_y = 8
    
    for (i=block_c_x; i < block_c_x + BLOCK_SIZE; i++)
    {
      iix = i/mb_cr_x_div4;
      mvb  = mv [jjx][iix][list_idx][ref][blocktype];
      
      // <FTRD Compatibility with APEC and chroma bug fix
#ifdef MV_COMPETITION
      if ((skip_mode== TRUE) &&  (input->mv_competition > 0))
      {
        mvb[0] = (short) mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][0];
        mvb[1] = (short) mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][1];
      }
#endif
      // FTRD>
      
      //(ii, jj) are coordinators in 4 times upsampled luma grid
      ii   = (i + img->opix_c_x)*f1_x + mvb[0];
      jj   = jpos + mvb[1];
      
      if (active_sps->chroma_format_idc == 1)
        jj  += list[ref]->chroma_vector_adjustment;

//   0  1  2  3
//   4  5  6  7
//   8  9 10 11
//  12 13 14 15
      if (!geometry)
      {
      if(m==0 || m==2 || m==8 || m==10)    
      {
          get_eighthpel_blockUV_direct(refimage, ii, jj, img, blockUV, 0, 2, 2);
        mpred[m+0] = blockUV[0][0];
          mpred[m+1] = blockUV[0][1];
          mpred[m+4] = blockUV[1][0];
        mpred[m+5] = blockUV[1][1];
      }
      m++;
      }
      else
      {
        get_eighthpel_blockUV_direct(refimage, ii, jj, img, blockUV, 1, 2, 2);
        *mpred++ = blockUV[0][0];
      }
    }
  }
}


/*!
************************************************************************
* \brief
*    Predict an intra chroma 4x4 block
************************************************************************
*/
void IntraChromaPrediction4x4 (int  uv,       // <-- colour component
                               int  block_x,  // <-- relative horizontal block coordinate of 4x4 block
                               int  block_y)  // <-- relative vertical   block coordinate of 4x4 block
{
  int mode = img->mb_data[img->current_mb_nr].c_ipred_mode;
  int j;
  
  //===== prediction =====
  for (j=block_y; j<block_y+4; j++)
    memcpy(&img->mpr[j][block_x],&img->mprr_c[uv][mode][j][block_x], BLOCK_MULTIPLE * sizeof(imgpel));
}



/*!
************************************************************************
* \brief
*    Predict one chroma 4x4 block
************************************************************************
*/
void
ChromaPrediction4x4 (int   uv,           // <-- colour component
                     int   block_x,      // <-- relative horizontal block coordinate of 4x4 block
                     int   block_y,      // <-- relative vertical   block coordinate of 4x4 block
                     int   p_dir,        // <-- prediction direction
                     int   fw_mode,      // <-- list0  prediction mode (1-7, 0=DIRECT if bw_mode=0)
                     int   bw_mode,      // <-- list1 prediction mode (1-7, 0=DIRECT if fw_mode=0)
                     short fw_ref_idx,   // <-- reference frame for list0 prediction (if (<0) -> intra prediction)
                     short bw_ref_idx)   // <-- reference frame for list1 prediction 
{
  static int fw_pred[MB_BLOCK_SIZE];
  static int bw_pred[MB_BLOCK_SIZE];
  
  int  i, j;
  int  block_x4  = block_x+4;
  int  block_y4  = block_y+4;
  int* fpred     = fw_pred;
  int* bpred     = bw_pred;
  short****** mv_array = img->all_mv;
  
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  
  int  apply_weights = ( (active_pps->weighted_pred_flag && (img->type == P_SLICE)) ||
    (active_pps->weighted_bipred_idc && (img->type == B_SLICE)));  
  
  if (currMB->bi_pred_me && fw_ref_idx == 0 && bw_ref_idx == 0 && p_dir == 2 && fw_mode==1 && bw_mode==1)
    mv_array = currMB->bi_pred_me == 1? img->bipred_mv1 : img->bipred_mv2 ;
  
  //===== INTRA PREDICTION =====
  if (p_dir==-1)
  {
    IntraChromaPrediction4x4 (uv, block_x, block_y);
    return;
  }
  
  //===== INTER PREDICTION =====
  if ((p_dir==0) || (p_dir==2))
  {
    OneComponentChromaPrediction4x4 (fw_pred, block_x, block_y, mv_array, LIST_0, fw_ref_idx, fw_mode, uv);
  }
  if ((p_dir==1) || (p_dir==2))
  {
    OneComponentChromaPrediction4x4 (bw_pred, block_x, block_y, mv_array, LIST_1, bw_ref_idx, bw_mode, uv);
  }
  
  if (apply_weights)
  {
    if (p_dir==2)
    {
      for (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          if(input->UseHPFilter != 0)
          {
#ifdef BROUND_FIX
            img->mpr[j][i] =  clip1a_chr((wbp_weight[0][fw_ref_idx][bw_ref_idx][uv+1] * *fpred++ + wbp_weight[1][fw_ref_idx][bw_ref_idx][uv+1] * *bpred++ 
                  + 2*wp_chroma_round*(wp_offset[0][fw_ref_idx][uv+1] + wp_offset[1][bw_ref_idx][uv+1])
                  + 2*wp_chroma_round*(1 - img->bipred_rounding_control)) >> (chroma_log_weight_denom + 1));
#else
            img->mpr[j][i] =  clip1a_chr(((wbp_weight[0][fw_ref_idx][bw_ref_idx][uv+1] * *fpred++ + wbp_weight[1][fw_ref_idx][bw_ref_idx][uv+1] * *bpred++ 
                  + 2*wp_chroma_round - img->bipred_rounding_control) >> (chroma_log_weight_denom + 1)) + ((wp_offset[0][fw_ref_idx][uv+1] + wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
#endif
          }
          else
          {
            img->mpr[j][i] =  clip1a_chr(((wbp_weight[0][fw_ref_idx][bw_ref_idx][uv+1] * *fpred++ + wbp_weight[1][fw_ref_idx][bw_ref_idx][uv+1] * *bpred++ 
                  + 2*wp_chroma_round) >> (chroma_log_weight_denom + 1)) + ((wp_offset[0][fw_ref_idx][uv+1] + wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
          }
    }
    else if (p_dir==0)
    {
      for (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = clip1a_chr(((wp_weight[0][fw_ref_idx][uv+1] * *fpred++ + wp_chroma_round) >> chroma_log_weight_denom) +  wp_offset[0][fw_ref_idx][uv+1]);
    }
    else // (p_dir==1)
    {
      for (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = clip1a_chr(((wp_weight[1][bw_ref_idx][uv+1] * *bpred++ + wp_chroma_round) >> chroma_log_weight_denom) + wp_offset[1][bw_ref_idx][uv+1]);
    }
  }
  else
  {
    if (p_dir==2)
    {
      for (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          if(input->UseHPFilter != 0)
          {
            img->mpr[j][i] = (*fpred++ + *bpred++ + 1 - img->bipred_rounding_control) >> 1; 
          }
          else
          {
            img->mpr[j][i] = (*fpred++ + *bpred++ + 1) >> 1; 
          }
    }
    else if (p_dir==0)
    {
      for (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = *fpred++;
    }
    else // (p_dir==1)
    {
      for (j=block_y; j<block_y4; j++)
        for (i=block_x; i<block_x4; i++)  
          img->mpr[j][i] = *bpred++;
    }
  }
}


/*!
************************************************************************
* \brief
*    Chroma residual coding for an macroblock
************************************************************************
*/
void ChromaResidualCoding (int mbi_x, int mbi_y, int mode, int mode_idx, int* cr_cbp)
{
  int   uv, block8, block_y, block_x, j, i;
  int   fw_mode, bw_mode;
  short p_dir, refframe, bw_ref;
  int   yuv = img->yuv_format - 1; //ADD-VG-15052004
  static imgpel backup_mpr_c[100][2][64][64];//[uv][y][x]
  int   block8x8_idx[3][4][4] =     //ADD-VG-15052004
  { { {0, 1, 0, 0}, 
  {2, 3, 0, 0},
  {0, 0, 0, 0},
  {0, 0, 0, 0}, },
  
  { {0, 1, 0, 0}, 
  {0, 1, 0, 0}, 
  {2, 3, 0, 0},
  {2, 3, 0, 0}  },
  
  { {0, 0, 1, 1},
  {0, 0, 1, 1},
  {2, 2, 3, 3},
  {2, 2, 3, 3}  }
  };

  int   skipped = (img->mb_data[img->current_mb_nr].mb_type == 0 && (img->type == P_SLICE));
#if P_DIRECT
        skipped = 0;
#endif

  for (*cr_cbp=0, uv=0; uv<2; uv++)
  {
    //===== prediction of chrominance blocks ===d==
    block8 = 0;
    for (block_y=0; block_y < img->mb_cr_size_y; block_y+=4)
      for (block_x=0; block_x < img->mb_cr_size_x; block_x+=4)
      {
        if(isfirstQP || mode<=0 || mode>=4) //ChromaResidualCoding is not used for submacroblock_mode_decision
        {
          block8 = block8x8_idx[yuv][block_y>>2][block_x>>2];
          SetModesAndRefframe (block8, &p_dir, &fw_mode, &bw_mode, &refframe, &bw_ref);

          ChromaPrediction4x4 (uv, block_x, block_y, p_dir, fw_mode, bw_mode, refframe, bw_ref);

          if(mode>0 && mode<4)
          {
            int mb64_x = mbi_x<<3;
            int mb64_y = mbi_y<<3;
            for(j=block_y; j<block_y+4; j++)
              for(i=block_x; i<block_x+4; i++)
                backup_mpr_c[mode_idx][uv][mb64_y+j][mb64_x+i] = img->mpr[j][i];
          }
        }
        else
        {//!isfirstQP && mode>0 && mode<=P8x8
          int mb64_x = mbi_x<<3;
          int mb64_y = mbi_y<<3;
          for(j=block_y; j<block_y+4; j++)
            for(i=block_x; i<block_x+4; i++)
              img->mpr[j][i] = backup_mpr_c[mode_idx][uv][mb64_y+j][mb64_x+i];

        }
      }

      // ==== set chroma residue to zero for skip Mode in SP frames 
      if (img->NoResidueDirect)
      {
        for (j=0; j<img->mb_cr_size_y; j++)
          memcpy(&enc_picture->imgUV[uv][img->pix_c_y+j][img->pix_c_x], img->mpr[j], img->mb_cr_size_x * sizeof(imgpel));             
      }
      else
      {
        if (skipped)    //this will never happen under P_DIRECT macro
        {
          for (j=0; j<img->mb_cr_size_y; j++)
            memcpy(&enc_picture->imgUV[uv][img->pix_c_y+j][img->pix_c_x], img->mpr[j], img->mb_cr_size_x * sizeof(imgpel));
        }
        else
        {
          for (j=0; j<img->mb_cr_size_y; j++)
            for (i=0; i<img->mb_cr_size_x; i++)
              img->m7[j][i] = imgUV_org[uv][img->opix_c_y+j][img->opix_c_x+i] - img->mpr[j][i];
        }
      }

        //===== DCT, Quantization, inverse Quantization, IDCT, and Reconstruction =====
        if (!img->NoResidueDirect && !skipped)
        {
          *cr_cbp=dct_chroma   (uv,*cr_cbp); 
        }
  }
  
  //===== update currMB->cbp =====
  img->mb_data[img->current_mb_nr].cbp += ((*cr_cbp)<<4);  
}

/*!
************************************************************************
* \brief
*    Intra prediction of the chrminance layers of one macroblock
************************************************************************
*/
void IntraChromaPrediction (int *mb_up, int *mb_left, int*mb_up_left)
{
  
  //Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int      s, s0, s1, s2, s3, i, j;
  pel_t**  image;
  int      block_x, block_y;
  int      mb_nr = img->current_mb_nr;
  int      mb_available_up;
  int      mb_available_left[2];
  int      mb_available_up_left;
  int      ih,iv;
  int      ib,ic,iaa;
  int      uv;
  imgpel   hline[16], vline[16];
  //int      best_mode = DC_PRED_8;  //just an initilaization here, should always be overwritten
  PixelPos up;        //!< pixel position  p(0,-1)
  PixelPos left[17];  //!< pixel positions p(-1, -1..15)
  int      cr_MB_x = img->mb_cr_size_x;
  int      cr_MB_y = img->mb_cr_size_y;
  
  //ADD-VG-07062004 
  int      blk_x;
  int      blk_y;
  int      b8,b4;
  int      yuv = img->yuv_format - 1;
  
  static int block_pos[3][4][4]= //[yuv][b8][b4]
  {
    { {0, 1, 2, 3},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
    { {0, 1, 2, 3},{2, 3, 2, 3},{0, 0, 0, 0},{0, 0, 0, 0}},
    { {0, 1, 2, 3},{1, 1, 3, 3},{2, 3, 2, 3},{3, 3, 3, 3}}
  };
  
  //ADD-VG-07062004-END 
  
  
  for (i=0;i<cr_MB_y+1;i++)
  {
    getNeighbour(mb_nr, -1 , i-1 , 0, &left[i]);
  }
  
  getNeighbour(mb_nr, 0 , -1 , 0, &up);
  
  
  mb_available_up                             = up.available;
  mb_available_up_left                        = left[0].available;
  mb_available_left[0] = mb_available_left[1] = left[1].available;
  
  if(input->UseConstrainedIntraPred)
  {
    mb_available_up = up.available ? img->intra_block[up.mb_addr] : 0;
    for (i=0, mb_available_left[0]=1; i<(cr_MB_y>>1);i++)
      mb_available_left[0]  &= left[i+1].available ? img->intra_block[left[i+1].mb_addr]: 0;
    for (i=(cr_MB_y>>1), mb_available_left[1]=1; i<cr_MB_y;i++)
      mb_available_left[1] &= left[i+1].available ? img->intra_block[left[i+1].mb_addr]: 0;
    mb_available_up_left = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
  }
  
  if (mb_up)
    *mb_up = mb_available_up;
  if (mb_left)
    *mb_left = mb_available_left[0] && mb_available_left[1];
  if (mb_up_left)
    *mb_up_left = mb_available_up_left;
  
  
  // compute all chroma intra prediction modes for both U and V
  for (uv=0; uv<2; uv++)
  {
    image = enc_picture->imgUV[uv];
    
    // DC prediction
    for(b8=0; b8<img->num_blk8x8_uv >> 1;b8++)
    {
      for (b4=0; b4<4; b4++)
      {
        block_y = subblk_offset_y[yuv][b8][b4];
        block_x = subblk_offset_x[yuv][b8][b4];
        blk_x = block_x;
        blk_y = block_y + 1;
        
        s=img->dc_pred_value_chroma;
        s0=s1=s2=s3=0;
        
        //===== get prediction value =====
        switch (block_pos[yuv][b8][b4])
        {
        case 0:  //===== TOP LEFT =====
          if      (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s0 += image[up.pos_y][up.pos_x + i];
          if      (mb_available_left[0])  for (i=blk_y;i<(blk_y+4);i++)  s2 += image[left[i].pos_y][left[i].pos_x];
          if      (mb_available_up && mb_available_left[0])  s  = (s0+s2+4) >> 3;
          else if (mb_available_up)                          s  = (s0   +2) >> 2;
          else if (mb_available_left[0])                     s  = (s2   +2) >> 2;
          break;
        case 1: //===== TOP RIGHT =====
          if      (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s1 += image[up.pos_y][up.pos_x + i];
          else if (mb_available_left[0])  for (i=blk_y;i<(blk_y+4);i++)  s2 += image[left[i].pos_y][left[i].pos_x];
          if      (mb_available_up)                          s  = (s1   +2) >> 2;
          else if (mb_available_left[0])                     s  = (s2   +2) >> 2;
          break;
        case 2: //===== BOTTOM LEFT =====
          if      (mb_available_left[1])  for (i=blk_y;i<(blk_y+4);i++)  s3 += image[left[i].pos_y][left[i].pos_x];
          else if (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s0 += image[up.pos_y][up.pos_x + i];
          if      (mb_available_left[1])                     s  = (s3   +2) >> 2;
          else if (mb_available_up)                          s  = (s0   +2) >> 2;
          break;
        case 3: //===== BOTTOM RIGHT =====
          if      (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s1 += image[up.pos_y][up.pos_x + i];
          if      (mb_available_left[1])  for (i=blk_y;i<(blk_y+4);i++)  s3 += image[left[i].pos_y][left[i].pos_x];
          if      (mb_available_up && mb_available_left[1])  s  = (s1+s3+4) >> 3;
          else if (mb_available_up)                          s  = (s1   +2) >> 2;
          else if (mb_available_left[1])                     s  = (s3   +2) >> 2;
          break;
        }
        
        //===== prediction =====
        for (j=block_y; j<block_y+4; j++)
          for (i=block_x; i<block_x+4; i++)
          {
            img->mprr_c[uv][DC_PRED_8][j][i] = s;
          }
      }
    }
    
    // vertical prediction
    if (mb_available_up)
    {
      memcpy(hline,&image[up.pos_y][up.pos_x], cr_MB_x * sizeof(imgpel));
      for (j=0; j<cr_MB_y; j++)
        memcpy(img->mprr_c[uv][VERT_PRED_8][j], hline, cr_MB_x * sizeof(imgpel));
    }
    
    // horizontal prediction 
    if (mb_available_left[0] && mb_available_left[1])
    {
      for (i=0; i<cr_MB_y; i++)
        vline[i] = image[left[i+1].pos_y][left[i+1].pos_x];
      for (i=0; i<cr_MB_x; i++)
        for (j=0; j<cr_MB_y; j++)
          img->mprr_c[uv][HOR_PRED_8][j][i] = vline[j];
    }
    
    // plane prediction
    if (mb_available_left[0] && mb_available_left[1] && mb_available_up && mb_available_up_left)
    {
      ih = (cr_MB_x>>1)*(hline[cr_MB_x-1] - image[left[0].pos_y][left[0].pos_x]);
      for (i=0;i<(cr_MB_x>>1)-1;i++)
        ih += (i+1)*(hline[(cr_MB_x>>1)+i] - hline[(cr_MB_x>>1)-2-i]);
      
      iv = (cr_MB_y>>1)*(vline[cr_MB_y-1] - image[left[0].pos_y][left[0].pos_x]);
      for (i=0;i<(cr_MB_y>>1)-1;i++)
        iv += (i+1)*(vline[(cr_MB_y>>1)+i] - vline[(cr_MB_y>>1)-2-i]);
      
      ib= ((cr_MB_x == 8?17:5)*ih+2*cr_MB_x)>>(cr_MB_x == 8?5:6);
      ic= ((cr_MB_y == 8?17:5)*iv+2*cr_MB_y)>>(cr_MB_y == 8?5:6);
      
      iaa=16*(hline[cr_MB_x-1]+vline[cr_MB_y-1]);
      for (j=0; j<cr_MB_y; j++)
        for (i=0; i<cr_MB_x; i++)
          img->mprr_c[uv][PLANE_8][j][i]=max(0,min(img->max_imgpel_value_uv,
          (iaa+(i-(cr_MB_x>>1)+1)*ib+(j-(cr_MB_y>>1)+1)*ic+16)>>5));
    }
  }
}


/*!
************************************************************************
* \brief
*    Check if all reference frames for a macroblock are zero
************************************************************************
*/
int
ZeroRef (Macroblock* currMB)
{
  int i,j;
  
  for (j=img->block_y; j<img->block_y + BLOCK_MULTIPLE; j++)
    for (i=img->block_x; i<img->block_x + BLOCK_MULTIPLE; i++)
    {
      if (enc_picture->ref_idx[LIST_0][j][i]!=0)
        return 0;
    }
    return 1;
}


/*!
************************************************************************
* \brief
*    Converts macroblock type to coding value
************************************************************************
*/
int MBType2Value (Macroblock* currMB)
{
#if !B_MBTYPE
  static const int dir1offset[3]    =   { 1,  2, 3};
  static const int dir2offset[3][3] = { { 0,  4,  8},   // 1. block forward
										                    { 6,  2, 10},   // 1. block backward
										                    {12, 14, 16}};  // 1. block bi-directional
#endif
  int mbtype, pdir0, pdir1;
  
  if (img->type!=B_SLICE)
  {
    if      (currMB->mb_type==I8MB ||currMB->mb_type==I4MB)
      return (img->type==I_SLICE ? 0 : 6);
    else if (currMB->mb_type==I16MB) return (img->type==I_SLICE ? 0 : 6) + img->i16offset;
    else if (currMB->mb_type==P8x8)
    {
#ifndef GROUP_REF
      if (input->symbol_mode==UVLC 
        && ZeroRef (currMB))         return 5;
      else      
#endif
        return 4;
    }
    else                             return currMB->mb_type;
  }
  else
  {
    mbtype = currMB->mb_type;
    pdir0  = currMB->b8pdir[0];
    pdir1  = currMB->b8pdir[3];
    
    if      (mbtype==0)       return 0;
    else if (mbtype==I4MB || mbtype==I8MB)
      return 23;
    else if (mbtype==I16MB)   return 23 + img->i16offset;
    else if (mbtype==P8x8)    return 22;
#if B_MBTYPE
    else if (mbtype==1)       return 1;
    else if (mbtype==2)       return 2;
    else                      return 3;
#else
    else if (mbtype==1)       return dir1offset[pdir0];
    else if (mbtype==2)       return 4 + dir2offset[pdir0][pdir1];
    else                      return 5 + dir2offset[pdir0][pdir1];
#endif
  }
}


/*!
************************************************************************
* \brief
*    Writes 4x4 intra prediction modes for a macroblock
************************************************************************
*/
int writeIntra4x4Modes(void)
{
  int i;
  Macroblock    *currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int           *bitCount   = currMB->bitcounter;
  Slice         *currSlice  = img->currentSlice;
  const int     *partMap    = assignSE2partition[input->partition_mode];
  DataPartition *dataPart   = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);
  
  int rate = 0;
  
  currMB->IntraChromaPredModeFlag = 1;
  
  for(i=0;i<16;i++)
  {
    currSE->context = i;
    currSE->value1  = currMB->intra_pred_modes[i];
    currSE->value2  = 0;
    
#if TRACE
    if (currSE->value1 < 0 )
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 4x4 mode  = predicted (context: %d)",currSE->context);
    else
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 4x4 mode  = %3d (context: %d)",currSE->value1,currSE->context);
#endif
    
    // set symbol type and function pointers
    currSE->type = SE_INTRAPREDMODE;
    
    // encode and update rate
    if (input->symbol_mode == UVLC)
    {
      writeSyntaxElement_Intra4x4PredictionMode(currSE, dataPart);
    }
    else
    {
      currSE->writing = writeIntraPredMode_CABAC;
      dataPart->writeSyntaxElement (currSE, dataPart);
    }
    
    bitCount[BITS_MB_MODE]+=currSE->len;
    rate += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }
  
  return rate;
}

/*!
************************************************************************
* \brief
*    Writes 8x8 intra prediction modes for a macroblock
************************************************************************
*/
int writeIntra8x8Modes(void)
{
  int block8x8;
  Macroblock    *currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int           *bitCount   = currMB->bitcounter;
  Slice         *currSlice  = img->currentSlice;
  const int     *partMap    = assignSE2partition[input->partition_mode];
  DataPartition *dataPart   = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);
  
  int rate = 0;
  
  currMB->IntraChromaPredModeFlag = 1;
  
  for(block8x8=0;block8x8<4;block8x8++)
  {
    
    currSE->context = block8x8<<2;
    currSE->value1  = currMB->intra_pred_modes8x8[(block8x8<<2)];
    currSE->value2  = 0;
    
#if TRACE
    if (currSE->value1 < 0 )
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 8x8 mode  = predicted (context: %d)",currSE->context);
    else
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 8x8 mode  = %3d (context: %d)",currSE->value1,currSE->context);
#endif
    
    // set symbol type and function pointers
    currSE->type = SE_INTRAPREDMODE;
    
    // encode and update rate
    if (input->symbol_mode == UVLC)    
    {
      writeSyntaxElement_Intra4x4PredictionMode(currSE, dataPart);
    }
    else
    {
      currSE->writing = writeIntraPredMode_CABAC;
      dataPart->writeSyntaxElement (currSE, dataPart);
    }
    
    bitCount[BITS_MB_MODE]+=currSE->len;
    rate += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }
  
  return rate;
}

int writeIntraModes(void)
{
  switch (img->mb_data[img->current_mb_nr].mb_type)
  {
  case I4MB:
    return writeIntra4x4Modes();
    break;
  case I8MB:
    return writeIntra8x8Modes();
    break;
  default:
    return 0;
    break;
  }
}

/*!
************************************************************************
* \brief
*    Converts 8x8 block type to coding value
************************************************************************
*/
int B8Mode2Value (int b8mode, int b8pdir)
{
  static const int b8start[8] = {0,0,0,0, 1, 4, 5, 10};
  static const int b8inc  [8] = {0,0,0,0, 1, 2, 2, 1};
  
  if (img->type!=B_SLICE)
  {
    return (b8mode-4);
  }
  else
  {
    return b8start[b8mode] + b8inc[b8mode] * b8pdir;
  }
}

/*!
************************************************************************
* \brief
*    Codes macroblock header
* \param rdopt
*    true for calls during RD-optimization
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
************************************************************************
*/
int writeMBLayer (int rdopt, int *coeff_rate)
{
  int             i,j;
  int             mb_nr      = img->current_mb_nr;
  Macroblock*     currMB     = &img->mb_data[mb_nr];
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[input->partition_mode];
  int             no_bits    = 0;
  int             mb_type;
//  Macroblock*     prevMB     = mb_nr ? (&img->mb_data[prev_mb_nr]) : NULL;
//  int             prev_mb_nr = FmoGetPreviousMBNr(img->current_mb_nr);
//  int             skip       = currMB->mb_type ? 0:((img->type == B_SLICE) ? !currMB->cbp:1);
//  int             prevMbSkipped = 0;
//  Macroblock      *topMB = NULL;
  
  currMB->IntraChromaPredModeFlag = IS_INTRA(currMB);

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
  
  if(img->type == I_SLICE)
  {
    //========= write mb_type (I_SLICE) =========
    currSE->value1  = MBType2Value (currMB);
    currSE->value2  = 0;
    currSE->type    = SE_MBTYPE;
    
    if (input->symbol_mode == UVLC)  
      currSE->mapping = ue_linfo;
    else
      currSE->writing = writeMB_typeInfo_CABAC;
    
    dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE,   "mb_type (I_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
#endif
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
    
  }
  // not I_SLICE, CABAC
  else if (input->symbol_mode == CABAC)
  {

#if defined MB32X32 && !defined TEST_ONLY_Z_SCAN
    if(MB32.write_mb32_mbtype)     //this is 1 for P8x8 mode in bigblocks
    {
      Macroblock mb32x32;
      mb32x32.mb_type = P8x8;

      mb32x32.geo_mode = -1;
      no_bits += write_mb_type(&mb32x32, 1);
      currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
    }
#endif

    if(currMB->geo_mode != -1)
      mb_type         = MBType2Value_geo (currMB);
    else
      mb_type         = MBType2Value (currMB);

    //========= write mb_skip_flag (CABAC) =========
    currSE->value1  = mb_type;
    currSE->value2  = currMB->cbp;
    currSE->type    = SE_MBTYPE;
    currSE->writing = writeMB_skip_flagInfo_CABAC;
    dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
    
    CheckAvailabilityOfNeighborsCABAC();
    
    //========= write mb_type (CABAC) =========
#if P_DIRECT    //mbtype!=0 for other modes and cbp!=0 is for P/B-Direct mode
    if (currMB->mb_type != 0 || currMB->cbp != 0)
#else
    if (currMB->mb_type != 0 || ((img->type == B_SLICE) && currMB->cbp != 0))
#endif
    {
      currSE->value1  = mb_type;
      currSE->value2  = 0;
      currSE->type    = SE_MBTYPE;
#if B_MBTYPE    //to see what value to set in dir1 and dir2, check input parameter in function MBType2Value()
      if(currMB->geo_mode != -1)
      {
        //mbtype = 3;//use B_8x16
        currSE->dir_partition1  = best8x8pdir_geo[currMB->geo_mode][0];
        currSE->dir_partition2  = best8x8pdir_geo[currMB->geo_mode][1];
      }
      else
      {
        currSE->dir_partition1 = currMB->b8pdir[0];//this is for Partition 1
        currSE->dir_partition2 = currMB->b8pdir[3];//this is for Partition 2
      }
#endif
      currSE->writing = writeMB_typeInfo_CABAC;

      dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
      if (img->type == B_SLICE) 
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_type (B_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
      else                      
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_type (P_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
#endif
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;

    if(currMB->mb_type == 3 || currMB->geo_mode != -1)
    {
      currSE->value1 = currMB->geo_mode;
      currSE->value2  = rho_bits[0]+THETA_BITS;
      currSE->writing = writeMB_geoInfo_CABAC;
      dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;
    }
    }
  }
  else if (currMB->mb_type != 0 || ((img->type == B_SLICE) && currMB->cbp != 0))
  {
#ifdef MB32X32
    if(no32x32) //tm_vlc
    {
#endif
    //===== Run Length Coding: Non-Skipped macroblock =====
    currSE->value1  = img->cod_counter;
    currSE->value2  = 0;
    currSE->mapping = ue_linfo;
    currSE->type    = SE_MBTYPE;
    dataPart->writeSyntaxElement( currSE, dataPart);
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
    
#ifdef MV_COMPETITION
    if ((input->mv_competition > 0)&& (rdopt==0))
      if ((img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))  
        no_bits  += write_predictor_index_for_skip_mode();
#endif

#ifdef MB32X32
    }
    else
    {
      //non-skip
      currSE->len = u_1 ("mb_skip_flag", 0, dataPart->bitstream);    
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;
    }
#endif

    
    // Reset cod counter
      img->cod_counter = 0;
      // Put out mb mode
      if(currMB->geo_mode != -1)
        currSE->value1 = MBType2Value_geo (currMB);
      else
      currSE->value1  = MBType2Value (currMB);//tm_vlc
      
      if (img->type != B_SLICE)
      {
        currSE->value1--;

        if(currSE->value1 == 4)
        {
          if(!ZeroRef_mvres (currMB))
            currSE->value1 = 3;
        }
      }
      currSE->mapping = ue_linfo;
      currSE->type    = SE_MBTYPE;
      currSE->value2  = 0;

      dataPart->writeSyntaxElement( currSE, dataPart);
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;

      if(currMB->mb_type == 3 || currMB->geo_mode != -1)
      {
        currSE->value1 = currMB->geo_mode;
        currSE->value2  = rho_bits[0]+THETA_BITS;
        currSE->len = writeMB_geoInfo_VLC(currSE, dataPart);
        bitCount[BITS_MB_MODE] += currSE->len;
        no_bits                += currSE->len;
        currSE++;
        currMB->currSEnr++;
      }
  }
  else
  {
    //Run Length Coding: Skipped macroblock
#ifdef MB32X32//tm_vlc
    if(!no32x32)
    {
      currSE->len = u_1 ("mb_skip_flag", 1, dataPart->bitstream);    
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;
    }
    else
#endif
      img->cod_counter++;


    currMB->skip_flag = 1;
    
    // CAVLC
    for (j=0; j < (4 + img->num_blk8x8_uv); j++)
      for (i=0; i < 4; i++)
        img->nz_coeff [img->current_mb_nr][i][j]=0;
      
#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
        currMB->send_index_predictor_skip = send_index_for_skip_prediction(currMB->best_predictor_for_skip);
#endif      
      
      //
      if(FmoGetNextMBNr(img->current_mb_nr) == -1 && img->cod_counter>0)
      {
        // Put out run
        currSE->value1  = img->cod_counter;
        currSE->value2  = 0;
        currSE->mapping = ue_linfo;
        currSE->type    = SE_MBTYPE;
        
        dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_run");
#endif
        bitCount[BITS_MB_MODE] += currSE->len;
        no_bits                += currSE->len;
        currSE++;
        currMB->currSEnr++;
        
#ifdef MV_COMPETITION
        if (input->mv_competition > 0)
        {
          if(rdopt==0)
          {
            if ((img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))
              no_bits  += write_predictor_index_for_skip_mode();
          }
        }
#endif     
        
        // Reset cod counter
        img->cod_counter = 0;
        
      }
  }
  
#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    if((rdopt==0)&&(input->symbol_mode == CABAC))  
    {  
      if ((currMB->mb_type == 0) && (img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))
        no_bits  += write_predictor_index_for_skip_mode();
    }
  }
#endif

  
  //init NoMbPartLessThan8x8Flag
  currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;
  
  //===== BITS FOR 8x8 SUB-PARTITION MODES =====
  if (IS_P8x8 (currMB))
  {
    dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
    for (i=0; i<4; i++)
    {
      if (input->symbol_mode==UVLC)   
        currSE->mapping = ue_linfo;
      else
      {
#if B_MBTYPE
        currSE->dir_partition1 = currMB->b8pdir[i];
#endif
        currSE->writing = writeB8_typeInfo_CABAC;
      }
      
      currSE->value1  = B8Mode2Value (currMB->b8mode[i], currMB->b8pdir[i]);
      currSE->value2  = 0;
      currSE->type    = SE_MBTYPE;
      dataPart->writeSyntaxElement (currSE, dataPart);
      bitCount[BITS_MB_MODE]+= currSE->len;
      no_bits               += currSE->len;
      currSE++;
      currMB->currSEnr++;
      
      //set NoMbPartLessThan8x8Flag for P8x8 mode
      currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || (currMB->b8mode[i]==4);
    }

#ifdef MV_COMPETITION
    if (input->mv_competition > 0)
      pass_with_writing = !rdopt;
#endif

    no_bits += writeMotionInfo2NAL  ();
    currSE   = &img->MB_SyntaxElements[currMB->currSEnr];
  }
  
  //============= Transform size flag for INTRA MBs =============
  //-------------------------------------------------------------
  //transform size flag for INTRA_4x4 and INTRA_8x8 modes
  if ((currMB->mb_type == I8MB || currMB->mb_type == I4MB) && input->Transform8x8Mode)
  {
    currSE->value1 = currMB->luma_transform_size_8x8_flag;
    currSE->type   = SE_HEADER;
    
    if( input->symbol_mode==UVLC)
    {
      currSE->mapping = ue_linfo;
      currSE->type    = SE_MBTYPE;
      currSE->bitpattern = currMB->luma_transform_size_8x8_flag;
      currSE->len = 1;
      writeSyntaxElement2Buf_Fixed(currSE, dataPart->bitstream);
    }
    else
    {
      currSE->writing = writeMB_transform_size_CABAC;
      dataPart->writeSyntaxElement(currSE, dataPart);
    }
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }
  
  
  //===== BITS FOR INTRA PREDICTION MODES ====
  no_bits += writeIntraModes();
  //===== BITS FOR CHROMA INTRA PREDICTION MODE ====
  if (currMB->IntraChromaPredModeFlag && img->yuv_format != YUV400)
    no_bits += writeChromaIntraPredMode();
  else if(!rdopt) //GB CHROMA !!!!!
    currMB->c_ipred_mode = DC_PRED_8; //setting c_ipred_mode to default is not the right place here
  //resetting in rdopt.c (but where ??)
  //with cabac and bframes maybe it could crash without this default
  //since cabac needs the right neighborhood for the later MBs
  
  //----- motion information -----
  if (currMB->mb_type !=0 && currMB->mb_type !=P8x8)
  {
#ifdef MV_COMPETITION
    if (input->mv_competition > 0)
      pass_with_writing = !rdopt;
#endif

    if(currMB->geo_mode != -1)
      no_bits  += writeMotionInfo2NAL_geo  (currMB->geo_mode, 0);
    else
      no_bits  += writeMotionInfo2NAL  ();
  }



#if P_DIRECT    //mbtype!=0 for other modes and cbp!=0 is for P/B-Direct mode
  if (currMB->mb_type != 0 || currMB->cbp != 0)
#else
  if ((currMB->mb_type!=0) || (img->type==B_SLICE && (currMB->cbp!=0)))
#endif
  {
      no_bits    += writeCBP_Transform_size (rdopt, 0);
      no_bits    += writeDeltaQP            (rdopt, img->mb_data[img->current_mb_nr].cbp);
      *coeff_rate = writeLumaCoeff          (rdopt, 0);

    if (img->yuv_format != YUV400)
      *coeff_rate  += writeChromaCoeff ();
    
    no_bits  += *coeff_rate;
  }

  return no_bits;
}

void write_terminating_bit (short bit)
{
  DataPartition*          dataPart;
  const int*              partMap   = assignSE2partition[input->partition_mode];
  EncodingEnvironmentPtr  eep_dp;
  
  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  dataPart = &(img->currentSlice->partArr[partMap[SE_MBTYPE]]);
  dataPart->bitstream->write_flag = 1;
  eep_dp                          = &(dataPart->ee_cabac);
  
  biari_encode_symbol_final(eep_dp, bit); 
#if TRACE
  fprintf (p_trace, "      CABAC terminating bit = %d\n",bit);
#endif
  
}


/*!
************************************************************************
* \brief
*    Write chroma intra prediction mode.
************************************************************************
*/
int writeChromaIntraPredMode(void)
{
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  int*            bitCount  = currMB->bitcounter;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             rate      = 0;
  DataPartition*  dataPart;
  
  //===== BITS FOR CHROMA INTRA PREDICTION MODES
  if (input->symbol_mode==UVLC)  
    currSE->mapping = ue_linfo;
  else                            
    currSE->writing = writeCIPredMode_CABAC;
  
  currSE->value1 = currMB->c_ipred_mode;
  currSE->value2 = 0;
  currSE->type = SE_INTRAPREDMODE;
  dataPart = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);
  
  dataPart->writeSyntaxElement (currSE, dataPart);
  bitCount[BITS_MB_MODE] += currSE->len;
  rate                    += currSE->len;
#if TRACE
  snprintf(currSE->tracestring, TRACESTRING_SIZE, "Chroma intra pred mode");
#endif
  currSE++;
  currMB->currSEnr++;
  
  return rate;
}


/*!
************************************************************************
* \brief
*    Set global last_dquant according to macroblock delta qp
************************************************************************
*/

extern int last_dquant;

void set_last_dquant(void)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  
#if P_DIRECT
  if (IS_INTERMV (currMB) || IS_INTRA (currMB) || ((img->type!=I_SLICE)  && currMB->cbp != 0))
#else
  if ((IS_INTERMV (currMB) || IS_INTRA (currMB)) 
    || ((img->type==B_SLICE)  && currMB->cbp != 0))
#endif
  {
    // non-skip
    last_dquant = currMB->delta_qp;
  }
  else
  {
    // skip
    last_dquant = 0;
  }
}


/*!
************************************************************************
* \brief
*    Passes the chosen syntax elements to the NAL
************************************************************************
*/
void write_one_macroblock (int eos_bit)
{
  Macroblock* currMB   = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;
  
  extern int cabac_encoding;
  //--- constrain intra prediction ---
  if(input->UseConstrainedIntraPred && (img->type==P_SLICE || img->type==B_SLICE))
  {
    if( !IS_INTRA(currMB) )
    {
      img->intra_block[img->current_mb_nr] = 0;
    }
    else
    {
      img->intra_block[img->current_mb_nr] = 1;
    }
  }
  
  //===== init and update number of intra macroblocks =====
  if (img->current_mb_nr==0)
    intras=0;
  
  if (IS_INTRA(currMB))
    intras++;
  
  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  if (input->symbol_mode==CABAC && img->current_mb_nr!=img->currentSlice->start_mb_nr && eos_bit)
  {
    write_terminating_bit (0);
  }
  
  cabac_encoding = 1;

  //--- write macroblock ---
  writeMBLayer (0, &i);  // i is temporary
  
  if (!((currMB->mb_type !=0 ) || ((img->type==B_SLICE) && currMB->cbp != 0) ))
  { 
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));  // CAVLC
  }
  
  set_last_dquant();
  
  //--- set total bit-counter ---
  
#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
      + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
      + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB]
      + bitCount[BITS_SKIP_PRED]
      + bitCount[BITS_MV_PRED];
  }
  else
#endif
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
      + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
      + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB];
  }  
  ;
  
  stats->bit_slice += bitCount[BITS_TOTAL_MB];
  
  cabac_encoding = 0;
}


/*!
************************************************************************
* \brief
*    Sets context for reference frame parameter
************************************************************************
*/
int BType2CtxRef (int btype)
{
  return (btype<4 ? 0 : 1);
}


/*!
************************************************************************
* \brief
*    Codes the reference frame
************************************************************************
*/
int writeReferenceFrame (int mode, int i, int j, int fwd_flag, int  ref)
{
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  int*            bitCount  = currMB->bitcounter;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             rate      = 0;
  DataPartition*  dataPart;
  int             num_ref   = ( fwd_flag ? listXsize[LIST_0+currMB->list_offset]: listXsize[LIST_1+currMB->list_offset]);
  int             flag_mode = 0;
  
  if( num_ref == 1 )
  {
    return 0;
  }
  
  if ( num_ref == 2 )
  {
    flag_mode = 1;
  }
  
  currSE->value1 = ref;
  currSE->value2  = 0;
  currSE->type   = SE_REFFRAME;
  
  dataPart = &(currSlice->partArr[partMap[currSE->type]]);
  if (input->symbol_mode == UVLC)
  {
    if(img->mv_res)
    {
      if(ref == 0 && mode <= 4)
        currSE->value2 = ref_idx_ctx(mode, i, j, !fwd_flag, ref, 0);
      write_ref_idx(mode, currSE, dataPart);
    }
    else
    {
      if( flag_mode )
      {
        currSE->bitpattern = 1 - currSE->value1;
        currSE->len = 1;
        writeSyntaxElement2Buf_Fixed(currSE, dataPart->bitstream);
      }
      else
      {
        currSE->mapping = ue_linfo;
        dataPart->writeSyntaxElement (currSE, dataPart);
      }
    }
  }
  else
  {
    currSE->context = BType2CtxRef (mode);
    img->subblock_x = i; // position used for context determination
    img->subblock_y = j; // position used for context determination
    currSE->writing = writeRefFrame_CABAC;
    currSE->value2 = (fwd_flag)? LIST_0:LIST_1;
    dataPart->writeSyntaxElement (currSE, dataPart);
  }
  
  bitCount[BITS_INTER_MB] += currSE->len;
  rate                    += currSE->len;
#if TRACE
  if (fwd_flag)
  {
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "ref_idx_l0 = %d", currSE->value1);
  }
  else
  {
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "ref_idx_l1 = %d", currSE->value1);
  }
#endif
  currSE++;
  currMB->currSEnr++;
  
  return rate;
}


/*!
************************************************************************
* \brief
*    Writes motion vectors of an 8x8 block
************************************************************************
*/
int writeMotionVector8x8 (int  i0,
                          int  j0,
                          int  i1,
                          int  j1,
                          int  refframe,
                          int  list_idx,
                          int  mv_mode)
{
  int            i, j, k, l, m;
  int            curr_mvd;
  DataPartition* dataPart;
  
  int            rate       = 0;
  int            step_h     = input->part_size[mv_mode][0];
  int            step_v     = input->part_size[mv_mode][1];
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement* currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*         currSlice  = img->currentSlice;
  int*           bitCount   = currMB->bitcounter;
  const int*     partMap    = assignSE2partition[input->partition_mode];
  int            refindex   = refframe;
  
  short******    all_mv     = img->all_mv;
  short******    pred_mv    = img->pred_mv;
  
  if (currMB->bi_pred_me && currMB->b8pdir[0]==2 && mv_mode == 1 && refindex == 0)
    all_mv = currMB->bi_pred_me == 1? img->bipred_mv1 : img->bipred_mv2 ;
  
  for (j=j0; j<j1; j+=step_v)
  {
    for (i=i0; i<i1; i+=step_h)
    {

      int mv_shift =0;  //1/8 pel as default 
      int currMvVec;
      int predMvVec;


#ifdef MV_COMPETITION
      
      // <FTRD KTA 1.8 compatibility with RDO_Q
      //  if((input->Transform8x8Mode)&&(pass_with_writing == 1) && (input->mv_competition > 0))
      if((input->Transform8x8Mode)&&(pass_with_writing == 1) && (input->mv_competition > 0) &&(!input->UseRDO_Q))
        // FTRD KTA 1.8 compatibility with RDO_Q>   
      {
        short best_prediction[2];
        SetMotionVectorPredictor_Competition (best_prediction, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j,input->blc_size[mv_mode][0],input->blc_size[mv_mode][1]);
        GetBestPredVector(best_prediction,
          all_mv[j][i][list_idx][refindex][mv_mode][0],
          all_mv[j][i][list_idx][refindex][mv_mode][1],
          img->lambda_mf[img->type][img->qp],refindex,list_idx,mv_mode,i*4,j*4); 
        pred_mv[j][i][list_idx][refindex][mv_mode][0]=best_prediction[0];
        pred_mv[j][i][list_idx][refindex][mv_mode][1]=best_prediction[1];
      }
#endif

#ifdef H264_COMPLIANT
      {
          short pred_mv_stored[2];
          SetMotionVectorPredictor (pred_mv_stored, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j, input->blc_size[mv_mode][0],input->blc_size[mv_mode][1]);
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=pred_mv_stored[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=pred_mv_stored[1];
      }
#endif

#ifdef RDO_Q
      if(input->UseRDO_Q)
      {
        // <FTRD KTA 1.8 compatibility with RDO_Q
#ifdef MV_COMPETITION
        if( (input->mv_competition > 0))    
        {
          short best_prediction[2];
          SetMotionVectorPredictor_Competition (best_prediction, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j,input->blc_size[mv_mode][0],input->blc_size[mv_mode][1]);
          GetBestPredVector(best_prediction,
            all_mv[j][i][list_idx][refindex][mv_mode][0],
            all_mv[j][i][list_idx][refindex][mv_mode][1],
            img->lambda_mf[img->type][img->qp],refindex,list_idx,mv_mode,i*4,j*4); 
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=best_prediction[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=best_prediction[1];
        }
        else
        {
#endif  //MV_COMPETITION
          // FTRD KTA 1.8 compatibility with RDO_Q>
          short pred_mv_stored[2];
          SetMotionVectorPredictor (pred_mv_stored, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j, input->blc_size[mv_mode][0],input->blc_size[mv_mode][1]);
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=pred_mv_stored[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=pred_mv_stored[1];
          // <FTRD KTA 1.8 compatibility with RDO_Q
#ifdef MV_COMPETITION
        }  
#endif //MV_COMPETITION
        // FTRD KTA 1.8 compatibility with RDO_Q>
      }
#endif   //RDO_Q
      

	  if (img->mv_res)
	  {
		  if (((all_mv[j][i][list_idx][refindex][mv_mode][0]& 1) ==0)&&((all_mv[j][i][list_idx][refindex][mv_mode][1]& 1))==0)
			  mv_shift =1; //switch to 1/4 pel
      if(input->symbol_mode == CABAC)
      {

		  img->subblock_x = i; // position used for context determination
		  img->subblock_y = j; // position used for context determination
		  currSE->value1 = mv_shift;
  		  currSE->type   = SE_MVD; 
		  currSE->value2 = 0;
		  currSE->writing  = writeMVRES_4_8_CABAC;
		  dataPart = &(currSlice->partArr[partMap[SE_MVD]]);
		  dataPart->writeSyntaxElement (currSE, dataPart);
#if TRACE
		  snprintf(currSE->tracestring, TRACESTRING_SIZE, "mv_shift  = %3d  ", mv_shift);
#endif
		  bitCount[BITS_INTER_MB] += currSE->len;
		  rate                    += currSE->len;
		  currSE++;  
		  currMB->currSEnr++;
      }


	  }
	  else
	  {
		mv_shift=0;
	  }


      for (k=0; k<2; k++) 
      {        

		if (img->mv_res && (mv_shift > 0))
		{
			int scaled_mvd;
			currMvVec=roundInt(all_mv[j][i][list_idx][refindex][mv_mode][k],mv_shift);
			predMvVec=roundInt(pred_mv[j][i][list_idx][refindex][mv_mode][k],mv_shift);
			curr_mvd = currMvVec - predMvVec;
			scaled_mvd = curr_mvd*(1<< mv_shift);
			//--- store (oversampled) mvd ---
			for (l=0; l < step_v; l++)
			  for (m=0; m < step_h; m++)
			  {
				currMB->mvd[list_idx][j+l][i+m][k] = scaled_mvd;
			  }
		}
		else

		{
        curr_mvd = all_mv[j][i][list_idx][refindex][mv_mode][k] - pred_mv[j][i][list_idx][refindex][mv_mode][k];
        
        //--- store (oversampled) mvd ---
        for (l=0; l < step_v; l++)
          for (m=0; m < step_h; m++)
          {
            currMB->mvd[list_idx][j+l][i+m][k] = curr_mvd;
          }
		}
          currSE->value1 = curr_mvd;
          currSE->value2 = 0;
          currSE->type   = SE_MVD;
          if (input->symbol_mode == UVLC)
          {
            currSE->mapping = se_linfo;
          }
          else
          {
            img->subblock_x = i; // position used for context determination
            img->subblock_y = j; // position used for context determination

		    currSE->value2  = (2*k+list_idx)*2+mv_shift;

            currSE->writing = writeMVD_CABAC;
          }  
          dataPart = &(currSlice->partArr[partMap[SE_MVD]]);
          dataPart->writeSyntaxElement (currSE, dataPart);
          bitCount[BITS_INTER_MB] += currSE->len;
          rate                    += currSE->len;
          currSE++;  
          currMB->currSEnr++;
      }
#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
      { 
        int maxmode=0;
        if (img->type == P_SLICE)
          maxmode = mv_comp.nb_mode_for_mvp;
        else if (img->type == B_SLICE)
          maxmode = mv_comp.nb_mode_for_mvb;
        if (send_index_mv_prediction[mv_mode][refindex][list_idx][j][i][maxmode] == 0)
          rate+=writeIndexForMotionVectorPrediction (refframe, list_idx, mv_mode, i, j);
      }
#endif
    }
  }
  return rate;
}

int ref_idx_ctx(int mv_mode, int i, int j, int list_idx, int ref, int mb_ext_level)
{
  int k;
  int a=0, b=0;
  PixelPos block_a, block_b;
  int mv_eighth;
  int mb_nr = img->current_mb_nr;
  int subblock_x=0, subblock_y=0; 

  if(mb_ext_level ==0)
  {    
    subblock_x = i;
    subblock_y = j;
  }
  else
  {  
    if(i==0 && j==0 )
    {
      subblock_x = 0;
      subblock_y = 0;
      mb_nr      = img->current_mb_nr;
    }
    else if(i==0 && j==(2<<mb_ext_level) )
    {
      subblock_x = 0;
      subblock_y = 0;
      mb_nr      = img->current_mb_nr + (img->PicWidthInMbs << (mb_ext_level-1));
    }
    else if(i==(2<<mb_ext_level) && j==0 )
    {
      subblock_x = 0;
      subblock_y = 0;
      mb_nr      = img->current_mb_nr + (1<<(mb_ext_level- 1));
    }
    else
    {
      error("subblock_x and subblock_y are out of the range\n", -1);
    }

  }

  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y, -1,  0, &block_a);
  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y,  0, -1, &block_b);

  if (block_b.available)
  {
    for(k=0; k<2; k++)
      b |= absm(img->mb_data[block_b.mb_addr].mvd[list_idx][block_b.y][block_b.x][k]) & 1;
  }
  else
    b=0;

  if (block_a.available)
  {
    for(k=0; k<2; k++)
      a |= absm(img->mb_data[block_a.mb_addr].mvd[list_idx][block_a.y][block_a.x][k]) & 1;
  }
  else
    a = 0;

  mv_eighth = ((img->all_mv[j][i][list_idx][ref][mv_mode][0]& 1) ==1) || ((img->all_mv[j][i][list_idx][ref][mv_mode][1]& 1)==1);

  if(a==0 && b==0)
    return mv_eighth;
  else
    return 2+mv_eighth;
}
int
group_ref (Macroblock* currMB, int mb_ext_level)
{
  int i0, j0, jj, k;
  int ref, ref_b0;//upper left 8x8 block
  int mv_eighth, mv_eighth_b0;//
  int   step_h0         = (input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][0] >> 2)<<mb_ext_level;
  int   step_v0         = (input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][1] >> 2)<<mb_ext_level;

  ref_b0 = enc_picture->ref_idx[LIST_0][img->block_y][img->block_x];

  if(ref_b0 == 0)
  {
    mv_eighth_b0 = ref_idx_ctx(currMB->b8mode[0], 0, 0, 0, ref_b0, 0) & 1;
    if(mv_eighth_b0)
      return 0;
  }
  else
  {
    return 0;
    //mv_eighth_b0 = 0;
  }

  for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
  {
    jj = img->block_y+j0;
    for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
    {
      k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));

      ref = enc_picture->ref_idx[LIST_0][jj][img->block_x+i0];
      
      if(ref != ref_b0)
        return 0;

      if(ref == 0)
      {
        mv_eighth = ref_idx_ctx(currMB->b8mode[k], i0, j0, 0, ref, 0) & 1;
        if(mv_eighth != mv_eighth_b0)
          return 0;
      }
    }
  }
  return 1;
}

int
ZeroRef_mvres (Macroblock* currMB)
{
  int i0, j0, jj, k, ref;
  int   step_h0         = (input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][0] >> 2);
  int   step_v0         = (input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][1] >> 2);
  int   zeroref = 1;  
  
  for (j0=0; j0<4; j0+=step_v0)
  {
    jj = img->block_y+j0;
    for (i0=0; i0<4; i0+=step_h0)
    {
      k=j0+(i0 >> 1);

      ref = enc_picture->ref_idx[LIST_0][jj][img->block_x+i0];
      if(ref == 0)
      {
        if(ref_idx_ctx(currMB->b8mode[k], i0, j0, 0, ref, 0) & 1)
          zeroref = 0;
      }
      else
        zeroref = 0;
    }
  }
  return zeroref;
}
/*!
************************************************************************
* \brief
*    Writes motion info
************************************************************************
*/
int writeMotionInfo2NAL (void)
{
  int k, j0, i0, refframe;
  int jj;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int             no_bits   = 0;
  int   bframe          = (img->type==B_SLICE);
  int   step_h0         = (input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][0] >> 2);
  int   step_v0         = (input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][1] >> 2);
  int   zeroref = 1;
  //=== If multiple ref. frames, write reference frame for the MB ===
  if (IS_INTERMV (currMB))
  {
    int group_flag=0;
    if(input->symbol_mode == UVLC)
    {
#ifdef GROUP_REF
		if(currMB->mb_type>1 && listXsize[LIST_0]>1)
		{
		  const int*      partMap   = assignSE2partition[input->partition_mode];
		  group_flag = group_ref (currMB, 0);
		  no_bits += u_1 ("group_flag", group_flag, (img->currentSlice->partArr[partMap[SE_REFFRAME]]).bitstream);
		}
#else
		if(IS_P8x8 (currMB))
		{
		  if(img->mv_res)
		  {
			zeroref = ZeroRef_mvres (currMB);
			//printf("mb_nr %d zeroref %d\n", img->current_mb_nr, zeroref);
		  }
		  else
		  {
			zeroref = ZeroRef (currMB);
		  }
		}
#endif
    }


    // if UVLC is turned on, a 8x8 macroblock with all ref=0 in a P-frame is signalled in macroblock mode
#ifdef GROUP_REF
    if((input->symbol_mode == UVLC && group_flag == 0) || input->symbol_mode==CABAC)
#else
    if (!IS_P8x8 (currMB) || !zeroref || input->symbol_mode==CABAC || bframe)
#endif
    {
      for (j0=0; j0<4; j0+=step_v0)
      {
        jj = img->block_y+j0;
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=j0+(i0 >> 1);
          
          if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has forward vector
          {
            no_bits += writeReferenceFrame (currMB->b8mode[k], i0, j0, 1, enc_picture->ref_idx[LIST_0][jj][img->block_x+i0]);
          }
		}
      }
      for (j0=0; j0<4; j0+=step_v0)
      {
        jj = img->block_y+j0;
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=j0+(i0 >> 1);
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has backward vector
          {
            no_bits += writeReferenceFrame (currMB->b8mode[k], i0, j0, 0, enc_picture->ref_idx[LIST_1][jj][img->block_x+i0]);
          }
        }
      }
    }
  }
  
  //===== write forward motion vectors =====
  if (IS_INTERMV (currMB))
  {
    for (j0=0; j0<4; j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=j0+(i0 >> 1);
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has forward vector
        {
          refframe  = enc_picture->ref_idx[LIST_0][jj][img->block_x+i0];
          no_bits  += writeMotionVector8x8 (i0, j0, i0+step_h0, j0+step_v0, refframe, LIST_0, currMB->b8mode[k]);
        }
      }
    }
  }
  
  
  //===== write backward motion vectors =====
  if (IS_INTERMV (currMB) && bframe)
  {
    for (j0=0; j0<4; j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=j0+(i0 >> 1);
        if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has backward vector
        {
          refframe  = enc_picture->ref_idx[LIST_1][jj][img->block_x+i0];
          no_bits  += writeMotionVector8x8 (i0, j0, i0+step_h0, j0+step_v0, refframe, LIST_1, currMB->b8mode[k]);
        }
      }
    }
  }
  return no_bits;
}



/*!
************************************************************************
* \brief
*    Writes chrominance coefficients
************************************************************************
*/
int writeChromaCoeff (void)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount  = currMB->bitcounter;
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k, uv;
  int   b8, b4, param;
  int*  ACLevel;
  int*  ACRun;
  int*  DCLevel;
  int*  DCRun;
  //ADD-VG-14052004
  int   chroma_dc_context[3]={CHROMA_DC, CHROMA_DC_2x4, CHROMA_DC_4x4};
  int   yuv = img->yuv_format - 1;
  
  static unsigned char chroma_ac_param[3][8][4] =
  {
    {{ 4, 20,  5, 21},
    {36, 52, 37, 53},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0}},
    
    {{ 4, 20,  5, 21},
    { 6, 22,  7, 23},
    {36, 52, 37, 53},
    {38, 54, 39, 55},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0},
    { 0,  0,  0,  0}},
    
    {{ 4, 20,  5, 21},
    {36, 52, 37, 53},
    { 6, 22,  7, 23},
    {38, 54, 39, 55},
    { 8, 24,  9, 25},
    {40, 56, 41, 57},
    {10, 26, 11, 27},
    {42, 58, 43, 59}}};
    //ADD-VG-14052004-END
    
    //=====
    //=====   D C - C O E F F I C I E N T S
    //=====
    if (cbp > 15)  // check if any chroma bits in coded block pattern is set
    {
      for (uv=0; uv < 2; uv++)
      {
        if (input->symbol_mode == UVLC)
        {
          param = uv;
          rate += writeCoeff4x4_CAVLC (CHROMA_DC, 0, 0, param);
        }
        else
        {
          
          DCLevel = img->cofDC[uv+1][0];
          DCRun   = img->cofDC[uv+1][1];
          
          level=1;
          for (k=0; k <= img->num_cdc_coeff && level != 0; ++k)
          {
            level = currSE->value1 = DCLevel[k]; // level
            run   = currSE->value2 = DCRun  [k]; // run
            
            currSE->writing = writeRunLevel_CABAC;
            
            currSE->context     = chroma_dc_context[yuv];
            currSE->type        = (IS_INTRA(currMB) ? SE_CHR_DC_INTRA : SE_CHR_DC_INTER);
            img->is_intra_block =  IS_INTRA(currMB);
            img->is_v_block     = uv;
            
            // choose the appropriate data partition
            dataPart = &(currSlice->partArr[partMap[currSE->type]]);
            dataPart->writeSyntaxElement (currSE, dataPart);
            bitCount[BITS_COEFF_UV_MB] += currSE->len;
            rate                       += currSE->len;
#if TRACE
            snprintf(currSE->tracestring, TRACESTRING_SIZE, "DC Chroma %2d: level =%3d run =%2d",k, level, run);
#endif
            // proceed to next SE 
            currSE++;  
            currMB->currSEnr++;
          }
        }
      }
    }
    
    //=====
    //=====   A C - C O E F F I C I E N T S
    //=====
    uv=-1;   
    if (cbp >> 4 == 2) // check if chroma bits in coded block pattern = 10b
    {  
      for (b8=4; b8 < (4+img->num_blk8x8_uv); b8++)
        for (b4=0; b4 < 4; b4++)
        {
          if (input->symbol_mode == UVLC)
          {
            param = chroma_ac_param[yuv][b8-4][b4];
            rate += writeCoeff4x4_CAVLC (CHROMA_AC, b8, b4, param);
          }
          else
          {
            
            ACLevel = img->cofAC[b8][b4][0];
            ACRun   = img->cofAC[b8][b4][1];
            
            level=1;
            uv++;
            
            img->subblock_y = subblk_offset_y[yuv][b8-4][b4]>>2;
            img->subblock_x = subblk_offset_x[yuv][b8-4][b4]>>2;
            
            for (k=0; k < 16 && level != 0; k++)
            {
              level = currSE->value1 = ACLevel[k]; // level
              run   = currSE->value2 = ACRun  [k]; // run
              
              if (input->symbol_mode == UVLC)   currSE->mapping = levrun_linfo_inter;
              else                              currSE->writing = writeRunLevel_CABAC;
              
              currSE->context     = CHROMA_AC;
              currSE->type        = (IS_INTRA(currMB) ? SE_CHR_AC_INTRA : SE_CHR_AC_INTER);
              img->is_intra_block =  IS_INTRA(currMB);
              img->is_v_block     = (uv>=(img->num_blk8x8_uv<<1));
              
              // choose the appropriate data partition
              dataPart = &(currSlice->partArr[partMap[currSE->type]]);
              dataPart->writeSyntaxElement (currSE, dataPart);
              bitCount[BITS_COEFF_UV_MB] += currSE->len;
              rate                       += currSE->len;
#if TRACE
              snprintf(currSE->tracestring, TRACESTRING_SIZE, "AC Chroma %2d: level =%3d run =%2d",k, level, run);
#endif
              
              // proceed to next SE 
              currSE++;  
              currMB->currSEnr++;
            }
          }
        }
    }
    
    return rate;
}



/*!
************************************************************************
* \brief
*    Writes Luma coeff of an 4x4 block
************************************************************************
*/
int writeLumaCoeff4x4_CABAC (int b8, int b4, int intra4x4mode)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC[b8][b4][0];
  int*  ACRun   = img->cofAC[b8][b4][1];

  img->subblock_x = ((b8&0x1)==0) ? (((b4&0x1)==0)? 0: 1) : (((b4&0x1)==0)? 2: 3); // horiz. position for coeff_count context
  img->subblock_y = (b8<2)        ? ((b4<2)       ? 0: 1) : ((b4<2)       ? 2: 3); // vert.  position for coeff_count context
  
  level=1; // get inside loop
  for(k=0; k<=16 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run
    
    currSE->writing = writeRunLevel_CABAC;
    
    currSE->context     = LUMA_4x4;
    currSE->type        = (k==0 
      ? (intra4x4mode ? SE_LUM_DC_INTRA : SE_LUM_DC_INTER) 
      : (intra4x4mode ? SE_LUM_AC_INTRA : SE_LUM_AC_INTER));
    img->is_intra_block = intra4x4mode;
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);          
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;
  }
  return rate;
}

/*!
************************************************************************
* \brief
*    Writes Luma coeff of an 8x8 block
************************************************************************
*/
int writeLumaCoeff8x8_CABAC (int b8, int intra_mode)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC[b8][0][0];
  int*  ACRun   = img->cofAC[b8][0][1];
  
  img->subblock_x = ((b8&0x1)==0)?0:2;  // horiz. position for coeff_count context
  img->subblock_y = (b8<2)?0:2;     // vert.  position for coeff_count context
  
  
  level=1; // get inside loop
  for(k=0; k<=64 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run
    
    currSE->writing = writeRunLevel_CABAC;
    
    currSE->context     = LUMA_8x8;
    currSE->type        = (k==0 
      ? (intra_mode ? SE_LUM_DC_INTRA : SE_LUM_DC_INTER) 
      : (intra_mode ? SE_LUM_AC_INTRA : SE_LUM_AC_INTER));
    img->is_intra_block = intra_mode;
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;
    
  }
  
  return rate;
}

/*!
************************************************************************
* \brief
*    Writes Luma Coeff of an 8x8 block
************************************************************************
*/
int writeLumaCoeff8x8 (int block8x8, int block_mode, int transform_size_flag)
{
  int  block4x4, rate = 0;
  int intra4x4mode = (block_mode==IBLOCK);
  
  if (block_mode == I8MB)
    assert(transform_size_flag == 1);
  
  
  if((!transform_size_flag) || input->symbol_mode == UVLC) // allow here if 4x4 or UVLC
  {
    if (input->symbol_mode == UVLC )
    {
      for (block4x4=0; block4x4<4; block4x4++) 
        rate += writeCoeff4x4_CAVLC (LUMA, block8x8, block4x4, (transform_size_flag)?(block_mode==I8MB):intra4x4mode);// CAVLC, pass new intra
    }
    else
    {
      for (block4x4=0; block4x4<4; block4x4++)  
        rate += writeLumaCoeff4x4_CABAC (block8x8, block4x4, intra4x4mode);
    }      
  }
  else 
    rate += writeLumaCoeff8x8_CABAC (block8x8, (block_mode == I8MB));
  
  return rate;
}

#ifdef USE_INTRA_MDDT
/*! 
*************************************************************************************
* \brief
*   write the luma coefficients of an INTRA16x16 macroblock
*
* \para writeLumaCoeff16x16_CABAC()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
int writeLumaCoeff16x16_CABAC ()
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC16[0];
  int*  ACRun   = img->cofAC16[1];
  
  img->subblock_x = 0;//img->pix_x;  // horiz. position for coeff_count context
  img->subblock_y = 0;//img->pix_y;  // vert.  position for coeff_count context
  
  level=1; // get inside loop
  for(k=0; k<=256 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run
    
    currSE->writing = writeRunLevel16x16_CABAC;
    
    currSE->context     = LUMA_16x16;
    currSE->type        = SE_LUM_AC_INTRA;
    img->is_intra_block = 1;    // only applicable to intra blocks
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;
    
  }
  
  return rate;
}

int writeLumaCoeff16x16 ( )
{
  int  rate = 0;
  
  if( input->symbol_mode == UVLC) // allow here if 4x4 or UVLC
  {
    assert(0);
  }
  else 
    rate += writeLumaCoeff16x16_CABAC ();
  
  return rate;
}
#endif 


  
#if (defined MB32X32)
//Rahul
int writeCBP_Big(int rdopt, int cbp)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  DataPartition*  dataPart;


    //=====   C B P   =====
    //---------------------
  if(input->symbol_mode == CABAC)
  {
    currSE->value1 = cbp;
    currSE->writing = writeCBP_CABAC32;
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
    dataPart->writeSyntaxElement(currSE, dataPart);
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "CBP (%2d,%2d) = %3d",img->mb_x, img->mb_y, cbp);
#endif
  }
  else//tm_vlc
  {
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
    currSE->len = u_1 ("CBP", cbp, dataPart->bitstream);    
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "CBP_Big (%2d,%2d) = %3d",img->mb_x, img->mb_y, cbp);
#endif  
  }


    bitCount[BITS_CBP_MB] += currSE->len;
    rate                  += currSE->len;
    // proceed to next SE
    currSE++;  
    currMB->currSEnr++;     

  return rate;
}
#endif





/*!
************************************************************************
* \brief
*    Writes CBP, DQUANT, and Luma Coefficients of an macroblock
*    mb32: 1-32x32 block
************************************************************************
*/
#ifdef USE_INTRA_MDDT
#ifdef MB32X32
int writeCBPandLumaCoeff (int rdopt, int mb32, int skipDquant)
#else
int writeCBPandLumaCoeff (int rdopt)
#endif
#else
int writeCBPandLumaCoeff (void)
#endif
{
  int             mb_x, mb_y, i, j, k;
  int             level, run;
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  int             need_transform_size_flag;   //ADD-VG-24062004
  
  int   b8, b4;
  int*  DCLevel = img->cofDC[0][0];
  int*  DCRun   = img->cofDC[0][1];
  int*  ACLevel;
  int*  ACRun;
  
  if (!IS_NEWINTRA (currMB))
  {
    //need_transform_size_flag is different from the original defination
#if P_DIRECT
    need_transform_size_flag = ((mb32 || (currMB->mb_type >= 0 && currMB->mb_type <= 3)||
        (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
        (currMB->NoMbPartLessThan8x8Flag))
        && currMB->mb_type != I8MB && currMB->mb_type != I4MB
        //&& (currMB->cbp&15)  - !!!
        && input->Transform8x8Mode);
#else
    need_transform_size_flag = ((mb32 || (currMB->mb_type >= 1 && currMB->mb_type <= 3)||
        (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
        (currMB->NoMbPartLessThan8x8Flag))
        && currMB->mb_type != I8MB && currMB->mb_type != I4MB
        //&& (currMB->cbp&15)  - !!!
        && input->Transform8x8Mode);
#endif

    if(need_transform_size_flag)
    {
      rate += writelumaCBPTransform_size(mb32);
    }


    if( need_transform_size_flag &&
      ((currMB->cbp&15) == 0 ||
       (currMB->luma_transform_size_8x8_flag==2 
       && (mb32 || currMB->mb_type == 1 || currMB->geo_mode!=-1 || (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag)))))
    {
      rate += write_chroma_cbp();
    }
    else

	{
    //=====   C B P   =====
    //---------------------
    currSE->value1 = cbp;
#ifdef MB32X32
    currSE->value2 = mb32;
#endif    
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
    {
      if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_intra;
      currSE->type = SE_CBP_INTRA;
    }
    else
    {
      if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_inter;
      currSE->type = SE_CBP_INTER;
    }
    if (input->symbol_mode == CABAC)   currSE->writing = writeCBP_CABAC;
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
    
    dataPart->writeSyntaxElement(currSE, dataPart);
    bitCount[BITS_CBP_MB] += currSE->len;
    rate                  += currSE->len;
    // proceed to next SE
    currSE++;  
    currMB->currSEnr++;
	}
  }

  //=====   DQUANT   =====
  //----------------------
#ifdef MB32X32
  delta_qp_sent = 0;
  if(img->type == I_SLICE || (!mb32 && !delta_qp_sent) )
#endif
  {
#if defined MB32X32
    if((cbp!=0 || IS_NEWINTRA (currMB)) && !mb32)
#else
    if (cbp!=0 || IS_NEWINTRA (currMB))
#endif
    {
    currSE->value1 = currMB->delta_qp;
    
    if (input->symbol_mode==UVLC)   currSE->mapping = se_linfo;
    else                            currSE->writing = writeDquant_CABAC;
    
    if (IS_INTER (currMB))  currSE->type = SE_DELTA_QUANT_INTER;
    else                    currSE->type = SE_DELTA_QUANT_INTRA;
    
    
    // choose the appropriate data partition
    dataPart = &(img->currentSlice->partArr[partMap[currSE->type]]);
    dataPart->writeSyntaxElement(  currSE, dataPart);
    bitCount[BITS_DELTA_QUANT_MB] += currSE->len;
    rate                          += currSE->len;
    // proceed to next SE
    currSE++;
    currMB->currSEnr++;
  }
  }  
  for (i=0; i < 4; i++)
    memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));
  
  if (!IS_NEWINTRA (currMB))
  {
    if(currMB->mb_type >= 0 && currMB->mb_type <= 3 && currMB->luma_transform_size_8x8_flag == 2)
    {
#ifdef MB32X32
      rate += writeLumaCoefBigBlock(currMB, mb32, rdopt);
#else
      rate += writeLumaCoefBigBlock(currMB, rdopt);
#endif
      return rate;
    }

    //=====  L U M I N A N C E   =====
    //--------------------------------
    for (i=0; i<4; i++)
      if (cbp & (1<<i))
      {
#ifdef USE_INTRA_MDDT
        if(!rdopt && currMB->mb_type == I8MB && input->UseIntraMDDT)
        {
          int pic_block_y, pic_block_x;
          int ipredmode; 
          byte (*scan)[2];
          
          pic_block_y = img->mb_y*4+2*(i>>1);
          pic_block_x = img->mb_x*4+2*(i &1); 
          ipredmode = img->ipredmode8x8[pic_block_y][pic_block_x];
          img->update8x8Count[ipredmode]++;
        
          if(input->symbol_mode == UVLC)
          {
            int blk4x4; 
            scan = img->scanOrder8x8[ipredmode];
            for(blk4x4 = 0; blk4x4 < 4; blk4x4++)
            {
              int*  ACLevel = img->cofAC[i][blk4x4][0];
              int*  ACRun   = img->cofAC[i][blk4x4][1];
              int  pos       = 0;
              int  level     = 1;
              int  k; 
              
              for (k=0; k < 16 && level != 0; k++)
              {
                level = ACLevel[k]; // level
                run   = ACRun  [k]; // run
                if(level == 0)
                  break;
                pos   += run;
                img->scanStats8x8[ipredmode][4*pos+blk4x4]++;
                pos++;
              }
            }
          }
          else
          {
            int*  ACLevel = img->cofAC[i][0][0];
            int*  ACRun   = img->cofAC[i][0][1];
            int  pos       = 0;
            int  level     = 1;
            int  k; 
            
            
            scan = img->scanOrder8x8[ipredmode];
            for (k=0; k < 64 && level != 0; k++)
            {
              level = ACLevel[k]; // level
              run   = ACRun  [k]; // run
              if(level == 0)
                break;
              pos   += run;
              img->scanStats8x8[ipredmode][pos]++;
              pos++;
            }
          }
        }
        else if(!rdopt && currMB->mb_type == I4MB && input->UseIntraMDDT)
        {
          int   b4; 
          
          for(b4 = 0; b4 < 4; b4++)
          {
            int*  ACLevel = img->cofAC[i][b4][0];
            int*  ACRun   = img->cofAC[i][b4][1];
            int  pos       = 0;
            int  level     = 1;
            int  k; 
            int pic_block_y, pic_block_x;
            int ipredmode; 
            byte (*scan)[2];
            
            pic_block_y = img->block_y+2*(i>>1)+(b4>>1);
            pic_block_x = img->block_x+2*(i &1)+(b4 &1);           
            ipredmode = img->ipredmode[pic_block_y][pic_block_x];
            
            scan = img->scanOrder4x4[ipredmode];
            img->update4x4Count[ipredmode]++;
            
            if(ACLevel[0])
            {
              for (k=0; k < 16 && level != 0; k++)
              {
                level = ACLevel[k]; // level
                run   = ACRun  [k]; // run
                if(level == 0)
                  break;
                pos   += run;
                img->scanStats4x4[ipredmode][pos]++;
                pos++;
              }
            }
          }
        }
#endif
        rate += writeLumaCoeff8x8 (i, currMB->b8mode[i], currMB->luma_transform_size_8x8_flag);
      }
#ifdef USE_INTRA_MDDT
      if(!rdopt && input->UseIntraMDDT)
      {
        int i; 
        for(i = 0; i < 9; i++)
        {
          if(img->update4x4Count[i] >= img->update4x4Thres[i])
          {
            img->update4x4[i] = 1;
            img->update4x4Count[i] = 0;
          }
          else img->update4x4[i] = 0;
          if(img->update8x8Count[i] >= img->update8x8Thres[i])
          {
            img->update8x8[i] = 1;
            img->update8x8Count[i] = 0;
          }
          else img->update8x8[i] = 0;
        }
        
        updateScanOrder(0);
        normalizeScanStats();
      }
#endif 
  }
  else
  {
    //=====  L U M I N A N C E   f o r   1 6 x 1 6   =====
#ifdef USE_INTRA_MDDT
    if(input->symbol_mode == CABAC && input->UseIntraMDDT)
    {
      if(currMB->cbp & 0xf)
        rate += writeLumaCoeff16x16 ();
    }
    else
    {
#endif
      //----------------------------------------------------
      // DC coeffs
    
      if (input->symbol_mode == UVLC)
      {
        rate += writeCoeff4x4_CAVLC (LUMA_INTRA16x16DC, 0, 0, 0);  // CAVLC
      }
      else
      {
        level=1; // get inside loop
        for (k=0; k<=16 && level!=0; k++)
        {
          level = currSE->value1 = DCLevel[k]; // level
          run   = currSE->value2 = DCRun  [k]; // run
          
          currSE->writing = writeRunLevel_CABAC;
          
          currSE->context     = LUMA_16DC;
          currSE->type        = SE_LUM_DC_INTRA;   // element is of type DC
          img->is_intra_block = 1;
          
          // choose the appropriate data partition
          dataPart = &(currSlice->partArr[partMap[currSE->type]]);
          dataPart->writeSyntaxElement (currSE, dataPart);
          bitCount[BITS_COEFF_Y_MB] += currSE->len;
          rate                      += currSE->len;
#if TRACE
          snprintf(currSE->tracestring, TRACESTRING_SIZE, "DC luma 16x16 sng(%2d) level =%3d run =%2d", k, level, run);
#endif
          // proceed to next SE
          currSE++;
          currMB->currSEnr++;
        }
      }
      
      // AC coeffs
      if (cbp & 15)
      {
        for (mb_y=0; mb_y < 4; mb_y += 2)
          for (mb_x=0; mb_x < 4; mb_x += 2)
            for (j=mb_y; j < mb_y+2; j++)
              for (i=mb_x; i < mb_x+2; i++)
              {
                b8      = 2*(j >> 1) + (i >> 1);
                b4      = 2*(j & 0x01) + (i & 0x01);
                if (input->symbol_mode == UVLC)
                {
                  rate += writeCoeff4x4_CAVLC (LUMA_INTRA16x16AC, b8, b4, 0);  // CAVLC
                }
                else
                {
                  ACLevel = img->cofAC[b8][b4][0];
                  ACRun   = img->cofAC[b8][b4][1];
                  
                  img->subblock_y = j;
                  img->subblock_x = i;
                  
                  level=1; // get inside loop
                  for (k=0;k<16 && level !=0;k++)
                  {
                    level = currSE->value1 = ACLevel[k]; // level
                    run   = currSE->value2 = ACRun  [k]; // run
                    
                    currSE->writing = writeRunLevel_CABAC;
                    currSE->context     = LUMA_16AC;
                    currSE->type        = SE_LUM_AC_INTRA;   // element is of type AC
                    img->is_intra_block = 1;
                    
                    // choose the appropriate data partition
                    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
                    dataPart->writeSyntaxElement (currSE, dataPart);
                    bitCount[BITS_COEFF_Y_MB] += currSE->len;
                    rate                      += currSE->len;
#if TRACE
                    snprintf(currSE->tracestring, TRACESTRING_SIZE, "AC luma 16x16 sng(%2d) level =%3d run =%2d", k, level, run);
#endif
                    // proceed to next SE
                    currSE++;
                    currMB->currSEnr++;
                  }
                }
              }
      }
#ifdef USE_INTRA_MDDT
    }
#endif
  }
  
  return rate;
}



//Rahul
int writeDeltaQP(int rdopt, int cbp)  
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  //Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  DataPartition*  dataPart;

  if(cbp != 0 || IS_NEWINTRA (currMB))
  {
      currSE->value1 = currMB->delta_qp;
      if (input->symbol_mode==UVLC)   currSE->mapping = se_linfo;//tm_vlc
      else                            currSE->writing = writeDquant_CABAC;
      
      if (IS_INTER (currMB))  currSE->type = SE_DELTA_QUANT_INTER;
      else                    currSE->type = SE_DELTA_QUANT_INTRA;
      
      // choose the appropriate data partition
      dataPart = &(img->currentSlice->partArr[partMap[currSE->type]]);
      dataPart->writeSyntaxElement(  currSE, dataPart);
      bitCount[BITS_DELTA_QUANT_MB] += currSE->len;
      rate                          += currSE->len;
  #if TRACE
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Delta QP (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->delta_qp);
  #endif
      // proceed to next SE
      currSE++;
      currMB->currSEnr++;
  }
  return rate;
}


int writeCBP_Transform_size (int rdopt, int mb32)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  int             need_transform_size_flag;   //ADD-VG-24062004
  
  if (!IS_NEWINTRA (currMB))
  {
    //need_transform_size_flag is different from the original defination
#if P_DIRECT
    need_transform_size_flag = ((mb32 || (currMB->mb_type >= 0 && currMB->mb_type <= 3)||
      (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
      (currMB->NoMbPartLessThan8x8Flag))
      && currMB->mb_type != I8MB && currMB->mb_type != I4MB
      //&& (currMB->cbp&15)  - !!!
      && input->Transform8x8Mode);
#else
    need_transform_size_flag = ((mb32 || (currMB->mb_type >= 1 && currMB->mb_type <= 3)||
      (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
      (currMB->NoMbPartLessThan8x8Flag))
      && currMB->mb_type != I8MB && currMB->mb_type != I4MB
      //&& (currMB->cbp&15)  - !!!
      && input->Transform8x8Mode);
#endif

    if(need_transform_size_flag)
    {
      rate += writelumaCBPTransform_size(mb32);
    }

    if(   need_transform_size_flag 
      &&  ((currMB->cbp&15) == 0 || (currMB->luma_transform_size_8x8_flag==2 && (mb32 || currMB->mb_type == 1 || currMB->geo_mode!=-1 || (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag)))))
    {
      rate += write_chroma_cbp();
    }
    else
    {
      //=====   C B P   =====
      //---------------------
      currSE->value1 = cbp;
#ifdef MB32X32
      currSE->value2 = mb32;
#endif    
      if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
      {
        if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_intra;
        currSE->type = SE_CBP_INTRA;
      }
      else
      {
        if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_inter;
        currSE->type = SE_CBP_INTER;
      }
      if (input->symbol_mode == CABAC)   currSE->writing = writeCBP_CABAC;

      // choose the appropriate data partition
      dataPart = &(currSlice->partArr[partMap[currSE->type]]);

      dataPart->writeSyntaxElement(currSE, dataPart);
      bitCount[BITS_CBP_MB] += currSE->len;
      rate                  += currSE->len;
#if TRACE
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "CBP (%2d,%2d) = %3d",img->mb_x, img->mb_y, cbp);
#endif
      // proceed to next SE
      currSE++;  
      currMB->currSEnr++;
    }
  }

  return rate;
}

int writeLumaCoeff (int rdopt, int mb32)
{
  int             mb_x, mb_y, i, j, k;
  int             level, run;
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  
  int   b8, b4;
  int*  DCLevel = img->cofDC[0][0];
  int*  DCRun   = img->cofDC[0][1];
  int*  ACLevel;
  int*  ACRun;
 
  if (input->symbol_mode == UVLC)//tm_vlc
  {
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));
  }

  if (!IS_NEWINTRA (currMB))
  {
    if(currMB->mb_type >= 0 && currMB->mb_type <= 3 && currMB->luma_transform_size_8x8_flag == 2)
    {
      rate += writeLumaCoefBigBlock(currMB, mb32, rdopt);
      return rate;
    }

    //=====  L U M I N A N C E   =====
    //--------------------------------
    for (i=0; i<4; i++)
      if (cbp & (1<<i))
      {
#ifdef USE_INTRA_MDDT
        if(!rdopt && currMB->mb_type == I8MB && input->UseIntraMDDT)
        {
          int pic_block_y, pic_block_x;
          int ipredmode; 
          byte (*scan)[2];
          
          pic_block_y = img->mb_y*4+2*(i>>1);
          pic_block_x = img->mb_x*4+2*(i &1); 
          ipredmode = img->ipredmode8x8[pic_block_y][pic_block_x];
          img->update8x8Count[ipredmode]++;
          
          if(input->symbol_mode == UVLC)
          {
            int blk4x4; 
            scan = img->scanOrder8x8[ipredmode];
            for(blk4x4 = 0; blk4x4 < 4; blk4x4++)
            {
              int*  ACLevel = img->cofAC[i][blk4x4][0];
              int*  ACRun   = img->cofAC[i][blk4x4][1];
              int  pos       = 0;
              int  level     = 1;
              int  k; 
              
              for (k=0; k < 16 && level != 0; k++)
              {
                level = ACLevel[k]; // level
                run   = ACRun  [k]; // run
                if(level == 0)
                  break;
                pos   += run;
                img->scanStats8x8[ipredmode][4*pos+blk4x4]++;
                pos++;
              }
            }
          }
          else
          {
            int*  ACLevel = img->cofAC[i][0][0];
            int*  ACRun   = img->cofAC[i][0][1];
            int  pos       = 0;
            int  level     = 1;
            int  k; 
            
            scan = img->scanOrder8x8[ipredmode];
            for (k=0; k < 64 && level != 0; k++)
            {
              level = ACLevel[k]; // level
              run   = ACRun  [k]; // run
              if(level == 0)
                break;
              pos   += run;
              img->scanStats8x8[ipredmode][pos]++;
              pos++;
            }
          }
        }
        else if(!rdopt && currMB->mb_type == I4MB && input->UseIntraMDDT)
        {
          int   b4; 
          
          for(b4 = 0; b4 < 4; b4++)
          {
            int*  ACLevel = img->cofAC[i][b4][0];
            int*  ACRun   = img->cofAC[i][b4][1];
            int  pos       = 0;
            int  level     = 1;
            int  k; 
            int pic_block_y, pic_block_x;
            int ipredmode; 
            byte (*scan)[2];
            
            pic_block_y = img->block_y+2*(i>>1)+(b4>>1);
            pic_block_x = img->block_x+2*(i &1)+(b4 &1);           
            ipredmode = img->ipredmode[pic_block_y][pic_block_x];
            
            scan = img->scanOrder4x4[ipredmode];
            img->update4x4Count[ipredmode]++;
            
            if(ACLevel[0])
            {
              for (k=0; k < 16 && level != 0; k++)
              {
                level = ACLevel[k]; // level
                run   = ACRun  [k]; // run
                if(level == 0)
                  break;
                pos   += run;
                img->scanStats4x4[ipredmode][pos]++;
                pos++;
              }
            }
          }
        }
#endif
        rate += writeLumaCoeff8x8 (i, currMB->b8mode[i], currMB->luma_transform_size_8x8_flag);
      }
#ifdef USE_INTRA_MDDT
      if(!rdopt && input->UseIntraMDDT)
      {
        int i; 
        for(i = 0; i < 9; i++)
        {
          if(img->update4x4Count[i] >= img->update4x4Thres[i])
          {
            img->update4x4[i] = 1;
            img->update4x4Count[i] = 0;
          }
          else img->update4x4[i] = 0;
          if(img->update8x8Count[i] >= img->update8x8Thres[i])
          {
            img->update8x8[i] = 1;
            img->update8x8Count[i] = 0;
          }
          else img->update8x8[i] = 0;
        }
        
        updateScanOrder(0);
        normalizeScanStats();
      }
#endif 
  }
  else
  {
    //=====  L U M I N A N C E   f o r   1 6 x 1 6  INTRA =====
#ifdef USE_INTRA_MDDT
    if(input->symbol_mode == CABAC && input->UseIntraMDDT)
    {
      if(currMB->cbp & 0xf)
        rate += writeLumaCoeff16x16 ();
    }
    else
    {
#endif
      //----------------------------------------------------
      // DC coeffs
    
      if (input->symbol_mode == UVLC)
      {
        rate += writeCoeff4x4_CAVLC (LUMA_INTRA16x16DC, 0, 0, 0);  // CAVLC
      }
      else
      {
        level=1; // get inside loop
        for (k=0; k<=16 && level!=0; k++)
        {
          level = currSE->value1 = DCLevel[k]; // level
          run   = currSE->value2 = DCRun  [k]; // run
          
          currSE->writing = writeRunLevel_CABAC;
          
          currSE->context     = LUMA_16DC;
          currSE->type        = SE_LUM_DC_INTRA;   // element is of type DC
          img->is_intra_block = 1;
          
          // choose the appropriate data partition
          dataPart = &(currSlice->partArr[partMap[currSE->type]]);
          dataPart->writeSyntaxElement (currSE, dataPart);
          bitCount[BITS_COEFF_Y_MB] += currSE->len;
          rate                      += currSE->len;
#if TRACE
          snprintf(currSE->tracestring, TRACESTRING_SIZE, "DC luma 16x16 sng(%2d) level =%3d run =%2d", k, level, run);
#endif
          // proceed to next SE
          currSE++;
          currMB->currSEnr++;
        }
      }
      
      // AC coeffs
      if (cbp & 15)
      {
        for (mb_y=0; mb_y < 4; mb_y += 2)
          for (mb_x=0; mb_x < 4; mb_x += 2)
            for (j=mb_y; j < mb_y+2; j++)
              for (i=mb_x; i < mb_x+2; i++)
              {
                b8      = 2*(j >> 1) + (i >> 1);
                b4      = 2*(j & 0x01) + (i & 0x01);
                if (input->symbol_mode == UVLC)
                {
                  rate += writeCoeff4x4_CAVLC (LUMA_INTRA16x16AC, b8, b4, 0);  // CAVLC
                }
                else
                {
                  ACLevel = img->cofAC[b8][b4][0];
                  ACRun   = img->cofAC[b8][b4][1];
                  
                  img->subblock_y = j;
                  img->subblock_x = i;
                  
                  level=1; // get inside loop
                  for (k=0;k<16 && level !=0;k++)
                  {
                    level = currSE->value1 = ACLevel[k]; // level
                    run   = currSE->value2 = ACRun  [k]; // run
                    
                    currSE->writing = writeRunLevel_CABAC;
                    currSE->context     = LUMA_16AC;
                    currSE->type        = SE_LUM_AC_INTRA;   // element is of type AC
                    img->is_intra_block = 1;
                    
                    // choose the appropriate data partition
                    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
                    dataPart->writeSyntaxElement (currSE, dataPart);
                    bitCount[BITS_COEFF_Y_MB] += currSE->len;
                    rate                      += currSE->len;
#if TRACE
                    snprintf(currSE->tracestring, TRACESTRING_SIZE, "AC luma 16x16 sng(%2d) level =%3d run =%2d", k, level, run);
#endif
                    // proceed to next SE
                    currSE++;
                    currMB->currSEnr++;
                  }
                }
              }
      }
#ifdef USE_INTRA_MDDT
    }
#endif
  }
  
  return rate;
}



/*!
************************************************************************
* \brief
*    Get the Prediction from the Neighboring Blocks for Number of Nonzero Coefficients 
*    
*    Luma Blocks
************************************************************************
*/
int predict_nnz(int i,int j)
{
  PixelPos pix;
  
  int pred_nnz = 0;
  int cnt      = 0;
  int mb_nr    = img->current_mb_nr;
  
  // left block
  getLuma4x4Neighbour(mb_nr, i, j, -1, 0, &pix);
  
  if (pix.available && active_pps->constrained_intra_pred_flag && ((input->partition_mode != 0) && !img->currentPicture->idr_flag))
  {
    pix.available &= img->intra_block[pix.mb_addr];
    cnt--;
  }
  
  if (pix.available)
  {
    pred_nnz = img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
    cnt++;
  }
  
  // top block
  getLuma4x4Neighbour(mb_nr, i, j, 0, -1, &pix);
  
  if (pix.available && active_pps->constrained_intra_pred_flag && ((input->partition_mode != 0) && !img->currentPicture->idr_flag))
  {
    pix.available &= img->intra_block[pix.mb_addr];
    cnt--;
  }
  
  if (pix.available)
  {
    pred_nnz += img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
    cnt++;
  }
  
  if (cnt==2)
  {
    pred_nnz++;
    pred_nnz>>=1; 
  }
  
  return pred_nnz;
}


/*!
************************************************************************
* \brief
*    Get the Prediction from the Neighboring Blocks for Number of Nonzero Coefficients 
*    
*    Chroma Blocks   
************************************************************************
*/
int predict_nnz_chroma(int i,int j)
{
  PixelPos pix;
  
  int pred_nnz = 0;
  int cnt      = 0;
  int mb_nr    = img->current_mb_nr;
  static int j_off_tab [12] = {0,0,0,0,4,4,4,4,8,8,8,8};
  int j_off = j_off_tab[j];
  
  
  if (img->yuv_format != YUV444)
  {
    //YUV420 and YUV422
    // left block
    getChroma4x4Neighbour(mb_nr, i & 0x01, j-4, -1, 0, &pix);
    
    if (pix.available && active_pps->constrained_intra_pred_flag && ((input->partition_mode != 0) && !img->currentPicture->idr_flag))
    {
      pix.available &= img->intra_block[pix.mb_addr];
      cnt--;
    }
    
    if (pix.available)
    {
      pred_nnz = img->nz_coeff [pix.mb_addr ][2 * (i >> 1) + pix.x][4 + pix.y];
      cnt++;
    }
    
    // top block
    getChroma4x4Neighbour(mb_nr, i & 0x01, j-4, 0, -1, &pix);
    
    if (pix.available && active_pps->constrained_intra_pred_flag && ((input->partition_mode != 0) && !img->currentPicture->idr_flag))
    {
      pix.available &= img->intra_block[pix.mb_addr];
      cnt--;
    }
    
    if (pix.available)
    {
      pred_nnz += img->nz_coeff [pix.mb_addr ][2 * (i >> 1) + pix.x][4 + pix.y];
      cnt++;
    }
  }
  else
  {
    //YUV444
    // left block
    getChroma4x4Neighbour(mb_nr, i, j-j_off, -1, 0, &pix);
    
    if (pix.available && active_pps->constrained_intra_pred_flag && ((input->partition_mode != 0) && !img->currentPicture->idr_flag))
    {
      pix.available &= img->intra_block[pix.mb_addr];
      cnt--;
    }
    
    if (pix.available)
    {
      pred_nnz = img->nz_coeff [pix.mb_addr ][pix.x][j_off + pix.y];
      cnt++;
    }
    
    // top block
    getChroma4x4Neighbour(mb_nr, i, j-j_off, 0, -1, &pix);
    
    if (pix.available && active_pps->constrained_intra_pred_flag && ((input->partition_mode != 0) && !img->currentPicture->idr_flag))
    {
      pix.available &= img->intra_block[pix.mb_addr];
      cnt--;
    }
    
    if (pix.available)
    {
      pred_nnz += img->nz_coeff [pix.mb_addr ][pix.x][j_off + pix.y];
      cnt++;
    }
    
  }
  
  if (cnt==2)
  {
    pred_nnz++;
    pred_nnz>>=1; 
  }
  
  return pred_nnz;
}

#ifdef RDO_Q//TREL_CAVLC
int estSyntaxElement_NumCoeffTrailingOnes(SyntaxElement *se, DataPartition *this_dataPart)
{
  static const int lentab[3][4][17] = 
  {
    {   
      { 1, 6, 8, 9,10,11,13,13,13,14,14,15,15,16,16,16,16},
      { 0, 2, 6, 8, 9,10,11,13,13,14,14,15,15,15,16,16,16},
      { 0, 0, 3, 7, 8, 9,10,11,13,13,14,14,15,15,16,16,16},
      { 0, 0, 0, 5, 6, 7, 8, 9,10,11,13,14,14,15,15,16,16},
    },                                                 
    {                                                  
      { 2, 6, 6, 7, 8, 8, 9,11,11,12,12,12,13,13,13,14,14},
      { 0, 2, 5, 6, 6, 7, 8, 9,11,11,12,12,13,13,14,14,14},
      { 0, 0, 3, 6, 6, 7, 8, 9,11,11,12,12,13,13,13,14,14},
      { 0, 0, 0, 4, 4, 5, 6, 6, 7, 9,11,11,12,13,13,13,14},
    },                                                 
    {                                                  
      { 4, 6, 6, 6, 7, 7, 7, 7, 8, 8, 9, 9, 9,10,10,10,10},
      { 0, 4, 5, 5, 5, 5, 6, 6, 7, 8, 8, 9, 9, 9,10,10,10},
      { 0, 0, 4, 5, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9,10,10,10},
      { 0, 0, 0, 4, 4, 4, 4, 4, 5, 6, 7, 8, 8, 9,10,10,10},
    },

  };

  int vlcnum;

  vlcnum = se->len;


  if (vlcnum == 3)
  {
    se->len = 6;  // 4 + 2 bit FLC
  }
  else
  {
    se->len = lentab[vlcnum][se->value2][se->value1];
    //se->inf = codtab[vlcnum][se->value2][se->value1];
  }

  if (se->len == 0)
  {
    printf("ERROR: (numcoeff,trailingones) not valid: vlc=%d (%d, %d)\n", 
      vlcnum, se->value1, se->value2);
    exit(-1);
  }


#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}


int estSyntaxElement_NumCoeffTrailingOnesChromaDC(SyntaxElement *se, DataPartition *this_dataPart)
{
  static const int lentab[3][4][17] = 
  {
    //YUV420
   {{ 2, 6, 6, 6, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 1, 6, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    { 0, 0, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    { 0, 0, 0, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
    //YUV422
   {{ 1, 7, 7, 9, 9,10,11,12,13, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 2, 7, 7, 9,10,11,12,12, 0, 0, 0, 0, 0, 0, 0, 0}, 
    { 0, 0, 3, 7, 7, 9,10,11,12, 0, 0, 0, 0, 0, 0, 0, 0}, 
    { 0, 0, 0, 5, 6, 7, 7,10,11, 0, 0, 0, 0, 0, 0, 0, 0}},
    //YUV444
   {{ 1, 6, 8, 9,10,11,13,13,13,14,14,15,15,16,16,16,16},
    { 0, 2, 6, 8, 9,10,11,13,13,14,14,15,15,15,16,16,16},
    { 0, 0, 3, 7, 8, 9,10,11,13,13,14,14,15,15,16,16,16},
    { 0, 0, 0, 5, 6, 7, 8, 9,10,11,13,14,14,15,15,16,16}}
  };

  static const int codtab[3][4][17] = 
  {
    //YUV420
   {{ 1, 7, 4, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 1, 6, 3, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 0, 1, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
    //YUV422
   {{ 1,15,14, 7, 6, 7, 7, 7, 7, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 1,13,12, 5, 6, 6, 6, 5, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 0, 1,11,10, 4, 5, 5, 4, 0, 0, 0, 0, 0, 0, 0, 0},
    { 0, 0, 0, 1, 1, 9, 8, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0}},
    //YUV444
   {{ 1, 5, 7, 7, 7, 7,15,11, 8,15,11,15,11,15,11, 7, 4}, 
    { 0, 1, 4, 6, 6, 6, 6,14,10,14,10,14,10, 1,14,10, 6}, 
    { 0, 0, 1, 5, 5, 5, 5, 5,13, 9,13, 9,13, 9,13, 9, 5}, 
    { 0, 0, 0, 3, 3, 4, 4, 4, 4, 4,12,12, 8,12, 8,12, 8}}
  
  };
  int yuv = img->yuv_format - 1;

  se->len = lentab[yuv][se->value2][se->value1];
  se->inf = codtab[yuv][se->value2][se->value1];

  if (se->len == 0)
  {
    printf("ERROR: (numcoeff,trailingones) not valid: (%d, %d)\n", 
      se->value1, se->value2);
    exit(-1);
  }


#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}


int estSyntaxElement_VLC(SyntaxElement *se, DataPartition *this_dataPart)
{

  //se->inf = se->value1;
  se->len = se->value2;

#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}


int estSyntaxElement_Level_VLC1(SyntaxElement *se)//, DataPartition *this_dataPart)
{
  int level, levabs, sign;

  level = se->value1;
  levabs = absm(level);
  sign = (level < 0 ? 1 : 0);

  
  if (levabs < 8)
  {
    se->len = levabs * 2 + sign - 1;
    //se->inf = 1;
  }
  else if (levabs < 8+8)
  {
    // escape code1
    se->len = 19;//14 + 1 + 4;
    //se->inf = (1 << 4) | ((levabs - 8) << 1) | sign;
  }
  else
  {
    // escape code2
    se->len = 28;//14 + 2 + 12;
    //se->inf = (0x1 << 12) | ((levabs - 16)<< 1) | sign;
  }


#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}


/*!
 ************************************************************************
 * \brief
 *    write VLC for Coeff Level
 ************************************************************************
 */
int estSyntaxElement_Level_VLCN(SyntaxElement *se, int vlc)//, DataPartition *this_dataPart)
{
  int iCodeword;
  int iLength;

  int level = se->value1;

  int levabs = absm(level);
  int sign = (level < 0 ? 1 : 0);  

  int shift = vlc-1;
  int escape = (15<<shift)+1;

  int numPrefix = (levabs-1)>>shift;

  int sufmask = ~((0xffffffff)<<shift);
  int suffix = (levabs-1)&sufmask;

  if (levabs < escape)
  {
    iLength = numPrefix + vlc + 1;
    iCodeword = (1<<(shift+1))|(suffix<<1)|sign;
  }
  else
  {
    iLength = 28;
    iCodeword = (1<<12)|((levabs-escape)<<1)|sign;
  }
  se->len = iLength;
 
#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}
int estSyntaxElement_TotalZeros(SyntaxElement *se, DataPartition *this_dataPart)
{
  static const int lentab[TOTRUN_NUM][16] = 
  {
    { 1,3,3,4,4,5,5,6,6,7,7,8,8,9,9,9},  
    { 3,3,3,3,3,4,4,4,4,5,5,6,6,6,6},  
    { 4,3,3,3,4,4,3,3,4,5,5,6,5,6},  
    { 5,3,4,4,3,3,3,4,3,4,5,5,5},  
    { 4,4,4,3,3,3,3,3,4,5,4,5},  
    { 6,5,3,3,3,3,3,3,4,3,6},  
    { 6,5,3,3,3,2,3,4,3,6},  
    { 6,4,5,3,2,2,3,3,6},  
    { 6,6,4,2,2,3,2,5},  
    { 5,5,3,2,2,2,4},  
    { 4,4,3,3,1,3},  
    { 4,4,2,1,3},  
    { 3,3,1,2},  
    { 2,2,1},  
    { 1,1},  
  };

  int vlcnum;

  vlcnum = se->len;

  // se->value1 : TotalZeros
  se->len = lentab[vlcnum][se->value1];
  //se->inf = codtab[vlcnum][se->value1];

  if (se->len == 0)
  {
    printf("ERROR: (TotalZeros) not valid: (%d)\n",se->value1);
    exit(-1);
  }


#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}
int estSyntaxElement_TotalZerosChromaDC(SyntaxElement *se, DataPartition *this_dataPart)
{
  static const int lentab[3][TOTRUN_NUM][16] = 
  {
    //YUV420
   {{ 1,2,3,3},
    { 1,2,2},
    { 1,1}},
    //YUV422
   {{ 1,3,3,4,4,4,5,5},
    { 3,2,3,3,3,3,3},
    { 3,3,2,2,3,3},
    { 3,2,2,2,3},
    { 2,2,2,2},
    { 2,2,1},   
    { 1,1}},
    //YUV444
   {{ 1,3,3,4,4,5,5,6,6,7,7,8,8,9,9,9},  
    { 3,3,3,3,3,4,4,4,4,5,5,6,6,6,6},  
    { 4,3,3,3,4,4,3,3,4,5,5,6,5,6},  
    { 5,3,4,4,3,3,3,4,3,4,5,5,5},  
    { 4,4,4,3,3,3,3,3,4,5,4,5},  
    { 6,5,3,3,3,3,3,3,4,3,6},  
    { 6,5,3,3,3,2,3,4,3,6},  
    { 6,4,5,3,2,2,3,3,6},  
    { 6,6,4,2,2,3,2,5},  
    { 5,5,3,2,2,2,4},  
    { 4,4,3,3,1,3},  
    { 4,4,2,1,3},  
    { 3,3,1,2},  
    { 2,2,1},  
    { 1,1}}  
  };

  static const int codtab[3][TOTRUN_NUM][16] = 
  {
    //YUV420
   {{ 1,1,1,0},
    { 1,1,0},
    { 1,0}},
    //YUV422
   {{ 1,2,3,2,3,1,1,0},
    { 0,1,1,4,5,6,7},
    { 0,1,1,2,6,7},
    { 6,0,1,2,7},
    { 0,1,2,3},
    { 0,1,1},   
    { 0,1}},
    //YUV444
   {{1,3,2,3,2,3,2,3,2,3,2,3,2,3,2,1},
    {7,6,5,4,3,5,4,3,2,3,2,3,2,1,0},
    {5,7,6,5,4,3,4,3,2,3,2,1,1,0},
    {3,7,5,4,6,5,4,3,3,2,2,1,0},
    {5,4,3,7,6,5,4,3,2,1,1,0},
    {1,1,7,6,5,4,3,2,1,1,0},
    {1,1,5,4,3,3,2,1,1,0},
    {1,1,1,3,3,2,2,1,0},
    {1,0,1,3,2,1,1,1,},
    {1,0,1,3,2,1,1,},
    {0,1,1,2,1,3},
    {0,1,1,1,1},
    {0,1,1,1},
    {0,1,1},
    {0,1}}  
  };
  int vlcnum;
  int yuv = img->yuv_format - 1;
  
  vlcnum = se->len;

  se->len = lentab[yuv][vlcnum][se->value1];
  se->inf = codtab[yuv][vlcnum][se->value1];

  if (se->len == 0)
  {
    printf("ERROR: (TotalZeros) not valid: (%d)\n",se->value1);
    exit(-1);
  }


#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}
int estSyntaxElement_Run(SyntaxElement *se, DataPartition *this_dataPart)
{
  static const int Run_lentab[TOTRUN_NUM][16] = 
  {
    {1,1},
    {1,2,2},
    {2,2,2,2},
    {2,2,2,3,3},
    {2,2,3,3,3,3},
    {2,3,3,3,3,3,3},
    {3,3,3,3,3,3,3,4,5,6,7,8,9,10,11},
  };

  int vlcnum;

  vlcnum = se->len;

  se->len = Run_lentab[vlcnum][se->value1];
  //se->inf = codtab[vlcnum][se->value1];

  if (se->len == 0)
  {
    printf("ERROR: (run) not valid: (%d)\n",se->value1);
    exit(-1);
  }


#if TRACE
  if (se->type <= 1)
    trace2out (se);
#endif

  return (se->len);
}

extern const byte SNGL_SCAN[16][2];

int estCAVLCbits (int m4[4][4], int level_to_enc[16], int nnz, int block_type, int b8, int b4, int param)
{
  int           no_bits    = 0;
  Macroblock    *currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  //int           *bitCount  = currMB->bitcounter;
  //DataPartition *dataPart;

  int coeff_ctr, scan_pos;
  int i, j, k,level,run,vlcnum;
  int numcoeff, lastcoeff, numtrailingones; 
  int numones, totzeros, zerosleft, numcoef;
  int numcoeff_vlc;
  int code, level_two_or_higher;
  int bitcounttype = 0;
  int max_coeff_num = 0;
  const byte (*pos_scan)[2] = SNGL_SCAN;
  

  static int incVlc[] = {0,3,6,12,24,48,32768};  // maximum vlc = 6

  int  pLevel[17] = {0};
  int  pRun[17] = {0};

  static const int Token_lentab[3][4][17] = 
  {
    {   
      { 1, 6, 8, 9,10,11,13,13,13,14,14,15,15,16,16,16,16},
      { 0, 2, 6, 8, 9,10,11,13,13,14,14,15,15,15,16,16,16},
      { 0, 0, 3, 7, 8, 9,10,11,13,13,14,14,15,15,16,16,16},
      { 0, 0, 0, 5, 6, 7, 8, 9,10,11,13,14,14,15,15,16,16},
    },                                                 
    {                                                  
      { 2, 6, 6, 7, 8, 8, 9,11,11,12,12,12,13,13,13,14,14},
      { 0, 2, 5, 6, 6, 7, 8, 9,11,11,12,12,13,13,14,14,14},
      { 0, 0, 3, 6, 6, 7, 8, 9,11,11,12,12,13,13,13,14,14},
      { 0, 0, 0, 4, 4, 5, 6, 6, 7, 9,11,11,12,13,13,13,14},
    },                                                 
    {                                                  
      { 4, 6, 6, 6, 7, 7, 7, 7, 8, 8, 9, 9, 9,10,10,10,10},
      { 0, 4, 5, 5, 5, 5, 6, 6, 7, 8, 8, 9, 9, 9,10,10,10},
      { 0, 0, 4, 5, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9,10,10,10},
      { 0, 0, 0, 4, 4, 4, 4, 4, 5, 6, 7, 8, 8, 9,10,10,10},
    },

  };

  static const int Totalzeros_lentab[TOTRUN_NUM][16] = 
  {
    { 1,3,3,4,4,5,5,6,6,7,7,8,8,9,9,9},  
    { 3,3,3,3,3,4,4,4,4,5,5,6,6,6,6},  
    { 4,3,3,3,4,4,3,3,4,5,5,6,5,6},  
    { 5,3,4,4,3,3,3,4,3,4,5,5,5},  
    { 4,4,4,3,3,3,3,3,4,5,4,5},  
    { 6,5,3,3,3,3,3,3,4,3,6},  
    { 6,5,3,3,3,2,3,4,3,6},  
    { 6,4,5,3,2,2,3,3,6},  
    { 6,6,4,2,2,3,2,5},  
    { 5,5,3,2,2,2,4},  
    { 4,4,3,3,1,3},  
    { 4,4,2,1,3},  
    { 3,3,1,2},  
    { 2,2,1},  
    { 1,1},  
  };
  static const int Run_lentab[TOTRUN_NUM][16] = 
  {
    {1,1},
    {1,2,2},
    {2,2,2,2},
    {2,2,2,3,3},
    {2,2,3,3,3,3},
    {2,3,3,3,3,3,3},
    {3,3,3,3,3,3,3,4,5,6,7,8,9,10,11},
  };

  //convert zigzag scan to (run, level) pairs
  run=-1;
  scan_pos=0;

  switch (block_type)
  {
    case LUMA:
      max_coeff_num = 16;
      bitcounttype = BITS_COEFF_Y_MB;

      break;
    case LUMA_INTRA16x16DC:
      max_coeff_num = 16;
      bitcounttype = BITS_COEFF_Y_MB;

      break;
    case LUMA_INTRA16x16AC:
      max_coeff_num = 15;
      bitcounttype = BITS_COEFF_Y_MB;

      break;
    case CHROMA_AC:
      max_coeff_num = 15;
      bitcounttype = BITS_COEFF_UV_MB;

      break;
    default:
      error("writeCoeff4x4_CAVLC: Invalid block type", 600);
      break;
  }

  for (coeff_ctr=0;coeff_ctr < max_coeff_num;coeff_ctr++)
  {
    i=pos_scan[coeff_ctr][0];
    j=pos_scan[coeff_ctr][1];
    
    run++;

    level = level_to_enc[coeff_ctr];

    if (level != 0)
    {
	  pLevel[scan_pos] = sign(level,m4[j][i]);

      pRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;                     // reset zero level counter
    }
  }
  pLevel[scan_pos] = 0;  

  numcoeff = 0;
  numtrailingones = 0;
  numones = 0;
  lastcoeff = 0;
  totzeros = 0;
  level = 1;

  for(k = 0; (k <= max_coeff_num)&& level !=0; k++)
  {
    level = pLevel[k]; // level
    run   = pRun[k];   // run

    if (level)
    {
      if (run)
        totzeros += run;
      if (absm(level) == 1)
      {
        numtrailingones ++;
        numones ++;
        if (numtrailingones > 3)
        {
          numtrailingones = 3; /* clip to 3 */
        }
      }
      else
      {
        numtrailingones = 0;
      }
      numcoeff ++;
      lastcoeff = k;
    }
  }



    if (nnz < 2)
    {
      numcoeff_vlc = 0;
    }
    else if (nnz < 4)
    {
      numcoeff_vlc = 1;
    }
    else if (nnz < 8)
    {
      numcoeff_vlc = 2;
    }
    else 
    {
      numcoeff_vlc = 3;
    }


  //currSE->type  = dptype;   

  currSE->value1 = numcoeff;
  currSE->value2 = numtrailingones;
  currSE->len = numcoeff_vlc; /* use len to pass vlcnum */



    if (currSE->len == 3)
      no_bits += 6;  // 4 + 2 bit FLC
    else
      no_bits += Token_lentab[currSE->len][currSE->value2][currSE->value1];

  if (!numcoeff)
    return no_bits;

  if (numcoeff)
  {
    code = 0;
    for (k = lastcoeff; k > lastcoeff-numtrailingones; k--)
    {
      level = pLevel[k]; // level
      if (absm(level) > 1)
      {
        printf("ERROR: level > 1\n");
        exit(-1);
      }
      code <<= 1;
      if (level < 0)
      {
        code |= 0x1;
      }
    }

    if (numtrailingones)
    {
      no_bits               +=numtrailingones;

    }

    // encode levels
    level_two_or_higher = 1;
    if (numcoeff > 3 && numtrailingones == 3)
      level_two_or_higher = 0;

    if (numcoeff > 10 && numtrailingones < 3)
      vlcnum = 1;
    else
      vlcnum = 0;

    for (k = lastcoeff - numtrailingones; k >= 0; k--)
    {
      level = pLevel[k]; // level

      currSE->value1 = level;
      //currSE->type  = dptype;   

          if (level_two_or_higher)
          {
            if (currSE->value1 > 0)
              currSE->value1 --;
            else
              currSE->value1 ++;
            level_two_or_higher = 0;
          }

      //    encode level
      if (vlcnum == 0)
        estSyntaxElement_Level_VLC1(currSE);//, dataPart);
      else
        estSyntaxElement_Level_VLCN(currSE, vlcnum);//, dataPart);

      // update VLC table
      if (absm(level)>incVlc[vlcnum])
        vlcnum++;

      if (k == lastcoeff - numtrailingones && absm(level)>3)
        vlcnum = 2;

      //bitCount[bitcounttype]+=currSE->len;
      no_bits               +=currSE->len;
    }

    // encode total zeroes
    if (numcoeff < max_coeff_num)
    {
      //currSE->type  = dptype;   
      currSE->value1 = totzeros;

      vlcnum = numcoeff-1;

      currSE->len = vlcnum;
      no_bits               +=Totalzeros_lentab[currSE->len][currSE->value1];
    }

    // encode run before each coefficient
    zerosleft = totzeros;
    numcoef = numcoeff;
    for (k = lastcoeff; k >= 0; k--)
    {
      run = pRun[k]; // run

      currSE->value1 = run;

      // for last coeff, run is remaining totzeros
      // when zerosleft is zero, remaining coeffs have 0 run
      if (numcoeff <= 1 || !zerosleft)
        break;

      if (numcoef > 1 && zerosleft) 
      {

        vlcnum = zerosleft - 1;
        if (vlcnum > RUNBEFORE_NUM-1)
          vlcnum = RUNBEFORE_NUM-1;

        currSE->len = vlcnum;

        no_bits               +=Run_lentab[currSE->len][currSE->value1];

        zerosleft -= run;
        numcoef --;
      }
    }
  }

  return no_bits;
}
#endif


/*!
************************************************************************
* \brief
*    Writes coeff of an 4x4 block (CAVLC)
*
* \author
*    Karl Lillevold <karll@real.com>
*    contributions by James Au <james@ubvideo.com>
************************************************************************
*/

int writeCoeff4x4_CAVLC (int block_type, int b8, int b4, int param)
{
  int           no_bits    = 0;
  Macroblock    *currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  int           *bitCount  = currMB->bitcounter;
  Slice         *currSlice = img->currentSlice;
  DataPartition *dataPart;
  int           *partMap   = assignSE2partition[input->partition_mode];
  
  int k,level,run,vlcnum;
  int numcoeff, lastcoeff, numtrailingones; 
  int numones, totzeros, zerosleft, numcoef;
  int numcoeff_vlc;
  int code, level_two_or_higher;
  int dptype = 0, bitcounttype = 0;
  int nnz, max_coeff_num = 0, cdc=0, cac=0;
  int subblock_x, subblock_y;
  char type[15];
  
  static int incVlc[] = {0,3,6,12,24,48,32768};  // maximum vlc = 6
  
  int*  pLevel = NULL;
  int*  pRun = NULL;
  
  switch (block_type)
  {
  case LUMA:
    max_coeff_num = 16;
    bitcounttype = BITS_COEFF_Y_MB;
    
    pLevel = img->cofAC[b8][b4][0];
    pRun   = img->cofAC[b8][b4][1];
    
    sprintf(type, "%s", "Luma");
    if (IS_INTRA (currMB))
    {
      dptype = SE_LUM_AC_INTRA;
    }
    else
    {
      dptype = SE_LUM_AC_INTER;
    }
    break;
  case LUMA_INTRA16x16DC:
    max_coeff_num = 16;
    bitcounttype = BITS_COEFF_Y_MB;
    
    pLevel = img->cofDC[0][0];
    pRun   = img->cofDC[0][1];
    
    sprintf(type, "%s", "Lum16DC");
    dptype = SE_LUM_DC_INTRA;
    break;
  case LUMA_INTRA16x16AC:
    max_coeff_num = 15;
    bitcounttype = BITS_COEFF_Y_MB;
    
    pLevel = img->cofAC[b8][b4][0];
    pRun   = img->cofAC[b8][b4][1];
    
    sprintf(type, "%s", "Lum16AC");
    dptype = SE_LUM_AC_INTRA;
    break;
    
  case CHROMA_DC:
    max_coeff_num = img->num_cdc_coeff;
    bitcounttype = BITS_COEFF_UV_MB;
    cdc = 1;
    
    pLevel = img->cofDC[param+1][0];
    pRun   = img->cofDC[param+1][1];
    
    sprintf(type, "%s", "ChrDC");
    if (IS_INTRA (currMB))
    {
      dptype = SE_CHR_DC_INTRA;
    }
    else
    {
      dptype = SE_CHR_DC_INTER;
    }
    break;
  case CHROMA_AC:
    max_coeff_num = 15;
    bitcounttype = BITS_COEFF_UV_MB;
    cac = 1;
    
    pLevel = img->cofAC[b8][b4][0];
    pRun   = img->cofAC[b8][b4][1];
    
    sprintf(type, "%s", "ChrAC");
    if (IS_INTRA (currMB))
    {
      dptype = SE_CHR_AC_INTRA;
    }
    else
    {
      dptype = SE_CHR_AC_INTER;
    }
    break;
  default:
    error("writeCoeff4x4_CAVLC: Invalid block type", 600);
    break;
  }
  
  dataPart = &(currSlice->partArr[partMap[dptype]]);
  
  numcoeff = 0;
  numtrailingones = 0;
  numones = 0;
  lastcoeff = 0;
  totzeros = 0;
  level = 1;
  
  for(k = 0; (k <= ((cdc)?img->num_cdc_coeff:16))&& level !=0; k++)
  {
    level = pLevel[k]; // level
    run   = pRun[k];   // run
    
    if (level)
    {
      if (run)
        totzeros += run;
      if (absm(level) == 1)
      {
        numtrailingones ++;
        numones ++;
        if (numtrailingones > 3)
        {
          numtrailingones = 3; /* clip to 3 */
        }
      }
      else
      {
        numtrailingones = 0;
      }
      numcoeff ++;
      lastcoeff = k;
    }
  }
  
  if (!cdc)
  {
    if (!cac)
    {
      // luma
      subblock_x = ((b8&0x1)==0)?(((b4&0x1)==0)?0:1):(((b4&0x1)==0)?2:3); 
      // horiz. position for coeff_count context
      subblock_y = (b8<2)?((b4<2)?0:1):((b4<2)?2:3); 
      // vert.  position for coeff_count context
      nnz = predict_nnz(subblock_x,subblock_y);
    }
    else
    {
      // chroma AC
      subblock_x = param >> 4;
      subblock_y = param & 15;
      nnz = predict_nnz_chroma(subblock_x,subblock_y);
    }
    
    img->nz_coeff [img->current_mb_nr ][subblock_x][subblock_y] = numcoeff;
    
    
    if (nnz < 2)
    {
      numcoeff_vlc = 0;
    }
    else if (nnz < 4)
    {
      numcoeff_vlc = 1;
    }
    else if (nnz < 8)
    {
      numcoeff_vlc = 2;
    }
    else 
    {
      numcoeff_vlc = 3;
    }
    
  }
  else
  {
    // chroma DC (has its own VLC)
    // numcoeff_vlc not relevant
    numcoeff_vlc = 0;
    
    subblock_x = param;
    subblock_y = param;
  }
  
  currSE->type  = dptype;   
  
  currSE->value1 = numcoeff;
  currSE->value2 = numtrailingones;
  currSE->len = numcoeff_vlc; /* use len to pass vlcnum */
  
#if TRACE
  snprintf(currSE->tracestring, 
    TRACESTRING_SIZE, "%s # c & tr.1s(%d,%d) vlc=%d #c=%d #t1=%d",
    type, subblock_x, subblock_y, numcoeff_vlc, numcoeff, numtrailingones);
#endif
  
  if (!cdc)
    writeSyntaxElement_NumCoeffTrailingOnes(currSE, dataPart);
  else
    writeSyntaxElement_NumCoeffTrailingOnesChromaDC(currSE, dataPart);
  
  bitCount[bitcounttype]+=currSE->len;
  no_bits               +=currSE->len;
  
  // proceed to next SE
  currSE++;
  currMB->currSEnr++;
  
  
  if (!numcoeff)
    return no_bits;
  
  if (numcoeff)
  {
    code = 0;
    for (k = lastcoeff; k > lastcoeff-numtrailingones; k--)
    {
      level = pLevel[k]; // level
      if (absm(level) > 1)
      {
        printf("ERROR: level > 1\n");
        exit(-1);
      }
      code <<= 1;
      if (level < 0)
      {
        code |= 0x1;
      }
    }
    
    if (numtrailingones)
    {
      currSE->type  = dptype;   
      
      currSE->value2 = numtrailingones;
      currSE->value1 = code;
      
#if TRACE
      snprintf(currSE->tracestring, 
        TRACESTRING_SIZE, "%s trailing ones sign (%d,%d)", 
        type, subblock_x, subblock_y);
#endif
      
      writeSyntaxElement_VLC (currSE, dataPart);
      bitCount[bitcounttype]+=currSE->len;
      no_bits               +=currSE->len;
      
      // proceed to next SE
      currSE++;
      currMB->currSEnr++;
    }
    
    // encode levels
    level_two_or_higher = 1;
    if (numcoeff > 3 && numtrailingones == 3)
      level_two_or_higher = 0;
    
    if (numcoeff > 10 && numtrailingones < 3)
      vlcnum = 1;
    else
      vlcnum = 0;
    
    for (k = lastcoeff - numtrailingones; k >= 0; k--)
    {
      level = pLevel[k]; // level
      
      currSE->value1 = level;
      currSE->type  = dptype;   
      
#if TRACE
      snprintf(currSE->tracestring, 
        TRACESTRING_SIZE, "%s lev (%d,%d) k=%d vlc=%d lev=%3d",
        type, subblock_x, subblock_y, k, vlcnum, level);
#endif
      
      if (level_two_or_higher)
      {
        if (currSE->value1 > 0)
          currSE->value1 --;
        else
          currSE->value1 ++;
        level_two_or_higher = 0;
      }
      
      //    encode level
      if (vlcnum == 0)
        writeSyntaxElement_Level_VLC1(currSE, dataPart);
      else
        writeSyntaxElement_Level_VLCN(currSE, vlcnum, dataPart);
      
      // update VLC table
      if (absm(level)>incVlc[vlcnum])
        vlcnum++;
      
      if (k == lastcoeff - numtrailingones && absm(level)>3)
        vlcnum = 2;
      
      bitCount[bitcounttype]+=currSE->len;
      no_bits               +=currSE->len;
      
      // proceed to next SE
      currSE++;
      currMB->currSEnr++;
    }
    
    // encode total zeroes
    if (numcoeff < max_coeff_num)
    {
      
      currSE->type  = dptype;   
      currSE->value1 = totzeros;
      
      vlcnum = numcoeff-1;
      
      currSE->len = vlcnum;
      
#if TRACE
      snprintf(currSE->tracestring, 
        TRACESTRING_SIZE, "%s totalrun (%d,%d) vlc=%d totzeros=%3d",
        type, subblock_x, subblock_y, vlcnum, totzeros);
#endif
      if (!cdc)
        writeSyntaxElement_TotalZeros(currSE, dataPart);
      else
        writeSyntaxElement_TotalZerosChromaDC(currSE, dataPart);
      
      bitCount[bitcounttype]+=currSE->len;
      no_bits               +=currSE->len;
      
      // proceed to next SE
      currSE++;
      currMB->currSEnr++;
    }
    
    // encode run before each coefficient
    zerosleft = totzeros;
    numcoef = numcoeff;
    for (k = lastcoeff; k >= 0; k--)
    {
      run = pRun[k]; // run
      
      currSE->value1 = run;
      currSE->type  = dptype;   
      
      // for last coeff, run is remaining totzeros
      // when zerosleft is zero, remaining coeffs have 0 run
      if (numcoeff <= 1 || !zerosleft)
        break;
      
      if (numcoef > 1 && zerosleft) 
      {
        
        vlcnum = zerosleft - 1;
        if (vlcnum > RUNBEFORE_NUM-1)
          vlcnum = RUNBEFORE_NUM-1;
        
        currSE->len = vlcnum;
        
#if TRACE
        snprintf(currSE->tracestring, 
          TRACESTRING_SIZE, "%s run (%d,%d) k=%d vlc=%d run=%2d",
          type, subblock_x, subblock_y, k, vlcnum, run);
#endif
        
        writeSyntaxElement_Run(currSE, dataPart);
        
        bitCount[bitcounttype]+=currSE->len;
        no_bits               +=currSE->len;
        
        zerosleft -= run;
        numcoef --;
        
        // proceed to next SE
        currSE++;
        currMB->currSEnr++;
      }
    }
  }
  
  return no_bits;
}

/*!
************************************************************************
* \brief
*    Find best 16x16 based intra mode
*
* \par Input:
*    Image parameters, pointer to best 16x16 intra mode
*
* \par Output:
*    best 16x16 based SAD
************************************************************************/
int find_sad_16x16(int *intra_mode)
{
  int current_intra_sad_2,best_intra_sad2;
  int M1[16][16],M0[4][4][4][4],M3[4],M4[4][4];
  
  int i,j,k;
  int ii,jj;
  int mb_nr = img->current_mb_nr;
  
  PixelPos up;          //!< pixel position p(0,-1)
  PixelPos left[17];    //!< pixel positions p(-1, -1..15)
  
  int up_avail, left_avail, left_up_avail;
  
  for (i=0;i<17;i++)
  {
    getNeighbour(mb_nr, -1 ,  i-1 , 1, &left[i]);
  }
  
  getNeighbour(mb_nr, 0     ,  -1 , 1, &up);
  
  if (!(input->UseConstrainedIntraPred))
  {
    up_avail   = up.available;
    left_avail = left[1].available;
    left_up_avail = left[0].available;
  }
  else
  {
    up_avail      = up.available ? img->intra_block[up.mb_addr] : 0;
    for (i=1, left_avail=1; i<17;i++)
      left_avail  &= left[i].available ? img->intra_block[left[i].mb_addr]: 0;
    left_up_avail = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
  }
  
  best_intra_sad2=MAX_VALUE;
  *intra_mode = DC_PRED_16;
  
  for (k=0;k<4;k++)
  {
    if (input->IntraDisableInterOnly == 0 || img->type != I_SLICE)
    {
      if (input->Intra16x16ParDisable && (k==VERT_PRED_16||k==HOR_PRED_16))
        continue;
      
      if (input->Intra16x16PlaneDisable && k==PLANE_16)
        continue;
    }
    //check if there are neighbours to predict from
    if ((k==0 && !up_avail) || (k==1 && !left_avail) || (k==3 && (!left_avail || !up_avail || !left_up_avail)))
    {
      ; // edge, do nothing
    }
    else
    {
      for (j=0;j<16;j++)
      {
        for (i=0;i<16;i++)
        {
          M1[j][i]=imgY_org[img->opix_y+j][img->opix_x+i]-img->mprr_2[k][j][i];
          M0[i & 0x03][i >> 2][j & 0x03][j >> 2]=M1[j][i];
        }
      }
      current_intra_sad_2=0;              // no SAD start handicap here
      for (jj=0;jj<4;jj++)
      {
        for (ii=0;ii<4;ii++)
        {
          for (j=0;j<4;j++)
          {
            M3[0]=M0[0][ii][j][jj]+M0[3][ii][j][jj];
            M3[1]=M0[1][ii][j][jj]+M0[2][ii][j][jj];
            M3[2]=M0[1][ii][j][jj]-M0[2][ii][j][jj];
            M3[3]=M0[0][ii][j][jj]-M0[3][ii][j][jj];
            
            M0[0][ii][j][jj]=M3[0]+M3[1];
            M0[2][ii][j][jj]=M3[0]-M3[1];
            M0[1][ii][j][jj]=M3[2]+M3[3];
            M0[3][ii][j][jj]=M3[3]-M3[2];
          }
          
          for (i=0;i<4;i++)
          {
            M3[0]=M0[i][ii][0][jj]+M0[i][ii][3][jj];
            M3[1]=M0[i][ii][1][jj]+M0[i][ii][2][jj];
            M3[2]=M0[i][ii][1][jj]-M0[i][ii][2][jj];
            M3[3]=M0[i][ii][0][jj]-M0[i][ii][3][jj];
            
            M0[i][ii][0][jj]=M3[0]+M3[1];
            M0[i][ii][2][jj]=M3[0]-M3[1];
            M0[i][ii][1][jj]=M3[2]+M3[3];
            M0[i][ii][3][jj]=M3[3]-M3[2];
            for (j=0;j<4;j++)
              if ((i+j)!=0)
                current_intra_sad_2 += absm(M0[i][ii][j][jj]);
          }
        }
      }
      
      for (j=0;j<4;j++)
        for (i=0;i<4;i++)
          M4[j][i]=M0[0][i][0][j]/4;
        
        // Hadamard of DC koeff
        for (j=0;j<4;j++)
        {
          M3[0]=M4[j][0]+M4[j][3];
          M3[1]=M4[j][1]+M4[j][2];
          M3[2]=M4[j][1]-M4[j][2];
          M3[3]=M4[j][0]-M4[j][3];
          
          M4[j][0]=M3[0]+M3[1];
          M4[j][2]=M3[0]-M3[1];
          M4[j][1]=M3[2]+M3[3];
          M4[j][3]=M3[3]-M3[2];
        }
        
        for (i=0;i<4;i++)
        {
          M3[0]=M4[0][i]+M4[3][i];
          M3[1]=M4[1][i]+M4[2][i];
          M3[2]=M4[1][i]-M4[2][i];
          M3[3]=M4[0][i]-M4[3][i];
          
          M4[0][i]=M3[0]+M3[1];
          M4[2][i]=M3[0]-M3[1];
          M4[1][i]=M3[2]+M3[3];
          M4[3][i]=M3[3]-M3[2];
          
          for (j=0;j<4;j++)
            current_intra_sad_2 += absm(M4[j][i]);
        }
        if(current_intra_sad_2 < best_intra_sad2)
        {
          best_intra_sad2=current_intra_sad_2;
          *intra_mode = k; // update best intra mode
        }
    }
  }
  best_intra_sad2 = best_intra_sad2/2;
  
  return best_intra_sad2;
  
}



